#!/usr/bin/env python

import os
import sys
import time
import atexit
import subprocess
import logging

HERE = os.path.abspath(os.path.dirname(__file__))

sys.path.insert(0, os.path.join(HERE, '..'))
from testing import venv


def check_lxc_mounted():
    out = subprocess.check_output(['mount'])
    return ' /var/lib/lxc ' in out


def _run_and_log_on_failure(cmd):
    '''Run the command and return True. On failure, log the output and return
    False.'''

    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        return True
    except subprocess.CalledProcessError as e:
        # Only print the output of the command if it failed.
        msg = 'Calling `%s` failed (with code %d):'
        logging.error(msg % (e.cmd, e.returncode))
        for line in e.output.splitlines():
            logging.error(line)
        return False


def bootstrap():
    '''Try to bootstrap, then confirm it succeeded by calling status.'''

    cmd = ['juju', 'bootstrap']
    if not _run_and_log_on_failure(cmd):
        logging.error('Bootstrapping failed.')
        return False

    cmd = ['juju', 'status']
    if not _run_and_log_on_failure(cmd):
        # XXX ev 2014-07-07: Workaround juju:
        # Error details: EOF
        time.sleep(15)
        if not _run_and_log_on_failure(cmd):
            logging.error('Bootstrapping suceeded but instance unreachable.')
            return False

    return True


def destroy_environment():
    '''Destroy the local environment. Return True if successful, False
    otherwise.'''

    cmd = ['juju', 'destroy-environment', '--force', '-y', 'local']
    return _run_and_log_on_failure(cmd)


def fake_out_branch():
    '''Work around tarmac using bzr export vs bzr branch (quoting dobey):

    tarmac used to just run the tests in the checked out branch, but it was
    changed to doing an export to a temp dir for security, so that rogue
    branches can't commit themselves and such.

    Return True on success, False on failure.'''

    kw = {'stderr': subprocess.PIPE}
    email = 'uci-bot <canonical-ci-engineering@lists.launchpad.net>'
    try:
        subprocess.check_output(['bzr', 'init'], **kw)
        subprocess.check_output(['bzr', 'whoami', email], **kw)
        subprocess.check_output(['bzr', 'add', '.'], **kw)
        subprocess.check_output(['bzr', 'commit', '-m', '""'], **kw)
    except subprocess.CalledProcessError as e:
        # Only print the output of the command if it failed.
        logging.error('Calling `%s` failed:' % e.cmd)
        for line in e.output.splitlines():
            logging.error(line)
            return False
    return True


def main():
    # XXX cprov 20140903: Sets logging level to INFO for the entire
    # procedure. It should never be lowered to DEBUG otherwise credentials
    # leak via python-keystone requests.
    logging.root.setLevel(logging.INFO)

    if hasattr(sys, 'real_prefix') and os.environ.get('REEXECED_UNDER_VENV'):
        # Inside the forked process. Clean up on exit.
        atexit.register(venv.delete)
        if not fake_out_branch():
            return 1

    if not check_lxc_mounted():
        logging.error('/var/lib/lxc not mounted.')
        return 1

    # TODO ev 2014-07-04 Delete all the swift containers created during
    # testing.

    # Run both the deployment and tests under a virtualenv.
    venv.install()

    from testing import run_tests
    # deploy.py isn't in our pythonpath when run as a script
    sys.path.append(os.path.abspath('juju-deployer'))
    import deploy

    try:
        with open('/dev/null', 'w') as null:
            subprocess.check_call(['juju', 'status'], stdout=null, stderr=null)
    except subprocess.CalledProcessError:
        # The environment doesn't exist, so `juju status` fails.
        pass
    else:
        # The environment wasn't cleaned up from the last run, but that's to be
        # expected. It's left around for debugging between runs.
        if not destroy_environment():
            return 1

    logging.info('Deploying.')
    t = time.time()
    # Bootstrap, then deploy.
    if not bootstrap() or deploy.main() != 0:
        return 1

    logging.info('Done in %.2fs' % (time.time() - t))
    return run_tests.main(sys.argv[1:], sys.stdout, sys.stderr)


if __name__ == '__main__':
    sys.exit(main())
