#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Branch Source Builder check script."""
import atexit
import logging
import optparse
import os
import shutil
import subprocess
import sys
import tempfile
import time


from bsbuilder.upload_package import (
    get_signing_keyring,
    MissingSigningKeyring,
)
from ci_utils import (
    amqp_utils,
    gpg,
    health_check,
    unit_config,
)
from ci_utils.testing import (
    get_test_file_path,
)


DESCRIPTION = 'In-Production BSB service check.'
VERSION = '1.0'

TESTING_SOURCE_VERSION = '0.1'
TESTING_PARAMS_TEMPLATE = {
    'subtickets': [{
        'id': 1,
        'resource_uri': 'IGNORE-ME',
        'source_package_upload': {
            'id': 1,
            'version': 'CHANGE-ME',
            'sourcepackage': {'name': 'citesting'}
        },
        'artifact': [],
    }],
    'series': 'CHANGE_ME',
    'ppa': 'CHANGE-ME',
    'archive': 'CHANGE-ME',
    'progress_trigger': 'bsb-test',
    'ticket_id': 'bsb-testing'
}


class TestDataStore(object):
    """Mimics datastore API from a temporary directory.

    A temporary directory is created on initialization and removed via
    a atexit hook.

    Implements only `get_file` and `put_file` currently.
    """
    def __init__(self):
        self.container_path = tempfile.mkdtemp()

        def cleanup(path):
            if os.path.exists(path):
                shutil.rmtree(path)
        atexit.register(cleanup, self.container_path)

    def get_file(self, filename):
        file_path = os.path.join(self.container_path, filename)
        with open(file_path) as fd:
            return fd.read()

    def put_file(self, filename, contents, content_type=None):
        file_path = os.path.join(self.container_path, filename)
        with open(file_path, 'w') as fd:
            fd.write(contents)
        return file_path


def prepare_source_upload(ds, keyid, version=None):
    """Creates a new source upload (citesting) for testing.

    Requires a functional 'devscripts' (dch and debuild) installation.

    Creates a source package upload inside the fake DataStore (on disk)
    with an optionally given version (or use the current timestamp).

    Returns the adjusted parameters to run `BSBuilderWorker.handle_request`.
    """
    # Create a copy of the testing parameters template for mangling.
    testing_params = TESTING_PARAMS_TEMPLATE.copy()
    testing_params['series'] = unit_config.get('series')
    subticket = testing_params['subtickets'][0]
    if version is None:
        version = str(int(time.time()))
    subticket['source_package_upload']['version'] = version

    # Copy and extract the testing tarball inside the datastore directory.
    source_name = subticket['source_package_upload']['sourcepackage']['name']
    tarball_filename = '{}_{}.tar.gz'.format(
        source_name, TESTING_SOURCE_VERSION)
    shutil.copy(
        get_test_file_path(tarball_filename), ds.container_path)
    subprocess.call(
        'tar zxf {}'.format(tarball_filename), shell=True,
        cwd=ds.container_path)

    # Update the package changelog with a new entry for the target suite.
    tarball_dir = '{}-{}'.format(source_name, TESTING_SOURCE_VERSION)
    source_path = os.path.join(ds.container_path, tarball_dir)
    subprocess.call(
        'DEBEMAIL="CI Arline <canonical-ci-engineering@list.launchpad.net>" '
        'dch -v {} --distribution {} --force-distribution '
        '\'Testing\''.format(
            version, testing_params['series']),
        shell=True, cwd=source_path)

    # Build the source package using `debuild`.
    source_path = os.path.join(
        ds.container_path, '{}-{}'.format(source_name, version))
    subprocess.call(
        'debuild -S -k0x{}'.format(keyid), shell=True, cwd=source_path)

    # Update the BSB request parameters with the new source information.
    artifacts = subticket['artifact']
    artifacts.append(
        {'name': '{}_{}_source.changes'.format(source_name, version)})
    artifacts.append({'name': '{}_{}.dsc'.format(source_name, version)})
    artifacts.append({'name': '{}_{}.tar.gz'.format(source_name, version)})

    return testing_params


def check(params, ds, log):
    """Exercises `BSBuilderWorker.handle_request`."""
    # Import BSBuilderWorker/amqp_worker because it messes with logging
    # configuration/format.
    from bsbuilder.run_worker import BSBuilderWorker
    worker = BSBuilderWorker()
    # Patch the worker data_store factory to return the testing one.
    worker._create_data_store = lambda ticket_id: ds

    try:
        # run `worker.handle_request` in an isolated GPG environment
        # so it has a chance to auto-retrieve missing keys from the
        # keyserver.
        with gpg.GnupgEnvironment() as gnupg:
            (cb, result) = worker.handle_request(params, log)
    except Exception as exc:
        message = 'BSB raised: {}'.format(exc.message)
        log.exception(message)
        return 1, message

    if cb is not amqp_utils.progress_completed:
        message = 'BSB failed: {}\n{}'.format(cb, result)
        log.error(message)
        return 1, message

    if result['message'] != 'The package build was successful.':
        message = 'Unexpected result: {}'.format(result)
        log.error(message)
        return 1, message

    artifacts = result['artifacts']
    if len(artifacts) != 1 or artifacts[0]['type'] != 'LOGS':
        message = 'Unexpected artifacts: {}'.format(result)
        log.error(message)
        return 1, message

    return 0, 'SUCCESS'


def main():
    parser = optparse.OptionParser(description=DESCRIPTION, version=VERSION)
    parser.add_option("-l", "--log-level",
                      action="store", dest="log_level", default='ERROR',
                      help="Logging level name.")
    (options, args) = parser.parse_args()

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(levelname)s %(message)s')
    log = logging.getLogger(__name__)
    log.info('{} - {}'.format(DESCRIPTION, VERSION))

    log_level = logging.getLevelName(options.log_level)
    try:
        log.setLevel(log_level)
    except ValueError as exc:
        logging.warn('Using log level INFO ({})'.format(exc.message))

    # Fail as soon as possible if the 'unit_config' is not reachable.
    try:
        unit_config.get('launchpad_user')
    except IOError:
        log.error('No unit_config found!')
        sys.exit(1)

    # Use the instance secret keyring (for production runs) or fallback
    # to the user's one (local testing).
    try:
        keyring_data, _ignored = get_signing_keyring()
    except MissingSigningKeyring:
        with open(os.path.expanduser('~/.gnupg/secring.gpg')) as fd:
            keyring_data = fd.read()
    except Exception as exc:
        log.error(str(exc))
        sys.exit(1)

    # Use a special DataStore and an isolated GnupgEnvironment for testing.
    testing_ds = TestDataStore()
    with gpg.GnupgEnvironment() as gnupg:
        gnupg.importKeyring(keyring_data)
        key = gnupg.listKeys(secret=True)[0]
        testing_params = prepare_source_upload(testing_ds, key['keyid'])

    # Adjust building and target destination according to the unit_config.
    testing_params['ppa'] = 'ppa:{}/ci-pool-001'.format(
        unit_config.get('launchpad_ppa_owner'))
    testing_params['archive'] = unit_config.get('master_ppa')

    with health_check.HealthCheck('bsbuilder') as hcheck:
        return_code, message = check(testing_params, testing_ds, log)
        if not return_code:
            hcheck.write_result_ok()
        else:
            hcheck.write_result_failed(message)


if __name__ == '__main__':
    main()
