# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import mock
import os
import shutil
import tempfile

from bsbuilder.tests import BsbTestBase
from ci_utils.testing import (
    TESTING_PRIVATE_KEYRING_DATA
)


class TestUploadPackage(BsbTestBase):
    '''Tests for the upload_package module.'''

    sample_subticket_list = [{
        'status': 'New',
        'current_workflow_step': 'New',
        'artifact': [{
            'id': 6,
            'type': 'SPU',
            'name': 'package_0.1_source.changes',
            'reference': 'https://swift.url/package_0.1_source.changes',
            'resource_uri': '/api/v1/fullsubticketartifact/6/'
        }, {
            'id': 5,
            'type': 'SPU',
            'name': 'package_0.1.tar.gz',
            'reference': 'https://swift.url/package_0.1.tar.gz',
            'resource_uri': '/api/v1/fullsubticketartifact/5/'
        }, {
            'id': 4,
            'type': 'SPU',
            'name': 'package_0.1.dsc',
            'reference': 'https://swift.url/package_0.1.dsc',
            'resource_uri': '/api/v1/fullsubticketartifact/4/'
        }],
        'assignee': 'user_id',
        'sourcepackage': {
            'id': 1,
            'name': 'package',
            'resource_uri': '/api/v1/sourcepackage/1/'
        },
        'source_package_uploads': [{
            'version': '0.1',
            'id': 2,
            'resource_uri': '/api/v1/spu/2/'
        }],
        'id': 2,
        'resource_uri': '/api/v1/fullsubticket/2/'
    }]

    def setUp(self):
        """Import 'upload_package' locally."""
        super(TestUploadPackage, self).setUp()
        # XXX cprov 20140911: it's necessary because loading
        # cupstream2distro requires unit_config (which is only mocked in
        # BsbTestBase) and the problem only manifest itself in the
        # jenkins-bot job setup.
        from bsbuilder import upload_package
        self.upload_package = upload_package

    @mock.patch('bsbuilder.upload_package.create_temp_dir')
    def test_get_source_files(self, temp_dir_mock):
        # 'get_source_files' downloads uploaded source files
        # and returns results in the expected format.
        temp_dir_mock.return_value = self.tmpdir
        mocked_ds = mock.Mock()
        mocked_ds.get_file.return_value = 'foo'

        results = self.upload_package.get_source_files(
            self.sample_subticket_list, mocked_ds)
        [result] = results
        # Source upload files were downloaded from swift.
        self.assertEqual(
            [mock.call('package_0.1_source.changes'),
             mock.call('package_0.1.tar.gz'),
             mock.call('package_0.1.dsc')],
            mocked_ds.get_file.call_args_list)
        # Downloads were stored in the location returned by
        # 'create_temp_dir'.
        self.assertEqual(self.tmpdir, result['location'])
        # Results points back to the original (given) 'fullsubticket'
        # resource_uri.
        self.assertEqual(
            self.sample_subticket_list[0]['resource_uri'],
            result['resource'])
        # Results point to the given source name and version.
        self.assertEqual(
            result['name'],
            self.sample_subticket_list[0]['sourcepackage']['name'])
        spu = self.sample_subticket_list[0]['source_package_uploads'][0]
        self.assertEqual(spu['version'], result['version'])
        # Result 'files' correspond to the given source package upload
        # files.
        artifacts = self.sample_subticket_list[0]['artifact']
        self.assertItemsEqual(
            [a['name'] for a in artifacts], result['files'])

    def mock_service_dir(self):
        """Mock 'ci_utils.get_service_dir' for the testing context.

        Since 'upload_package' module is locally imported, it requires
        a *special* dance to mock 'get_service_dir' only in that context.
        """
        _p = mock.patch.object(self.upload_package, 'get_service_dir')
        self.addCleanup(_p.stop)
        service_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, service_dir)
        mocked_sd = _p.start()
        mocked_sd.return_value = service_dir
        return service_dir

    def test_get_signing_keyring(self):
        # 'get_signing_keyring' returns a valid signing keyring and the
        # python-gnupg key information (dict) when it's available in the
        # instance environment.
        service_dir = self.mock_service_dir()
        # Emulates the 'key-secret-subordinate' keys/ with the testing
        # secret keyring.
        keys_dir = os.path.join(service_dir, 'keys')
        os.mkdir(keys_dir)
        with open(os.path.join(keys_dir, 'gpg.sec'), 'w') as fd:
            fd.write(TESTING_PRIVATE_KEYRING_DATA)

        keyring_data, key_data = self.upload_package.get_signing_keyring()

        self.assertEqual(
            TESTING_PRIVATE_KEYRING_DATA, keyring_data)
        self.assertEqual(
            '52DC2178A5ECADD2F213DD84D9A2A8312ABEFCDB',
            key_data['fingerprint'])

    def test_get_signing_keyring_unavailable(self):
        # 'get_signing_keyring' raises 'MissingSigningKeyring' when
        # the keyring is not available.
        self.mock_service_dir()

        exc_class = self.upload_package.MissingSigningKeyring
        with self.assertRaises(exc_class) as error:
            self.upload_package.get_signing_keyring()

        self.assertEqual(
            'Secret keyring not available.',
            str(error.exception))

    def test_get_signing_keyring_invalid(self):
        # 'get_signing_keyring' raises 'MissingSigningKeyring' when
        # the keyring does not contain a single secret key.
        service_dir = self.mock_service_dir()
        # Creates an invalid secret keyring.
        keys_dir = os.path.join(service_dir, 'keys')
        os.mkdir(keys_dir)
        with open(os.path.join(keys_dir, 'gpg.sec'), 'w') as fd:
            fd.write('Zoing!')

        exc_class = self.upload_package.MissingSigningKeyring
        with self.assertRaises(exc_class) as error:
            self.upload_package.get_signing_keyring()

        self.assertEqual(
            'Secret keyring does not contain a valid signing key.',
            str(error.exception))

    def test_resign_upload(self):
        # `resigned_upload` re-signs a source uploads (changes and DSC
        # files) in a given directory with the given signing key.

        # Copy a foobar upload to the testing dir with sample unsigned
        # changes and DSC files.
        files = [
            'foobar_0.1-1_unsigned.changes',
            'foobar_0.1-1_dsc.changes',
            'foobar_0.1.orig.tar.gz',
            'foobar_0.1-1.debian.tar.gz'
        ]
        self.copy_test_files(files)
        self.rename_test_files(
            'foobar_0.1-1_unsigned.changes', 'foobar_0.1-1_source.changes')
        self.rename_test_files(
            'foobar_0.1-1_dsc.changes', 'foobar_0.1-1.dsc')

        self.upload_package.resign_upload(
            self.tmpdir, TESTING_PRIVATE_KEYRING_DATA)

        self.assertSigned('foobar_0.1-1_source.changes')
        self.assertSigned('foobar_0.1-1.dsc')

    def test_resign_upload_no_sources(self):
        # `resigned_upload` raises `UploadResigningError` if there is
        # no source upload in the given directory.
        exc_class = self.upload_package.UploadResigningError
        with self.assertRaises(exc_class) as error:
            self.upload_package.resign_upload(
                self.tmpdir, TESTING_PRIVATE_KEYRING_DATA)
        self.assertEqual(
            'Invalid upload directory: no sources uploads found.',
            str(error.exception))

    def test_resign_upload_multiple_sources(self):
        # `resigned_upload` raises `UploadResigningError` if there are
        # more than one source upload in the given directory.
        files = [
            'foobar_0.1-1_source.changes',
            'barfoo_0.1-1_source.changes',
        ]
        self.copy_test_files(files)

        exc_class = self.upload_package.UploadResigningError
        with self.assertRaises(exc_class) as error:
            self.upload_package.resign_upload(
                self.tmpdir, TESTING_PRIVATE_KEYRING_DATA)
        self.assertEqual(
            'Invalid upload directory: expected single source upload and '
            'found 2 (barfoo_0.1-1_source.changes, '
            'foobar_0.1-1_source.changes).',
            str(error.exception))

    def test_resign_upload_invalid_or_missing_key(self):
        # `resigned_upload` raises `MissingSigningKeyring` if the given
        # signing keyring is not valid (single secret key signing-capable).
        exc_class = self.upload_package.MissingSigningKeyring
        with self.assertRaises(exc_class) as error:
            self.upload_package.resign_upload(self.tmpdir, 'zoing')
        self.assertEqual(
            'Secret keyring does not contain a valid signing key.',
            str(error.exception))

    @mock.patch('subprocess.check_output')
    def test_resign_upload_missing_devscripts(self, mocked_check_output):
        # `resigned_upload` raises `UploadResigningError` if 'devscripts'
        # (`debsign`) is not installed.
        files = [
            'foobar_0.1-1_source.changes',
            'foobar_0.1-1.dsc',
            'foobar_0.1.orig.tar.gz',
            'foobar_0.1-1.debian.tar.gz'
        ]
        self.copy_test_files(files)

        mocked_check_output.side_effect = [
            OSError('Not found!')
        ]

        exc_class = self.upload_package.UploadResigningError
        with self.assertRaises(exc_class) as error:
            self.upload_package.resign_upload(
                self.tmpdir, TESTING_PRIVATE_KEYRING_DATA)
        self.assertEqual(
            'Failed to resign upload: `devscripts` is not installed.',
            str(error.exception))
