# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import mock

from bsbuilder.tests import BsbTestBase
from ci_utils.testing import (
    TESTING_PRIVATE_KEYRING_DATA,
)


class TestBsbWorker(BsbTestBase):

    fake_request_params = {
        'ticket_id': 1,
        'subtickets': [],
        'series': 'a_series',
        'ppa': 'a_ppa',
        'archive': 'an_archive',
        'progress_trigger': 'A_TRIGGER'
    }

    fake_subticket_status = [{
        'name': 'the_source',
        'version': '1.0',
        'status_text': 'A_STATUS',
        'message': 'SUB_MESSAGE',
    }]

    def setUp(self):
        super(TestBsbWorker, self).setUp()
        # XXX cprov 20140911: it's necessary because loading
        # cupstream2distro requires unit_config (which is only mocked in
        # BsbTestBase) and the problem only manifest itself in the
        # jenkins-bot job setup.
        from bsbuilder import run_worker
        self.worker = run_worker.BSBuilderWorker()

        # Patch common infrastruture method/function.
        common_patches = (
            'ci_utils.amqp_worker.AMQPWorker._create_data_store',
            'bsbuilder.run_worker.watch_ppa.check_ppa',
            'bsbuilder.run_worker.watch_ppa.watch_ppa',
        )
        for spec in common_patches:
            attr_name = spec.split('.')[-1]
            _m = mock.patch(spec)
            setattr(self, attr_name, _m.start())
            self.addCleanup(_m.stop)

        # Adjust common mocking.
        self.log = mock.MagicMock()
        self.ds = mock.MagicMock()
        self.ds.put_file.return_value = 'swift_ref'
        self._create_data_store.return_value = self.ds

    @mock.patch('ci_utils.amqp_utils.progress_completed')
    def test_handle_request(self, mocked_progress_completed):
        # `BSBuilderWorker.handle_request` uses watch_ppa helpers
        # to build all sources related to a ticket.

        # Creating successful watch_ppa mocks.
        self.check_ppa.return_value = (False, self.fake_subticket_status)
        self.watch_ppa.return_value = (0, self.fake_subticket_status)

        # Process a testing BSB request.
        cb, result = self.worker.handle_request(
            self.fake_request_params, self.log)
        # 'progress_complete' is returned at the end of the processing
        # with context information to the lander and the log messages
        # stored in Swift.
        self.assertEqual(mocked_progress_completed, cb)
        self.assertEquals(
            {'artifacts': [{'type': 'LOGS',
                            'name': 'package_build.output.log',
                            'reference': 'swift_ref'}],
             'message': 'The package build was successful.',
             'subticket_status': [{'message': 'SUB_MESSAGE',
                                   'version': '1.0',
                                   'name': 'the_source',
                                   'status_text': 'A_STATUS'}],
             'ppa': 'a_ppa',
             'archive': 'an_archive'},
            result)
        # A ticket data_store acessor is created.
        self._create_data_store.assert_called_once_with(
            self.fake_request_params['ticket_id'])

        # log.warning() is called pointing to the missing signing keyring.
        self.assertEqual(
            [mock.call('Secret keyring not available.')],
            self.log.warning.call_args_list)

        # log.info() is called for announcing the package was
        # successfully built.
        self.log.info.assert_called_with(
            'The package build was successful.')

    @mock.patch('ci_utils.amqp_utils.progress_failed')
    def test_handle_request_check_failure(self, mocked_progress_failed):
        # `BSBuilderWorker.handle_request` failures store enough
        # tracing information.

        # Creating successful watch_ppa mocks.
        self.check_ppa.return_value = (True, self.fake_subticket_status)

        # Process a testing BSB request.
        cb, result = self.worker.handle_request(
            self.fake_request_params, self.log)

        # 'progress_failed' is called informing the lander of the issue.
        self.assertEqual(mocked_progress_failed, cb)
        self.assertEqual(
            {'artifacts': [{'type': 'LOGS',
                            'name': 'package_build.output.log',
                            'reference': 'swift_ref'}],
             'message': 'Source package upload failed.',
             'subticket_status': [{'message': 'SUB_MESSAGE',
                                   'version': '1.0',
                                   'name': 'the_source',
                                   'status_text': 'A_STATUS'}],
             'ppa': 'a_ppa',
             'archive': 'an_archive'},
            result)

        # log.error() is called once stating the problem.
        self.log.error.assert_called_once_with(
            'Source package upload failed.')

    @mock.patch('bsbuilder.upload_package.upload_source_packages')
    @mock.patch('bsbuilder.upload_package.create_upload_list')
    @mock.patch('bsbuilder.upload_package.get_signing_keyring')
    @mock.patch('ci_utils.amqp_utils.progress_completed')
    def test_handle_request_resigning_uploads(
            self, mocked_progress_completed, mocked_keyring,
            mocked_upload_list, mocked_upload):
        # `BSBuilderWorker.handle_request` re-signs uploads if a valid
        # signing keyring is available.

        # Copy testing upload files to the upload_dir and rename
        # unsigned testing changes and DSC to their expected names.
        test_files = [
            'foobar_0.1-1_unsigned.changes',
            'foobar_0.1-1_dsc.changes',
            'foobar_0.1.orig.tar.gz',
            'foobar_0.1-1.debian.tar.gz'
        ]
        self.copy_test_files(test_files)
        self.rename_test_files(
            'foobar_0.1-1_unsigned.changes', 'foobar_0.1-1_source.changes')
        self.rename_test_files(
            'foobar_0.1-1_dsc.changes', 'foobar_0.1-1.dsc')

        # Use a fake upload_list corresponding to the copied testing files.
        mocked_upload_list.return_value = [{
            'name': 'foobar',
            'version': '1.0-1',
            'location': self.tmpdir,
            'files': [
                'foobar_0.1-1_source.changes',
                'foobar_0.1-1.dsc',
                'foobar_0.1.orig.tar.gz',
                'foobar_0.1-1.debian.tar.gz'
            ],
        }]

        # Creating successful watch_ppa mocks.
        self.check_ppa.return_value = (False, self.fake_subticket_status)
        self.watch_ppa.return_value = (0, self.fake_subticket_status)

        # Use the testing keyring for signing.
        mocked_keyring.return_value = (
            TESTING_PRIVATE_KEYRING_DATA,
            {'uids': ['A_NAME'], 'fingerprint': 'A_FPR'},
        )

        cb, result = self.worker.handle_request(
            self.fake_request_params, self.log)

        # BSB succeeded and uploaded (re-)signed files.
        self.assertEqual(mocked_progress_completed, cb)
        self.assertSigned('foobar_0.1-1_source.changes')
        self.assertSigned('foobar_0.1-1.dsc')

    @mock.patch('bsbuilder.upload_package.upload_source_packages')
    @mock.patch('bsbuilder.upload_package.create_upload_list')
    @mock.patch('bsbuilder.upload_package.get_signing_keyring')
    @mock.patch('bsbuilder.upload_package.resign_upload')
    @mock.patch('ci_utils.amqp_utils.progress_failed')
    def test_handle_request_resigning_fails_gracefully(
            self, mocked_progress_failed, mocked_keyring, mocked_resign,
            mocked_upload_list, mocked_upload):
        # `BSBuilderWorker.handle_request` fails gracefully if upload
        # re-signing cannot be executed.

        # Use a fake upload_list corresponding to the copied testing files.
        mocked_upload_list.return_value = [{
            'name': 'foobar',
            'version': '1.0-1',
            'location': self.tmpdir,
            'files': [
                'foobar_0.1-1_source.changes',
                'foobar_0.1-1.dsc',
                'foobar_0.1.orig.tar.gz',
                'foobar_0.1-1.debian.tar.gz'
            ],
        }]

        # Creating successful watch_ppa mocks.
        self.check_ppa.return_value = (False, self.fake_subticket_status)
        self.watch_ppa.return_value = (0, self.fake_subticket_status)

        # Use the testing keyring for signing.
        mocked_keyring.return_value = (
            TESTING_PRIVATE_KEYRING_DATA,
            {'uids': ['A_NAME'], 'fingerprint': 'A_FPR'},
        )

        # Fail when re-signing uploads.
        from bsbuilder import upload_package
        mocked_resign.side_effect = [
            upload_package.UploadResigningError('I am dead, sorry!')
        ]

        cb, result = self.worker.handle_request(
            self.fake_request_params, self.log)

        # BSB fails gracefully, returning tracing information (dict) and
        # complete logging stored in swift.
        self.assertEqual(mocked_progress_failed, cb)
        self.assertEqual(
            {'archive': 'an_archive',
             'artifacts': [
                 {'name': 'package_build.output.log',
                  'reference': 'swift_ref',
                  'type': 'LOGS'}
             ],
             'message': 'I am dead, sorry!',
             'ppa': 'a_ppa',
             'subticket_status': [
                 {'message': 'SUB_MESSAGE',
                  'name': 'the_source',
                  'status_text': 'A_STATUS',
                  'version': '1.0'}
             ]}, result)
