#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import atexit
from dput.changes import parse_changes_file
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile

from ci_utils import (
    get_service_dir,
    gpg,
)
from cupstream2distro import packagemanager


def parse_arguments():
    parser = argparse.ArgumentParser(
        description='Source package upload handler.')
    parser.add_argument('-p', '--ppa',
                        required=True,
                        help='Target PPA for source package upload.')
    parser.add_argument('-s', '--source',
                        required=True,
                        action='append',
                        help='A source package file to include in the upload.')
    return parser.parse_args()


def cleanup(directory):
    if os.path.exists(directory) and os.path.isdir(directory):
        logging.info('Removing temp directory: {}'.format(directory))
        shutil.rmtree(directory)


def create_temp_dir():
    directory = tempfile.mkdtemp()
    atexit.register(cleanup, directory)
    return directory


def get_source_files(subtickets, datastore):
    '''Retrieves the source package files and creates initial upload_list.'''
    upload_list = []
    for subticket in subtickets:
        package = {}
        # XXX cprov 20141017: operating only on the latest SPU until we can
        # sort the relationship between SPU and its specific artifacts.
        spu = subticket['source_package_uploads'][-1]
        package['id'] = spu['id']
        package['name'] = subticket['sourcepackage']['name']
        package['version'] = spu['version']
        package['resource'] = subticket['resource_uri']
        package['location'] = create_temp_dir()
        package['files'] = []
        for artifact in subticket['artifact']:
            file_name = artifact['name']
            logging.info('Retrieving source file: {}'.format(file_name))
            content = datastore.get_file(file_name)
            file_path = os.path.join(package['location'], file_name)
            with open(file_path, 'w') as out_file:
                out_file.write(content)
            package['files'].append(file_name)
        upload_list.append(package)
    return upload_list


def parse_dsc_file(dsc_file):
    regexp = re.compile("^Architecture: (.*)\n")
    with open(dsc_file) as in_file:
        for line in in_file:
            arch_lists = regexp.findall(line)
            if arch_lists:
                return arch_lists[0]


def parse_source_files(upload_list):
    '''Perform a second pass of the upload list and parse out more info.

    The purpose of this is to make sure all source package files have
    been downloaded and identify the source package architecture.
    This must be done after all files in a subticket have been acquired.'''
    for package in upload_list:
        for source in package['files']:
            source_name = os.path.join(package['location'], source)
            if source.endswith('changes'):
                changes = parse_changes_file(filename=source_name,
                                             directory=package['location'])
                for package_file in changes.get_files():
                    if os.path.exists(package_file):
                        if package_file.endswith('.dsc'):
                            package['architecture'] = parse_dsc_file(
                                package_file)
                    else:
                        raise EnvironmentError(
                            'Not found: {}'.format(package_file))


def create_upload_list(subtickets, datastore):
    '''Parses the subtickets into the upload list.'''
    upload_list = get_source_files(subtickets, datastore)
    parse_source_files(upload_list)
    return upload_list


def upload_source_packages(ppa, upload_list):
    '''Attempts source file upload into the PPA.'''
    logging.info('Upload to the ppa: {}'.format(ppa))
    for package in upload_list:
        packagemanager.upload_package(package['name'], package['version'],
                                      ppa, package['location'])

        directory = package['location']
        if os.path.exists(directory) and os.path.isdir(directory):
            logging.info('Removing temp directory: {}'.format(directory))
            shutil.rmtree(directory)


class MissingSigningKeyring(Exception):
    """Raised if no valid secret signing key are available."""


class UploadResigningError(Exception):
    """Raised if there was an error while re-signing uploads."""


def _check_signing_keyring(gnupg_env):
    """Returns the valid single secret key data (python-gnupg dict).

    Raises `MissingSigningKeyring` if the keyring does not contain any
    secret key or `UploadResigningError` if it contains multiple secret
    keys.
    """
    secret_keys = gnupg_env.listKeys(secret=True)

    if not secret_keys:
        raise MissingSigningKeyring(
            'Secret keyring does not contain a valid signing key.')

    if len(secret_keys) > 1:
        raise UploadResigningError(
            'Invalid secret keyring: multiple ({}) secret keys '
            'detected.'.format(len(secret_keys)))

    return secret_keys[0]


def get_signing_keyring():
    """Return the instance secret keyring data, and secret key information.

    It only works on instances using the 'key-secret-subordinate' charm
    properly configured, which lays a valid secret GNUPG keyring at:
    <service_dir>/keys/gpg.sec

    Checks if the existing keyring is valid and raises an appropriate error
    if it isn't (see `_check_signing_keyring`), otherwise returns the valid
    keyring content and key information.
    """
    keyring_path = os.path.join(get_service_dir(), 'keys', 'gpg.sec')

    # No secret keyring available, the charm is not available.
    if not os.path.exists(keyring_path):
        raise MissingSigningKeyring('Secret keyring not available.')

    # Import the given signing keyring and check it's valid.
    keyring_data = open(keyring_path).read()
    with gpg.GnupgEnvironment() as gnupg_env:
        gnupg_env.importKeyring(keyring_data)
        key_data = _check_signing_keyring(gnupg_env)

    return keyring_data, key_data


def resign_upload(upload_dir, signing_keyring):
    """Re-sign a single upload found in the given path (using `debsign`).

    Raises `MissingSigningKeyring` if the given signing_keyring does not
    contain signing key. This should be treated in callsites as critical
    if it's a production environment and possibly as a warning on
    development setups.

    Raises `UploadResigningError` if multiple uploads were found in the
    given 'upload_dir' or `debsign` failed to re-sign the upload. In any of
    these cases, callsites should discard the upload files for safety
    (since some might be resigned and some not) and possibly retry the job
    in another instance.
    """
    with gpg.GnupgEnvironment() as gnupg_env:
        # Import the given signing keyring and check it's valid (again,
        # because it might have changed since `get_signing_keyring`).
        gnupg_env.importKeyring(signing_keyring)
        key_data = _check_signing_keyring(gnupg_env)
        # Find source uploads and re-sign them with `debsign`.
        sources_changes = [
            fname for fname in sorted(os.listdir(upload_dir))
            if fname.endswith('_source.changes')]
        if not sources_changes:
            raise UploadResigningError(
                'Invalid upload directory: no sources uploads found.')
        if len(sources_changes) > 1:
            raise UploadResigningError(
                'Invalid upload directory: expected single source upload '
                'and found {} ({}).'.format(
                    len(sources_changes), ', '.join(sources_changes)))
        [source_changes] = sources_changes
        try:
            subprocess.check_output(
                ['debsign', '-S', '--re-sign',
                 '-k0x{}'.format(key_data['keyid']), source_changes],
                cwd=upload_dir)
        except OSError as exc:
            raise UploadResigningError(
                'Failed to resign upload: `devscripts` is not installed.')
        except subprocess.CalledProcessError as exc:
            raise UploadResigningError(
                'Failed to resign upload due to:\n{}({})\n{}'.format(
                    exc.cmd, exc.returncode, exc.output))


def main():
    logging.basicConfig(level=logging.INFO,
                        format="%(asctime)s %(levelname)s %(message)s")
    args = parse_arguments()
    upload_source_packages(args.ppa, args.source)


if __name__ == '__main__':
    sys.exit(main())
