# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""The proxy to post a britney request from outside the uci-engine."""


import argparse
import sys


from britney import queues


class PostRequestArgParser(argparse.ArgumentParser):

    def __init__(self, prog=None, description=None):
        if prog is None:
            prog = 'uci-britney-request'
        if description is None:
            description = 'Request a test run for britney.'
        super(PostRequestArgParser, self).__init__(prog=prog,
                                                   description=description)
        # FIXME: Most of the time this will be the current development series
        # and it would be nice to get the value from python-distro-info (as
        # suggested by pitti) -- vila 2014-10-15.
        self.add_argument(
            'series', metavar='SERIES',
            help='The SERIES for which the test should be run.')
        self.add_argument(
            'architecture', metavar='ARCHITECTURE',
            help='The ARCHITECTURE for which the test should be run.')
        self.add_argument(
            'package', metavar='PACKAGE',
            help='The PACKAGE for which the test should be run.')
        # Optional arguments
        self.add_argument(
            '--output', '-O', metavar='OUTPUT_QUEUE',
            help='Send progress/results from the test runner to the queue'
            ' named OUTPUT_QUEUE.')
        self.add_argument(
            '--report', '-R', metavar='REPORT_QUEUE',
            help='Send result report message to the queue named REPORT_QUEUE.')

    def parse_args(self, args=None, out=None, err=None):
        """Parse arguments, overridding stdout/stderr if provided.

        Overridding stdout/stderr is provided for tests to address arpgarse
        hardcoding sys.{stdout,stderr} usage.

        :params args: The arguments to the script.

        :param out: Default to sys.stdout.

        :param err: Default to sys.stderr.

        :return: The populated namespace.

        """
        out_orig = sys.stdout
        err_orig = sys.stderr
        try:
            if out is not None:
                sys.stdout = out
            if err is not None:
                sys.stderr = err
            return super(PostRequestArgParser, self).parse_args(args)
        finally:
            sys.stdout = out_orig
            sys.stderr = err_orig


def post(queue, series, architecture, package, output=None, report=None):
    msg = dict(series=series, architecture=architecture, package=package)
    if output is not None:
        msg['output_queue'] = output
    if report is not None:
        msg['report_queue'] = report
    # FIXME: Error checking to guarantee that we at least give a proper error
    # message. Known failures: bad host (can't connect), bad credentials (can't
    # connect). Possible failures: rabbit down. -- vila 2014-10-24
    queue.publish(msg)


def cli_run(args, stdout, stderr):
    """Post a britney test request from the command line."""
    parser = PostRequestArgParser()
    opts = parser.parse_args(args, out=stdout, err=stderr)
    queue = queues.PublisherQueue('britney.requests')
    post(queue, opts.series, opts.architecture, opts.package,
         output=opts.output, report=opts.report)
    return 0
