#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals


import argparse
import errno
import os
import json
import logging
import sys


from ci_utils import amqp_utils


class BritneyReportHandler(object):
    """Process a britney test report.

    This is fed by messages received on a queue or directly for tests by
    calling the ``handle()`` method.
    """

    def __init__(self, output_dir):
        self.output_dir = output_dir
        try:
            # Try to create the needed dirs
            os.makedirs(self.output_dir)
        except OSError as e:
            # They are already there, no worries
            if e.errno != errno.EEXIST:
                raise

    def handle(self, request):
        """Handle a report request.

        :param request: The request describing a britney result to process.

        This creates a file in the 'output_dir' directory under a name and with
        the content expected by britney.

        The name is something like:
        {series}_{arch}_{package}-{DATE}_{HOUR}[_{whatever}].result

        The content is (all fields on a single line separated by spaces):
        {series} {arch} {package} {version} {PASS|FAILED} [{dep} {dep_version]*

        If an error occured while setting up the testbed (by adt-run) we get
        only:
         {series} {arch} {package}
        """

        # From 'result_at' we get info for the path
        store_prefix = request['result_at']
        # The prefix has the timestamp
        # Sample: trusty/amd64/libp/libpng/20141001_000000_localhost
        stamp = store_prefix.split('/')[4]
        day, hour = stamp.split('_')[:2]

        # From 'result' we get the series, arch and package parts
        result = request['britney_result']
        # The first 3 fields are guaranteed to exist.
        series, arch, package = result.split()[:3]
        logging.info('{} {} {}'.format(series, arch, package))

        path = os.path.join(self.output_dir, '{}_{}_{}_{}-{}.result'.format(
            series, arch, package, day, hour))
        with open(path, 'wb') as f:
            f.write(result)


class BritneyReportsQueue(object):
    """The queue receiving the reports for the britney results."""

    def __init__(self, queue_name, request_handler, max_reports=0):
        """Britney reports queue.

        :param queue_name: The queue name where the reports are received.

        :param request_handler: The handler that processes the reports.

        :param max_reports: Max number of reports to handle before
            quitting. The default value (0) is used to never quit.
        """
        self.queue_name = queue_name
        self.request_handler = request_handler
        self.max_reports = max_reports
        self.nb_reports = 0

    def consume(self):
        conf = amqp_utils.get_config()
        amqp_utils.process_queue(conf, self.queue_name, self.handle_request)

    def handle_request(self, msg):
        logging.debug('Will handle {}'.format(msg.body))
        body = json.loads(msg.body)
        ret = self.request_handler.handle(body)
        msg.channel.basic_ack(msg.delivery_tag)
        logging.debug('Handling returned: {} for {}'.format(ret, msg.body))
        self.nb_reports += 1
        if self.max_reports and self.nb_reports >= self.max_reports:
            # Leave the loop
            msg.channel.basic_cancel(msg.consumer_tag)
        return ret


class ProcessReportArgParser(argparse.ArgumentParser):

    def __init__(self, prog=None, description=None):
        if prog is None:
            prog = 'uci-britney-report'
        if description is None:
            description = 'Process a test report from britney.'
        super(ProcessReportArgParser, self).__init__(prog=prog,
                                                     description=description)
        self.add_argument(
            'report', metavar='REPORT_QUEUE',
            help='The REPORT_QUEUE publishing the test results.')
        self.add_argument(
            'output', metavar='OUTPUT',
            help='The OUTPUT directory where the results are stored.')
        # Optional arguments
        self.add_argument(
            '--nb-reports', '-n', metavar='REPORTS', default=1, type=int,
            help='The number of reports to process from the queue'
            ' named REPORT_QUEUE.')

    def parse_args(self, args=None, out=None, err=None):
        """Parse arguments, overridding stdout/stderr if provided.

        Overridding stdout/stderr is provided for tests to address arpgarse
        hardcoding sys.{stdout,stderr} usage.

        :params args: The arguments to the script.

        :param out: Default to sys.stdout.

        :param err: Default to sys.stderr.

        :return: The populated namespace.

        """
        out_orig = sys.stdout
        err_orig = sys.stderr
        try:
            if out is not None:
                sys.stdout = out
            if err is not None:
                sys.stderr = err
            return super(ProcessReportArgParser, self).parse_args(args)
        finally:
            sys.stdout = out_orig
            sys.stderr = err_orig


def cli_run(args, stdout, stderr):
    """Process britney test reports from the command line."""
    parser = ProcessReportArgParser()
    opts = parser.parse_args(args, out=stdout, err=stderr)
    request_handler = BritneyReportHandler(opts.output)
    queue = BritneyReportsQueue(opts.report, request_handler,
                                max_reports=opts.nb_reports)
    if opts.nb_reports == 0:
        sys.stderr.write('Waiting for reports. ^C to exit.')
    queue.consume()
    return 0
