#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals


import json
import logging
import sys
import uuid


from britney import queues
from ci_utils import (
    amqp_utils,
    dump_stack,
    image_store,
    unit_config,
)


class BritneyRequestHandler(object):
    """Process a britney test request.

    This is fed by messages received on a queue or directly for tests by
    calling the ``handle()`` method.
    """

    def __init__(self, test_runner_queue, output_queue):
        """Britney requests handler.

        :param test_runner_queue: The queue where the test runner requests
            should be sent.

        :param output_queue_name: The rabbit queue name where the test runner
            will send progress and results.

        """
        super(BritneyRequestHandler, self).__init__()
        self.test_runner_queue = test_runner_queue
        self.output_queue = output_queue
        self.auth_conf = unit_config.get_auth_config()

    def get_image_id(self, series, arch):
        return image_store.uci_image_name('britney', series, arch)

    def handle(self, params):
        series = params['series']
        architecture = params['architecture']
        package = params['package']
        # The output queue can be overridden for a given request (for tests and
        # debug purposes).
        output_queue = params.get('output_queue', self.output_queue)
        report_queue = params.get('report_queue', None)
        params = dict(ticket_id=str(uuid.uuid4()),
                      progress_trigger=output_queue,
                      series=series,
                      architecture=architecture,
                      image_id=self.get_image_id(series, architecture),
                      adt_opts=['--apt-upgrade', '--apt-pocket=proposed'],
                      package_list=[package])
        if report_queue is not None:
            params['report_queue'] = report_queue
        self.test_runner_queue.publish(params)


class BritneyRequestsQueue(object):
    """The queue receiving the britney test requests.

    The requests are forwarded to the request handler.
    """

    def __init__(self, queue_name, request_handler):
        """Britney test requests queue.

        :param queue_name: The queue name where the test requests are received.

        :param request_handler: The handler that processes the requests.
        """
        self.queue_name = queue_name
        self.request_handler = request_handler

    def consume(self):
        conf = amqp_utils.get_config()
        # Process all messages and delete the queue when done
        amqp_utils.process_queue(conf, self.queue_name, self.handle_request,
                                 delete=True)

    def handle_request(self, msg):
        logging.info('Will handle {}'.format(msg.body))
        body = json.loads(msg.body)
        ret = self.request_handler.handle(body)
        msg.channel.basic_ack(msg.delivery_tag)
        logging.info('Handling returned: {} for {}'.format(ret, msg.body))
        return ret


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    logging.basicConfig(level=logging.INFO,
                        format="%(asctime)s %(levelname)s %(message)s")
    logging.info('Waiting for messages. ^C to exit.')
    tr_queue = queues.PublisherQueue(amqp_utils.TEST_RUNNER_QUEUE)
    request_handler = BritneyRequestHandler(tr_queue, 'britney.results')
    # Build the britney requests queue subscribing to the rabbit queue and
    # proving the handler that will forward to the test runner queue.
    queue = BritneyRequestsQueue('britney.requests', request_handler)
    queue.consume()
    sys.exit(0)
