#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals


import datetime
import json
import logging
import os
import sys


from britney import queues
from ci_utils import (
    amqp_utils,
    data_store,
    dump_stack,
    unit_config,
)


class BritneyTestRunnerArtifactMapper(object):

    def __init__(self, release, architecture, package, time_stamp, host_name):
        if package.startswith('lib'):
            prefix = package[0:4]
        else:
            prefix = package[0]
        # Sample: trusty/amd64/libp/libpng/20141001_000000_localhost
        self.dir_name = os.path.join(release, architecture, prefix, package,
                                     '{}_{}'.format(time_stamp, host_name))

    def get_artifact_name(self, name):
        return os.path.join(self.dir_name, name)


# FIXME: We do store *all* britney test results in a single swift
# container. This won't scale but we'll revisit the layout when we get a better
# understanding of the britney needs (and numbers !) -- vila 2014-10-03
class BritneyDataStore(data_store.DataStore):

    def file_name(self, filename):
        # Allow paths to be used as object names in the container by overriding
        # the filtering in the base class.
        return filename


class BritneyResultHandler(object):
    """Process a britney test request.

    This is fed by messages received on a queue or directly for tests by
    calling the ``handle()`` method.
    """

    def __init__(self, britney_data_store, tr_ds_factory,
                 _time_stamp=None):
        """Britney results handler.

        :param britney_data_store: The data store where the test results are
            stored for britney.

        :param tr_ds_factory: A callable to create the test runner data store
            for a given request.

        :param _time_stamp: The date and time the test result is
            processed. This is intended to be used by tests and default to
            utcnow().

        """
        self.britney_data_store = britney_data_store
        self.tr_ds_factory = tr_ds_factory
        self.tr_data_store = None
        self._time_stamp = _time_stamp

    def handle_progress(self, params):
        # We just ignore the progress messages
        pass

    def copy_artifact(self, name):
        # FIXME: If swift becomes unavailable we can't acquire the test results
        # from the test runner swift container nor can we upload them to the
        # britney container. It's quite unlikely though that swift becomes
        # unavailable *after* the test runner processed the request so just
        # logging an exception should be enough. -- vila 2014-10-02
        value = self.tr_data_store.get_file(name)
        target = self.mapper.get_artifact_name(name)
        self.britney_data_store.put_file(target, value, 'text/plain')

    def handle_result(self, params):
        # Copy the produced results from the test request swift container into
        # the britney swift container.
        request = params['request']
        ticket_id = request['ticket_id']
        series = request['series']
        architecture = request['architecture']
        self.tr_data_store = self.tr_ds_factory(ticket_id)
        time_stamp = self._time_stamp
        if time_stamp is None:
            time_stamp = datetime.datetime.utcnow().strftime('%Y%m%d_%H%M%S')
        package = request['package_list'][0]
        self.mapper = BritneyTestRunnerArtifactMapper(
            series, architecture, package, time_stamp,
            # uci-engine testbed hostnames are meaningless, so we just use a
            # generic name here.
            'uci-testbed')
        self.copy_artifact('{}.log'.format(package))
        self.copy_artifact('{}.britney.results'.format(package))
        # The test runner swift container has been processed and can die
        self.tr_data_store.delete(recursive=True)
        report_queue = request.get('report_queue', None)
        if report_queue is not None:
            # Publish a message to report where the results have been stored
            queue = queues.PublisherQueue(report_queue)
            msg = dict(result_at=self.mapper.dir_name,
                       britney_result=self.tr_data_store.get_file(
                           '{}.britney.results'.format(package)))
            queue.publish(msg)

    def handle(self, params):
        state = params['state']
        if state == 'STATUS':
            self.handle_progress(params)
        elif state == 'COMPLETED':
            self.handle_result(params)
        else:
            # This indicates a programming error in the test runner producing
            # an unexpected message
            logging.error('Unable to handle {}'.format(params))
            # FIXME: properly testing the above will require a logger object
            # for the handler that can be overriden by tests -- vila 2014-09-30


class BritneyResultsQueue(object):
    """The queue receiving the test runner results for the britney requests."""

    def __init__(self, queue_name, request_handler):
        """Britney test results queue.

        :param queue_name: The queue name where the test results are received.

        :param request_handler: The handler that processes the results.
        """
        self.queue_name = queue_name
        self.request_handler = request_handler

    def consume(self):
        conf = amqp_utils.get_config()
        # Process all messages and delete the queue when done
        amqp_utils.process_queue(conf, self.queue_name, self.handle_request,
                                 delete=True)

    def handle_request(self, msg):
        logging.info('Will handle {}'.format(msg.body))
        body = json.loads(msg.body)
        ret = self.request_handler.handle(body)
        msg.channel.basic_ack(msg.delivery_tag)
        logging.info('Handling returned: {} for {}'.format(ret, msg.body))
        return ret


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    logging.basicConfig(level=logging.INFO,
                        format="%(asctime)s %(levelname)s %(message)s")
    logging.info('Waiting for messages. ^C to exit.')
    britney_data_store = BritneyDataStore(
        'britney.results', unit_config.get_auth_config(), public=True)

    def ticket_data_store_factory(ticket_id):
        """Data stores from the test runner are built from the ticket_id."""
        return data_store.create_for_ticket(ticket_id,
                                            unit_config.get_auth_config())

    request_handler = BritneyResultHandler(britney_data_store,
                                           ticket_data_store_factory)
    queue = BritneyResultsQueue('britney.results', request_handler)
    queue.consume()
    sys.exit(0)
