# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from cStringIO import StringIO
import os
import unittest


from ucitests import fixtures as fixtures


from britney import process_reports


class TestBritneyReportHandlerErrors(unittest.TestCase):

    def setUp(self):
        super(TestBritneyReportHandlerErrors, self).setUp()
        self.handler = process_reports.BritneyReportHandler('unused')

    def assertError(self, exc_class, exc_msg, request):
        with self.assertRaises(exc_class) as cm:
            self.handler.handle(request)
        self.assertEqual(exc_msg, cm.exception.message)

    def test_empty_message(self):
        self.assertError(KeyError, 'result_at', dict())

    def test_wrong_prefix(self):
        self.assertError(IndexError, 'list index out of range',
                         dict(result_at='xxx'))

    def test_wrong_date(self):
        self.assertError(ValueError, 'need more than 1 value to unpack',
                         dict(result_at='trusty/amd64/libp/libpng/xxx',
                              britney_result='trusty amd64 libpng'))

    def test_empty_result(self):
        self.assertError(
            ValueError, 'need more than 0 values to unpack',
            dict(result_at='trusty/amd64/libp/libpng/yymmdd_hhmmss',
                 britney_result=''))


class TestBritneyReportHandler(unittest.TestCase):

    def setUp(self):
        super(TestBritneyReportHandler, self).setUp()
        fixtures.set_uniq_cwd(self)
        prefix = 'trusty/amd64/'
        self.handler = process_reports.BritneyReportHandler(prefix)

    def test_result_file_created(self):
        result = 'trusty amd64 libpng'
        self.handler.handle(
            dict(result_at='trusty/amd64/libp/libpng/yymmdd_hhmmss',
                 britney_result=result))
        result_path = 'trusty/amd64/trusty_amd64_libpng_yymmdd-hhmmss.result'
        self.assertTrue(os.path.exists(result_path),
                        '{} is not there'.format(result_path))
        with open(result_path) as f:
            self.assertEqual(result, f.read())


class TestCliRun(unittest.TestCase):
    "Smoke blackbox tests."""

    def setUp(self):
        super(TestCliRun, self).setUp()
        self.out = StringIO()
        self.err = StringIO()

    def test_run_no_args_errors(self):
        with self.assertRaises(SystemExit):
            process_reports.cli_run([], self.out, self.err)


class TestOptionParsing(unittest.TestCase):

    def setUp(self):
        super(TestOptionParsing, self).setUp()
        self.out = StringIO()
        self.err = StringIO()

    def parse_args(self, args):
        ns = process_reports.ProcessReportArgParser().parse_args(
            args, self.out, self.err)
        return ns

    def test_default_values(self):
        ns = self.parse_args(['reports', 'output'])
        self.assertEqual('output', ns.output)
        self.assertEqual('reports', ns.report)
        self.assertEqual(1, ns.nb_reports)
        self.assertEqual('', self.out.getvalue())
        self.assertEqual('', self.err.getvalue())

    def test_nb_reports(self):
        ns = self.parse_args(['reports', 'output', '-n0'])
        self.assertEqual(0, ns.nb_reports)
