#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import os
import re
import shutil
import sys
import textwrap

from charmhelpers import (
    core,
    fetch
)

hooks = core.hookenv.Hooks()


def juju_info(msg, level='INFO'):
    core.hookenv.log(msg, level)


def _service_name():
    unit = core.hookenv.service_name()
    for x in (':', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return unit


def _service_dir():
    unit = core.hookenv.service_name()
    for x in (':', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return os.path.join(core.hookenv.config()['install_root'], unit)


def _code_dir(revno='current'):
    return os.path.join(_service_dir(), 'code/%s' % revno)


def _upstart_log_dir():
    return os.path.join(_service_dir(), 'logs')


def _lander_logs_dir():
    return os.path.join(_service_dir(), 'lander-logs')


def _relation_file():
    relation = core.hookenv.relation_id()
    return os.path.join(_service_dir(), 'juju-relations', relation)


def add_dependencies():
    juju_info('Adding dependencies.')
    config = core.hookenv.config()
    fetch.configure_sources(update=True)
    for package in config['packages'].split():
        fetch.apt_install(package, fatal=True)


def _get_revno(src):
    revno = os.path.basename(src).split('.tgz')[0]
    revfile = os.path.join(src, '.uploaded_version_info')
    if os.path.exists(revfile):
        # revline will be: branch@617 - 2014-06-25 11:50:44 -0500
        revline = open(revfile).read()
        match = re.match(r'.*@(\d+)', revline)
        if match:
            revno = 'r' + match.group(1) + '_' + revno
    return revno


def _payload_local(payload):
    '''Figure out if the payload is already available locally'''
    name = payload.rsplit('/', 1)[1]
    name = name.split('.tgz')[0]
    for dirname in os.listdir(_code_dir('')):
        if dirname == 'current':
            continue
        if name == dirname.split('_', 1)[1]:
            return dirname


def _payload_extract(payload):
    juju_info('grabbing %s' % payload)
    src = fetch.install_remote(payload)

    revno = _get_revno(src)

    core.host.mkdir(_code_dir('..'))

    # charmhelpers puts the payload in an odd place, move it to our code_dir
    dst = _code_dir(revno)
    shutil.move(src, dst)


def _payloads_not_configured(avail):
    '''returns a list of payloads no longer configured for the charm'''
    configured_dirs = [_payload_local(x) for x in avail]
    all_dirs = [x for x in os.listdir(_code_dir('')) if x != 'current']
    return set(all_dirs) - set(configured_dirs)


def handle_code():
    '''deploy all new payloads configured for the charm. return list of
       payloads to cleanup.'''
    config = core.hookenv.config()
    avail = [x for x in config['available_code'].split(' ') if x.strip()]
    new_current = config['current_code']
    if new_current not in avail:
        raise ValueError(
            '"current_code" not configured to a value in "available_code"')

    for payload in avail:
        if not _payload_local(payload):
            _payload_extract(payload)

    current = _code_dir()
    if os.path.exists(current):
        os.unlink(current)
    dst = _payload_local(new_current)
    juju_info('linking %s as current' % dst)
    os.symlink(dst, current)

    return _payloads_not_configured(avail)


def create_unit_config():
    unit_config = core.hookenv.config().get('unit-config')
    if unit_config:
        with open(os.path.join(_service_dir(), 'unit_config'), 'w') as f:
            f.write(base64.b64decode(unit_config))


def create_upstart():
    config = core.hookenv.config()
    template = textwrap.dedent('''
    #--------------------------------------------------------------
    # This file is managed by the lander Juju charm
    #--------------------------------------------------------------
    description "starts a lander worker"
    start on (local-filesystems and net-device-up IFACE=eth0)
    stop on runlevel [!12345]

    # If the process quits unexpectedly trigger respawn it
    respawn
    # unless it fails 15 times within 5 seconds
    respawn limit 15 5

    env JUJU_UNIT_IDX={idx}

    chdir {cdir}

    # setuid/setgid don't work because we need to do mkdir as root
    # this is the easiest work-around:
    script
      [ -d {logdir} ] || mkdir {logdir}
      [ -d {lander_logs} ] || mkdir {lander_logs}
      chown {uid}:{gid} {logdir} {lander_logs}
      chmod 755 {logdir} {lander_logs}
      exec su -c "{main}" {uid} >>{logdir}/upstart.log 2>&1
    end script
    ''')
    params = {
        'main': config['main'],
        'cdir': _code_dir(),
        'uid': config.get('uid', 'nobody'),
        'gid': config.get('gid', 'nogroup'),
        'idx': core.hookenv.local_unit().split('/')[1],
        'logdir': _upstart_log_dir(),
        'lander_logs': _lander_logs_dir()
    }
    with open('/etc/init/%s.conf' % _service_name(), 'w') as f:
        f.write(template.format(**params))
        # need this so the file is ready for the subprocess call
        f.flush()
        os.fsync(f.fileno())


def restart_upstart():
    if core.host.service_restart(_service_name()):
        # if it wasn't running, restart fails, so just try to start
        if core.host.service_start(_service_name()):
            sys.exit('unable to start service')


@hooks.hook()
def config_changed():
    core.host.mkdir(_code_dir(''))

    add_dependencies()
    create_unit_config()
    create_upstart()
    todelete = handle_code()

    restart_upstart()

    for x in todelete:
        shutil.rmtree(_code_dir(x))


@hooks.hook()
@hooks.hook('lander-relation-changed')
def lander_relation_joined():
    rfile = _relation_file()
    rdir = os.path.dirname(rfile)
    if not os.path.exists(rdir):
        os.mkdir(rdir)
    with open(rfile, 'w') as f:
        unit = core.hookenv.remote_unit().split('/')[0]
        addr = core.hookenv.relation_get('private-address')
        f.write('%s: %s' % (unit, addr))


@hooks.hook()
def lander_relation_broken():
    rfile = _relation_file()
    juju_info('removing relation file: %s' % rfile)
    if os.path.exists(rfile):
        os.unlink(rfile)


@hooks.hook()
def amqp_relation_joined():
    config = core.hookenv.config()
    core.hookenv.relation_set(None, {
        'username': config['amqp-user'],
        'vhost': config['amqp-vhost'],
    })


@hooks.hook()
def amqp_relation_changed():
    config = core.hookenv.config()
    addr = core.hookenv.relation_get('private-address')
    password = core.hookenv.relation_get('password')
    with open(os.path.join(_service_dir(), 'amqp_config.py'), 'w') as f:
        f.write('# DO NOT EDIT. Generated by lander charm hook\n')
        f.write('AMQP_USER = "%s"\n' % config['amqp-user'])
        f.write('AMQP_VHOST = "%s"\n' % config['amqp-vhost'])
        f.write('AMQP_HOST = "%s"\n' % addr)
        f.write('AMQP_PASSWORD = "%s"\n' % password)


@hooks.hook()
def amqp_relation_broken():
    os.unlink(os.path.join(_service_dir(), 'amqp_config.py'))


@hooks.hook('json_status-relation-joined')
def json_status_relation_joined():
    host = core.hookenv.unit_get('public-address')
    if not host:
        juju_info('no public address found, using private')
        host = core.hookenv.unit_private_ip()
    url = 'http://%s/json_status' % host
    core.hookenv.relation_set(None, {'status-url': url})


if __name__ == "__main__":
    hooks.execute(sys.argv)
