# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import mock
import os
import unittest
import shutil
import tempfile

import yaml

import hooks

HERE = os.path.abspath(os.path.dirname(__file__))

HTTP_REL = {
    'hostname': 'remote.com',
    'port': '8080',
}


INFO_REL = {
    'private-address': '10.0.0.1',
}


PATCHES = [
    'charmhelpers.core.hookenv.config',
    'charmhelpers.core.hookenv.local_unit',
    'charmhelpers.core.hookenv.log',
    'charmhelpers.core.hookenv.remote_unit',
    'charmhelpers.core.hookenv.relation_get',
    'charmhelpers.core.hookenv.relation_id',
    'charmhelpers.core.hookenv.relation_set',
    'charmhelpers.core.hookenv.unit_get',
    'charmhelpers.core.host.mkdir',
    'charmhelpers.fetch.apt_install',
    'charmhelpers.fetch.configure_sources',
    'charmhelpers.fetch.install_remote',
]


class JujuTestCase(unittest.TestCase):
    def setUp(self):
        """Setup a testing hook environment."""
        super(JujuTestCase, self).setUp()
        self.tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmpdir)

        for spec in PATCHES:
            attr_name = spec.split('.')[-1]
            _m = mock.patch(spec)
            setattr(self, attr_name, _m.start())
            self.addCleanup(_m.stop)

        self.local_unit.return_value = 'local-test'
        self.remote_unit.return_value = 'remote-test'
        self.relation_get.side_effect = (
            lambda rel_name: HTTP_REL.get(rel_name))
        self.relation_id.return_value = '1'

        self._set_default_config()

    def _set_default_config(self):
        defaults = {}
        config = os.path.join(HERE, '../config.yaml')
        if os.path.exists(config):
            with open(config) as f:
                config = yaml.safe_load(f)
                for name, option in config['options'].iteritems():
                    defaults[name] = option.get('default')
        self.config.return_value = defaults
        self.config.return_value['install_root'] = self.tmpdir


class TestInstallHook(JujuTestCase):
    def setUp(self):
        super(TestInstallHook, self).setUp()
        m = mock.patch('hooks._service_dir')
        m.start().return_value = self.tmpdir
        self.addCleanup(m.stop)
        os.mkdir(os.path.join(self.tmpdir, 'code'))

    def test_payload_local(self):
        self.assertIsNone(hooks._payload_local('http://no_such_sha1.tgz'))
        name = 'r123_sha1hash'
        os.mkdir(hooks._code_dir(name))
        self.assertEqual(name, hooks._payload_local('http://sha1hash.tgz'))

    def test_payload_not_configured(self):
        local = ['r1_aa', 'r2_bb', 'r3_cc', 'r4_dd']
        configured = ['http://aa.tgz', 'http://cc.tgz', 'http://dd.tgz']
        os.mkdir(hooks._code_dir())  # create a directory for "current"
        for x in local:
            os.mkdir(hooks._code_dir(x))
        not_configured = hooks._payloads_not_configured(configured)
        self.assertEquals(not_configured, set(['r2_bb']))

    @mock.patch('hooks._payload_extract')
    def test_handle_code_noop(self, payload_extract):
        local = ['r1_aa', 'r2_bb']
        self.config.return_value['current_code'] = 'http://bb.tgz'
        self.config.return_value['available_code'] = '''
            http://aa.tgz http://bb.tgz'''
        for x in local:
            os.mkdir(hooks._code_dir(x))
        hooks.handle_code()
        self.assertEqual(0, payload_extract.call_count)
        self.assertEqual('r2_bb', os.readlink(hooks._code_dir()))

    @mock.patch('hooks._payload_extract')
    def test_handle_code(self, payload_extract):
        self.config.return_value['current_code'] = 'http://bb.tgz'
        self.config.return_value['available_code'] = 'http://bb.tgz'

        payload_extract.side_effect = lambda x: os.mkdir(
            hooks._code_dir('r2_bb'))
        self.assertEqual(set([]), hooks.handle_code())
        self.assertEqual(1, payload_extract.call_count)
        self.assertEqual('r2_bb', os.readlink(hooks._code_dir()))


class TestLanderHook(JujuTestCase):
    def setUp(self):
        super(TestLanderHook, self).setUp()
        p = os.path.join(self.tmpdir, self.local_unit.return_value)
        os.mkdir(p)
        self.relations = os.path.join(p, 'juju-relations')

    def test_joined_broken(self):
        hooks.hooks.execute(['hooks/lander-relation-joined'])
        self.assertTrue(os.path.exists(self.relations))
        rfile = os.path.join(self.relations, self.relation_id.return_value)
        self.assertTrue(os.path.exists(rfile))

        hooks.hooks.execute(['hooks/lander-relation-broken'])
        self.assertFalse(os.path.exists(rfile))


class TestAmqpHook(JujuTestCase):
    def setUp(self):
        super(TestAmqpHook, self).setUp()
        p = os.path.join(self.tmpdir, self.local_unit.return_value)
        os.mkdir(p)
        self.amqp_file = os.path.join(p, 'amqp_config.py')
        with open(self.amqp_file, 'w') as f:
            f.write('\n')

    def test_joined(self):
        hooks.hooks.execute(['hooks/amqp-relation-joined'])
        self.assertEqual(1, self.relation_set.call_count)

    def test_changed(self):
        hooks.hooks.execute(['hooks/amqp-relation-changed'])
        required = ['AMQP_USER', 'AMQP_VHOST', 'AMQP_HOST', 'AMQP_PASSWORD']
        with open(self.amqp_file) as f:
            for line in f.readlines():
                key = line.split('=')[0].strip()
                if key in required:
                    required.remove(key)
        self.assertEqual([], required)

    def test_broken(self):
        hooks.hooks.execute(['hooks/amqp-relation-broken'])
        self.assertFalse(os.path.exists(self.amqp_file))
