#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import os
import re
import shutil
import subprocess
import sys
import tempfile
import textwrap

from charmhelpers.contrib.charmsupport import nrpe

import charmhelpers.core.hookenv
import charmhelpers.core.host
import charmhelpers.fetch
import charmhelpers.payload.execd

hooks = charmhelpers.core.hookenv.Hooks()

required_pkgs = ['python-amqplib', 'python-jinja2', 'python-pip']


def juju_info(msg):
    charmhelpers.core.hookenv.log(msg, 'INFO')


def _service_name():
    unit = charmhelpers.core.hookenv.service_name()
    for x in (':', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return unit


def _service_dir():
    unit = charmhelpers.core.hookenv.service_name()
    for x in (':', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return os.path.join(
        charmhelpers.core.hookenv.config()['install_root'], unit)


def _code_dir(revno='current'):
    return os.path.join(_service_dir(), 'code/%s' % revno)


def _log_dir():
    return os.path.join(_service_dir(), 'logs')


def _nagios_dir():
    return os.path.join(_service_dir(), 'nagios')


def _get_revno(src):
    revno = os.path.basename(src).split('.tgz')[0]
    revfile = os.path.join(src, '.uploaded_version_info')
    if os.path.exists(revfile):
        # revline will be: branch@617 - 2014-06-25 11:50:44 -0500
        revline = open(revfile).read()
        match = re.match(r'.*@(\d+)', revline)
        if match:
            revno = 'r' + match.group(1) + '_' + revno
    return revno


def _payload_local(payload):
    '''Figure out if the payload is already available locally'''
    name = payload.rsplit('/', 1)[1]
    name = name.split('.tgz')[0]
    for dirname in os.listdir(_code_dir('')):
        if dirname == 'current':
            continue
        if name == dirname.split('_', 1)[1]:
            return dirname


def _payload_extract(payload):
    juju_info('grabbing %s' % payload)
    src = charmhelpers.fetch.install_remote(payload)

    revno = _get_revno(src)

    charmhelpers.core.host.mkdir(_code_dir('..'))

    # charmhelpers puts the payload in an odd place, move it to our code_dir
    dst = _code_dir(revno)
    shutil.move(src, dst)


def _payloads_not_configured(avail):
    '''returns a list of payloads no longer configured for the charm'''
    configured_dirs = [_payload_local(x) for x in avail]
    all_dirs = [x for x in os.listdir(_code_dir('')) if x != 'current']
    return set(all_dirs) - set(configured_dirs)


def handle_code():
    '''deploy all new payloads configured for the charm. return list of
       payloads to cleanup.'''
    config = charmhelpers.core.hookenv.config()
    avail = [x for x in config['available_code'].split(' ') if x.strip()]
    new_current = config['current_code']
    if new_current not in avail:
        raise ValueError(
            '"current_code" not configured to a value in "available_code"')

    for payload in avail:
        if not _payload_local(payload):
            _payload_extract(payload)

    current = _code_dir()
    if os.path.exists(current):
        os.unlink(current)
    dst = _payload_local(new_current)
    juju_info('linking %s as current' % dst)
    os.symlink(dst, current)

    return _payloads_not_configured(avail)


def pip_install(package):
    cmd_line = ['pip', 'install', '-b', '/tmp/']
    if package.startswith('svn+') or package.startswith('git+') or \
       package.startswith('hg+') or package.startswith('bzr+'):
        cmd_line.append('-e')
    cmd_line.append(package)
    return(subprocess.call(cmd_line))


@hooks.hook()
def install(upgrade=False):
    config = charmhelpers.core.hookenv.config()
    # Support Canonistack/Prodstack basenode subordinate charm interface.
    charmhelpers.payload.execd.execd_preinstall()
    _create_nagios_dir()

    if upgrade:
        charmhelpers.fetch.apt_update()
    else:
        charmhelpers.fetch.configure_sources(update=True)

    pkgs = [x for x in config.get('packages', '').split(' ') if x]
    pkgs.extend(required_pkgs)

    juju_info('installing apt packages...')
    charmhelpers.fetch.apt_install(pkgs, fatal=True)

    pip_pkgs = [x for x in config.get('pip-packages', '').split(' ') if x]
    if pip_pkgs:
        juju_info('installing pip packages...')
        for package in pip_pkgs:
            pip_install(package.strip())


@hooks.hook()
def upgrade_charm():
    install(True)
    try:
        subprocess.check_call(['/sbin/restart', _service_name()])
    except subprocess.CalledProcessError:
        # if it wasn't running, restart fails, so just try to start
        subprocess.check_call(['/sbin/start', _service_name()])


def _cron_root():
    return '/etc/cron.d'


def _get_template(name):
    from jinja2 import (
        Environment,
        FileSystemLoader,
    )
    template_dir = os.path.join(
        charmhelpers.core.hookenv.charm_dir(), 'templates')
    # XXX cprov 20140719: once we move to Jinja2 > 2.7 (trusty) we can pass
    # 'keep_trailing_newline=True' and remove the hack below.
    template_env = Environment(loader=FileSystemLoader(template_dir))
    return template_env.get_template(name)


def _update_cron_config():
    """Create a /etc/cron.d/<service> according to the cron_* options."""
    service = charmhelpers.core.hookenv.service_name()
    config = charmhelpers.core.hookenv.config()
    cron_path = os.path.join(_cron_root(), service)
    # Removing old cron configuration.
    juju_info('Removing old cron configuration ({})'.format(cron_path))
    if os.path.exists(cron_path):
        os.unlink(cron_path)
    # Dismiss if no schedule or command was configured.
    if not config['cron_schedule'] or not config['cron_cmd']:
        charmhelpers.core.host.service_restart('cron')
        juju_info('No cron scheduling or command information available. '
                  'Nothing installed!')
        return
    # Populate cron configuration. 'user' used is the service owner.
    params = {
        'schedule': config['cron_schedule'],
        'user': config['uid'],
        'logs_dir': _log_dir(),
        'code_dir': _code_dir(),
        'command': config['cron_cmd'],
    }
    template = _get_template('cron.tmpl')
    with open(cron_path, 'w') as fp:
        fp.write(template.render(**params))
        # XXX cprov 20140719: add trailing newline manually, since it's not
        # supported on precise's Jinja and required by cron. See above.
        fp.write('\n')
    charmhelpers.core.host.service_restart('cron')
    juju_info('Cron configuration updated ({}).'.format(cron_path))


def _create_nagios_dir():
    config = charmhelpers.core.hookenv.config()
    ndir = _nagios_dir()
    if not os.path.exists(ndir):
        charmhelpers.core.host.mkdir(ndir, config['uid'], config['gid'], 0755)


def _update_nrpe_config():
    nrpe_compat = nrpe.NRPE()
    health_script_path = os.path.join(
        _code_dir(), 'bin', 'engine_health.py')
    conf = nrpe_compat.config
    check_http_params = conf.get('nagios_check_http_params')
    if check_http_params:
        nrpe_compat.add_check(
            shortname='rabbitmq-worker',
            description='Check RabbitMQ Service',
            check_cmd='check_http %s' % check_http_params
        )
    check_health_params = conf.get('nagios_check_health_params')
    if check_health_params:
        nrpe_compat.add_check(
            shortname='engine_health',
            description='Check Engine Health',
            check_cmd='%s %s' % (health_script_path, check_health_params)
        )
    nrpe_compat.write()


@hooks.hook()
def config_changed():
    charmhelpers.core.host.mkdir(_code_dir(''))

    _create_unit_config()
    _create_upstart()
    todelete = handle_code()
    _update_cron_config()
    _update_nrpe_config()

    _restart_upstart()

    for x in todelete:
        shutil.rmtree(_code_dir(x))


def _create_unit_config():
    unit_config = charmhelpers.core.hookenv.config().get('unit-config')
    if unit_config:
        with open(os.path.join(_service_dir(), 'unit_config'), 'w') as f:
            f.write(base64.b64decode(unit_config))


def _restart_upstart():
    try:
        subprocess.check_call(['/sbin/restart', _service_name()])
    except subprocess.CalledProcessError:
        # if it wasn't running, restart fails, so just try to start
        subprocess.check_call(['/sbin/start', _service_name()])


def _create_upstart():
    template = textwrap.dedent('''
    #--------------------------------------------------------------
    # This file is managed by the rabbitmq-worker Juju charm
    #--------------------------------------------------------------
    description "starts a rabbitmq worker"
    start on (local-filesystems and net-device-up IFACE=eth0)
    stop on runlevel [!12345]

    # use sigint so python code just needs to catch KeyboardInterrupt
    kill signal SIGINT

    # If the process quits unexpectedly trigger respawn it
    respawn
    # unless it fails 15 times within 5 seconds
    respawn limit 15 5

    chdir {cdir}

    # setuid/setgid don't work because we need to do mkdir as root
    # this is the easiest work-around:
    script
      [ -d {logdir} ] || mkdir {logdir}
      chown {uid}:{gid} {logdir}
      chmod 755 {logdir}
      exec su -c "{main}" {uid} >>{logdir}/upstart.log 2>&1
    end script
    ''')
    config = charmhelpers.core.hookenv.config()
    params = {
        'main': config['main'],
        'cdir': _code_dir(),
        'sdir': _service_dir(),
        'uid': config.get('uid', 'nobody'),
        'gid': config.get('gid', 'nogroup'),
        'logdir': _log_dir(),
    }
    path = '/etc/init/{}.conf'.format(_service_name())
    with tempfile.NamedTemporaryFile('w', delete=False) as f:
        f.write(template.format(**params))
    # Rename to final destination
    os.rename(f.name, path)
    os.chmod(path, 0644)


def _create_amqp_config(config):
    path = os.path.join(_service_dir(), 'amqp_config.py')
    relation = charmhelpers.core.hookenv.relation_get()
    if 'password' not in relation:
        return path  # not yet ready

    with tempfile.NamedTemporaryFile('w', delete=False) as f:
        f.write('# DO NOT EDIT. Generated by rabbitmq-worker charm hook\n')
        f.write('AMQP_USER = "%s"\n' % config['amqp-user'])
        f.write('AMQP_VHOST = "%s"\n' % config['amqp-vhost'])
        f.write('AMQP_HOST = "%s"\n' % relation['private-address'])
        f.write('AMQP_PASSWORD = "%s"\n' % relation['password'])
    # Now that the file is complete (and closed) rename it to its final
    # destination. This avoids races where another process try to import that
    # file as a python module (creating a .pyc) and ends up having close enough
    # time stamps to fool python (which won't reload the completed file).
    os.rename(f.name, path)
    os.chmod(path, 0644)
    return path


@hooks.hook()
def amqp_relation_joined():
    config = charmhelpers.core.hookenv.config()
    charmhelpers.core.hookenv.relation_set(
        None, {'username': config['amqp-user'], 'vhost': config['amqp-vhost']})


@hooks.hook()
def amqp_relation_changed():
    config = charmhelpers.core.hookenv.config()
    if not os.path.exists(_create_amqp_config(config)):
        return
    _restart_upstart()


@hooks.hook()
def amqp_relation_broken():
    subprocess.call(['/sbin/stop', _service_name()])
    os.unlink(os.path.join(_service_dir(), 'amqp_config.py'))


@hooks.hook('json_status-relation-joined')
def json_status_relation_joined():
    config = charmhelpers.core.hookenv.config()
    host = charmhelpers.core.hookenv.unit_private_ip()
    url = 'http://%s:%d/json_status' % (host, config['json_status_port'])
    charmhelpers.core.hookenv.relation_set(None, {'status-url': url})


@hooks.hook('nrpe-external-master-relation-changed')
@hooks.hook()
def nrpe_external_master_relation_joined():
    _update_nrpe_config()


@hooks.hook('local-monitors-relation-changed')
@hooks.hook()
def local_monitors_relation_joined():
    _update_nrpe_config()


if __name__ == '__main__':
    hooks.execute(sys.argv)
