#!/usr/bin/env python

import os
import shutil
import subprocess
import sys

import charmhelpers.fetch
from charmhelpers.core.hookenv import log, config, relation_set


def _service_dir(charmconfig):
    unit = os.environ['JUJU_UNIT_NAME'].split('/')[0]
    for x in (':', '-', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return os.path.join(charmconfig['install_root'], unit)


def _install_from_tarball(charmconfig, upgrade):
    #Do we need this?
    pass


def _install_from_bzr(charmconfig, upgrade):
    log('grabbing service from bzr...', 'INFO')
    sdir = _service_dir(charmconfig)
    rev = charmconfig.get('revno', '')

    if upgrade:
        args = ['bzr', 'pull']
        if rev:
            args.extend(['-r', rev])
        subprocess.check_call(args, cwd=sdir)
        return
    if os.path.exists(sdir):
        log('deleting pre-existing service directory: {}'.format(sdir),
            'INFO')
        shutil.rmtree(sdir)

    args = ['bzr', 'branch']
    if rev:
        args.extend(['-r', rev])
    args.append(charmconfig['branch'])
    args.append(sdir)
    subprocess.check_call(args)


def _write_system_image_config(charmconfig):
    sdir = _service_dir(charmconfig)
    config_file = os.path.join(sdir, 'etc', 'config')
    from jinja2 import Environment, FileSystemLoader
    template_env = Environment(loader=FileSystemLoader(os.path.join(
        os.environ['CHARM_DIR'], 'data')))
    svars = {
        'base_path': sdir,
    }
    template = template_env.get_template('config.tmpl').render(svars)
    log("Writing system-image config to {}".format(config_file), 'INFO')
    with open(config_file, 'w') as conf:
        conf.write(str(template))


def _create_stable_channel(charmconfig):
    log('creating stable channel...', 'INFO')
    libpath = os.path.join(_service_dir(charmconfig), 'lib')
    if libpath not in sys.path:
        sys.path.insert(0, libpath)

    from systemimage import tree, config as si_config
    conf = si_config.Config()
    pub = tree.Tree(conf)
    pub.create_channel('stable')
    pub.create_device('stable', 'mako')
    for keyring in ("archive-master", "image-master", "image-signing"):
        pub.publish_keyring(keyring)


def _generate_keys(charmconfig):
    log('generating keys...', 'INFO')
    #haveged is just an entropy generator
    subprocess.check_output(['haveged'])
    srvdir = _service_dir(charmconfig)
    #the system-image tools unfortunately really want to run from this dir
    os.chdir(srvdir)
    out = subprocess.check_output(['{}/tests/generate-keys'.format(srvdir)])
    log(out, 'INFO')
    #copy the keys we just generated into the proper location
    shutil.rmtree('{}/secret/gpg/keys/'.format(srvdir))
    shutil.copytree('{}/tests/keys'.format(srvdir), '{}/secret/gpg/keys/'.format(srvdir))
    subprocess.check_output(['{}/bin/generate-keyrings'.format(srvdir)])
    log(out, 'INFO')


def install(charmconfig, upgrade=False):
    pkgs = ['bzr', 'abootimg', 'android-tools-fsutils', 'python-gnupg',
            'python-gpgme', 'fakeroot', 'pxz', 'pep8', 'pyflakes',
            'python-mock', 'haveged']
    log('installing apt packages...', 'INFO')
    charmhelpers.fetch.configure_sources(update=True)
    charmhelpers.fetch.apt_install(pkgs)

    if charmconfig.get('vcs') == 'tarball':
        _install_from_tarball(charmconfig, upgrade)
    else:
        _install_from_bzr(charmconfig, upgrade)

    _write_system_image_config(charmconfig)
    _generate_keys(charmconfig)
    _create_stable_channel(charmconfig)


def website_relation_joined(charmconfig):
    host = subprocess.check_output(['unit-get', 'private-address']).strip()
    relation_set(None, {'port': charmconfig["port"], 'hostname': host})


def website_relation_changed(charmconfig):
    return website_relation_joined(charmconfig)


def website_relation_broken(charmconfig):
    pass


def main():
    hook = os.path.basename(sys.argv[0])
    log("Running hook: {}".format(hook), 'INFO')

    hook_py = hook.replace('-', '_')
    funcs = globals()
    if hook_py not in funcs:
        print("Unknown hook: {}".format(hook))
        return 1

    charmconfig = config()
    try:
        return funcs[hook_py](charmconfig)
    except subprocess.CalledProcessError as e:
        log('Error running: {}: {}'.format(e.cmd, e.output), 'INFO')
        return e.returncode


def start(charmconfig):
    pass


def stop(charmconfig):
    pass


def upgrade_charm(charmconfig):
    install(charmconfig, True)
    start(charmconfig)

if __name__ == '__main__':
    sys.exit(main())
