#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import fcntl
import json
import os
import sys

import charmhelpers.contrib.charmsupport.nrpe
import charmhelpers.core.hookenv
import charmhelpers.payload.execd

from charmhelpers.core import host

hooks = charmhelpers.core.hookenv.Hooks()
log = charmhelpers.core.hookenv.log


###############################################################################
# Global variables
###############################################################################
proxy_confs = '/etc/apache2'


def health_path():
    return charmhelpers.core.hookenv.config()['health_path']


def json_status_file():
    return os.path.join(health_path(), 'status_urls.json')


def _add_proxy(url_base, remote):
    with open(_conf_file(), 'w') as f:
        f.write(
            'RewriteRule ^/%s/(.*)$ %s$1 [P]\n' % (url_base, remote))
        f.write('ProxyPassReverse /%s %s\n' % (url_base, remote))
    host.service_restart('apache2')


def _remove_proxy():
    cf = _conf_file()
    if os.path.exists(cf):
        os.unlink(cf)
        host.service_restart('apache2')


@contextlib.contextmanager
def status_urls():
    '''enforce exclusive access to the file to support concurrent updates'''
    with open(json_status_file(), 'r+') as f:
        try:
            fcntl.lockf(f, fcntl.LOCK_EX)
            f.seek(0)
            data = json.load(f)
            yield data

            # clear the file before writing back data as it appears the new
            # dump doesn't set the end-of-file correctly.
            f.truncate(0)
            f.seek(0)
            json.dump(data, f)
        finally:
            fcntl.lockf(f, fcntl.LOCK_UN)


@hooks.hook('json_status-relation-joined')
@hooks.hook('json_status-relation-changed')
def json_status_relation_joined():
    url = charmhelpers.core.hookenv.relation_get('status-url')
    unit = charmhelpers.core.hookenv.remote_unit()
    proxy_base = 'json_status-' + unit
    log('status URL is: %s' % url)

    with status_urls() as data:
        data[unit] = {
            'url': '/' + proxy_base + '/',
            'unit': unit,
        }
    _add_proxy(proxy_base, url)


@hooks.hook('json_status-relation-departed')
def json_status_relation_departed():
    if not os.path.exists(json_status_file()):
        return

    unit = charmhelpers.core.hookenv.remote_unit()

    with status_urls() as data:
        if unit in data:
            del data[unit]
    _remove_proxy()


@hooks.hook()
def install():
    if not os.path.exists(health_path()):
        os.mkdir(health_path())
    if not os.path.exists(json_status_file()):
        with open(json_status_file(), 'w') as f:
            f.write('{}')


def _conf_file():
    unit = charmhelpers.core.hookenv.remote_unit().replace('/', '_')
    relation = charmhelpers.core.hookenv.relation_type()
    return os.path.join(proxy_confs, 'juju-' + relation + '-' + unit + '.conf')


@hooks.hook('proxy_logs-relation-joined')
def proxy_logs_relation_joined():
    name = charmhelpers.core.hookenv.remote_unit()
    addr = charmhelpers.core.hookenv.relation_get('private-address')

    _add_proxy('logs-' + name, 'http://' + addr + '/')


@hooks.hook('proxy_logs-relation-departed')
def proxy_logs_relation_departed():
    _remove_proxy()


if __name__ == "__main__":
    hooks.execute(sys.argv)
