# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import mock
import os
import unittest
import shutil
import tempfile

import yaml

import hooks

HERE = os.path.abspath(os.path.dirname(__file__))

HTTP_REL = {
    'hostname': 'remote.com',
    'port': '8080',
}


INFO_REL = {
    'private-address': '10.0.0.1',
}


PATCHES = [
    'charmhelpers.core.hookenv.config',
    'charmhelpers.core.hookenv.relation_get',
    'charmhelpers.core.hookenv.remote_unit',
]


class RestishTestCase(unittest.TestCase):
    def setUp(self):
        """Setup a testing hook environment.

        Patches the required charmhelper and intercom objects and
        initializes few mocks for tests.
        """
        super(RestishTestCase, self).setUp()
        self.tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmpdir)

        for spec in PATCHES:
            attr_name = spec.split('.')[-1]
            _m = mock.patch(spec)
            setattr(self, attr_name, _m.start())
            self.addCleanup(_m.stop)

        hooks.log = mock.Mock()
        self.remote_unit.return_value = 'remote-test'
        self._set_default_config()

    def _set_default_config(self):
        defaults = {}
        config = os.path.join(HERE, '../config.yaml')
        if os.path.exists(config):
            with open(config) as f:
                config = yaml.safe_load(f)
                for name, option in config['options'].iteritems():
                    defaults[name] = option.get('default')
        defaults['health_path'] = self.tmpdir
        self.config.return_value = defaults


class TestJSONStatusHooks(RestishTestCase):

    def test_status_urls_flocking(self):
        expected_data = {'foo': 'bar'}
        with open(hooks.json_status_file(), 'w') as f:
            json.dump(expected_data, f)

        with hooks.status_urls() as data:
            self.assertEqual(expected_data, data)
            data['blah'] = 123
            expected_data = data

        with open(hooks.json_status_file()) as f:
            data = json.load(f)
            self.assertEqual(expected_data, data)

    @mock.patch('hooks._add_proxy')
    def test_json_status_joined(self, add_proxy):
        with open(hooks.json_status_file(), 'w') as f:
            f.write('{}')

        self.relation_get.return_value = 'foo'
        hooks.hooks.execute(['hooks/json_status-relation-joined'])
        add_proxy.assert_called_once_with('json_status-remote-test', 'foo')
        with hooks.status_urls() as data:
            self.assertIn('remote-test', data)

    @mock.patch('hooks._add_proxy')
    def test_json_status_departed(self, add_proxy):
        with open(hooks.json_status_file(), 'w') as f:
            f.write('{}')

        self.relation_get.return_value = 'foo'
        hooks.hooks.execute(['hooks/json_status-relation-joined'])
        add_proxy.assert_called_once_with('json_status-remote-test', 'foo')
        with hooks.status_urls() as data:
            self.assertIn('remote-test', data)

        with hooks.status_urls() as data:
            del data['remote-test']

        with hooks.status_urls() as data:
            self.assertEqual({}, data)

