#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import json
import os
import re
import shutil
import socket
import subprocess
import sys
import tempfile
import time

import charmhelpers.core
import charmhelpers.fetch
import charmhelpers.payload.execd

from charmhelpers.contrib.charmsupport.nrpe import NRPE

required_pkgs = ['python-jinja2']

hooks = charmhelpers.core.hookenv.Hooks()


def juju_log(msg):
    charmhelpers.core.hookenv.log(msg, 'INFO')


def _service_dir():
    unit = charmhelpers.core.hookenv.service_name()
    for x in (':', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return os.path.join(charmhelpers.core.hookenv.config()['install_root'],
                        unit)


def _code_dir(revno='current'):
    return os.path.join(_service_dir(), 'code/%s' % revno)


def _log_dir():
    return os.path.join(_service_dir(), 'logs')


def _nagios_dir():
    return os.path.join(_service_dir(), 'nagios')


def add_dependencies():
    juju_log('Adding dependencies.')
    config = charmhelpers.core.hookenv.config()

    # compile packages list from config
    # `if x` is for possible multiple spaces in `config['packages']`
    pkgs = [x for x in config['packages'].split(' ') if x]
    pkgs.extend(required_pkgs)

    framework = config['framework']
    if framework == 'restish':
        pkgs.append('python-restish')
    elif framework == 'pyramid':
        pkgs.append('python3-pyramid')
    elif framework == 'django':
        pkgs.append('python-django')
    else:
        juju_log('Invalid framework specified: {}'.format(framework))
        sys.exit(1)
    juju_log('installing apt packages...')
    unique_pkgs = set(pkgs)
    charmhelpers.fetch.configure_sources(update=True)
    charmhelpers.fetch.apt_install(unique_pkgs, fatal=True)


def _get_revno(src):
    revno = os.path.basename(src).split('.tgz')[0]
    revfile = os.path.join(src, '.uploaded_version_info')
    if os.path.exists(revfile):
        # revline will be: branch@617 - 2014-06-25 11:50:44 -0500
        revline = open(revfile).read()
        match = re.match(r'.*@(\d+)', revline)
        if match:
            revno = 'r' + match.group(1) + '_' + revno
    return revno


def _payload_local(payload):
    '''Figure out if the payload is already available locally'''
    name = os.path.basename(payload)
    name = os.path.splitext(name)[0]
    for dirname in os.listdir(_code_dir('')):
        if dirname == 'current':
            continue
        if name == dirname.split('_', 1)[1]:
            return dirname


def _payload_extract(payload):
    juju_log('grabbing %s' % payload)
    src = charmhelpers.fetch.install_remote(payload)
    revno = _get_revno(src)

    charmhelpers.core.host.mkdir(_code_dir('..'))

    # charmhelpers puts the payload in an odd place, move it to our code_dir
    dst = _code_dir(revno)
    shutil.move(src, dst)


def _payloads_not_configured(avail):
    '''returns a list of payloads no longer configured for the charm'''
    configured_dirs = [_payload_local(x) for x in avail]
    all_dirs = [x for x in os.listdir(_code_dir('')) if x != 'current']
    return set(all_dirs) - set(configured_dirs)


def handle_code():
    charmhelpers.payload.execd.execd_preinstall()
    config = charmhelpers.core.hookenv.config()
    _create_nagios_dir()
    avail = [x for x in config['available_code'].split(' ') if x.strip()]
    new_current = config['current_code']
    if new_current not in avail:
        raise ValueError(
            '"current_code" not configured to a value in "available_code"')
    add_dependencies()
    for payload in avail:
        if not _payload_local(payload):
            _payload_extract(payload)

    current = _code_dir()
    if os.path.exists(current):
        os.unlink(current)
    dst = _payload_local(new_current)
    juju_log('linking %s as current' % dst)
    os.symlink(dst, current)

    return _payloads_not_configured(avail)


def _wsgi_reload():
    # use upstart if possible or in the context of a non-wsgi relation
    if os.path.exists('/etc/init/gunicorn.conf'):
        charmhelpers.core.host.service_restart('gunicorn')
        return

    # Trigger WSGI reloading
    for relid in charmhelpers.core.hookenv.relation_ids():
        charmhelpers.core.hookenv.relation_set(
            relid, {'wsgi_timestamp': time.time()})


def _cron_root():
    return '/etc/cron.d'


def _get_template(name):
    from jinja2 import (
        Environment,
        FileSystemLoader,
    )
    template_dir = os.path.join(
        charmhelpers.core.hookenv.charm_dir(), 'templates')
    # XXX cprov 20140719: once we move to Jinja2 > 2.7 (trusty) we can pass
    # 'keep_trailing_newline=True' and remove the hack below.
    template_env = Environment(loader=FileSystemLoader(template_dir))
    return template_env.get_template(name)


def _update_cron_config():
    """Create a /etc/cron.d/<service> according to the cron_* options."""
    service = charmhelpers.core.hookenv.service_name()
    config = charmhelpers.core.hookenv.config()
    cron_path = os.path.join(_cron_root(), service)
    # Removing old cron configuration.
    juju_log('Removing old cron configuration ({})'.format(cron_path))
    if os.path.exists(cron_path):
        os.unlink(cron_path)
    # Dismiss if no schedule or command was configured.
    if not config['cron_schedule'] or not config['cron_cmd']:
        charmhelpers.core.host.service_restart('cron')
        juju_log('No cron scheduling or command information available. '
                 'Nothing installed!')
        return
    # Populate cron configuration. 'user' used is the service owner.
    params = {
        'schedule': config['cron_schedule'],
        'user': config['wsgi_user'],
        'logs_dir': _log_dir(),
        'code_dir': _code_dir(),
        'command': config['cron_cmd'],
    }
    template = _get_template('cron.tmpl')
    with open(cron_path, 'w') as fp:
        fp.write(template.render(**params))
        # XXX cprov 20140719: add trailing newline manually, since it's not
        # supported on precise's Jinja and required by cron. See above.
        fp.write('\n')
    charmhelpers.core.host.service_restart('cron')
    juju_log('Cron configuration updated ({}).'.format(cron_path))


def _set_logging():
    config = charmhelpers.core.hookenv.config()
    log_level = config.get('log_level', 'INFO')

    data = dict(
        log_dir=_log_dir(),
        log_level=log_level,
    )
    with open(os.path.join(_service_dir(), 'logging.json'), 'w') as f:
        json.dump(data, f)


def _get_reverseproxies():
    reverseproxies = []
    relids = charmhelpers.core.hookenv.relation_ids('website')

    for rid in relids:
        remote_units = charmhelpers.core.hookenv.related_units(rid)
        for unit in remote_units:
            relation = charmhelpers.core.hookenv.relation_get(
                rid=rid,
                unit=unit,
            )
            if relation is not None:
                reverseproxie = relation.get('private-address')
                if reverseproxie is not None:
                    reverseproxies.append(relation['private-address'])

    return reverseproxies


def _set_allowed_hosts():
    config = charmhelpers.core.hookenv.config()
    ips = [
        socket.gethostname(),
        socket.getfqdn(),
        charmhelpers.core.hookenv.unit_get('public-address'),
        charmhelpers.core.hookenv.unit_private_ip(),
        'localhost',
    ]

    # include any reverseproxies as valid hosts
    reverseproxies = _get_reverseproxies()

    juju_log("adding reverseproxies: {}".format(reverseproxies))
    ips += reverseproxies

    host = config.get('hostname')
    if host:
        ips.append(host)

    root_dir = _service_dir()
    with open(os.path.join(root_dir, 'allowed_hosts.json'), 'w') as f:
        json.dump(ips, f)


def _create_nagios_dir():
    config = charmhelpers.core.hookenv.config()
    ndir = _nagios_dir()
    if not os.path.exists(ndir):
        charmhelpers.core.host.mkdir(
            ndir, config['wsgi_user'], config['wsgi_group'], 0755)


def create_unit_config():
    unit_config = charmhelpers.core.hookenv.config().get('unit-config')
    if unit_config:
        with open(os.path.join(_service_dir(), 'unit_config'), 'w') as f:
            f.write(base64.b64decode(unit_config))


@hooks.hook('upgrade-charm')
@hooks.hook()
def config_changed():
    charmhelpers.core.host.mkdir(_code_dir(''))
    add_dependencies()
    create_unit_config()
    todelete = handle_code()
    _update_cron_config()
    if os.path.exists(_code_dir('current')):
        _db_migrate()

    relids = charmhelpers.core.hookenv.relation_ids('oauth-server')
    for relid in relids:
        for unit in charmhelpers.core.hookenv.related_units(relid):
            relation = charmhelpers.core.hookenv.relation_get(
                rid=relid, unit=unit)
            if relation and 'url' in relation:
                juju_log("creating client credentials "
                         "for {}".format(relation['url']))
                _create_client(relation['url'])

    _set_allowed_hosts()
    update_nrpe_config()
    _set_logging()
    _wsgi_reload()
    for x in todelete:
        shutil.rmtree(_code_dir(x))


@hooks.hook('website-relation-changed')
@hooks.hook()
def website_relation_joined():
    config = charmhelpers.core.hookenv.config()
    host = charmhelpers.core.hookenv.unit_get('private-address')
    relation_id = charmhelpers.core.hookenv.relation_id()
    charmhelpers.core.hookenv.relation_set(
        relation_id, {'port': config["port"], 'hostname': host})
    _set_allowed_hosts()


@hooks.hook('json_status-relation-joined')
def json_status_relation_joined():
    config = charmhelpers.core.hookenv.config()
    path = config.get('json_status_path')
    if path:
        port = config['port']
        host = charmhelpers.core.hookenv.unit_get('public-address')
        if not host:
            juju_log('no public address found, using private')
            host = charmhelpers.core.hookenv.unit_private_ip()
        url = 'http://%s:%s/%s' % (host, port, path)
        charmhelpers.core.hookenv.relation_set(None, {'status-url': url})


@hooks.hook('wsgi-relation-changed')
@hooks.hook()
def wsgi_relation_joined():
    config = charmhelpers.core.hookenv.config()
    cdir = _code_dir()
    ldir = _log_dir()

    if not os.path.exists(ldir):
        charmhelpers.core.host.mkdir(
            ldir, config['wsgi_user'], config['wsgi_group'], 0755)

    full_path = []
    path = config.get('python_path')
    if path:
        full_path = []
        for p in path.split(':'):
            if p[0] != '/':
                p = os.path.join(cdir, p)
            full_path.append(p)

    charmhelpers.core.hookenv.relation_set(relation_settings={
        'working_dir': cdir,
        'wsgi_log_file': os.path.join(ldir, 'gunicorn.log'),
        'wsgi_access_logfile': os.path.join(ldir, 'access.log'),
        'port': config['port'],
        'python_path': ':'.join(full_path),
    })

    charmhelpers.core.hookenv.open_port(config['port'])


@hooks.hook()
def wsgi_relation_broken():
    port = charmhelpers.core.hookenv.config()['port']
    charmhelpers.core.hookenv.close_port(port)


@hooks.hook('nrpe-external-master-relation-changed')
@hooks.hook()
def nrpe_external_master_relation_joined():
    update_nrpe_config()


@hooks.hook('local-monitors-relation-changed')
@hooks.hook()
def local_monitors_relation_joined():
    update_nrpe_config()


def update_nrpe_config():
    nrpe_compat = NRPE()
    health_script_path = os.path.join(
        _code_dir(), 'bin', 'engine_health.py')
    conf = nrpe_compat.config
    check_http_params = conf.get('nagios_check_http_params')
    if check_http_params:
        nrpe_compat.add_check(
            shortname='restish',
            description='Check Restish Server',
            check_cmd='check_http %s' % check_http_params
        )
    check_health_params = conf.get('nagios_check_health_params')
    if check_health_params:
        nrpe_compat.add_check(
            shortname='engine_health',
            description='Check Engine Health',
            check_cmd='%s %s' % (health_script_path, check_health_params)
        )
    nrpe_compat.write()


def _db_migrate():
    config = charmhelpers.core.hookenv.config()
    if not os.path.exists(_pgsql_file()):
        # the pgsql relation doesn't exist
        return

    relation = charmhelpers.core.hookenv.relation_get()
    cmd = config.get('db_migration_cmd')

    # some relation calls will just have "private-address". We need to ensure
    # we've gotten all the db credentials before trying to migrate
    if relation and (not relation.get('user') or not relation.get('host')):
        return

    if cmd:
        # retry to avoid a race condition with the postgres charm which
        # stops and starts the DB to mount external storage.
        retries = 0
        MAX_RETRIES = 5
        SLEEP_INTERVAL = 5
        while retries < MAX_RETRIES:
            try:
                subprocess.check_call([cmd], shell=True, cwd=_code_dir())
                juju_log("migration command succeeded")
                break
            except subprocess.CalledProcessError:
                juju_log("migration command failed, retrying...")
                pass
            time.sleep(SLEEP_INTERVAL)
            retries += 1
        if retries == MAX_RETRIES:
            juju_log("migration command failed")
            raise Exception("Migration failed")


def _pgsql_file():
    return os.path.join(_service_dir(), 'pgsql.json')


@hooks.hook()
def oauth_client_relation_joined():
    host = charmhelpers.core.hookenv.unit_private_ip()
    port = charmhelpers.core.hookenv.config().get('port')
    path = ''
    url = 'http://%s:%s/%s' % (host, port, path)
    charmhelpers.core.hookenv.relation_set(
        None, {
            'url': url,
        })


@hooks.hook()
def oauth_client_relation_changed():
    relation = charmhelpers.core.hookenv.relation_get()

    client_data = relation.get('client_data')

    if client_data is None:
        return

    client_data = json.loads(client_data)

    juju_log("got client credentials")

    with open(os.path.join(_service_dir(), 'oauth_credentials.json'),
              'w') as fp:
        json.dump(client_data, fp)


def _create_client(client_redirect_uri):
    code_dir = _code_dir()

    data = None
    try:
        data = subprocess.check_output(
            ['./run-python', 'ticket_system/manage.py', 'create_client',
             client_redirect_uri],
            cwd=code_dir,
        ).strip()
        juju_log("OAuth2 client created for {}".format(client_redirect_uri))
    except subprocess.CalledProcessError as e:
        # XXX: Don't fail the deployment here, but we need to find a better
        # way to let the deployer know that there is missing functionality.
        # Evan's suggestion was a nagios check to ensure that a client was
        # created, but it's possible to have multiple clients so that check
        # will need to be able to determine which clients weren't successfully
        # created.
        charmhelpers.core.hookenv.log(
            "create_client failed: {}".format(e), 'WARN')
        return

    if data is not None and ':' in data:
        client_key, client_secret = data.split(':')

        relids = charmhelpers.core.hookenv.relation_ids('oauth-server')
        unit = charmhelpers.core.hookenv.local_unit()
        for rid in relids:
            relation = charmhelpers.core.hookenv.relation_get(
                rid=rid, unit=unit,)
            if relation:
                charmhelpers.core.hookenv.relation_set(
                    rid, {
                        'client_data': json.dumps({
                            'client_key': client_key,
                            'client_secret': client_secret,
                        }),
                    })
    else:
        raise Exception("Couldn't parse create_client data: {}".format(data))


@hooks.hook('oauth-server-relation-joined')
@hooks.hook('oauth-server-relation-changed')
def oauth_server_relation_joined_changed():
    relation = charmhelpers.core.hookenv.relation_get()

    client_redirect_uri = relation.get('url')

    if client_redirect_uri is None:
        return

    juju_log("got client redirect url: {}".format(client_redirect_uri))
    config_changed()


@hooks.hook('pgsql-relation-joined')
@hooks.hook('pgsql-relation-changed')
def pgsql_relation_joined_changed():
    packages = ["python-psycopg2", "postgresql-client"]
    charmhelpers.fetch.apt_install(packages, fatal=True)

    relation = charmhelpers.core.hookenv.relation_get()
    with open(_pgsql_file(), 'w') as f:
        json.dump(relation, f)

    # some relation calls will just have "private-address". We need to ensure
    # we've gotten all the db credentials before trying to migrate
    if relation.get('user') and relation.get('host'):
        _db_migrate()

    config_changed()
    _wsgi_reload()


@hooks.hook()
def pgsql_relation_broken():
    settings = _pgsql_file()
    if os.path.exists(settings):
        os.unlink(settings)

    _wsgi_reload()


def _create_amqp_config():
    path = os.path.join(_service_dir(), 'amqp_config.py')
    relation = charmhelpers.core.hookenv.relation_get()
    if 'password' not in relation:
        return path  # not yet ready

    with tempfile.NamedTemporaryFile('w', delete=False) as f:
        f.write('# Automatically-generated by Juju. DO NOT EDIT MANUALLY.\n')
        f.write('AMQP_USER = "%s"\n'
                % charmhelpers.core.hookenv.config()['amqp-user'])
        f.write('AMQP_VHOST = "%s"\n'
                % charmhelpers.core.hookenv.config()['amqp-vhost'])
        f.write('AMQP_HOST = "%s"\n' % relation['private-address'])
        f.write('AMQP_PASSWORD = "%s"\n' % relation['password'])
    # Now that the file is complete (and closed) rename it to its final
    # destination. This avoids races where another process try to import that
    # file as a python module (creating a .pyc) and ends up having close enough
    # time stamps to fool python (which won't reload the completed file).
    os.rename(f.name, path)
    os.chmod(path, 0644)
    return path


@hooks.hook()
def amqp_relation_joined():
    config = charmhelpers.core.hookenv.config()
    charmhelpers.core.hookenv.relation_set(
        None, {
            'username': config['amqp-user'],
            'vhost': config['amqp-vhost'],
            'admin': config['amqp-admin'],
        })


@hooks.hook()
def amqp_relation_changed():
    if not os.path.exists(_create_amqp_config()):
        return
    _wsgi_reload()


@hooks.hook()
def amqp_relation_departed():
    os.unlink(os.path.join(_service_dir(), 'amqp_config.py'))
    _wsgi_reload()


if __name__ == "__main__":
    hooks.execute(sys.argv)
