# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import mock
import os
import unittest
import shutil
import tempfile

import yaml

import hooks
import intercom


HERE = os.path.abspath(os.path.dirname(__file__))

HTTP_REL = {
    'hostname': 'remote.com',
    'port': '8080',
}


WEBSITE_REL = {
    'website:1': {
        'private-address': '10.0.0.2',
    }
}


INFO_REL = {
    'private-address': '10.0.0.1',
}


PATCHES = [
    'charmhelpers.core.hookenv.close_port',
    'charmhelpers.core.hookenv.charm_dir',
    'charmhelpers.core.hookenv.config',
    'charmhelpers.core.hookenv.local_unit',
    'charmhelpers.core.hookenv.log',
    'charmhelpers.core.hookenv.open_port',
    'charmhelpers.core.hookenv.related_units',
    'charmhelpers.core.hookenv.relation_get',
    'charmhelpers.core.hookenv.relation_id',
    'charmhelpers.core.hookenv.relation_ids',
    'charmhelpers.core.hookenv.relation_set',
    'charmhelpers.core.hookenv.remote_unit',
    'charmhelpers.core.hookenv.unit_get',
    'charmhelpers.core.host.mkdir',
    'charmhelpers.core.host.service_restart',
    'charmhelpers.fetch.apt_install',
    'charmhelpers.fetch.configure_sources',
    'charmhelpers.fetch.install_remote',
    'charmhelpers.payload.execd.execd_preinstall',
    'intercom.log',
    'intercom.get_service_dir',
    'os.environ',
]


class RestishTestCase(unittest.TestCase):
    def setUp(self):
        """Setup a testing hook environment.

        Patches the required charmhelper and intercom objects and
        initializes few mocks for tests.
        """
        super(RestishTestCase, self).setUp()
        self.tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmpdir)

        for spec in PATCHES:
            attr_name = spec.split('.')[-1]
            _m = mock.patch(spec)
            setattr(self, attr_name, _m.start())
            self.addCleanup(_m.stop)

        self.local_unit.return_value = 'local-test'
        self.remote_unit.return_value = 'remote-test'
        self.relation_get.side_effect = (
            lambda rel_name: HTTP_REL.get(rel_name))
        self.relation_id.return_value = '1'
        self.relation_ids.side_effect = (
            lambda rel_id: ["{}:{}".format(rel_id, self.relation_id())])
        self.get_service_dir.return_value = self.tmpdir
        self.charm_dir.return_value = os.path.abspath(
            os.path.join(HERE, '..'))

        def env_get(name):
            return {
                'JUJU_UNIT_NAME': 'restish/0',
            }[name]
        self.environ.__getitem__.side_effect = env_get

        self.unit_get.return_value = '127.0.0.1'

        self._set_default_config()

    def _set_default_config(self):
        defaults = {}
        config = os.path.join(HERE, '../config.yaml')
        if os.path.exists(config):
            with open(config) as f:
                config = yaml.safe_load(f)
                for name, option in config['options'].iteritems():
                    defaults[name] = option.get('default')
        self.config.return_value = defaults


class TestInstallHook(RestishTestCase):
    def setUp(self):
        super(TestInstallHook, self).setUp()
        m = mock.patch('hooks._service_dir')
        m.start().return_value = self.tmpdir
        self.addCleanup(m.stop)
        os.mkdir(os.path.join(self.tmpdir, 'code'))
        self.install_remote.return_value = os.path.join(self.tmpdir, 'foo.tgz')
        os.mkdir(self.install_remote.return_value)
        revfile = os.path.join(
            self.install_remote.return_value, '.uploaded_version_info')
        with open(revfile, 'w') as f:
            f.write('branch@1234 - 2014-06-25 11:50:44 -0500')
        self.revno = 'r1234_foo'

    def test_framework_restish(self):
        local = ['r1_aa', 'r2_bb']
        self.config.return_value['current_code'] = 'http://bb.tgz'
        self.config.return_value['available_code'] = '''
            http://aa.tgz http://bb.tgz'''
        for x in local:
            os.mkdir(hooks._code_dir(x))
        hooks.handle_code()
        pkgs = self.apt_install.call_args_list[0][0][0]
        self.assertIn('python-restish', pkgs)

    def test_framework_pyramid(self):
        self.config.return_value['framework'] = 'pyramid'
        local = ['r1_aa', 'r2_bb']
        self.config.return_value['current_code'] = 'http://bb.tgz'
        self.config.return_value['available_code'] = '''
            http://aa.tgz http://bb.tgz'''
        for x in local:
            os.mkdir(hooks._code_dir(x))
        hooks.handle_code()
        pkgs = self.apt_install.call_args_list[0][0][0]
        self.assertIn('python3-pyramid', pkgs)

    @mock.patch('hooks._payload_extract')
    def test_handle_code_noop(self, payload_extract):
        local = ['r1_aa', 'r2_bb']
        self.config.return_value['current_code'] = 'http://bb.tgz'
        self.config.return_value['available_code'] = '''
            http://aa.tgz http://bb.tgz'''
        for x in local:
            os.mkdir(hooks._code_dir(x))
        hooks.handle_code()
        self.assertEqual(0, payload_extract.call_count)
        self.assertEqual('r2_bb', os.readlink(hooks._code_dir()))

    @mock.patch('hooks._payload_extract')
    def test_handle_code(self, payload_extract):
        self.config.return_value['current_code'] = 'http://bb.tgz'
        self.config.return_value['available_code'] = 'http://bb.tgz'
        payload_extract.side_effect = lambda x: os.mkdir(
            hooks._code_dir('r2_bb'))
        self.assertEqual(set([]), hooks.handle_code())
        self.assertEqual(1, payload_extract.call_count)
        self.assertEqual('r2_bb', os.readlink(hooks._code_dir()))


class TestConfigHooks(RestishTestCase):
    def setUp(self):
        super(TestConfigHooks, self).setUp()

        local_patches = (
            'hooks._service_dir',
            'hooks._wsgi_reload',
            'hooks.update_nrpe_config',
            'hooks._cron_root',
        )

        for spec in local_patches:
            attr_name = spec.split('.')[-1]
            _m = mock.patch(spec)
            setattr(self, attr_name, _m.start())
            self.addCleanup(_m.stop)

        self._service_dir.return_value = self.tmpdir
        cron_root = os.path.join(self.tmpdir, 'cron.d')
        os.mkdir(cron_root)
        self._cron_root.return_value = cron_root
        os.mkdir(os.path.join(self.tmpdir, 'code'))
        self.relation_get.side_effect = lambda: {'user': 'data', 'host': 'h'}

    @mock.patch('hooks._payload_extract')
    def test_cron_unconfigured(self, payload_extract):
        # If any of the cron_* configuration options are empty, existing
        # cron file is removed and cron is restarted.
        self.config.return_value['current_code'] = 'http://bb.tgz'
        self.config.return_value['available_code'] = 'http://bb.tgz'
        payload_extract.side_effect = lambda x: os.mkdir(
            hooks._code_dir('r2_bb'))
        cron_file_path = os.path.join(self._cron_root(), 'local-test')
        with open(cron_file_path, 'w') as fd:
            fd.write('removed')
        hooks.hooks.execute(['hooks/config-changed'])
        self.assertEqual([], os.listdir(self._cron_root()))
        self.service_restart.assert_called_once_with('cron')

    @mock.patch('hooks._payload_extract')
    def test_cron_good(self, payload_extract):
        # Once configured, cron job is installed.
        self.config.return_value['current_code'] = 'http://bb.tgz'
        self.config.return_value['available_code'] = 'http://bb.tgz'
        payload_extract.side_effect = lambda x: os.mkdir(
            hooks._code_dir('r2_bb'))
        self.config.return_value['cron_schedule'] = '*/10 * * * *'
        self.config.return_value['cron_cmd'] = 'foo.sh'
        self.config.return_value['wsgi_user'] = 'bar'
        hooks.hooks.execute(['hooks/config-changed'])
        self.assertEqual(['local-test'], os.listdir(self._cron_root()))
        code_dir = os.path.join(self.tmpdir, 'code')
        logs_dir = os.path.join(self.tmpdir, 'logs')
        with open(os.path.join(self._cron_root(), 'local-test')) as fd:
            cron_content = fd.read()
        expected_cron_line = (
            '*/10 * * * * bar cd {}/current && foo.sh >>{}/cron.log 2>&1')
        self.assertEqual(
            expected_cron_line.format(code_dir, logs_dir),
            cron_content.splitlines()[-1])
        # Cron-files must terminate with a newline (before EOF), be sure
        # our templating system is configured accordingly.
        self.assertTrue(cron_content.endswith('\n'))
        # System cron service is restarted.
        self.service_restart.assert_called_once_with('cron')

    @mock.patch('hooks._get_reverseproxies')
    def test_allowed_hosts(self, reverse_proxy):
        reverse_proxy.return_value = []
        hooks._set_allowed_hosts()
        self.assertTrue(
            os.path.exists(os.path.join(self.tmpdir, 'allowed_hosts.json')))


class TestWebsiteHooks(RestishTestCase):

    def setUp(self):
        super(TestWebsiteHooks, self).setUp()
        m = mock.patch('hooks._service_dir')
        m.start().return_value = self.tmpdir
        self.addCleanup(m.stop)
        self.reverseproxie_address = WEBSITE_REL[
            "website:{}".format(self.relation_id())]['private-address']
        self.relation_get.side_effect = (
            lambda rid, unit: WEBSITE_REL.get(rid))
        self.relation_ids.side_effect = (
            lambda rel_id: ["{}:{}".format(rel_id, self.relation_id())])
        self.related_units.return_value = ['a_unit']

    def test_reverseproxies(self):
        hooks.hooks.execute(['hooks/website-relation-changed'])
        self.assertTrue(
            os.path.exists(os.path.join(self.tmpdir, 'allowed_hosts.json')))
        content = open(os.path.join(self.tmpdir, 'allowed_hosts.json')).read()
        self.assertIn(self.reverseproxie_address, content)


class TestIntercomHooks(RestishTestCase):

    def get_intercom_content(self):
        """Utility for getting the stored relationships."""
        content = open(os.path.join(self.tmpdir, 'intercom.json')).read()
        return json.loads(content)

    def set_intercom_content(self, registry):
        """Utility for populating relationships."""
        content = json.dumps(registry)
        with open(os.path.join(self.tmpdir, 'intercom.json'), 'w') as fd:
            fd.write(content)

    def test_intercom_changed(self):
        # New 'intercom' relations are stored in the JSON registry
        # as <hostname>:<port>
        intercom.hooks.execute(['hooks/intercom-relation-changed'])
        self.log.assert_called_with(
            "Intercom unit 'remote-test' (remote.com) joined.")
        self.assertEquals(
            {'1': {'hostname': 'remote.com',
                   'port': '8080',
                   'unit': 'remote-test'}},
            self.get_intercom_content())

    def test_intercom_broken(self):
        # Broken 'intercom' relations are removed from the JSON registry.
        # 'remote_unit' and any relation are not available for 'broken'.
        self.remote_unit.return_value = ''
        self.relation_get.side_effect = lambda rel_name: ''
        # Create an initial intercom.json content.
        self.set_intercom_content(
            {'1': 'to be gone',
             '2': {'hostname': '10.0.0.2',
                   'port': '9000',
                   'unit': 'preserved'}})
        intercom.hooks.execute(['hooks/intercom-relation-broken'])
        self.log.assert_called_with("Intercom unit '' () departed.")
        self.assertEquals(
            {'2': {'hostname': '10.0.0.2',
                   'port': '9000',
                   'unit': 'preserved'}},
            self.get_intercom_content())


class TestIntercomServiceDir(unittest.TestCase):

    @mock.patch('os.environ')
    def test_intercom_get_service_dir(self, mocked_env):
        # Intercom get_service_dir() utility, returns a path in /srv
        # derived from the unit name. No content replacement.
        from intercom import get_service_dir

        def env_get(name):
            return {
                'JUJU_UNIT_NAME': 'ci-airline-gatekeeper-restish/0',
            }[name]
        mocked_env.__getitem__.side_effect = env_get

        self.assertEqual(
            '/srv/ci-airline-gatekeeper-restish', get_service_dir())


class TestWSGIHooks(RestishTestCase):

    def test_wsgi_joined(self):
        hooks.hooks.execute(['hooks/wsgi-relation-joined'])
        self.open_port.assert_called_once_with(8080)
        expected = {
            'python_path': '',
            'working_dir': '/srv/local-test/code/current',
            'port': 8080,
            'wsgi_access_logfile': '/srv/local-test/logs/access.log',
            'wsgi_log_file': '/srv/local-test/logs/gunicorn.log',
        }
        settings = self.relation_set.call_args_list[0][1]['relation_settings']
        self.assertEqual(expected, settings)
        self.mkdir.assert_called_once_with(
            '/srv/local-test/logs', 'www-data', 'www-data', 0755)

    def test_wsgi_broken(self):
        hooks.hooks.execute(['hooks/wsgi-relation-broken'])
        self.close_port.assert_called_once_with(8080)


class TestJSONStatusHooks(RestishTestCase):

    def test_json_status_joined(self):
        self.config.return_value['json_status_path'] = 'foo'
        hooks.hooks.execute(['hooks/json_status-relation-joined'])
        self.assertIn('status-url', self.relation_set.call_args_list[0][0][1])


class TestAMQPHooks(RestishTestCase):
    def setUp(self):
        super(TestAMQPHooks, self).setUp()
        m = mock.patch('hooks._service_dir')
        m.start().return_value = self.tmpdir
        self.addCleanup(m.stop)

        m = mock.patch('hooks._wsgi_reload')
        self.wsgi_reload = m.start()
        self.addCleanup(m.stop)

        self.relation_get.side_effect = lambda: {
            'password': 'foo',
            'private-address': '1.1.1.1',
        }

    def test_joined(self):
        hooks.hooks.execute(['hooks/amqp-relation-joined'])
        self.assertEqual(1, self.relation_set.call_count)

    def test_changed(self):
        hooks.hooks.execute(['hooks/amqp-relation-changed'])
        self.wsgi_reload.assert_called_once_with()
        self.assertTrue(
            os.path.exists(os.path.join(self.tmpdir, 'amqp_config.py')))

    def test_departed(self):
        with open(os.path.join(self.tmpdir, 'amqp_config.py'), 'w') as f:
            f.write('')
        hooks.hooks.execute(['hooks/amqp-relation-departed'])
        self.wsgi_reload.assert_called_once_with()


class TestPGSQLHooks(RestishTestCase):
    def setUp(self):
        super(TestPGSQLHooks, self).setUp()
        m = mock.patch('hooks._service_dir')
        m.start().return_value = self.tmpdir
        self.addCleanup(m.stop)

        m = mock.patch('hooks._wsgi_reload')
        self.wsgi_reload = m.start()
        self.addCleanup(m.stop)

        mock.patch('hooks.update_nrpe_config').start()

        self.relation_get.side_effect = lambda: {'user': 'data', 'host': 'h'}

    @mock.patch('hooks.handle_code')
    def test_pgsql_joined(self, h_code):
        cdir = os.path.join(self.tmpdir, 'code/current')
        os.makedirs(cdir)
        hooks.hooks.execute(['hooks/pgsql-relation-joined'])
        self.assertTrue(
            os.path.exists(os.path.join(self.tmpdir, 'pgsql.json')))
        self.assertEqual(2, self.wsgi_reload.call_count)

    def _do_migration(self, hook, runs=True):
        cdir = os.path.join(self.tmpdir, 'code/current')
        os.makedirs(cdir)
        with open(os.path.join(cdir, 'tmp.sh'), 'w') as f:
            f.write('rm $0   # just delete ourself')
            f.flush()
        self.config.return_value['db_migration_cmd'] = 'dash ./tmp.sh'
        hooks.hooks.execute([hook])
        if runs:
            self.assertFalse(os.path.exists(os.path.join(cdir, 'tmp.sh')))
        else:
            self.assertTrue(os.path.exists(os.path.join(cdir, 'tmp.sh')))

    @mock.patch('hooks.config_changed')
    @mock.patch('hooks.handle_code')
    def test_pgsql_joined_migration(self, h_code, c_changed):
        self._do_migration('hooks/pgsql-relation-joined')
        self.assertEqual(1, self.wsgi_reload.call_count)

    @mock.patch('hooks.handle_code')
    def test_migration_on_config_nochange(self, h_code):
        self.relation_get.side_effect = lambda: {
            'not_userd': 'data',
            'not_hostd': 'h'
        }
        self._do_migration('hooks/config-changed', False)

    @mock.patch('hooks.handle_code')
    def test_migration_on_config_change(self, h_code):
        with open(os.path.join(self.tmpdir, 'pgsql.json'), 'w') as f:
            f.write('{}')
        self._do_migration('hooks/config-changed')

    def test_pgsql_broken(self):
        hooks.hooks.execute(['hooks/pgsql-relation-broken'])
        self.assertEqual(1, self.wsgi_reload.call_count)
