# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import socket
import time

from amqplib import client_0_8 as amqp

BSBUILDER_QUEUE = 'bsbuilder'
IMAGE_BUILDER_QUEUE = 'imagebuilder'
TEST_RUNNER_QUEUE = 'test_runner'
PUBLISHER_QUEUE = 'publisher'
LANDER_QUEUE = 'lander'
PPA_CREATOR_QUEUE = 'ppacreator'
VALIDATOR_QUEUE = 'validator'
COVERAGE_RETRIEVER_QUEUE = 'coverage_retriever'

MONITORED_QUEUES = [
    PPA_CREATOR_QUEUE,
    BSBUILDER_QUEUE,
    IMAGE_BUILDER_QUEUE,
    TEST_RUNNER_QUEUE,
    PUBLISHER_QUEUE,
    LANDER_QUEUE,
    PPA_CREATOR_QUEUE,
    VALIDATOR_QUEUE,
]

log = logging.getLogger(__name__)


HERE = os.path.abspath(os.path.dirname(__file__))


def get_config():
    '''Load the rabbit config created by the charm'''
    config = None
    try:
        f = os.path.join(HERE, '../../../../amqp_config.py')
        # TODO remove once all charms have converted to new code layout
        if not os.path.exists(f):
            f = os.path.join(HERE, '../../amqp_config.py')
        if os.path.exists(f):
            import imp
            config = imp.load_source('amqp_config', f)
        else:
            log.warn('No amqp_config found at: %s' % os.path.abspath(f))
    except:
        log.exception('ERROR detecting rabbit args')
    return config


def connection(rabbit_config):
    return amqp.Connection(
        userid=rabbit_config.AMQP_USER,
        virtual_host=rabbit_config.AMQP_VHOST,
        host=rabbit_config.AMQP_HOST,
        password=rabbit_config.AMQP_PASSWORD
    )


def declare_queue(queue_name, config=None):
    if config is None:
        config = get_config()
    conn = channel = None
    try:
        conn = connection(config)
        channel = conn.channel()
        channel.queue_declare(queue=queue_name, durable=True,
                              auto_delete=False)
    except Exception as e:
        if channel:
            channel.close()
        if conn:
            conn.close()
        raise e
    return conn, channel


def send(queue_name, msg, raise_errors=False):
    '''send a message to the given queue.

    :param raise_errors: There are two users of this function. Some users are
    sending non-critical messages in which errors can safely be ignored. Other
    users need to catch the exception when this function fails.
    '''
    config = get_config()
    if not config:
        return 'rabbitmq settings not available.'
    conn = channel = None
    try:
        conn, channel = declare_queue(queue_name, config)
        body = amqp.Message(msg)
        body.properties['delivery_mode'] = 2  # Persistent
        channel.basic_publish(body, exchange='', routing_key=queue_name)
    except Exception as e:
        if raise_errors:
            raise
        logging.exception('unable to queue up message: %s', e)
        return str(e)
    finally:
        if channel:
            channel.close()
        if conn:
            conn.close()
    return None


def _run_forever(channel, queue, callback, retry_period=120):
    logging.info('Waiting for messages. ^C to exit.')
    tag = channel.basic_consume(callback=callback, queue=queue)
    try:
        timeout = time.time()
        while channel.callbacks and time.time() < timeout + retry_period:
            try:
                channel.wait()
                timeout = time.time()
            except (amqp.AMQPConnectionException, socket.error):
                logging.error('lost connection to Rabbit')
                # TODO metrics.meter('lost_rabbit_connection')
                # Don't probe immediately, give the network/process
                # time to come back.
                time.sleep(0.1)
        if channel.callbacks:
            logging.error('Rabbit did not reappear quickly enough.')
    except KeyboardInterrupt:
        pass
    finally:
        if channel and channel.callbacks and channel.is_open:
            channel.basic_cancel(tag)


def process_queue(config, queue_name, callback, delete=False):
    conn = channel = None
    try:
        conn, channel = declare_queue(queue_name, config)
        # We want a 'Fair dispatch' (see
        # https://www.rabbitmq.com/tutorials/tutorial-two-python.html). Rabbit
        # dispatch messsages as they arrive in the queue. That is, all workers
        # prefetch the messages in a round-robin fashion irrelevant of whether
        # they process them immediately. That's good to achieve the lowest
        # possible latency as far as messages are concerned but if they are not
        # processed immediately, they stay in the worker queues. If new workers
        # are added or finish processing their messages, the server has no more
        # messages to prefetch. Therefore, each worker should prefetch a single
        # message in order to receive a new one when it can process it. This
        # provides an overall better flow across the workers.
        channel.basic_qos(prefetch_size=0, prefetch_count=1, a_global=False)
        _run_forever(channel, queue_name, callback)
    finally:
        if channel:
            if delete:
                channel.queue_delete(queue_name)
            channel.close()
        if conn:
            conn.close()


def progress_waiting(progress_trigger, data=None):
    if data is None:
        data = {}
    data['state'] = 'WAITING'
    return send(progress_trigger, json.dumps(data))


_ratelimit = {'trigger': None, 'timestamp': 0, 'delay': 5}


def progress_update(progress_trigger, data=None, ratelimit=False):
    if ratelimit:
        now = time.time()
        if _ratelimit['trigger'] == progress_trigger:
            if now < _ratelimit['timestamp'] + _ratelimit['delay']:
                # too soon
                return
            else:
                # we'll allow the message to be logged, but we should back
                # off a little more before showing the same thing:
                if _ratelimit['delay'] < 20 * 60:  # max throttle = 20 min
                    _ratelimit['delay'] *= 2
        _ratelimit['trigger'] = progress_trigger
        _ratelimit['timestamp'] = now

    if data is None:
        data = {}
    data['state'] = 'STATUS'
    return send(progress_trigger, json.dumps(data))


def progress_completed(progress_trigger, data=None):
    if data is None:
        data = {}
    data['state'] = 'COMPLETED'
    data['result'] = 'PASSED'
    data['exit'] = True
    return send(progress_trigger, json.dumps(data))


def progress_failed(progress_trigger, data=None):
    if data is None:
        data = {}
    data['state'] = 'COMPLETED'
    data['result'] = 'FAILED'
    data['exit'] = True
    return send(progress_trigger, json.dumps(data))
