# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

from swiftclient import client
from swiftclient.client import ClientException


class DataStoreException(Exception):
    pass


def create_for_ticket(ticket_id, auth_config, public=True):
    return DataStore('ticket-{}'.format(ticket_id), auth_config, public=public)


class DataStore(object):
    """Store and retrieve files into a swift container."""

    def __init__(self, component, auth_config, public=False):
        self.container_id = component
        self.auth_config = auth_config
        self.public = public
        self.client = None

    @staticmethod
    def validate_auth_config(auth_config):
        required_fields = ['auth_user', 'auth_password', 'auth_tenant_name',
                           'auth_url', 'auth_region']

        missing_fields = []

        for field in required_fields:
            if field not in auth_config.keys():
                missing_fields.append(field)

        if len(missing_fields) > 0:
            raise DataStoreException(
                "missing fields: {}".format(', '.join(missing_fields))
            )

    def ensure_swift_client(self):
        if self.client is not None:
            return self.client

        self.validate_auth_config(self.auth_config)
        config = self.auth_config
        self.client = client.Connection(
            authurl=config.get('auth_url'),
            user=config.get('auth_user'),
            key=config.get('auth_password'),
            os_options={'tenant_name': config.get('auth_tenant_name'),
                        'region_name': config.get('auth_region')},
            auth_version=('/v1.' in config.get('auth_url') and '1.0' or '2.0'))
        try:
            self._create_container(self.container_id)
        except client.ClientException:
            logging.exception('Unable to connect to swift')
            raise DataStoreException("Missing or invalid authentication info.")
        self.change_visibility(public=self.public)
        self.container_url = "{}/{}".format(self.client.url, self.container_id)

    def list_files(self):
        self.ensure_swift_client()
        _, objects = self.client.get_container(self.container_id)
        return [x['name'] for x in objects]

    def put_file(self, filename, contents, content_type=None):
        self.ensure_swift_client()
        name = self.file_name(filename)
        try:
            self.client.put_object(self.container_id, obj=name,
                                   contents=contents,
                                   content_type=content_type)
        except ClientException as e:
            raise DataStoreException(
                "Failed to upload file: {}, Error: {}".format(filename, e)
            )

        return self.file_path(filename)

    def get_file(self, filename):
        self.ensure_swift_client()
        name = self.file_name(filename)

        contents = None
        try:
            hdr, contents = self.client.get_object(self.container_id, obj=name)
        except ClientException as e:
            raise DataStoreException(
                "Failed to get file: {}, Error: {}".format(filename, e)
            )
        return contents

    def change_visibility(self, public=False):
        self.ensure_swift_client()
        if public:
            read_acl = '.r:*'
        else:
            read_acl = ''
        self.client.post_container(self.container_id,
                                   {'X-Container-Read': read_acl})
        self.public = public

    def delete_file(self, filename):
        self.ensure_swift_client()
        name = self.file_name(filename)
        try:
            self.client.delete_object(self.container_id, obj=name)
        except ClientException as e:
            raise DataStoreException(
                "Failed to delete file: {}, Error: {}".format(filename, e)
            )

    def clear(self):
        self.ensure_swift_client()
        try:
            files = self.client.get_container(self.container_id)[1]

            for f in files:
                try:
                    self.client.delete_object(self.container_id,
                                              obj=f['name'])
                except ClientException as e:
                    raise DataStoreException(
                        "Failed to delete file: {}".format(e)
                    )
        except ClientException as e:
            raise DataStoreException(
                "Failed to get container: {}".format(e)
            )

    def delete(self, recursive=False):
        self.ensure_swift_client()
        try:
            if recursive:
                self.clear()

            self.client.delete_container(self.container_id)
        except ClientException as e:
            raise DataStoreException(
                "Failed to delete container: {}".format(e)
            )

    def file_name(self, filename):
        """Returns the file name that is used inside the data store.

        :param filename: The file name used to build the object name inside the
            container.
        """
        return os.path.basename(filename)

    def file_path(self, filename):
        name = self.file_name(filename)

        return "{}/{}".format(self.container_url, name)

    def _create_container(self, container):
        self.client.put_container(self.container_id)
