# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""Gnupg Environment handler."""
from __future__ import unicode_literals

import atexit
import gnupg
import os
import shutil
import tempfile
import urllib2

from ci_utils import unit_config


class GnupgEnvironmentError(Exception):
    """Raised where there was an error during the GNUPGHOME setup."""


def load_keyring_uri(keyring_uri):
    """Return the contents of the specified keyring URI

    It currently supports local and remote URI loading by the
    default `urllib2` loaders.

    Swift resources could be read via the swift http/s frontend
    if stored as public. Although, native swift loading can be
    easily implemented if needed.
    """
    try:
        contents = urllib2.urlopen(keyring_uri).read()
    except (ValueError, urllib2.URLError) as err:
        raise GnupgEnvironmentError(str(err))
    return contents


class GnupgEnvironment(object):
    """Gnupg environment handler.

    Provides safe isolation and restoring of GNUPG environment and
    proxies keyring actions through `gnupg` module.

    Also works as context manager (with GnupgEnvironment(): something())
    """

    def __init__(self, prefix='ci-gnupg-'):
        self.prefix = prefix
        self.home = None
        self._env = None

    def _setupHome(self):
        """Configure the isolated GNUPGHOME according to unit_config.

        If 'keyring_uri' is specified it will load and import the
        configured keyring.
        If 'keyserver_url' is specified it will be configured to
        'auto-retrive' needed keys from the configured keyserver.
        """
        # Load related unit_config option, by design (yaml templates)
        # they will be empty if not used, instead of simply absent,
        # thus "if <conf_value>:" and not "if <conf_value> is not None"
        config = unit_config.get_auth_config()
        keyring_uri = config.get('keyring_uri')
        keyserver_url = config.get('keyserver_url')

        if keyring_uri:
            # Load and import the configured keyring.
            keyring_data = load_keyring_uri(keyring_uri)
            self.importKeyring(keyring_data)

        if keyserver_url:
            # set needed GPG options, 'auto-key-retrieve' is necessary for
            # automatically retrieve from the keyserver unknown key when
            # verifying signatures and 'no-auto-check-trustdb' avoid wasting
            # time verifying the local keyring consistence.
            confpath = os.path.join(self.home, 'gpg.conf')
            with open(confpath, 'w') as conf:
                conf.write(
                    'keyserver hkp://{}\n'
                    'keyserver-options auto-key-retrieve\n'
                    'no-auto-check-trustdb\n'.format(keyserver_url))

    def isolate(self, context_manager=False):
        """Isolate GNUPG environment.

        If 'context_manager' is False (default) the isolated environment
        has to be explicitly removed using restore() or rely on a `atexit`
        task for removing the GNUPG home and restoring previous environment.
        Otherwise, this is used as a context manager with will automaticaly
        call restore() when exiting the context.

        Also see `_setupHome` for its configurations.
        """
        self._env = os.environ.get('GNUPGHOME')
        self.home = tempfile.mkdtemp(prefix=self.prefix)
        self._setupHome()
        os.environ['GNUPGHOME'] = self.home
        if not context_manager:
            atexit.register(self.restore)

    def restore(self):
        """Restore previous GNUPG environment."""
        if self.home is not None and os.path.exists(self.home):
            shutil.rmtree(self.home)
        self.home = None
        if self._env is not None:
            os.environ['GNUPGHOME'] = self._env
        else:
            if os.environ.get('GNUPGHOME'):
                del os.environ['GNUPGHOME']

    def __enter__(self):
        """Isolates GNUPGHOME."""
        self.isolate(context_manager=True)
        return self

    def __exit__(self, exc, value, tb):
        """Restore previous GNUPGHOME."""
        self.restore()

    @property
    def gnupg(self):
        """The initialized gnupg for the current context."""
        return gnupg.GPG(gnupghome=self.home)

    def importKeyring(self, keyring_data):
        """Import the given keyring data into the isolated keyring."""
        return self.gnupg.import_keys(keyring_data)

    def listKeys(self, *args, **kwargs):
        """Returns a list of matching keys."""
        return self.gnupg.list_keys(*args, **kwargs)
