# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import urlparse

import glanceclient
import keystoneclient.v2_0.client as ksclient


class ImageStoreException(Exception):
    pass


class ImageStore(object):

    ImageStoreException = ImageStoreException

    def __init__(self, auth_config):
        GLANCE_API_VERSION = '1'
        self._setup_auth(auth_config)
        region = auth_config.get('glance_auth_region')
        endpoints = self.keystone_client.service_catalog.get_endpoints(
            service_type='image', endpoint_type='publicURL')
        glance_endpoint = [
            x.get('publicURL') for x in endpoints.get('image')
            if x.get('region') == region
        ]
        if not glance_endpoint:
            raise ImageStoreException(
                'No valid glance endpoints found for the specified region. '
                'Check that the cloud credentials specified are valid')
        # Otherwise, there should be only one valid endpoint
        # Strip the version path in the url
        urlbits = urlparse.urlparse(glance_endpoint[0])
        self.glance_endpoint = urlbits.scheme + '://' + urlbits.netloc
        self.glance = glanceclient.Client(
            GLANCE_API_VERSION, self.glance_endpoint,
            token=self.keystone_client.auth_token)

    def _setup_auth(self, config):
        try:
            self.keystone_client = ksclient.Client(
                auth_url=config['glance_auth_url'],
                username=config['glance_auth_user'],
                password=config['glance_auth_password'],
                tenant_name=config['glance_auth_tenant_name']
            )
        except KeyError as exc:
            raise ImageStoreException(
                "Missing authentication info: {}".format(exc.message))
        except (ksclient.exceptions.Unauthorized,
                ksclient.exceptions.AuthorizationFailure) as exc:
            logging.exception('Unable to connect keystone')
            raise ImageStoreException(
                "Unable to contact keystone in {} ({})".format(
                    config['auth_url'], exc.message))

    def image_create(self, filename, image, is_public=False,
                     disk_format='qcow2', container_format='bare'):
        with open(image) as imgdata:
            try:
                image = self.glance.images.create(
                    name=filename,
                    is_public=is_public,
                    disk_format=disk_format,
                    container_format=container_format,
                    data=imgdata)
            except glanceclient.exc.HTTPException as exc:
                raise ImageStoreException(
                    "Unable to create an image in {} ({})".format(
                        self.glance.endpoint, str(exc)))
        return image

    def image_delete(self, image_id):
        self.glance.images.delete(image_id)


# the images referred to below are currently created manually as described in
# ../../../britney_proxy/README.images
# During review, it was mention that we want to take the distribution (debian,
# ubuntu-rtm) into account in the name space.
def uci_image_name(domain, series, architecture):
    """Returns an image name.

    Some images are uploaded to glance for uci-engine specific needs.

    :param domain: 'cloudimg' or 'britney'.

    :param series: The ubuntu series (precise, utopic, etc).

    :param architecture: The processor architecture ('amd64', i386, etc).
    """
    if domain not in ('cloudimg', 'britney'):
        raise ValueError('Invalid image domain')
    return 'uci/{}/{}-{}.img'.format(domain, series, architecture)
