# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import BaseHTTPServer
import json
import logging
import os
import threading
import urllib2
import urlparse
import time

HERE = os.path.abspath(os.path.dirname(__file__))
LAST_REVISION = os.path.join(HERE, '../../.uploaded_version_info')

# We have a few checks that take a while. The implementations for them use
# caching logic. This gives us a consistent place to manage this value.
CACHE_TIMEOUT = 60


class JSONStatus(object):
    def __init__(self):
        self.results = []

    def _add_result(self, name, value, status):
        self.results.append({
            'label': name,
            'value': value,
            'status': status,
        })

    def add_okay(self, name, value):
        self._add_result(name, value, 'okay')

    def add_fail(self, name, value):
        self._add_result(name, value, 'fail')

    def add_true_false(self, name, value, condition):
        '''adds an 'okay' result if condition is true, 'fail' otherwise'''
        if condition:
            self.add_okay(name, value)
        else:
            self.add_fail(name, value)

    def add_deployed_revision(self):
        if os.path.exists(LAST_REVISION):
            with open(LAST_REVISION) as f:
                self.add_okay('revision', f.read())
        else:
            self.add_fail('revision', 'Missing revision information')

    def add_url_test(self, name, url):
        try:
            urllib2.urlopen(url).read()
            self.add_okay(name, 'connected')
        except Exception as e:
            self.add_fail(name, str(e))

    def add_rabbit_configured(self):
        '''adds a result status for status of rabbit-configuration'''
        from ci_utils import amqp_utils
        if amqp_utils.get_config() is not None:
            self.add_okay('rabbit configured', True)
        else:
            self.add_fail('rabbit configured', False)

    def add_rabbit_worker_health(self, queue):
        '''checks if there are any workers attached to the queue.'''
        from ci_utils import amqp_utils
        config = amqp_utils.get_config()
        if config is None:
            # we already report rabbit isn't configured, so no sense adding
            # another failure for this
            return

        mgr = urllib2.HTTPPasswordMgrWithDefaultRealm()
        base = 'http://{}:55672'.format(config.AMQP_HOST)
        # NOTE: juju leaves this guest/guest and doesn't seem to expose
        # anything in the relation to override it
        mgr.add_password(None, base, 'guest', 'guest')
        handler = urllib2.HTTPBasicAuthHandler(mgr)

        try:
            resp = urllib2.build_opener(handler).open(base + '/api/queues')
            content = resp.read()
            data = json.loads(content)
            consumers = None
            for q in data:
                if q['name'] == queue:
                    consumers = q['consumers']
                    break
            if consumers is None:
                self.add_fail('workers-online', 'no queue defined')
            else:
                self.add_true_false('workers-online', consumers, consumers > 0)
        except:
            self.add_fail('workers-online', 'unable to check')

    def add_data_store_health(self, ticket=1):
        from ci_utils import unit_config, data_store
        try:
            ds = data_store.create_for_ticket(
                ticket, unit_config.get_auth_config())
            ds.ensure_swift_client()
            self.add_okay('data-store', 'connected')
            return ds
        except Exception as e:
            self.add_fail('data-store', str(e))

    def add_image_store_health(self):
        from ci_utils import unit_config, image_store
        try:
            imgstore = image_store.ImageStore(unit_config.get_auth_config())
            # create and delete an image
            imgstore.glance.images.create(
                name='json_status test', data='json_status', disk_format='raw',
                container_format="bare").delete()
            self.add_okay('image-store', 'connected')
        except Exception as e:
            self.add_fail('image-store', str(e))

    def _gatekeeper_test(self, ticket):
        import requests
        uploads = [
            os.path.join(HERE, 'testing/data/foobar_0.1-1_source.changes'),
            os.path.join(HERE, 'testing/data/foobar_0.1-1.dsc'),
            os.path.join(HERE, 'testing/data/foobar_0.1.orig.tar.gz'),
            os.path.join(HERE, 'testing/data/foobar_0.1-1.debian.tar.gz'),
        ]
        base = 'http://localhost:8080'
        create_url = base + '/api/v1/ticket/%s/sandbox/' % ticket

        # create sandbox
        resp = requests.post(create_url)
        if resp.status_code != 201:
            raise Exception('Unable to create sandbox: ' + resp.content)
        sandbox_url = base + resp.content

        # upload files to sandbox
        for f in uploads:
            dest_url = sandbox_url + '/' + os.path.basename(f)
            resp = requests.put(dest_url)
            if resp.status_code != 200:
                raise Exception('Unable to get tempurl: ' + resp.content)

            resp = requests.put(resp.content, data=open(f))
            if resp.status_code != 201:
                raise Exception('Unable to upload ' + f + ': ' + resp.content)

        # validate files
        resp = requests.post(sandbox_url)
        if resp.status_code != 201:
            raise Exception('Unable to validate sandbox: ' + resp.content)

    def add_gatekeeper_test(self):
        ticket = 'gk-test-%d' % time.time()
        ds = self.add_data_store_health(ticket)
        if ds:
            try:
                self._gatekeeper_test(ticket)
                self.add_okay('gatekeeper-upload', True)
            except Exception as e:
                logging.exception('failed to check upload')
                self.add_fail('gatekeeper-upload', e.message)
            finally:
                ds.delete(True)  # remove the test container

    def add_signing_key_health(self):
        """Add signing-key health-check.

        See `bsbuilder.upload_package.get_signing_keyring`.
        """
        from bsbuilder.upload_package import get_signing_keyring
        try:
            keyring_data, key_data = get_signing_keyring()
            key_label = '{} ({})'.format(
                key_data['uids'][0], key_data['fingerprint'])
            self.add_okay('signing-key', key_label)
        except Exception as exp:
            self.add_fail('signing-key', str(exp))

    def add_de_virt_ppa_capabilities(self):
        from ci_utils import unit_config, launchpad
        try:
            lp = launchpad.lp_login()
        except Exception as exp:
            self.add_fail('lp-login', str(exp))

        self.add_okay('login to launchpad', "logged in")

        user_name = unit_config.get('launchpad_user')
        ppa_admins = 'launchpad-ppa-self-admins'
        if lp.people[user_name] in lp.people[ppa_admins].members_details:
            self.add_okay('ppa-devirtualization', "possible")
        else:
            self.add_fail("ppa-devirtualization", "system is not capable")


class StatusHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    # must be static because this object is create for each request
    _cache = (None, 0)

    def __init__(self, *args, **kwargs):
        BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def _get_cached_response(self):
        query = urlparse.urlparse(self.path).query
        query = urlparse.parse_qs(query)

        resp, timeout = StatusHandler._cache
        if query.get('refresh', ['0'])[0] == '1' or time.time() > timeout:
            resp = json.dumps(self.health_cb().results)
            StatusHandler._cache = (resp, time.time() + CACHE_TIMEOUT)

        if 'jsonp' in query.get('format', []):
            cb = query.get('callback', ['cb'])[0]
            resp = cb + '(' + resp + ')'

        return resp

    def do_GET(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/json')
        self.end_headers()
        resp = self._get_cached_response()
        self.wfile.write(resp)


def run_server(health_cb, port=8080, threaded=True):
    httpd = BaseHTTPServer.HTTPServer(('0.0.0.0', port), StatusHandler)
    httpd.RequestHandlerClass.health_cb = health_cb
    if threaded:
        t = threading.Thread(target=httpd.serve_forever)
        t.daemon = True
        t.start()
        return httpd
    else:
        httpd.serve_forever()
