# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import sys

import setuptools
from setuptools.package_index import PackageIndex


HERE = os.path.abspath(os.path.dirname(__file__))
# A branch with all the ez-install deps in a single directory, ie:
#   lp:~canonical-ci-engineering/uci-engine/deps
PINNED_BRANCH = os.environ.get('CI_DEPS_BRANCH', '')
_deps = os.path.join(HERE, '../../.deps')
PINNED_LOCAL = os.environ.get('CI_DEPS_LOCAL', os.path.abspath(_deps))


class HackPI(PackageIndex):
    '''Ensure we download our dependencies from where we want

    We want to be able to pin down the location we download our easy-install
    dependencies from. This ensures we pull them from the place we desire.
    '''
    branch_updated = False

    def __init__(self, *args, **kwargs):
        super(HackPI, self).__init__(*args, **kwargs)
        if len(PINNED_BRANCH) and not HackPI.branch_updated:
            self.info('CI_DEPS_BRANCH({}) requested at {}'.format(
                PINNED_BRANCH, PINNED_LOCAL))
            if os.path.exists(PINNED_LOCAL):
                self.info('local deps exist, checking for updates')
                subprocess.check_call(['bzr', 'pull'], cwd=PINNED_LOCAL)
            else:
                self.info('Creating new branch for deps')
                subprocess.check_call(
                    ['bzr', 'branch', PINNED_BRANCH, PINNED_LOCAL])
            HackPI.branch_updated = True

    def _download_url(self, scheme, url, tmpdir):
        err = '_download_url should not be called with CI_DEPS_BRANCH enabled'
        if len(PINNED_BRANCH):
            raise RuntimeError(err)
        return super(HackPI, self)._download_url(scheme, url, tmpdir)

    def process_url(self, url, retrieve=False):
        if not len(PINNED_BRANCH):
            return super(HackPI, self).process_url(url, retrieve)
        self.process_filename(PINNED_LOCAL)


# setuptools is built upon the monkey-patching (probably because it has
# to override distutils in ways it wasn't designed for). So to hack
# setuptools we have to monkey patch it.
import setuptools.command.easy_install
setuptools.command.easy_install.easy_install.create_index = HackPI


def _local_pip_file(dep, ver):
    HackPI()  # ensure our local deps are ready
    for d in os.listdir(PINNED_LOCAL):
        if d.startswith('%s-%s' % (dep, ver)):
            return os.path.join(PINNED_LOCAL, d)
    raise RuntimeError('Missing dependency: %s' % d)


def _handle_openstack():
    # this stuff is required because the pbr library has a bug:
    #  https://bitbucket.org/pypa/setuptools/issue/73/
    # in addition the openstack packages like python-glanceclient
    # fail because they use pbr which does dirty things to setuptools
    # We do a no-deps option to try and keep what we need for local
    # development down to a minimum
    output = subprocess.check_output(['pip', 'freeze'])
    installed = [x.split('==')[0] for x in output.split('\n')]
    pips = [
        'Babel==1.3',
        'pbr==0.6',
        'iso8601==0.1.8',
        'prettytable==0.7.2',
        'python-glanceclient==0.12.0',
        'python-keystoneclient==0.6.0',
        'python-neutronclient==2.3.4',
        'python-novaclient==2.15.0',
        'python-swiftclient==1.8.0',
        'requests==2.2.1',
        'six==1.5.2',
    ]
    for p in pips:
        dep, ver = p.split('==')
        if dep in installed:
            continue
        if len(PINNED_BRANCH):
            print('using local copy of pip dependency')
            p = _local_pip_file(dep, ver)
        subprocess.check_call(['pip', 'install', '--no-deps', p])


def setup(name, test_suite, install_requires, **kwargs):
    '''A simple helper for our components to all call setuptools properly'''

    # ensure find_packages works if our current directory isn't the component
    component_dir = os.path.abspath(os.path.join(HERE, '..', '..', name))
    if os.path.exists(component_dir):
        os.chdir(component_dir)
    else:
        # We're probably in the component dir (like when we're called from the
        # Debian package build).
        component_dir = os.getcwd()

    if len(sys.argv) > 1 and sys.argv[1] in ('test', 'develop'):
        if kwargs.get('requires_openstack'):
            del kwargs['requires_openstack']
            _handle_openstack()

    args = {
        'name': name,
        'version': '0.1',
        'description': '%s component of Ubuntu CI Engine' % name,
        'author': 'Canonical CI Engineering Team',
        'license': 'AGPL',
        'packages': setuptools.find_packages(component_dir),
        'test_suite': test_suite,
        'install_requires': install_requires,
    }
    for key, val in kwargs.iteritems():
        args[key] = val
    return setuptools.setup(**args)
