
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import os
import time


HERE = os.path.abspath(os.path.dirname(__file__))
TXSTATSD_ENABLED = True

try:
    from txstatsd.client import UdpStatsDClient
    from txstatsd.metrics.extendedmetrics import ExtendedMetrics
except ImportError as e:
    TXSTATSD_ENABLED = False

log = logging.getLogger(__name__)


def get_config():
    '''Load the stats config created by the txstatsd charm'''
    config = None
    try:
        # try and find the config file, should be in the root of our bzr branch
        f = os.path.join(HERE, '../../stats_config.py')
        if os.path.exists(f):
            import imp
            config = imp.load_source('stats_config', f)
        else:
            log.warn('No stats_config found at: %s' % os.path.abspath(f))
    except:
        log.exception('ERROR detecting stats args')
    return config


_config = get_config()


class HasNotStarted(Exception):
    pass


class HasNotStopped(Exception):
    pass


class BadName(Exception):
    pass


class Metric(object):
    defaults = dict(
        namespace="ci-engine",
        prefix="misc",
        port=8125,
    )

    def __init__(self, name, **kwargs):
        global TXSTATSD_ENABLED

        txstatsd_address = _config.TXSTATSD_ADDRESS if _config else None

        self.name = name
        self.server = kwargs.get('server', txstatsd_address)

        # if there isn't a server specified we can't proceed.
        if not self.server:
            TXSTATSD_ENABLED = False

        self.port = kwargs.get('port', self.defaults['port'])
        self.prefix = kwargs.get('prefix', self.defaults['prefix'])
        self.namespace = '.'.join([self.defaults['namespace'], self.prefix])

        self.start_time = time.time()
        self.duration = None
        self.lap_duration = None
        self.lap_start_time = self.start_time

    def stop(self):

        if not self.start_time:
            raise HasNotStarted()

        self.duration = time.time() - self.start_time

        self._publish()

    def lap(self, name):

        current_time = time.time()
        self.lap_duration = current_time - self.lap_start_time
        self.lap_start_time = current_time

        if not self.duration:
            self.duration = self.lap_duration

        self._publish(name=name, duration=self.lap_duration)

    def _publish(self, **kwargs):
        """
        Publish metrics to txstatsd.
        """

        name = kwargs.get('name', self.name)
        duration = kwargs.get('duration', self.duration)

        if not name:
            raise BadName(name)

        if not duration:
            raise HasNotStopped

        if not TXSTATSD_ENABLED:
            return

        statsd_client = UdpStatsDClient(self.server, self.port)
        metrics = ExtendedMetrics(
            connection=statsd_client,
            namespace=self.namespace,
        )
        statsd_client.connect()
        metrics.timing(name, duration)

    def republish(self, name=None):
        """
        Re-publish an already stopped metric with an optional new name.
        """

        self._publish(name)
