# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import errno
import os
import subprocess

import swiftclient
import yaml

from novaclient import exceptions as nova_exceptions
from novaclient.v1_1 import client as nova_client

from ucitests import features

from ci_utils import unit_config

# Provide a shorcut so users don't have to import from ucitests
requires = features.requires


from ci_utils import juju


class Swift(features.Feature):

    def _probe(self):
        conn = self.get_connection()
        if conn is None:
            return False
        try:
            conn.head_account()
        except swiftclient.exceptions.ClientException:
            return False
        return True

    def get_connection(self):
        try:
            auth_ver = '/v1.' in os.environ['OS_AUTH_URL'] and '1.0' or '2.0'
            return swiftclient.client.Connection(
                authurl=os.environ['OS_AUTH_URL'],
                user=os.environ['OS_USERNAME'],
                key=os.environ['OS_PASSWORD'],
                os_options={'tenant_name': os.environ['OS_TENANT_NAME'],
                            'region_name': os.environ['OS_REGION_NAME']},
                auth_version=auth_ver)
        except KeyError:
            # If we miss some env vars, we can't connect
            return None


# The single instance shared by all tests
swift = Swift()


class NovaCompute(features.Feature):

    def _probe(self):
        client = self.get_client()
        if client is None:
            return False
        try:
            # can transiently fail with requests.exceptions.ConnectionError
            # (converted from MaxRetryError).
            client.authenticate()
        except nova_exceptions.ClientException:
            return False
        return True

    def get_client(self):
        try:
            return nova_client.Client(
                os.environ['GLANCE_OS_USERNAME'],
                os.environ['GLANCE_OS_PASSWORD'],
                project_id=os.environ['GLANCE_OS_TENANT_NAME'],
                auth_url=os.environ['GLANCE_OS_AUTH_URL'],
                region_name=os.environ['GLANCE_OS_REGION_NAME'],
                service_type='compute')
        except KeyError:
            # If we miss some env vars, we can't get a client
            return None
        except nova_exceptions.Unauthorized:
            # If the credentials are wrong, we can't get a client
            return None


# The single instance shared by all tests
nova_compute = NovaCompute()


class NovaCredentials(features.Feature):

    def __init__(self):
        super(NovaCredentials, self).__init__()

    def _probe(self):
        try:
            unit_config.get_auth_config()
        except IOError as e:
            if e.errno == errno.ENOENT:
                return False
            raise
        # We don't attempt to validate the credentials for now
        return True

    def feature_name(self):
        return 'Valid nova credentials file'


# The single instance shared by all tests
nova_creds = NovaCredentials()


class BootstrappedJuju(features.Feature):

    def __init__(self):
        super(BootstrappedJuju, self).__init__()
        self.env = None
        self.swift = swift
        self.nova = nova_compute

    def _probe_local(self):
        # We can connect to swift
        if not self.swift.available():
            return False

        try:
            subprocess.check_output(['juju', 'status'], stderr=subprocess.PIPE)
        except:
            return False

        return True

    def _probe_openstack(self):
        # There is a bucket
        control_bucket_id = self.get_control_bucket_id()
        if control_bucket_id is None:
            return False
        # We can connect to swift
        if not self.swift.available():
            return False
        # juju put a provider-state file in its control bucket
        provider_state = self.get_provider_state()
        if provider_state is None:
            return False
        # which contains the nova id of the state server node
        nova_id = self.get_nova_id(provider_state)
        if nova_id is None:
            return False
        if not self.nova.available():
            return False
        # if it's still alive, get the nova instance
        instance = self.get_nova_instance(nova_id)
        if instance is None:
            return False
        nova_ip = self.get_nova_ip(instance)
        if nova_ip is None:
            return False
        return True

    def _probe(self):
        # There is a defined juju env
        env_name = juju.environment_name()
        if env_name is None:
            return False
        self.env = juju.environment(env_name)
        if self.env is None:
            return False
        env_type = self.env.get('type', None)
        if env_type is None:
            return False
        if env_type == 'local':
            return self._probe_local()
        elif env_type == 'openstack':
            return self._probe_openstack()

        # Unknown provider: we don't know how to check for a bootstrap node
        return False

    def get_control_bucket_id(self):
        try:
            return self.env['control-bucket']
        except (KeyError, TypeError):
            # The yaml file is malformed or incomplete
            return None

    def get_provider_state(self):
        container_id = self.get_control_bucket_id()
        conn = self.swift.get_connection()
        try:
            # get_object return a tuple of (metadata, data)
            provider_state = conn.get_object(container_id, 'provider-state')[1]
            return yaml.safe_load(provider_state)
        except swiftclient.exceptions.ClientException:
            return None

    def get_nova_id(self, provider_state):
        try:
            return provider_state['state-instances'][0]
        except (TypeError, KeyError, IndexError):
            # The yaml file is malformed or incomplete
            return None

    def get_nova_instance(self, nova_id):
        try:
            return self.nova.get_client().servers.get(nova_id)
        except nova_exceptions.NotFound:
            return None

    def get_nova_ip(self, nova_instance):
        networks = nova_instance.networks.values()
        if not networks:
            return None
        # We use a single network and don't care how it's named
        net = networks[0]
        if not net:
            return None
        # Either we have a single address or the last one has been
        # added to make the instance reachable (floating or public).
        return net[-1]


# The single instance shared by all tests
bootstrapped_juju = BootstrappedJuju()
