# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import os

from ucitests import fixtures

from ci_utils import data_store
from ci_utils.json_status import LAST_REVISION


# A data store defines a namespace for objects. The namespace is rooted at the
# container id and objects are identified by a path inside the container.
# It defines an API to store and retrieve objects


class FakeDataStore(data_store.DataStore):
    """A fake data store to avoid requiring a swift server.

    This stores data on disk and as such should be
    isolated. ucitests.fixtures.set_uniq_cwd() provides such isolation and
    cleanup. It's the test responsibility to properly isolate itself.
    """

    def __init__(self, component, auth_config=None, public=False):
        super(FakeDataStore, self).__init__(
            # We use the container id as our backend directory and just store
            # files there.
            os.path.abspath(component),
            # We don't use the config so we inject a fake but valid one
            {
                'auth_url': 'http://example.com',
                'auth_user': 'user',
                'auth_password': 'pass',
                'auth_tenant_name': 'tenant',
                'auth_region': 'region',
            },
            public=public
        )
        # The following call is unguarded to fail if one attempts to create the
        # same data store twice.
        os.mkdir(self.container_id)

    def list_files(self):
        return sorted(os.listdir(self.container_id))

    def put_file(self, filename, contents, content_type=None):
        if os.sep in filename:
            base_dir = os.path.join(self.container_id,
                                    os.path.dirname(filename))
            try:
                # Try to create needed dirs
                os.makedirs(base_dir)
            except OSError as e:
                # They are already there, no worries
                if e.errno != errno.EEXIST:
                    raise
        full_path = os.path.join(self.container_id, filename)
        with open(full_path, 'w') as f:
            f.write(contents)
        return self.file_path(filename)

    def get_file(self, filename):
        full_path = os.path.join(self.container_id, filename)
        with open(full_path) as f:
            return f.read()

    def delete_file(self, filename):
        full_path = os.path.join(self.container_id, filename)
        os.remove(full_path)

    def clear(self):
        for f in self.list_files():
            self.delete_file(f)

    def delete(self, recursive=False):
        if recursive:
            self.clear()
        os.rmdir(self.container_id)

    def file_path(self, filename):
        return '{}{}/{}'.format('file://', self.container_id, filename)


class FakeTaskQueue(object):

    def __init__(self, queue_name, config=None):
        self.msgs = []
        self.queue_name = queue_name
        # We don't use the config which contains the credentials to access a
        # real rabbit server.

    def publish(self, msg):
        self.msgs.append(msg)

    def subscribe(self, func, *args, **kwargs):
        returns = []
        while not self.is_empty():
            ret = func(self.msgs.pop(0), *args, **kwargs)
            returns.append(ret)
        return returns

    def is_empty(self):
        return not self.msgs


def override_last_revision(test, content):
    """Test fixture for overriding the branch revision information.

    :param test: `TestCase` context used for adding the original content
                 restoring task.
    :param content: revision text for tests.
    """
    if os.path.exists(LAST_REVISION):
        with open(LAST_REVISION) as fd:
            original_content = fd.read()
    else:
        original_content = None
    with open(LAST_REVISION, 'w') as fd:
        fd.write(content)

    def restore():
        if original_content is None:
            os.remove(LAST_REVISION)
            return
        with open(LAST_REVISION, 'w') as fd:
            fd.write(original_content)

    test.addCleanup(restore)


def isolate_from_juju_env(test):
    """Isolate a test from juju environment definitions.

    This is usually called in setUp for tests that want to modify the juju
    environement definitions and restore them after the test is run.

    :param test: A test instance.
    """
    # Juju environments are defined in ${JUJU_HOME}/environments.yaml and can
    # be specified via ${JUJU_ENV}
    fixtures.set_uniq_cwd(test)
    fixtures.isolate_from_env(
        test, dict(HOME=test.uniq_dir, JUJU_HOME='.', JUJU_ENV=None))
