# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import socket
import time
import unittest

import mock

from ci_utils import amqp_utils


class TestAMQP(unittest.TestCase):
    @mock.patch('ci_utils.amqp_utils.connection')
    @mock.patch('ci_utils.amqp_utils.get_config')
    def testConnectFailed(self, get_config, connect):
        '''Ensure a failed queue connection returns an HTTP 503 error'''
        get_config.return_value = mock.Mock()
        error = 'mocked test exception'
        connect.side_effect = RuntimeError(error)
        r = amqp_utils.send('fake_queue', 'fake_message')
        self.assertIsNotNone(r)
        self.assertEqual(error, r)

    @mock.patch('ci_utils.amqp_utils.connection')
    @mock.patch('ci_utils.amqp_utils.get_config')
    def testSent(self, get_config, connect):
        '''Test a successful send returns nothing

        There's not much you can test in isolation here, but this gives
        a bit of a sanity check.
        '''
        get_config.return_value = mock.Mock()
        connect.return_value = mock.Mock()
        r = amqp_utils.send('fake_queue', 'fake_message')
        self.assertIsNone(r)

        connect.side_effect = RuntimeError('testSent')
        r = amqp_utils.send('fake_queue', 'fake message')
        self.assertIsNotNone(r)

        with self.assertRaises(RuntimeError):
            amqp_utils.send('fake_queue', 'fake message', True)

    @mock.patch('ci_utils.amqp_utils._run_forever')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testProcessQueue(self, connection, run_forever):
        '''Ensure we close the connection if something fails'''
        conn = mock.Mock()
        connection.return_value = conn
        run_forever.side_effect = RuntimeError
        with self.assertRaises(RuntimeError):
            amqp_utils.process_queue(None, None, None)
        self.assertTrue(conn.close.called)

    def testRunForever(self):
        '''Ensure this times out after the right amount of time'''
        channel = mock.Mock()
        callback = mock.Mock()
        channel.wait.side_effect = socket.error()
        start = time.time()
        retry_period = 0.1
        amqp_utils._run_forever(channel, 'foo', callback, retry_period)
        # see if it was within 1/10 of a second of the retry_period
        self.assertAlmostEqual(start + retry_period, time.time(), places=1)


class TestProgressTrigger(unittest.TestCase):
    @mock.patch('ci_utils.amqp_utils.send')
    def testProgress(self, send):
        trigger = 'foo_trigger'
        amqp_utils.progress_waiting(trigger)
        send.assert_called_with(trigger, json.dumps({'state': 'WAITING'}))

        send.reset_mock()
        params = {'foo': 'bar'}
        amqp_utils.progress_completed(trigger, params)
        params['state'] = 'COMPLETED'
        params['result'] = 'PASSED'
        params['exit'] = True
        send.assert_called_with(trigger, json.dumps(params))

        send.reset_mock()
        params = {'foo': 'bar'}
        amqp_utils.progress_failed(trigger, params)
        params['state'] = 'COMPLETED'
        params['result'] = 'FAILED'
        params['exit'] = True
        send.assert_called_with(trigger, json.dumps(params))
