# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import subprocess
import time
import unittest

import mock

from ci_utils import amqp_worker


class _worker(amqp_worker.AMQPWorker):
    '''Simple class to help mock the datastore api used in the worker.'''

    def __init__(self, amqp_utils, retval, exception=None, error_msg=None):
        super(_worker, self).__init__('queue-name')
        self.store = mock.Mock()
        if amqp_utils:
            self.amqp_utils = amqp_utils
        self.retval = retval
        self.exception = exception
        self.error_msg = error_msg

    def _create_data_store(self, ticket):
        # Always use our own store so the tests can assert on it
        return self.store

    def handle_request(self, params, log):
        log.info('called')
        self.params = params
        if self.exception:
            raise self.exception(self.error_msg)
        if self.error_msg:
            return self.amqp_utils.progress_failed, self.retval
        if params.get('cancel_url'):
            time.sleep(0.1)
            self.retval['cancelled'] = params['cancelled']
        return self.amqp_utils.progress_completed, self.retval


class TestAMQPWorker(unittest.TestCase):
    @mock.patch('ci_utils.amqp_worker.amqp_utils')
    def testOnMessageSimple(self, amqp_utils):
        '''Test simple on message logic'''
        msg = mock.Mock()
        msg.delivery_tag = 'foo'
        msg.body = json.dumps({
            'ticket_id': 1,
            'progress_trigger': 'queue-name',
            'param1': 'foo',
            'param2': 42,
        })
        retval = {'msg': 'this is the worker return value'}
        worker = _worker(amqp_utils, retval)
        worker._on_message(msg)
        self.assertEqual(worker.params['param1'], 'foo')
        self.assertEqual(worker.params['param2'], 42)
        self.assertEqual(1, amqp_utils.progress_update.call_count)
        msg.channel.basic_ack.assert_called_once_with('foo')
        self.assertEqual(0, amqp_utils.progress_failed.call_count)
        self.assertEqual(1, worker.store.put_file.call_count)
        amqp_utils.progress_completed.assert_called_once_with(
            'queue-name', retval)
        # assert_called_once_with works, but doesn't check the values
        # added to retval
        retval = amqp_utils.progress_completed.call_args[0][1]
        self.assertEqual(1, len(retval['artifacts']))
        self.assertEqual('queue-name.output.log',
                         retval['artifacts'][0]['name'])

    @mock.patch('ci_utils.amqp_worker.amqp_utils')
    def testOnMessageFail(self, amqp_utils):
        '''Test on message logic for a failure handled by the worker'''
        msg = mock.Mock()
        msg.delivery_tag = 'foo'
        msg.body = json.dumps({
            'ticket_id': 1,
            'progress_trigger': 'queue-name',
            'param1': 'foo',
            'param2': 42,
        })
        retval = {'msg': 'this is the worker return value'}
        worker = _worker(amqp_utils, retval, None, True)
        worker._on_message(msg)
        self.assertEqual(worker.params['param1'], 'foo')
        self.assertEqual(worker.params['param2'], 42)
        self.assertEqual(1, amqp_utils.progress_update.call_count)
        msg.channel.basic_ack.assert_called_once_with('foo')
        amqp_utils.progress_failed.assert_called_once_with(
            'queue-name', retval)
        self.assertEqual(0, amqp_utils.progress_completed.call_count)
        # assert_called_once_with works, but doesn't check the values
        # added to retval
        retval = amqp_utils.progress_failed.call_args[0][1]
        self.assertEqual(1, len(retval['artifacts']))
        self.assertEqual('queue-name.output.log',
                         retval['artifacts'][0]['name'])

    @mock.patch('ci_utils.amqp_worker.amqp_utils')
    def testOnMessageUnexpected(self, amqp_utils):
        '''Test on message logic for an unexpected failure'''
        msg = mock.Mock()
        msg.delivery_tag = 'foo'
        msg.body = json.dumps({
            'ticket_id': 1,
            'progress_trigger': 'queue-name',
        })
        err = 'this is the worker error'
        worker = _worker(None, None, RuntimeError, err)
        worker._on_message(msg)
        self.assertEqual(1, amqp_utils.progress_update.call_count)
        msg.channel.basic_ack.assert_called_once_with('foo')
        self.assertEqual(1, amqp_utils.progress_failed.call_count)
        self.assertEqual(
            'queue-name', amqp_utils.progress_failed.call_args[0][0])
        self.assertIn(
            err, amqp_utils.progress_failed.call_args[0][1]['traceback'])
        self.assertEqual(1, worker.store.put_file.call_count)

    @mock.patch('ci_utils.amqp_worker.amqp_utils')
    def testOnMessageCalledProcessError(self, amqp_utils):
        '''Test on message logic nicely describes failed subprocess calls'''
        msg = mock.Mock()
        msg.delivery_tag = 'foo'
        msg.body = json.dumps({
            'ticket_id': 1,
            'progress_trigger': 'queue-name',
        })
        err = 'this is the worker error'

        class _calledprocess_worker(_worker):
            rc = 1234
            cmd = ['/bin/foo', 'bar']
            output = 'i can\'t believe its not butter'

            def handle_request(self, params, log):
                raise subprocess.CalledProcessError(
                    self.rc, self.cmd, self.output)

        worker = _calledprocess_worker(None, None, None, None)
        worker._on_message(msg)

        self.assertEqual(1, amqp_utils.progress_update.call_count)
        msg.channel.basic_ack.assert_called_once_with('foo')
        self.assertEqual(1, amqp_utils.progress_failed.call_count)
        self.assertEqual(
            'queue-name', amqp_utils.progress_failed.call_args[0][0])
        self.assertIn('CalledProcessError',
                      amqp_utils.progress_failed.call_args[0][1]['traceback'])
        err = amqp_utils.progress_failed.call_args[0][1]['error_message']
        self.assertIn(str(worker.cmd), err)
        self.assertIn(str(worker.output), err)
        self.assertEqual(1, worker.store.put_file.call_count)

    @mock.patch('ci_utils.amqp_worker.amqp_utils')
    def testOnMessageKilled(self, amqp_utils):
        '''Test on message logic for handling ctrl-c (upstart stopping)'''
        msg = mock.Mock()
        msg.delivery_tag = 'foo'
        msg.body = json.dumps({
            'ticket_id': 1,
            'progress_trigger': 'queue-name',
        })
        worker = _worker(None, None, KeyboardInterrupt)
        with self.assertRaises(KeyboardInterrupt):
            worker._on_message(msg)
        self.assertEqual(1, amqp_utils.progress_update.call_count)
        msg.channel.basic_ack.assert_called_once_with('foo')
        self.assertEqual(1, amqp_utils.progress_failed.call_count)
        self.assertEqual(
            'queue-name', amqp_utils.progress_failed.call_args[0][0])
        self.assertIn('KeyboardInterrupt',
                      amqp_utils.progress_failed.call_args[0][1]['traceback'])
        self.assertEqual(1, worker.store.put_file.call_count)

    @mock.patch('ci_utils.amqp_worker.amqp_utils')
    def testNoTicket(self, amqp_utils):
        '''Ensure we can gracefully deal with a bad message in the queue'''
        msg = mock.Mock()
        msg.delivery_tag = 'foo'
        msg.body = json.dumps({
            'progress_trigger': 'queue-name',
        })
        worker = _worker(amqp_utils, 1)
        worker._on_message(msg)
        msg.channel.basic_ack.assert_called_once_with('foo')
        self.assertEqual(1, amqp_utils.progress_failed.call_count)
        self.assertEqual(
            'queue-name', amqp_utils.progress_failed.call_args[0][0])

    @mock.patch('ci_utils.amqp_worker.amqp_utils')
    def testNoQueue(self, amqp_utils):
        '''Ensure we can gracefully deal with a bad message in the queue'''
        msg = mock.Mock()
        msg.delivery_tag = 'foo'
        msg.body = json.dumps({
        })
        worker = _worker(amqp_utils, 1)
        worker._on_message(msg)
        # all we can check is that the message got acked
        msg.channel.basic_ack.assert_called_once_with('foo')

    @mock.patch('ci_utils.amqp_worker.amqp_utils')
    @mock.patch('ci_utils.amqp_worker._get')
    def testCancel(self, get, amqp_utils):
        '''Ensure workers get the "cancel" message'''
        msg = mock.Mock()
        msg.delivery_tag = 'foo'
        msg.body = json.dumps({
            'ticket_id': 1,
            'progress_trigger': 'queue-name',
            'param1': 'foo',
            'param2': 42,
            'cancel_url': 'foo'
        })
        get.return_value = {'building': False}
        retval = {'msg': 'this is the worker return value'}
        worker = _worker(amqp_utils, retval)
        worker.cancel_interval = 0.01
        worker._on_message(msg)
        msg.channel.basic_ack.assert_called_once_with('foo')
        amqp_utils.progress_completed.assert_called_once_with(
            'queue-name', retval)
        retval = amqp_utils.progress_completed.call_args[0][1]
        self.assertTrue(retval['cancelled'])

    def testSaveLastRun(self):
        results = {'foo': 'bar'}
        worker = _worker(None, None)
        worker.save_last_check(results)
        self.addCleanup(os.unlink, worker.last_check_file)
        check = worker.get_last_check()
        self.assertIn('last_run', check)
        self.assertEqual(results['foo'], check['foo'])


class TestTimer(unittest.TestCase):
    def testCanCancel(self):
        def cb():
            raise RuntimeError('cancel failed')
        t = amqp_worker._timer(1, cb)
        t.start()
        t.cancel()

    def testCBRuns(self):
        def cb():
            self.run = True
        self.run = False
        t = amqp_worker._timer(0.01, cb)
        t.start()
        time.sleep(.02)
        t.cancel()
        self.assertTrue(self.run)
