# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest


from ucitests import fixtures as uci_fixtures
import yaml


from ci_utils import juju
from ci_utils.testing import (
    features,
    fixtures,
)


class TestSwift(unittest.TestCase):

    def test_no_env_vars_no_swift(self):
        uci_fixtures.isolate_from_env(self, dict(OS_AUTH_URL=None,
                                                 OS_USERNAME=None,
                                                 OS_PASSWORD=None,
                                                 OS_TENANT_NAME=None,
                                                 OS_REGION_NAME=None))
        swift = features.Swift()
        self.assertFalse(swift.available())


class TestNovaCompute(unittest.TestCase):

    def test_no_env_vars_no_nova(self):
        uci_fixtures.isolate_from_env(self, dict(GLANCE_OS_USERNAME=None,
                                                 GLANCE_OS_PASSWORD=None,
                                                 GLANCE_OS_TENANT_NAME=None,
                                                 GLANCE_OS_AUTH_URL=None,
                                                 GLANCE_OS_REGION_NAME=None))
        nova = features.NovaCompute()
        self.assertFalse(nova.available())


class TestJujuGetEnvsFromYaml(unittest.TestCase):

    def setUp(self):
        super(TestJujuGetEnvsFromYaml, self).setUp()
        fixtures.isolate_from_juju_env(self)
        self.bsj = features.BootstrappedJuju()

    def test_broken_environments_file(self):
        # a juju file that defines no environment
        uci_fixtures.build_tree('''file: environments.yaml
default: foo
''')
        self.assertFalse(self.bsj.available())

    def test_inconsistent_environments_file(self):
        # a juju file that defines a default environment that doesn't exist
        uci_fixtures.build_tree('''file: environments.yaml
default: foo
environments:
    bar:
        type: local
''')
        self.assertFalse(self.bsj.available())


class TestJujuControlBucket(unittest.TestCase):

    def setUp(self):
        super(TestJujuControlBucket, self).setUp()
        fixtures.isolate_from_juju_env(self)
        self.bsj = features.BootstrappedJuju()

    def test_no_control_bucket(self):
        uci_fixtures.build_tree('''file: environments.yaml
default: foo
environments:
    foo:
''')
        self.bsj.env = juju.environment(juju.environment_name())
        self.assertEqual(None, self.bsj.get_control_bucket_id())


class TestProviderType(unittest.TestCase):

    def setUp(self):
        super(TestProviderType, self).setUp()
        fixtures.isolate_from_juju_env(self)
        # We don't want to test the specific providers, we only care about
        # finding their type. So we patch the type specific checks to be always
        # successful and test available() from a give environments.yaml file.

        def probe_true(feature):
            return True

        uci_fixtures.patch(self, features.BootstrappedJuju,
                           '_probe_local', probe_true)
        uci_fixtures.patch(self, features.BootstrappedJuju,
                           '_probe_openstack', probe_true)
        self.bsj = features.BootstrappedJuju()

    def test_no_env(self):
        self.assertFalse(self.bsj.available())

    def test_local_provider(self):
        uci_fixtures.build_tree('''file: environments.yaml
default: bar
environments: {bar: {type: local}}
''')
        self.assertTrue(self.bsj.available())

    def test_openstack_provider(self):
        uci_fixtures.build_tree('''file: environments.yaml
default: bar
environments: {bar: {type: openstack}}
''')
        self.assertTrue(self.bsj.available())

    def test_unknown_provider(self):
        uci_fixtures.build_tree('''file: environments.yaml
default: bar
environments: {bar: {type: vaporware}}
''')
        self.assertFalse(self.bsj.available())


class TestJujuNova(unittest.TestCase):

    def setUp(self):
        super(TestJujuNova, self).setUp()
        self.bsj = features.BootstrappedJuju()

    def test_get_nova_id_well_formed(self):
        provider_state = yaml.safe_load('''state-instances:
- a8c77d73-55b5-4b21-8c43-bf5f60ad5ecb
characteristics:
- arch: amd64
  mem: 1024
  rootdisk: 10240
  cpucores: 1
''')
        self.assertEqual('a8c77d73-55b5-4b21-8c43-bf5f60ad5ecb',
                         self.bsj.get_nova_id(provider_state))

    def test_get_nova_id_malformed(self):
        self.assertEqual(None, self.bsj.get_nova_id(None))
        self.assertEqual(None,
                         self.bsj.get_nova_id({'state-instances': []}))

    def assertIP(self, expected, networks):

        class Nova(object):

            def __init__(self, networks):
                self.networks = networks

        nova = Nova(dict(a_name=networks))
        self.assertEqual(expected, self.bsj.get_nova_ip(nova))

    def test_nova_ip_malformed(self):
        self.assertIP(None, None)
        self.assertIP(None, [])

    def test_nova_ip_well_formed(self):
        self.assertIP('15.125.112.72', ['192.168.0.1', '15.125.112.72'])
