# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from dput.changes import (
    Changes,
    ChangesFileException,
)

from ci_utils.file_handler import (
    ChangesProcessor,
    ChangesFileNotFound,
    ChangesParseError,
    ChangesValidationError,
    FileToUploadNotFound,
    UploadDirNotFound
)
from ci_utils import testing


class ChangesProcessorTestCase(testing.TestCaseWithGnupg):

    def setUp(self):
        super(ChangesProcessorTestCase, self).setUp()
        self.create_test_info()

    def create_test_info(self, source_name="foobar", source_version="0.1",
                         architecture="any", packaging_version="1"):
        self.source_name = source_name
        self.source_version = source_version
        self.architecture = architecture
        self.packaging_version = packaging_version
        self.full_version = "%s-%s" % (self.source_version,
                                       self.packaging_version)
        self.changes_file = (
            "%s_%s-%s_source.changes" % (self.source_name, self.source_version,
                                         self.packaging_version))
        self.dsc_file = (
            "%s_%s-%s.dsc" % (self.source_name, self.source_version,
                              self.packaging_version))
        self.debian_tar_gz_file = (
            "%s_%s-%s.debian.tar.gz" % (self.source_name, self.source_version,
                                        self.packaging_version))
        self.orig_file = "%s_%s.orig.tar.gz" % (self.source_name,
                                                self.source_version)

    def test_parse(self):
        tf_path = testing.get_test_file_path
        processor = ChangesProcessor(tf_path(self.changes_file))
        processor.parse()
        self.assertIsInstance(processor.changes, Changes)
        self.assertEqual(processor.changes.get_package_name(),
                         self.source_name)

        self.assertEqual(processor.source_package_name, self.source_name)
        self.assertEqual(processor.source_version, self.full_version)

        self.assertEqual(tf_path(self.dsc_file), processor.changes.get_dsc())
        self.assertEqual(tf_path(self.debian_tar_gz_file),
                         processor.changes.get_diff())
        self.assertTrue(tf_path(self.orig_file)
                        in processor.changes.get_files())

    def test_parse_changes_file_not_found(self):
        with self.assertRaises(ChangesFileNotFound):
            ChangesProcessor("barbaz_source.changes")

    def test_parse_changes_file_is_a_dsc(self):
        # We're only able to tell a file is a real .changes during parsing.
        processor = ChangesProcessor(
            testing.get_test_file_path("foobar_0.1-1_dsc.changes"))
        with self.assertRaises(ChangesParseError):
            processor.parse()

    def test_parse_changes_file_is_something_else(self):
        # We're only able to tell a file is a real .changes during parsing.
        processor = ChangesProcessor(
            testing.get_test_file_path("foobar_0.1-1_random.changes"))
        with self.assertRaises(ChangesFileException):
            processor.parse()

    def test_upload_dir_not_found(self):
        with self.assertRaises(UploadDirNotFound):
            ChangesProcessor(testing.get_test_file_path(self.changes_file),
                             "foo/bar/baz")

    def test_files_to_upload_dir_found(self):
        processor = ChangesProcessor(
            testing.get_test_file_path(self.changes_file),
            files_to_upload_dir=testing.TEST_DATA_ROOT)
        self.assertEquals(testing.TEST_DATA_ROOT,
                          processor.files_to_upload_dir)

    def test_file_to_upload_not_found(self):
        processor = ChangesProcessor(
            testing.get_test_file_path(self.changes_file),
            files_to_upload_dir="/tmp")
        self.assertRaises(FileToUploadNotFound, processor.parse)

    def test_all_files_found(self):
        processor = ChangesProcessor(
            testing.get_test_file_path(self.changes_file))
        self.assertEqual(0, len(processor.files_to_upload))
        processor.parse()
        # The three files (dsc, debian.tar.gz and orig) plus changes file
        self.assertEqual(4, len(processor.files_to_upload))
        processor.validate()
        self.assertTrue(
            processor.changes.get_dsc() in processor.files_to_upload)
        self.assertTrue(
            processor.changes.get_diff() in processor.files_to_upload)
        self.assertTrue(
            testing.get_test_file_path(self.orig_file) in
            processor.files_to_upload)

    def test_files_to_upload_dir_not_provided(self):
        processor = ChangesProcessor(
            testing.get_test_file_path(self.changes_file))
        self.assertEqual(testing.TEST_DATA_ROOT, processor.files_to_upload_dir)

    def test_signature_check_fails(self):
        processor = ChangesProcessor(
            testing.get_test_file_path("foobar_0.1-1_unsigned.changes"))
        processor.parse()
        self.assertRaises(ChangesValidationError, processor.validate)

    def test_validate_distribution_unreleased(self):
        processor = ChangesProcessor(
            testing.get_test_file_path("foobar_0.1-1_unreleased.changes"))
        processor.parse()
        with self.assertRaises(ChangesValidationError):
            processor.validate()

    def test_validate_distribution_unstable(self):
        # Changesfiles targeted to 'unstable' (debian) do not pass
        # validation because they won't be accepted in LP PPAs.
        processor = ChangesProcessor(
            testing.get_test_file_path("foobar_0.1-1_unstable.changes"))
        processor.parse()
        with self.assertRaises(ChangesValidationError):
            processor.validate()

    def test_validate_series_unsupported(self):
        processor = ChangesProcessor(
            testing.get_test_file_path(
                "foobar-unsupported_0.1-1ubuntu1_source.changes"))
        processor.parse()
        with self.assertRaises(ChangesValidationError):
            processor.validate()

    def test_process(self):
        # Simple test to guarantee minimum integrity of this method, as the
        # parts are being tested elsewhere.
        processor = ChangesProcessor(
            testing.get_test_file_path(self.changes_file))
        processor.process()
        self.assertIsInstance(processor.changes, Changes)
        self.assertEqual(processor.changes.get_package_name(),
                         self.source_name)

        self.assertEqual(4, len(processor.files_to_upload))
