# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import os
import unittest

from ucitests import fixtures as uci_fixtures

from ci_utils.testing import fixtures


class TestFakeDataStore(unittest.TestCase):
    """Test the FakeDataStore implementation.

    Since the target is tests themselves focusing on happy paths, no failures
    are tested here.
    """

    def setUp(self):
        super(TestFakeDataStore, self).setUp()
        uci_fixtures.set_uniq_cwd(self)
        self.ds_path = os.path.join(self.uniq_dir, 'data_store')
        self.ds = fixtures.FakeDataStore(self.ds_path, None)

    def test_put_get_file(self):
        self.assertEqual([], self.ds.list_files())
        url = self.ds.put_file('foo', 'foo content\n')
        self.assertTrue(url.startswith('file:///'))
        self.assertTrue(url.endswith(os.path.join(self.ds_path, 'foo')))
        self.assertEqual(['foo'], self.ds.list_files())
        self.assertEqual('foo content\n', self.ds.get_file('foo'))

    def test_put_file_with_slashes(self):
        url = self.ds.put_file('foo/bar/baz', 'baz content\n')
        self.assertTrue(url.endswith(os.path.join(self.ds_path,
                                                  'foo/bar/baz')))

    def test_multiple_put_file_with_slashes(self):
        url_baz = self.ds.put_file('foo/bar/baz', 'baz content\n')
        self.assertTrue(url_baz.endswith(os.path.join(self.ds_path,
                                                      'foo/bar/baz')))
        # Creating a file higher in the hierarchy also succeeds
        url_qux = self.ds.put_file('foo/qux', 'qux content\n')
        self.assertTrue(url_qux.endswith(os.path.join(self.ds_path,
                                                      'foo/qux')))
        # So does creating a file lower in the hierarchy
        url_foo = self.ds.put_file('foo/bar/quux/foo', 'qux content\n')
        self.assertTrue(url_foo.endswith(os.path.join(self.ds_path,
                                                      'foo/bar/quux/foo')))

    def test_list_files_empty(self):
        self.assertEqual([], self.ds.list_files())

    def test_list_files_some_files(self):
        self.ds.put_file('foo', 'foo content\n')
        self.ds.put_file('bar', 'bar content\n')
        self.assertEqual(sorted(['foo', 'bar']), self.ds.list_files())

    def test_delete_file(self):
        self.ds.put_file('foo', 'foo content\n')
        self.assertEqual(['foo'], self.ds.list_files())
        self.ds.delete_file('foo')
        self.assertEqual([], self.ds.list_files())

    def test_clear(self):
        self.ds.put_file('foo', 'foo content\n')
        self.ds.put_file('bar', 'bar content\n')
        self.assertEqual(sorted(['foo', 'bar']), self.ds.list_files())
        self.ds.clear()
        self.assertEqual([], self.ds.list_files())

    def test_delete(self):
        self.assertTrue(os.path.exists(self.ds_path))
        self.ds.put_file('foo', 'foo content\n')
        self.ds.delete(recursive=True)
        self.assertFalse(os.path.exists(self.ds_path))

    def test_delete_not_empty(self):
        self.assertTrue(os.path.exists(self.ds_path))
        self.ds.put_file('foo', 'foo content\n')
        with self.assertRaises(OSError) as cm:
            self.ds.delete()
        # FIXME: Ideally the fake data store and the real one should throw the
        # same exception in this case, not a big deal -- vila 2014-10-29
        self.assertEqual(errno.ENOTEMPTY, cm.exception.args[0])


class TestUsingFakeDataStore(unittest.TestCase):

    def setUp(self):
        super(TestUsingFakeDataStore, self).setUp()
        uci_fixtures.set_uniq_cwd(self)
        self.ds_path = os.path.join(self.uniq_dir, 'data_store')
        self.ds = fixtures.FakeDataStore(self.ds_path, None)

    def test_create_same_data_store(self):
        # Trying to create the same data store twice fails
        with self.assertRaises(OSError) as cm:
            fixtures.FakeDataStore(self.ds_path, None)
        self.assertEqual(errno.EEXIST, cm.exception.errno)
        self.assertEqual(self.ds_path, cm.exception.filename)
