# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""`GnupgEnvironment` tests."""
from __future__ import unicode_literals

import mock
import os
from StringIO import StringIO
import tempfile
from ucitests import fixtures as uci_fixtures
import unittest
import urllib2

from ci_utils.gpg import (
    GnupgEnvironment,
    GnupgEnvironmentError,
    load_keyring_uri,
)
from ci_utils.testing import (
    TESTING_KEYRING_DATA,
    TESTING_KEYRING_PATH,
)


class TestKeyringLoader(unittest.TestCase):

    @mock.patch('urllib2.urlopen')
    def test_returns_descriptor_contents(self, mock_urlopen):
        # `load_keyring_uri()` returns the given keyring contents (read()).
        mock_urlopen.return_value = StringIO('foo')
        self.assertEquals('foo', load_keyring_uri('a_path'))

    def test_unknown_location(self):
        # `load_keyring_uri` raises GnupgEnvironmentError if it
        # does not know how to load the given location.
        self.assertRaises(
            GnupgEnvironmentError, load_keyring_uri, 'not_a_uri')

    @mock.patch('urllib2.urlopen')
    def test_unreachable_error(self, mock_urlopen):
        # `load_keyring_uri()` raises GnupgEnvironmentError if it
        # cannot reach the given location
        def urlopen_error(url):
            raise urllib2.URLError('a_error')
        mock_urlopen.side_effect = urlopen_error
        self.assertRaises(
            GnupgEnvironmentError, load_keyring_uri, 'unreachable')

    def test_local_files(self):
        # Local keyrings can be loaded using 'file://' prefix.
        path = tempfile.mktemp('test-')
        self.addCleanup(os.unlink, path)
        open(path, 'w').write('foo')
        keyring_path = 'file://' + path
        self.assertEquals('foo', load_keyring_uri(keyring_path))


class TestGnupgEnvironment(unittest.TestCase):

    @mock.patch('ci_utils.unit_config.get_auth_config')
    def test_isolation_default(self, mock_config):
        # Exercises `GnupgEnvironment` with no extra options.
        mock_config.return_value = {}
        gpg_env = GnupgEnvironment()
        self.assertIsNone(gpg_env.home)
        # Simulate the existance of a previous user-define GNUPGHOME
        # and jump to a new isolated one.
        uci_fixtures.override_env(self, 'GNUPGHOME', 'previous')
        gpg_env.isolate()
        # The new GNUPGHOME is set, exists and is empty.
        self.assertEquals(gpg_env.home, os.environ.get('GNUPGHOME'))
        self.assertEquals([], os.listdir(gpg_env.home))
        self.assertEquals(0, len(gpg_env.listKeys()))
        # Save the created home path for later checks.
        home = gpg_env.home
        # Destroying the isolated GNUPG enviroment restores original
        # scenario.
        gpg_env.restore()
        self.assertEquals('previous', os.environ.get('GNUPGHOME'))
        self.assertFalse(os.path.exists(home))

    @mock.patch('ci_utils.unit_config.get_auth_config')
    def test_isolation_context_manager(self, mock_config):
        # `GnupgEnvironment` also works as a context manager.
        mock_config.return_value = {}
        uci_fixtures.override_env(self, 'GNUPGHOME', 'previous')
        # Create a context with an isolated GNUPGHOME.
        with GnupgEnvironment() as gpg_env:
            self.assertEquals(gpg_env.home, os.environ.get('GNUPGHOME'))
            self.assertEquals([], os.listdir(gpg_env.home))
            self.assertEquals(0, len(gpg_env.listKeys()))
            # Save the created home path for later checks.
            home = gpg_env.home
        # Once the context is gone the original scenario is restored.
        self.assertEquals('previous', os.environ.get('GNUPGHOME'))
        self.assertFalse(os.path.exists(home))

    @mock.patch('ci_utils.unit_config.get_auth_config')
    def test_isolation_initial_keyring(self, mock_config):
        # `GnupgEnvironment` supports loading a initial keyring
        # according to a unit_config option.
        mock_config.return_value = {
            'keyring_uri': 'file://' + TESTING_KEYRING_PATH,
        }
        # The `GnupgEnvironment` loaded the initial keyring but no
        # custom gpg.conf was created.
        with GnupgEnvironment() as gpg_env:
            self.assertEquals(1, len(gpg_env.listKeys()))
            self.assertEquals(
                ['pubring.gpg', 'pubring.gpg~', 'secring.gpg',
                 'trustdb.gpg'],
                sorted(os.listdir(gpg_env.home)))

    @mock.patch('ci_utils.unit_config.get_auth_config')
    def test_isolation_keyserver(self, mock_config):
        # `GnupgEnvironment` supports creating a custom gpg configuration
        # for auto-retrieving the need keys
        mock_config.return_value = {
            'keyserver_url': 'keyserver.internal'
        }
        # The `GnupgEnvironment` did not load any keys, instead a
        # a custom configuration file was created.
        with GnupgEnvironment() as gpg_env:
            self.assertEquals(0, len(gpg_env.listKeys()))
            self.assertEquals(
                ['gpg.conf', 'pubring.gpg', 'trustdb.gpg'],
                sorted(os.listdir(gpg_env.home)))
            self.assertEquals(
                ['keyserver hkp://keyserver.internal\n',
                 'keyserver-options auto-key-retrieve\n',
                 'no-auto-check-trustdb\n'],
                open(os.path.join(gpg_env.home, 'gpg.conf')).readlines())

    @mock.patch('ci_utils.unit_config.get_auth_config')
    def test_operations(self, mock_config):
        # `GnupgEnvironment` supports key importing and listing
        # which are proxies to the corresponding `gnupg` methods
        # running within the isolated environment.
        mock_config.return_value = {}

        with GnupgEnvironment() as gpg_env:
            # Listing public keys, nothing was imported.
            self.assertEquals([], gpg_env.listKeys())
            # Importing new keys return the original `gnupg` result object.
            result = gpg_env.importKeyring(TESTING_KEYRING_DATA)
            self.assertEquals('1 imported', result.summary())
            # Listing the just-imported key as the usual `gnupg`
            # key-dictionary.
            [key] = gpg_env.listKeys()
            self.assertEquals(
                ['Foo Bar Test <foobar@example.com>'], key.get('uids'))
            # Given arguments are passed to `gnupg`
            self.assertEquals([], gpg_env.listKeys(secret=True))
