# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import mock
import os
import shutil
import tempfile
import unittest

from ci_utils.health_check import HealthCheck


PATCHES = [
    'ci_utils.get_service_dir',
]


class TestHealthCheck(unittest.TestCase):

    def setUp(self):
        self.root_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.root_dir)
        self.service_dir = os.path.join(self.root_dir, 'service-dir')
        os.mkdir(self.service_dir)
        self.nagios_dir = os.path.join(self.service_dir, 'nagios')
        os.mkdir(self.nagios_dir)

        for spec in PATCHES:
            attr_name = spec.split('.')[-1]
            _m = mock.patch(spec)
            setattr(self, attr_name, _m.start())
            self.addCleanup(_m.stop)

        self.get_service_dir.return_value = self.service_dir

    def _get_result_file_contents(self, check):
        result_file = open(check.result_file_path).readlines()
        return '\n'.join(result_file)

    def test_basic(self):
        with HealthCheck('my_component') as check:
            self.assertEqual(check.service_name, 'my_component')
            self.assertEqual(
                check.result_file_path,
                os.path.join(self.nagios_dir, 'my_component.health'))

    def test_write_result_ok(self):
        with HealthCheck('my_component') as check:
            check.write_result_ok()
        self.assertEqual(self._get_result_file_contents(check), 'OK')

    def test_write_result_failed(self):
        with HealthCheck('my_component') as check:
            check.write_result_failed()
        self.assertEqual(self._get_result_file_contents(check), 'FAILED')

    def test_write_result_failed_with_content(self):
        output = "foobar error!"
        with HealthCheck('my_component') as check:
            check.write_result_failed(output)
        self.assertEqual(
            self._get_result_file_contents(check),
            'FAILED: {}'.format(output))
