# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""`intercom` charm relation wrapper."""
from __future__ import unicode_literals

import mock
import os
import simplejson as json
from ucitests import fixtures as uci_fixtures
import unittest

from ci_utils import intercom


class TestIntercom(unittest.TestCase):

    sample = {
        'intercom:666': {
            'hostname': '192.168.1.1',
            'port': '8000',
            'unit': 'ci-airline-bozo/0',
        },
        'intercom:111': {
            'hostname': '192.168.2.2',
            'port': '10000',
            'unit': 'ci-airline-butchie/1',
        },
    }

    def setUp(self):
        """Perform tests in a controlled environment.

        Setup a jail in a temporary directory and mocks '_get_intercom_path'
        to return controlled contents.
        """
        super(TestIntercom, self).setUp()
        uci_fixtures.set_uniq_cwd(self)
        self.test_intercom_path = os.path.join(os.getcwd(), 'intercom.json')
        _p = mock.patch('ci_utils.intercom._get_intercom_path')
        _m = _p.start()
        _m.return_value = self.test_intercom_path
        self.addCleanup(_p.stop)

    def test_missing_file(self):
        # `get_url` raises IOError if <service-dir>/intercom.json
        # is missing.
        with self.assertRaises(IOError):
            intercom.get_url('ci-airline-ts-django')

    def fake_intercom(self, content=None):
        """Fakes intercom content."""
        if content is None:
            content = self.sample.copy()
        with open(self.test_intercom_path, 'w') as fd:
            json.dump(content, fd)

    def test_get_url_works(self):
        # `get_url` lookups match unit name & number (defaults to ZERO).
        self.fake_intercom()
        self.assertEqual(
            'http://192.168.1.1:8000/',
            intercom.get_url('ci-airline-bozo'))
        self.assertEqual(
            'http://192.168.2.2:10000/',
            intercom.get_url('ci-airline-butchie', unit=1))
        # Misleading or missing lookups result in 'None' returned.
        self.assertIsNone(
            intercom.get_url('ci-airline-bozo', unit=1))
        self.assertIsNone(
            intercom.get_url('ci-airline-butchie'))
