# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import unittest
import urllib2

import mock

from ci_utils.json_status import JSONStatus, run_server


class TestRabbitStatus(unittest.TestCase):
    @mock.patch('ci_utils.amqp_utils.get_config')
    def testNoConfig(self, get_config):
        get_config.return_value = None
        status = JSONStatus()
        status.add_rabbit_worker_health('foo')
        self.assertEqual([], status.results)

    @mock.patch('ci_utils.amqp_utils.get_config')
    @mock.patch('urllib2.build_opener')
    def testCantReach(self, opener, get_config):
        opener.side_effect = RuntimeError('foo')
        status = JSONStatus()
        status.add_rabbit_worker_health('foo')
        expected = {
            'status': 'fail',
            'value': 'unable to check',
            'label': 'workers-online'
        }
        self.assertEqual([expected], status.results)

    @mock.patch('ci_utils.amqp_utils.get_config')
    @mock.patch('urllib2.build_opener')
    def _testRabbitHealth(self, queue, data, opener, get_config):
        fdfake = mock.Mock()
        fdfake.read.return_value = json.dumps(data)
        con = mock.Mock()
        con.open.return_value = fdfake
        opener.return_value = con
        status = JSONStatus()
        status.add_rabbit_worker_health(queue)
        return status.results

    def testWorkers(self):
        data = [
            {'name': 'bla', 'consumers': 2},
            {'name': 'foo', 'consumers': 1},
            {'name': 'bar', 'consumers': 1},
        ]
        expected = {
            'status': 'okay',
            'value': 1,
            'label': 'workers-online'
        }
        results = self._testRabbitHealth('foo', data)
        self.assertEqual([expected], results)

    def testNoWorkers(self):
        data = [
            {'name': 'bla', 'consumers': 2},
            {'name': 'foo', 'consumers': 0},
            {'name': 'bar', 'consumers': 1},
        ]
        expected = {
            'status': 'fail',
            'value': 0,
            'label': 'workers-online'
        }
        results = self._testRabbitHealth('foo', data)
        self.assertEqual([expected], results)

    def testNoQueue(self):
        data = [
            {'name': 'bla', 'consumers': 2},
            {'name': 'bar', 'consumers': 1},
        ]
        expected = {
            'status': 'fail',
            'value': 'no queue defined',
            'label': 'workers-online'
        }
        results = self._testRabbitHealth('foo', data)
        self.assertEqual([expected], results)


class TestHealthServer(unittest.TestCase):
    def get_resp(self, path='/'):
        server = run_server(self.cb, 0)
        try:
            url = 'http://localhost:%d%s' % (server.server_port, path)
            return urllib2.urlopen(url)
        finally:
            server.shutdown()

    def cb(self):
        self.cb_called = True
        status = JSONStatus()
        status.add_okay('name', 'value')
        return status

    def test_plain(self):
        resp = self.get_resp()
        self.assertEqual(200, resp.code)
        self.assertEqual(self.cb().results, json.loads(resp.read()))

    def test_jsonp(self):
        resp = self.get_resp('/?format=jsonp&callback=foo')
        self.assertEqual(200, resp.code)
        expected = 'foo(' + json.dumps(self.cb().results) + ')'
        self.assertEqual(expected, resp.read())

    def test_cache(self):
        self.cb_called = False
        resp = self.get_resp()
        self.assertEqual(200, resp.code)
        self.assertTrue(self.cb_called)

        self.cb_called = False
        resp = self.get_resp()
        self.assertEqual(200, resp.code)
        self.assertFalse(self.cb_called)

        resp = self.get_resp('/?refresh=1')
        self.assertEqual(200, resp.code)
        self.assertTrue(self.cb_called)
