# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest


from ucitests import fixtures as uci_fixtures
import yaml


from ci_utils import juju
from ci_utils.testing import fixtures


class TestGetEnvironment(unittest.TestCase):

    def setUp(self):
        super(TestGetEnvironment, self).setUp()
        uci_fixtures.set_uniq_cwd(self)
        uci_fixtures.isolate_from_env(
            self, dict(HOME=self.uniq_dir, JUJU_HOME=None, JUJU_ENV=None))
        os.mkdir('.juju')
        self.juju_home = os.path.join(os.environ['HOME'], '.juju')

    def test_get_juju_home(self):
        juju_home = '~/.juju/'
        os.environ['JUJU_HOME'] = juju_home
        self.assertEquals(juju_home, juju.home())

    def test_get_juju_home_unset(self):
        self.assertEquals(self.juju_home, juju.home())

    def test_get_juju_env_os_environ_set(self):
        env = 'hp'
        os.environ['JUJU_ENV'] = env
        self.assertEquals(env, juju.environment_name())

    def test_get_juju_env_current_env(self):
        env = 'hpcloud'
        with open(os.path.join(
                  self.juju_home, 'current-environment'), 'w') as f:
            f.write(env)
        self.assertEquals(env, juju.environment_name())

    def test_get_juju_env_default(self):
        env = {
            'default': 'hp',
            'environments': {
                'hp': {
                    'type': 'openstack',
                    'default-series': 'precise',
                    'use-floating-ip': False,
                },
            },
        }
        with open(os.path.join(
                  self.juju_home, 'environments.yaml'), 'w') as f:
            yaml.safe_dump(env, f)
        self.assertEquals('hp', juju.environment_name())


class TestJujuHome(unittest.TestCase):

    def setUp(self):
        super(TestJujuHome, self).setUp()
        fixtures.isolate_from_juju_env(self)

    def test_no_juju_home(self):
        del os.environ['JUJU_HOME']
        self.assertEqual(os.path.join(self.uniq_dir, '.juju'),
                         juju.home())

    def test_juju_home_set(self):
        os.environ['JUJU_HOME'] = 'foo'
        self.assertEqual('foo', juju.home())


class TestJujuGetEnvName(unittest.TestCase):

    def setUp(self):
        super(TestJujuGetEnvName, self).setUp()
        fixtures.isolate_from_juju_env(self)

    def test_no_env_name(self):
        self.assertEqual(None, juju.environment_name())

    def test_env_name_from_os_env(self):
        os.environ['JUJU_ENV'] = 'foo'
        self.assertEqual('foo', juju.environment_name())

    def test_no_env_name_in_env_yaml(self):
        uci_fixtures.build_tree('''file: environments.yaml
''')
        self.assertEqual(None, juju.environment_name())

    def test_env_name_from_current_file(self):
        uci_fixtures.build_tree('''file: current-environment
foo
''')
        self.assertEqual('foo', juju.environment_name())

    def test_env_name_from_env_yaml(self):
        uci_fixtures.build_tree('''file: environments.yaml
default: foo
''')
        self.assertEqual('foo', juju.environment_name())

    def test_os_env_override_current_file(self):
        uci_fixtures.build_tree('''file: environments.yaml
default: foo
''')
        os.environ['JUJU_ENV'] = 'bar'
        self.assertEqual('bar', juju.environment_name())

    def test_current_override_yaml_file(self):
        uci_fixtures.build_tree('''file: environments.yaml
default: foo
file: current-environment
bar
''')
        self.assertEqual('bar', juju.environment_name())


class TestJujuGetEnvsFromYaml(unittest.TestCase):

    def setUp(self):
        super(TestJujuGetEnvsFromYaml, self).setUp()
        fixtures.isolate_from_juju_env(self)

    def test_no_environments_file(self):
        self.assertEqual(None, juju.environments_from_yaml())


class TestJujuGetEnv(unittest.TestCase):

    def setUp(self):
        super(TestJujuGetEnv, self).setUp()
        fixtures.isolate_from_juju_env(self)

    def test_no_env(self):
        self.assertEqual(None, juju.environment('foo'))

    def test_broken_env_file(self):
        uci_fixtures.build_tree('''file: environments.yaml
''')
        self.assertEqual(None, juju.environment('foo'))

    def test_unknown_env(self):
        uci_fixtures.build_tree('''file: environments.yaml
environments:
    bar:
        type: local
''')
        self.assertEqual(None, juju.environment('foo'))

    def test_get_one_env(self):
        uci_fixtures.build_tree('''file: environments.yaml
environments:
    bar: {}
''')
        self.assertEqual({}, juju.environment('bar'))
