# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from functools import wraps
import re
import time
import xml.etree.ElementTree as etree


class CoverageRetrieverException(ValueError):
    """Bad things happen to good retrievers."""


def snip_coverage_xml(log_text):
    """Extract coverage.xml from a specially-marked log.

    :param log_text: bytestring representing build log to parse
    """
    regex = re.compile(
        """===== BEGIN coverage.xml =====
(.+?)
===== END coverage.xml =====""",
        re.DOTALL)
    match = regex.search(log_text)
    try:
        coverage_xml = match.group(1)
        etree.fromstring(coverage_xml)
        return coverage_xml
    except AttributeError as e:
        # unlikely that this e will be helpful but we include
        raise CoverageRetrieverException(
            "Failed to parse build log: no coverage.xml; {}".format(
                e.message))
    except etree.ParseError as e:
        raise CoverageRetrieverException(
            "Failed to parse build log: {}".format(e.message))


def retry(ExceptionToCheck, tries=4, delay=3, backoff=2, logger=None):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    Original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :param tries: number of times to try (not retry) before giving up
    :param delay: initial delay between retries in seconds
    :param backoff: backoff multiplier e.g. value of 2 will double the delay
        each retry
    :param logger: logger.Logger to use.
    """
    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            for i in range(tries):
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    d = delay * backoff ** i
                    if logger:
                        logger.warning(
                            "%s, retrying in %d seconds. . . ." % (e, d))
                    time.sleep(d)
            return f(*args, **kwargs)
        return f_retry  # true decorator
    return deco_retry
