#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ci_utils import amqp_utils
from ci_utils.data_store import DataStoreException
from coverageretriever import (
    CoverageRetrieverException,
    retry,
    snip_coverage_xml,
)

SUCCESS_TEMPLATE = u'Deposited coverage.xml into swift for ticket %s.'
MISSING_PARAMS_TEMPLATE = (u'Failed to handle request: missing build log '
                           'filename for ticket %s')
PARSE_FAILURE_TEMPLATE = u'Failed to parse build log for ticket %s: %s'
SWIFT_GET_FAILURE_TEMPLATE = (u'Failed to retrieve build log from swift '
                              'for ticket %s: %s')
SWIFT_PUT_FAILURE_TEMPLATE = (u'Failed to put build log into swift for '
                              'ticket %s: %s')


def generate_and_put_artifact(
        data_store,
        name='coverage-retriever.output.log',
        type='LOGS',
        message=None,
        reference=None):
    """Return a uci-artifact generated and deposited in swift.

    NOTE that if no reference is supplied, the artifact is deposited
    into swift and the resulting link is reported in the returned
    artifact; provided for convenience in logging.

    :param data_store: DataStore into which to put
    :param name: filename of artifact in swift
    :param type: 'LOGS' for logs or 'RESULTS' for coverage.xml
    :param message: a log message
    :param reference: swift link; supplied upon deposit if not specified
    """
    artifact = {
        'name': name,
        'type': type,
    }
    if message is not None:
        artifact['message'] = message
    if reference is None:
        reference = data_store.put_file(name, str(artifact), 'text/plain')
    artifact['reference'] = reference
    return artifact


class RetrievalFailed(Exception):
    """Return a dict upon operation failure.."""

    def __init__(self, return_tuple):
        self.return_tuple = return_tuple


class Retrieval(object):

    def __init__(self, ticket_id, logger, data_store):
        self.ticket_id = ticket_id
        self.logger = logger
        self.data_store = data_store
        self.result = dict(artifacts=[])

    def try_extract_build_log_path(self, params):
        """Extract build log from dict of params.

        :param params: dict, hopefully with 'build_log_path' as key
        """
        try:
            return params['build_log_path']
        except KeyError:
            self.report_failed(
                MISSING_PARAMS_TEMPLATE,
                self.ticket_id
            )

    def try_get_build_log_from_path(self, build_log_path):
        """Get build log at given path from swift.

        :param build_log_path: swift path from which to get
        """
        try:
            @retry(DataStoreException, logger=self.logger)
            def get_build_log_actual():
                return self.data_store.get_file(build_log_path)
            return get_build_log_actual()
        except DataStoreException as e:
            self.report_failed(
                SWIFT_GET_FAILURE_TEMPLATE,
                self.ticket_id,
                e.message
            )

    def try_extract_coverage_xml(self, log_text):
        """Snip coverage.xml from a given log text.

        :param log_text: build log string
        """
        try:
            return snip_coverage_xml(log_text)
        except CoverageRetrieverException as e:
            self.report_failed(
                PARSE_FAILURE_TEMPLATE,
                self.ticket_id,
                e.message
            )

    def try_put_coverage_xml(self, coverage_xml_text):
        """Put coverage.xml into swift.

        :param coverage_xml_text: coverage.xml string
        """
        try:
            return self.data_store.put_file(
                'coverage.xml',
                coverage_xml_text,
                'text/xml',
            )
        except DataStoreException as e:
            self.report_failed(
                SWIFT_PUT_FAILURE_TEMPLATE,
                self.ticket_id,
                e.message
            )

    def report_failed(self, format_str, *str_args):
        """Convenience function to raise and log a failiure.

        :param format_str: template to log
        :str_args str_args: items with which to format format_str
        """
        message = format_str % str_args
        self.logger.error(message)
        self.result['message'] = message
        self.result['artifacts'].append(
            generate_and_put_artifact(self.data_store, message=message)
        )
        raise RetrievalFailed(
            (amqp_utils.progress_failed, self.result)
        )

    def report_succeeded(self, ticket_id, coverage_xml_link):
        """Deposit a result artifact and return a log artifact.

        :param ticket_id: ticket for which to report retrieval success
        :param coverage_xml_link: deposited coverage.xml swift link
        """
        message = SUCCESS_TEMPLATE % ticket_id
        # NOTE that logging compels us to pass params to info (can't
        # reuse message declared above)
        self.logger.info(SUCCESS_TEMPLATE, ticket_id)
        # create our coverage.xml artifact
        self.result['artifacts'].append(
            generate_and_put_artifact(
                self.data_store,
                name='coverage.xml',
                type='RESULTS',
                reference=coverage_xml_link,
            )
        )
        # create our success message
        self.result['artifacts'].append(
            generate_and_put_artifact(
                self.data_store,
                message=message
            )
        )
        self.result['message'] = message
        return amqp_utils.progress_completed, self.result
