# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager
import unittest
import xml.etree.ElementTree as etree

from mock import patch


from ucitests import fixtures as uci_fixtures


from ci_utils.amqp_worker import AMQPWorker
from ci_utils.testing import fixtures

vanilla_build_log = u"""
===== BEGIN coverage.xml =====
<?xml version="1.0" ?>
<!DOCTYPE coverage
  SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-03.dtd'>
<coverage branch-rate="1.0" line-rate="1.0" timestamp="1410387089" version="gcovr 2.4 (r2774)">
	<sources>
		<source>/tmp/tmpz8yrh4tz/dummy</source>
	</sources>
	<packages>
		<package branch-rate="1.0" complexity="0.0" line-rate="1.0" name="">
			<classes>
				<class branch-rate="1.0" complexity="0.0" filename="fibonacci.cpp" line-rate="1.0" name="fibonacci_cpp">
					<lines>
						<line branch="false" hits="10" number="4"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="10" number="6">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="3" number="7"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="7" number="8">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="4" number="9"/>
						<line branch="false" hits="3" number="10"/>
					</lines>
				</class>
			</classes>
		</package>
	</packages>
</coverage>

===== END coverage.xml =====
"""  # NOQA

vanilla_coverage_xml = u"""
<?xml version="1.0" ?>
<!DOCTYPE coverage
  SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-03.dtd'>
<coverage branch-rate="1.0" line-rate="1.0" timestamp="1410387089" version="gcovr 2.4 (r2774)">
	<sources>
		<source>/tmp/tmpz8yrh4tz/dummy</source>
	</sources>
	<packages>
		<package branch-rate="1.0" complexity="0.0" line-rate="1.0" name="">
			<classes>
				<class branch-rate="1.0" complexity="0.0" filename="fibonacci.cpp" line-rate="1.0" name="fibonacci_cpp">
					<lines>
						<line branch="false" hits="10" number="4"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="10" number="6">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="3" number="7"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="7" number="8">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="4" number="9"/>
						<line branch="false" hits="3" number="10"/>
					</lines>
				</class>
			</classes>
		</package>
	</packages>
</coverage>

"""  # NOQA


class CoverageRetrieverTestCase(unittest.TestCase):

    def setUp(self):
        super(CoverageRetrieverTestCase, self).setUp()
        # We'll use fixtures.FakeDataStore that requires disk resources
        uci_fixtures.set_uniq_cwd(self)

    @contextmanager
    def patch_create_data_store(self):
        """Patch in a FakeDataStore for testing."""
        self.fake_data_store = fixtures.FakeDataStore('fake_ticket_id')
        self.addCleanup(self.fake_data_store.delete, recursive=True)
        with patch.object(
                AMQPWorker,
                '_create_data_store',
                return_value=self.fake_data_store) as create_data_store_patch:
            yield create_data_store_patch

    def put_fake_build_log(self, data_store, build_log=vanilla_build_log):
        """Put a build log into our fake data store.

        :param data_store: FakeDataStore into which to put
        :param build_log: str defaults to vanilla_build_log
        """
        data_store.put_file(
            'build.log',
            build_log,
            content_type='text/plain'
        )

    def validate_coverage_xml(self, coverage_xml, line_rate=1.0):
        """Convenience function to parse and verify coverage.xml.

        :param coverage_xml: retrieved coverage.xml bytestring
        :param line_rate: expected line rate (float)
        """
        root_coverage_node = etree.fromstring(coverage_xml)
        self.assertAlmostEqual(
            line_rate,
            float(root_coverage_node.attrib['line-rate']))
