# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

from mock import Mock

from ci_utils import amqp_utils
from coverageretriever.tests import CoverageRetrieverTestCase
from coverageretriever.retrieval import (
    MISSING_PARAMS_TEMPLATE,
    SUCCESS_TEMPLATE,
)
from coverageretriever.run_worker import CoverageRetrieverWorker


class HandleRequestTestCase(CoverageRetrieverTestCase):

    def test_vanilla_handle_request(self):
        """Large integration test for coverage retrieval."""
        fake_logger = Mock(spec_set=logging.Logger)
        with self.patch_create_data_store():
            self.put_fake_build_log(self.fake_data_store)
            worker = CoverageRetrieverWorker()
            fake_params = {'build_log_path': 'build.log',
                           'ticket_id': 'fake_ticket_id'}
            amqp_result, result_data = worker.handle_request(
                fake_params,
                fake_logger,
            )
        self.assertEqual(amqp_utils.progress_completed, amqp_result)
        self.assertEqual(
            SUCCESS_TEMPLATE % 'fake_ticket_id',
            result_data['message'],
        )
        fake_logger.info.assert_called_with(
            SUCCESS_TEMPLATE, 'fake_ticket_id'
        )
        self.assertEqual(2, len(result_data['artifacts']))
        # first artifact should be coverage.xml RESULTS link
        result_artifact = result_data['artifacts'][0]
        self.assertEqual('RESULTS', result_artifact['type'])
        self.assertEqual(
            'file://' + os.path.join(
                os.path.abspath(os.curdir),
                'fake_ticket_id/coverage.xml'),
            result_artifact['reference']
        )
        # second artifact should be a success log message
        log_artifact = result_data['artifacts'][1]
        self.assertEqual(
            SUCCESS_TEMPLATE % 'fake_ticket_id',
            log_artifact['message'],
        )
        self.assertEqual('LOGS', log_artifact['type'])
        self.assertEqual(
            'file://' + os.path.join(
                os.path.abspath(os.curdir),
                'fake_ticket_id/coverage-retriever.output.log'),
            log_artifact['reference']
        )
        coverage_xml = self.fake_data_store.get_file('coverage.xml')
        self.validate_coverage_xml(coverage_xml)

    def test_exception_returns_failure(self):
        """An exception in our imperatives returns a failure.

        Actuate a KeyError in try_extract_build_log_path, inspect our
        results to make sure we're logging, depositing a log in swift.
        """
        fake_logger = Mock(spec_set=logging.Logger)
        with self.patch_create_data_store() as fake_data_store:
            self.put_fake_build_log(fake_data_store)
            worker = CoverageRetrieverWorker()
            # actuate a KeyError in try_extract_build_log_path
            malevolent_params = {'gibberish': 'build.log',
                                 'ticket_id': 'fake_ticket_id'}
            amqp_result, result_data = worker.handle_request(
                malevolent_params,
                fake_logger
            )
        self.assertEqual(amqp_utils.progress_failed, amqp_result)
        self.assertEqual(
            MISSING_PARAMS_TEMPLATE % 'fake_ticket_id',
            result_data['message']
        )
        self.assertEqual(1, len(result_data['artifacts']))
        artifact = result_data['artifacts'][0]
        self.assertEqual(
            'file://' + os.path.join(
                os.path.abspath(os.curdir),
                'fake_ticket_id/coverage-retriever.output.log'),
            artifact['reference'])
        self.assertEqual(
            MISSING_PARAMS_TEMPLATE % 'fake_ticket_id',
            result_data['message']
        )
        self.assertEqual('LOGS', artifact['type'])
