#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from mock import patch, Mock

from ci_utils.data_store import DataStoreException
from coverageretriever import CoverageRetrieverException
from coverageretriever.tests import (
    CoverageRetrieverTestCase,
    vanilla_build_log,
    vanilla_coverage_xml,
)
from coverageretriever.retrieval import (
    Retrieval,
    RetrievalFailed,
    PARSE_FAILURE_TEMPLATE,
    SWIFT_GET_FAILURE_TEMPLATE,
    SWIFT_PUT_FAILURE_TEMPLATE,
)


class RetrievalTestCase(CoverageRetrieverTestCase):

    def test_extract_build_log_path(self):
        """We get the build log path from params."""
        with self.patch_create_data_store() as fake_data_store:
            params = {'build_log_path': 'build.log'}
            ret = Retrieval('fake-ticket-id', 'fake-logger', fake_data_store)
            self.assertEqual(
                'build.log',
                ret.try_extract_build_log_path(params)
            )

    @patch.object(Retrieval, 'report_failed')
    def test_extract_build_log_path_failure(self, report_failed):
        """Failure to raise the build log path from params fails."""
        with self.patch_create_data_store() as fake_data_store:
            malevolent_params = {'gibberish': 'build.log'}
            ret = Retrieval('fake-ticket-id', 'fake-logger', fake_data_store)
            ret.try_extract_build_log_path(malevolent_params)
        report_failed.assert_called()

    def test_try_get_build_log_from_path(self):
        """Retrieve our build log from swift."""
        with self.patch_create_data_store() as fake_data_store:
            fake_data_store.get_file.return_value = vanilla_build_log
            ret = Retrieval('fake-ticket-id', 'fake-logger', fake_data_store)
            self.assertEqual(
                vanilla_build_log,
                ret.try_get_build_log_from_path('build.log'),
            )

    def test_try_get_build_log_from_path_retries(self):
        """Retry if swift fails to yield our build log."""
        fake_logger = Mock(spec_set=logging.Logger)
        with self.patch_create_data_store() as fake_data_store:
            ret = Retrieval('fake-ticket-id', fake_logger, fake_data_store)
            fake_data_store.get_file.side_effect = [
                DataStoreException("No log!"),
                DataStoreException("Still no log!"),
                vanilla_build_log
            ]
            with patch('coverageretriever.time.sleep'):  # for the impatient
                self.assertEqual(
                    vanilla_build_log,
                    ret.try_get_build_log_from_path('build.log')
                )
        self.assertEqual(3, fake_data_store.get_file.call_count)
        fake_logger.warning.assert_called()

    @patch.object(Retrieval, 'report_failed')
    def test_try_get_build_log_from_path_raises(self, report_failed):
        """Return failure if swift fails to yield our build log."""
        fake_logger = Mock(spec_set=logging.Logger)
        with self.patch_create_data_store() as fake_data_store:
            ret = Retrieval('fake-ticket-id', fake_logger, fake_data_store)
            fake_data_store.get_file.side_effect = DataStoreException(
                'Drool...'
            )
            with patch('coverageretriever.time.sleep'):  # for the impatient
                ret.try_get_build_log_from_path('build.log')
        report_failed.assert_called_with(
            SWIFT_GET_FAILURE_TEMPLATE,
            'fake-ticket-id',
            'Drool...',
        )

    @patch('coverageretriever.retrieval.snip_coverage_xml')
    def test_extract_coverage_xml(self, snip_coverage_xml):
        """Vanilla snip_coverage_xml."""
        snip_coverage_xml.return_value = vanilla_coverage_xml
        with self.patch_create_data_store() as fake_data_store:
            ret = Retrieval('fake-ticket-id', 'fake-logger', fake_data_store)
            self.assertEqual(
                vanilla_coverage_xml,
                ret.try_extract_coverage_xml(vanilla_build_log)
            )

    @patch('coverageretriever.retrieval.snip_coverage_xml')
    @patch.object(Retrieval, 'report_failed')
    def test_extract_coverage_xml_raises(self,
                                         report_failed,
                                         snip_coverage_xml):
        """Fail if coverage.xml not found in build log."""
        with self.patch_create_data_store() as fake_data_store:
            ret = Retrieval('fake-ticket-id', 'fake-logger', fake_data_store)
            snip_coverage_xml.side_effect = CoverageRetrieverException(
                'Ruh-roh!')
            ret.try_extract_coverage_xml(vanilla_build_log)
        report_failed.assert_called_with(
            PARSE_FAILURE_TEMPLATE,
            'fake-ticket-id',
            'Ruh-roh!',
        )

    def test_try_put_coverage_xml(self):
        """Put coverage.xml into swift."""
        with self.patch_create_data_store() as fake_data_store:
            ret = Retrieval('fake-ticket-id', 'fake-logger', fake_data_store)
            ret.try_put_coverage_xml(vanilla_coverage_xml)

    @patch.object(Retrieval, 'report_failed')
    def test_try_put_coverage_xml_raises(self, report_failed):
        """Return failure if we fail to put coverage.xml into swift."""
        with self.patch_create_data_store() as fake_data_store:
            ret = Retrieval('fake-ticket-id', 'fake-logger', fake_data_store)
            with patch.object(fake_data_store, 'put_file') as put_file:
                put_file.side_effect = DataStoreException('Full!')
                ret.try_put_coverage_xml(vanilla_coverage_xml)
        report_failed.assert_called_with(
            SWIFT_PUT_FAILURE_TEMPLATE,
            'fake-ticket-id',
            'Full!',
        )

    def test_report_failed_raises(self):
        """Internal catch-all handler raises RetrievalFailed."""
        with self.patch_create_data_store() as fake_data_store:
            fake_logger = Mock(spec_set=logging.Logger)
            ret = Retrieval('fake-ticket-id', fake_logger, fake_data_store)
            self.assertRaises(
                RetrievalFailed,
                lambda: ret.report_failed("%s %s", 'foo', 'bar')
            )
        fake_logger.error.assert_called_with("foo bar")
