# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import xml.etree.ElementTree as etree

from coverageretriever import (
    snip_coverage_xml,
    CoverageRetrieverException,
)
from coverageretriever.tests import vanilla_build_log


# here we refuse to close our <sources> tag :|
invalid_xml_build_log = u"""
===== BEGIN coverage.xml =====
<?xml version="1.0" ?>
<!DOCTYPE coverage
  SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-03.dtd'>
<coverage branch-rate="1.0" line-rate="1.0" timestamp="1410387089" version="gcovr 2.4 (r2774)">
	<sources>
		<source>/tmp/tmpz8yrh4tz/dummy</source>
	<packages>
		<package branch-rate="1.0" complexity="0.0" line-rate="1.0" name="">
			<classes>
				<class branch-rate="1.0" complexity="0.0" filename="fibonacci.cpp" line-rate="1.0" name="fibonacci_cpp">
					<lines>
						<line branch="false" hits="10" number="4"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="10" number="6">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="3" number="7"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="7" number="8">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="4" number="9"/>
						<line branch="false" hits="3" number="10"/>
					</lines>
				</class>
			</classes>
		</package>
	</packages>
</coverage>

===== END coverage.xml =====
"""  # NOQA

# NOTE a *baseball* dingbat in sources filepath
unicode_build_log = unicode("""
===== BEGIN coverage.xml =====
<?xml version="1.0" ?>
<!DOCTYPE coverage
  SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-03.dtd'>
<coverage branch-rate="1.0" line-rate="1.0" timestamp="1410387089" version="gcovr 2.4 (r2774)">
	<sources>
		<source>/tmp/\u26BE/dummy</source>
	</sources>
	<packages>
		<package branch-rate="1.0" complexity="0.0" line-rate="1.0" name="">
			<classes>
				<class branch-rate="1.0" complexity="0.0" filename="fibonacci.cpp" line-rate="1.0" name="fibonacci_cpp">
					<lines>
						<line branch="false" hits="10" number="4"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="10" number="6">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="3" number="7"/>
						<line branch="true" condition-coverage="100% (2/2)" hits="7" number="8">
							<conditions>
								<condition coverage="100%" number="0" type="jump"/>
							</conditions>
						</line>
						<line branch="false" hits="4" number="9"/>
						<line branch="false" hits="3" number="10"/>
					</lines>
				</class>
			</classes>
		</package>
	</packages>
</coverage>

===== END coverage.xml =====
""")  # NOQA


class SnipCoverageXmlTestCase(unittest.TestCase):

    def test_snip_coverage_xml(self):
        """Snip coverage from a vanilla build log bytestring."""
        coverage_xml = snip_coverage_xml(vanilla_build_log)
        root = etree.fromstring(coverage_xml)
        node = root.find(".//*[@filename='fibonacci.cpp']")
        self.assertAlmostEqual(1.0, float(node.attrib['line-rate']))

    def test_not_found_raises(self):
        """Failure to find a coverage.xml raises an exception."""
        self.assertRaises(
            CoverageRetrieverException,
            snip_coverage_xml,
            'gibberish')

    def test_invalid_xml_raises(self):
        """We raise an exception if we find invalid XML."""
        self.assertRaises(
            CoverageRetrieverException,
            snip_coverage_xml,
            invalid_xml_build_log)

    def test_unicode_parses(self):
        """Snip for a build log with unicode characters."""
        coverage_xml = snip_coverage_xml(unicode_build_log)
        root = etree.fromstring(coverage_xml)
        node = root.find(".//*[@filename='fibonacci.cpp']")
        self.assertAlmostEqual(1.0, float(node.attrib['line-rate']))
