# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Gatekeeper SourcePackageUpload adapter."""
from __future__ import unicode_literals

from dput.changes import ChangesFileException
import os

from ci_utils import TmpDir
from ci_utils.file_handler import (
    ChangesParseError,
    ChangesProcessor,
    ChangesValidationError,
    FileToUploadNotFound,
    NotAChangesFileError,
)
from ci_utils.gpg import GnupgEnvironment


class SourcePackageUploadFailure(Exception):
    """Raised when SPU content does not pass validation."""


def get_changesfile_path(path):
    """Return the absolute path for the changesfile in the given path.

    Raises `SourcePackageUploadFailure` if there is none or more than one
    changesfile ('*.changes').
    """
    changes = [filename for filename in os.listdir(path)
               if filename.endswith('.changes')]
    if len(changes) == 0:
        raise SourcePackageUploadFailure('No changesfile found.')
    if len(changes) > 1:
        raise SourcePackageUploadFailure('More than one changesfile found.')
    [changesfilename] = changes
    return os.path.join(path, changesfilename)


def compare_contents(path, processed):
    """Fails if the given path has files not processed.

    Raises `SourcePackageUploadFailure` if there are filename in the
    given path that are not in the processed list.
    """
    processed_filenames = set(
        [os.path.basename(filename) for filename in processed])
    sandbox_filenames = set(os.listdir(path))
    diff = sandbox_filenames.difference(processed_filenames)
    if diff:
        raise SourcePackageUploadFailure(
            'The following uploaded files are not part of the source '
            'package: {}'.format(','.join(diff)))


class SourcePackageUpload(object):
    """Sandbox adapter for creating SourcePackageUploads."""

    def __init__(self, sandbox):
        self.sandbox = sandbox

    def validate(self):
        """Validates the context `Sandbox` contents as a single SPU.

        Dumps the sandbox content into a temporary directory a performs
        validation using the CLI `ChangesProcessor`.
        """
        with TmpDir() as tmpdir:
            self.sandbox.dump(tmpdir)
            changesfile_path = get_changesfile_path(tmpdir)
            try:
                changes = ChangesProcessor(changesfile_path)
                with GnupgEnvironment():
                    processed_filenames = changes.process()
            except FileToUploadNotFound as error:
                message = 'Missing file: {}'.format(
                    os.path.basename(error.message))
                raise SourcePackageUploadFailure(message)
            except ChangesFileException as error:
                tmpdir_fullpath = tmpdir + '/'
                message = error.message.replace(tmpdir_fullpath, '')
                raise SourcePackageUploadFailure(message)
            except (ChangesParseError, ChangesValidationError,
                    NotAChangesFileError) as error:
                raise SourcePackageUploadFailure(error.message)
            compare_contents(tmpdir, processed_filenames)

    def process(self, ticket_id):
        """Creates a ticket with the sandbox contents."""
        dest_container = 'ticket-{}'.format(ticket_id)
        try:
            self.validate()
            # XXX cprov 2014-05-29: no Ticket-System verification yet.
            for obj in self.sandbox.objects:
                obj.copy(dest_container)
        except SourcePackageUploadFailure as error:
            raise error
        finally:
            self.sandbox.drop()

        ticket_url = '/ticket.html?ticket_id={}'.format(ticket_id)
        return ticket_url
