# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""SourcePackageUpload test suite."""
from __future__ import unicode_literals

import mock
import os
import shutil
import unittest

from ci_utils.testing import (
    get_test_file_path,
    TESTING_KEYRING_PATH,
)
from gatekeeper.adapters.sourcepackageupload import (
    SourcePackageUpload,
    SourcePackageUploadFailure,
)


class FakeObject(object):

    def __init__(self, filename):
        self.filename = filename
        self.destination = None

    def copy(self, destination):
        self.destination = destination


class FakeSandbox(object):
    """Fake `Sandbox` object with special dump() implementation.

    It copies given filenames from the CLI test_data direcory to
    the dumping path.
    Executes 'after_dump_hook' passing the dumping path if it is given.
    This hook is useful for tampering upload contents for tests.
    """

    def __init__(self, filenames, after_dump_hook=None):
        self.filenames = filenames
        self.dropped = False
        self.after_dump_hook = after_dump_hook
        self.objects = [FakeObject(f) for f in filenames]

    def drop(self):
        self.dropped = True

    def dump(self, path):
        for filename in self.filenames:
            shutil.copy(get_test_file_path(filename),
                        os.path.join(path, filename))
        if callable(self.after_dump_hook):
            self.after_dump_hook(path)


class SourcePackageUploadTest(unittest.TestCase):
    """Tests for the `SourcePackageUpload` `Sandbox` adapter."""

    TESTING_UNIT_CONFIG = {
        'keyring_uri': 'file://' + TESTING_KEYRING_PATH,
    }

    def setUp(self):
        """Setting a mocking `unit_test` for these tests.

        Also silence `gnupg` stdout messages by mocking `sys.stdout`.
        """
        super(SourcePackageUploadTest, self).setUp()
        config_patcher = mock.patch('ci_utils.unit_config.get_auth_config')
        mock_config = config_patcher.start()
        mock_config.return_value = self.TESTING_UNIT_CONFIG
        self.addCleanup(config_patcher.stop)
        stdout_patcher = mock.patch('sys.stdout')
        stdout_patcher.start()
        self.addCleanup(stdout_patcher.stop)

    def test_validation_empty(self):
        # `SourcePackageUpload` validation fails for empty sandboxes.
        sandbox = FakeSandbox([])
        spu = SourcePackageUpload(sandbox)
        with self.assertRaises(SourcePackageUploadFailure) as raised:
            spu.validate()
        self.assertEquals('No changesfile found.', raised.exception.message)

    def test_validation_multiple_changesfiles(self):
        # `SourcePackageUpload` validation fails for multiple changesfiles.
        sandbox = FakeSandbox(
            ['foobar_0.1-1_source.changes',
             'barfoo_0.1-1_source.changes'])
        spu = SourcePackageUpload(sandbox)
        with self.assertRaises(SourcePackageUploadFailure) as raised:
            spu.validate()
        self.assertEquals(
            'More than one changesfile found.', raised.exception.message)

    def test_validation_extra_files(self):
        # `SourcePackageUpload` fails if there are more files in the
        # sandbox than declared on the changesfile.
        sandbox = FakeSandbox(
            ['foobar_0.1-1_source.changes',
             'foobar_0.1-1.dsc',
             'foobar_0.1.orig.tar.gz',
             'foobar_0.1-1.debian.tar.gz',
             'barfoo_0.1-1.debian.tar.gz'])
        spu = SourcePackageUpload(sandbox)
        with self.assertRaises(SourcePackageUploadFailure) as raised:
            spu.validate()
        self.assertEquals(
            'The following uploaded files are not part of the '
            'source package: barfoo_0.1-1.debian.tar.gz',
            raised.exception.message)

    def test_validation_missing_files(self):
        # `SourcePackageUpload` fails if there are missing files detected
        # by the `ChangesProcessor`.
        sandbox = FakeSandbox(
            ['foobar_0.1-1_source.changes',
             'foobar_0.1-1.dsc',
             'foobar_0.1.orig.tar.gz'])
        spu = SourcePackageUpload(sandbox)
        with self.assertRaises(SourcePackageUploadFailure) as raised:
            spu.validate()
        self.assertEquals(
            'Missing file: foobar_0.1-1.debian.tar.gz',
            raised.exception.message)

    def test_process_validation_fails(self):
        # `SourcePackageUpload` process validates the sandbox contents
        # re-raising validation errors (obfuscating the server processing
        # path) and wipes the sandbox.
        def break_tar(path):
            # `FakeSandbox.after_dump_hook` to break the upload DSC
            # contents. It should trigger a real 'dput' validation error.
            tar_path = os.path.join(path, 'foobar_0.1.orig.tar.gz')
            open(tar_path, 'w').write('broken')
        sandbox = FakeSandbox(
            ['foobar_0.1-1_source.changes',
             'foobar_0.1-1.dsc',
             'foobar_0.1.orig.tar.gz',
             'foobar_0.1-1.debian.tar.gz'],
            after_dump_hook=break_tar)
        spu = SourcePackageUpload(sandbox)
        with self.assertRaises(SourcePackageUploadFailure) as raised:
            spu.validate()
        self.assertEquals(
            'Checksum mismatch for file foobar_0.1.orig.tar.gz: '
            '0b8a1caec23d75d1154b8d9bef9cec6c03697638 != '
            'eb1f6019c6fa8b948207ecf507580cef9d41430e',
            raised.exception.message)

    def test_validation_works(self):
        # `SourcePackageUpload` validation works.
        sandbox = FakeSandbox(
            ['foobar_0.1-1_source.changes',
             'foobar_0.1-1.dsc',
             'foobar_0.1.orig.tar.gz',
             'foobar_0.1-1.debian.tar.gz'])
        spu = SourcePackageUpload(sandbox)
        self.assertIsNone(spu.validate())

    def test_process_works_preliminary(self):
        # `SourcePackageUpload` process only validates the sandbox
        # contents, copy it to the ticket container and drop it.
        # The affected ticket url is returned.
        sandbox = FakeSandbox(
            ['foobar_0.1-1_source.changes',
             'foobar_0.1-1.dsc',
             'foobar_0.1.orig.tar.gz',
             'foobar_0.1-1.debian.tar.gz'])
        spu = SourcePackageUpload(sandbox)
        ticket_url = spu.process('1')
        self.assertEquals(
            set(['ticket-1']),
            set([obj.destination for obj in sandbox.objects]))
        self.assertTrue(sandbox.dropped)
        self.assertEquals('/ticket.html?ticket_id=1', ticket_url)
