# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Swift supervised sandboxes mixin."""
from __future__ import unicode_literals

import datetime
import os
import pytz
import uuid

from swiftclient import client

from ci_utils import unit_config

from tempurl import (
    OSTempUrlProvider,
    HPTempUrlProvider,
)


class SandboxNotFound(Exception):
    """Raised when a swift sandbox could not be found."""


class SandboxObject(object):
    """Swift sandbox object."""

    def __init__(self, **kwargs):
        self.url = kwargs.get('url')
        self.token = kwargs.get('token')
        self.container = kwargs.get('container')
        self.name = kwargs.get('name')
        self.size = kwargs.get('bytes')
        self.ctype = kwargs.get('content_type')

    def __unicode__(self):
        return self.title

    @property
    def title(self):
        return '{} ({} bytes, {})'.format(self.name, self.size, self.ctype)

    def get(self):
        """Get object contents from Swift."""
        return client.get_object(
            self.url, self.token, self.container, self.name)

    def copy(self, new_container):
        """Copy this object to a given container."""
        source = b'/{}/{}'.format(self.container, self.name)
        headers = [
            (b'Content-Length', b'0'),
            (b'X-Copy-From', source),
        ]
        client.put_object(self.url, self.token, new_container, self.name,
                          headers=headers)

    def drop(self):
        """Drop object contents from Swift."""
        client.delete_object(
            self.url, self.token, self.container, self.name)


class SandboxContainer(object):
    """Swift sandbox utility with tempurl support. """

    PREFIX = 'sandbox-'

    def __init__(self, url, token, options, **kwargs):
        self.url = url
        self.token = token
        self.options = options
        self.name = kwargs.get('name')
        self.size = kwargs.get('bytes')
        self.count = kwargs.get('count')

    def __unicode__(self):
        return self.title

    @property
    def identifier(self):
        """Return the UUIDv1 part of the sandbox container name."""
        return self.name.replace(self.PREFIX, '')

    @property
    def title(self):
        """Return a text line describing this sandbox."""
        return '{} ({} bytes, {} objects)'.format(
            self.identifier, self.size, self.count)

    def _now(self):
        """Local wrapper for datetime.now() for facilitating tests."""
        return datetime.datetime.now(tzinfo=pytz.utc)

    @property
    def age(self):
        """Returns the age of this sandbox based on its name.

        Calculates the age of this sandbox based on the UUIDv1 used
        as its identifier.

        param: return: TimeDelta corresponding to this sandbox age.
        """
        u = uuid.UUID(self.identifier)
        assert u.version == 1, 'Only UUIDv1 is supportted'
        uuid_timestamp = (u.time - 0x01b21dd213814000L) / 1e7
        created = datetime.datetime.fromtimestamp(uuid_timestamp, pytz.utc)
        return self._now() - created

    @property
    def objects(self):
        """Iterator of `SandboxObjects` within this sandbox."""
        header, contents = client.get_container(
            self.url, self.token, self.name)
        return [
            SandboxObject(url=self.url, token=self.token,
                          container=self.name, **c) for c in contents]

    def drop(self):
        """Remove the sandbox and its contents."""
        for obj in self.objects:
            obj.drop()
        client.delete_container(self.url, self.token, self.name)

    def _get_tempurl_provider(self):
        """Return the tempurl provider appropriate for this context."""
        if unit_config.is_hpcloud(self.options.auth_url):
            return HPTempUrlProvider(self.url, self.token, self.options)
        return OSTempUrlProvider(self.url, self.token, self.options)

    def get_temp_url(self, object_name, method='GET'):
        """Return a tempurl for the given object_name and method."""
        provider = self._get_tempurl_provider()
        provider.set_temp_key(self.options.secret_key)
        return provider.get_temp_url(self.name, object_name, method=method)

    def dump(self, path):
        """Dump sandbox contents to the given path."""
        for obj in self.objects:
            obj_filepath = os.path.join(path, obj.name)
            with open(obj_filepath, 'wb') as fd:
                headers, content = obj.get()
                fd.write(content)


class SupervisedSandboxesMixin(object):
    """Swift datastore mixin for supervised sandboxes utility."""

    def _get_container_name(self, identifier):
        """Return the container name for a given sandbox identifier."""
        return SandboxContainer.PREFIX + identifier

    def _get_sandboxes(self, name=None):
        """Iterator of existing sandboxes as `SandboxContainer`s.

        If name is given, user it to filter the list.
        """
        if name is None:
            prefix = SandboxContainer.PREFIX
        else:
            prefix = name
        header, contents = client.get_account(
            self.url, self.token, prefix=prefix)
        return [SandboxContainer(self.url, self.token, self.options, **c)
                for c in contents]

    @property
    def sandboxes(self):
        """Iterator on all existing `SandboxContainer`s."""
        return self._get_sandboxes()

    def get_sandbox(self, identifier):
        """Return an specific `SandboxContainer`."""
        if not identifier:
            raise SandboxNotFound()
        container_name = self._get_container_name(identifier)
        sandboxes = list(self._get_sandboxes(name=container_name))
        if len(sandboxes) == 0:
            raise SandboxNotFound()
        return sandboxes[0]

    def _get_sandbox_identifier(self):
        """Local wrapper to UUID for facilitating tests."""
        return str(uuid.uuid1())

    def create_sandbox(self):
        """Create a new `SandboxContainer`."""
        identifier = self._get_sandbox_identifier()
        container_name = self._get_container_name(identifier)
        client.put_container(self.url, self.token, container_name)
        return self.get_sandbox(identifier)
