# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Datastore TempURL provider classes"""
from __future__ import unicode_literals

import hashlib
import hmac
import time

from swiftclient import client


class OSTempUrlProvider(object):
    """OpenStack tempurl provider."""

    def __init__(self, url, token, options):
        """Manage local 'temp_key' and 'options' values."""
        self.url = url
        self.token = token
        self.options = options
        self.temp_key = None

    def set_temp_key(self, key):
        """Set a temporary key for this context.

        Sets 'X-Account-Meta-Temp-URL-Key'.
        """
        assert isinstance(key, bytes), "Key must be string/bytes"
        temp_key_headers = {
            'X-Account-Meta-Temp-URL-Key': key,
        }
        client.post_account(self.url, self.token, temp_key_headers)
        self.temp_key = key

    def _get_path(self, container, object_name):
        auth_url = self.url.split('/')[-1]
        return '/v1/{}/{}/{}'.format(auth_url, container, object_name)

    def _get_temp_signature(self, method, expires, path):
        assert self.temp_key is not None, 'TempKey is not set'
        body = '%s\n%s\n%s' % (method, expires, path)
        return hmac.new(self.temp_key, body, hashlib.sha1).hexdigest()

    def get_temp_url(self, container, object_name, expires=None,
                     method='GET'):
        """Assembly a Swift tempurl for the given object path.

        It allow access from unauthenticated clients.
        """
        if expires is None:
            expires = int(self.options.expires)
        if self.options.public_host:
            host = self.options.public_host
        else:
            host = '/'.join(self.url.split('/')[:3])
        path = self._get_path(container, object_name)
        expires = int(time.time() + expires)
        sig = self._get_temp_signature(method, expires, path)
        url = '{host}{path}?temp_url_sig={sig}&temp_url_expires={expires}'
        return url.format(host=host, path=path, sig=sig,
                          expires=expires)


class HPTempUrlProvider(OSTempUrlProvider):
    """HP tempurl provider."""

    def set_temp_key(self, key):
        """HP smartly reuses the secret key for signing tempurls.

        It means the manager can't simply refresh the tempurl key to
        invalidate old tempurls without causing a access *panic* in
        the system.
        """
        self.temp_key = str(self.options.secret_key)

    def _get_temp_signature(self, method, expires, path):
        """HP prepends '<tenant_id>:<access_key_id>:' in the signature.

        Simply because "they can".
        """
        assert self.temp_key is not None, 'TempKey is not set'
        body = '%s\n%s\n%s' % (method, expires, path)
        sig = hmac.new(self.temp_key, body, hashlib.sha1).hexdigest()
        return '{}:{}:{}'.format(
            self.options.tenant_id, self.options.access_key_id, sig)
