# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Swift datastore sandbox mixin tests."""
from __future__ import unicode_literals

from datetime import datetime
from mock import patch
import os
import pytz
import shutil
import tempfile
from testtools import TestCase

from gatekeeper.datastore.tests import FakeOptions
from gatekeeper.datastore.sandbox import (
    SandboxContainer,
    SandboxNotFound,
    SandboxObject,
    SupervisedSandboxesMixin,
)
from gatekeeper.datastore.tempurl import (
    OSTempUrlProvider,
    HPTempUrlProvider,
)


def patch_it(context, name):
    """Nested patcher helper."""
    patcher = patch(name)
    thing = patcher.start()
    context.addCleanup(patcher.stop)
    return thing


class TestSandboxObject(TestCase):
    """Tests for `SandboxObject`."""

    def setUp(self):
        # Creates a sample `SandboxObject`.
        super(TestSandboxObject, self).setUp()
        contents = {
            'url': 'http://fake.com/AUTH_1234',
            'token': '4321',
            'container': 'CTN',
            'name': 'foo.txt',
            'bytes': 11,
            'content_type': 'text/plain',
        }
        self.object = SandboxObject(**contents)

    def test_title(self):
        # `SandboxObject` implements 'title'.
        self.assertEquals(
            'foo.txt (11 bytes, text/plain)', self.object.title)

    def test_get(self):
        # `SandboxObject` implements 'get'.
        test_content = (['header'], 'CONTENT')
        mock_go = patch_it(self, 'swiftclient.client.get_object')
        mock_go.return_value = test_content
        # the headers and content are returned.
        self.assertEquals(test_content, self.object.get())
        # 'get_object' is called once for the object.
        mock_go.assert_called_once_with(
            self.object.url, self.object.token, self.object.container,
            'foo.txt')

    def test_copy(self):
        # `SandboxObject` implements server-side 'copy'.
        mock_do = patch_it(self, 'swiftclient.client.put_object')
        self.object.copy('wonderland')
        # 'put_object' is called once for the object.
        expected_headers = [
            (b'Content-Length', b'0'),
            (b'X-Copy-From', b'/CTN/foo.txt'),
        ]
        mock_do.assert_called_once_with(
            self.object.url, self.object.token, 'wonderland', 'foo.txt',
            headers=expected_headers)

    def test_drop(self):
        # `SandboxObject` implements 'drop'.
        mock_do = patch_it(self, 'swiftclient.client.delete_object')
        self.object.drop()
        # 'delete_object' is called once for the object.
        mock_do.assert_called_once_with(
            self.object.url, self.object.token, self.object.container,
            'foo.txt')


class TestSandboxContainer(TestCase):
    """Tests for `SandboxObject`."""

    def setUp(self):
        # Creates a sample `SandboxContainer`.
        super(TestSandboxContainer, self).setUp()
        contents = {
            'name': 'sandbox-02c5fc18-ca73-11e3-88d9-001c4229f9ac',
            'bytes': 17,
            'count': 1,
        }
        self.url = 'http://fake.com/AUTH_1234'
        self.token = '4321'
        self.options = FakeOptions()
        self.sandbox = SandboxContainer(
            self.url, self.token, self.options, **contents)

    def test_prefix(self):
        # `SandboxContainer` implements 'PREFIX' class attribute.
        self.assertEquals(
            'sandbox-', SandboxContainer.PREFIX)

    def test_representation(self):
        # `SandboxContainer` representation.
        self.assertEquals(
            '02c5fc18-ca73-11e3-88d9-001c4229f9ac (17 bytes, 1 objects)',
            unicode(self.sandbox))

    def test_title(self):
        # `SandboxContainer` implements 'title'.
        self.assertEquals(
            '02c5fc18-ca73-11e3-88d9-001c4229f9ac (17 bytes, 1 objects)',
            self.sandbox.title)

    def test_identifier(self):
        # `SandboxContainer` implements 'identifier'.
        self.assertEquals(
            '02c5fc18-ca73-11e3-88d9-001c4229f9ac',
            self.sandbox.identifier)

    def test_age(self):
        # `SandboxContainer` implements 'age'.
        with patch.object(self.sandbox, '_now') as mock_now:
            mock_now.return_value = datetime(2014, 04, 23, tzinfo=pytz.utc)
            age = self.sandbox.age
        self.assertEquals('0:51:33.217380', str(age))

    def test_objects(self):
        # `SandboxContainer` contained objects iterator.
        with patch('swiftclient.client.get_container') as mock_container:
            mock_container.return_value = (
                'ignored',
                [{'name': 'one', 'bytes': 1, 'content_type': 'um'},
                 {'name': 'two', 'bytes': 2, 'content_type': 'dois'}])
            sandbox_objects = self.sandbox.objects

        self.assertEquals(
            ['one (1 bytes, um)',
             'two (2 bytes, dois)'],
            [o.title for o in sandbox_objects])

    def test_drop(self):
        # `SandboxContainer` drop method delete the sandbox and its contents.
        mock_gc = patch_it(self, 'swiftclient.client.get_container')
        mock_do = patch_it(self, 'swiftclient.client.delete_object')
        mock_dc = patch_it(self, 'swiftclient.client.delete_container')
        # Simulates a sandbox with a single object.
        mock_gc.return_value = (
            'ignored', [{'name': 'one', 'bytes': 1, 'content_type': 'um'}])
        self.sandbox.drop()
        # 'delete_object' is called once for the contained object.
        mock_do.assert_called_once_with(
            self.url, self.token, self.sandbox.name, 'one')
        # 'delete_container' is called once for the sandbox container.
        mock_dc.assert_called_once_with(
            self.url, self.token, self.sandbox.name)

    def test_get_tempurl_provider_os(self):
        # `SandboxContainer` uses appropriate tempurl provider for OS.
        self.assertIsInstance(
            self.sandbox._get_tempurl_provider(), OSTempUrlProvider)

    def test_get_tempurl_provider_hp(self):
        # `SandboxContainer` uses appropriate tempurl provider for HP
        hp_options = FakeOptions(
            auth_url='http://hpcloud.com', secret_key=b'secret')
        hp_sandbox = SandboxContainer(self.url, self.token, hp_options)
        self.assertIsInstance(
            hp_sandbox._get_tempurl_provider(), HPTempUrlProvider)

    def test_get_temp_url(self):
        # `SandboxContainer` wraps the tempurl provider.
        # Known expiration time.
        mock_time = patch_it(self, 'time.time')
        mock_time.return_value = 1397702769
        # Suppress external access.
        patch_it(self, 'swiftclient.client.post_account')
        # And finaly builds a predictable tempurl.
        self.assertEquals(
            'http://fake.com/v1/AUTH_1234/'
            'sandbox-02c5fc18-ca73-11e3-88d9-001c4229f9ac/a_object?'
            'temp_url_sig=d9e11bcb0dbd4f6d4033f7c48a9cf978d58385ac&'
            'temp_url_expires=1397703069',
            self.sandbox.get_temp_url('a_object'))

    def mock_for_dump(self, filenames, test_content):
        # Auxiliary method for mocking swift client for dumping
        # sandbox contents.
        mock_gc = patch_it(self, 'swiftclient.client.get_container')
        mock_gc.return_value = (
            'ignored',
            [{'name': name, 'bytes': len(test_content),
              'content_type': 'plain/text'}
             for name in filenames])
        mock_go = patch_it(self, 'swiftclient.client.get_object')
        mock_go.return_value = ([], test_content)

    def test_dump(self):
        # `SandboxContainer` can be dumped to the local filesystem.
        tmp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmp_dir)
        self.mock_for_dump(['one', 'two'], b'A_CONTENT')
        self.sandbox.dump(tmp_dir)
        self.assertEquals(['one', 'two'], sorted(os.listdir(tmp_dir)))
        self.assertEquals(
            'A_CONTENT', open(os.path.join(tmp_dir, 'one')).read())
        self.assertEquals(
            'A_CONTENT', open(os.path.join(tmp_dir, 'two')).read())


class TestSupervisedSandboxesMixin(TestCase):
    """Tests for `SupervisedSandboxesMixin`."""

    def setUp(self):
        # Creates a sample `SupervisedSandboxesMixin`.
        super(TestSupervisedSandboxesMixin, self).setUp()
        self.mixin = SupervisedSandboxesMixin()
        self.mixin.url = 'http://a-url'
        self.mixin.token = 'a-token'
        self.mixin.options = FakeOptions()

    def test_get_container_name(self):
        # `SandboxContainer.PREFIX` is used to build container names.
        self.assertEquals(
            'sandbox-foo', self.mixin._get_container_name('foo'))

    def test_sandboxes(self):
        # `SupervisedSandboxesMixin` implements a sandbox iterator.
        with patch('swiftclient.client.get_account') as mock_account:
            mock_account.return_value = (
                'ignored',
                [{'name': 'one', 'bytes': 1, 'count': 1},
                 {'name': 'two', 'bytes': 2, 'count': 2}])
            sandboxes = self.mixin.sandboxes
        self.assertEquals(
            ['one (1 bytes, 1 objects)',
             'two (2 bytes, 2 objects)'],
            [s.title for s in sandboxes])

    def test_get_sandbox(self):
        # `SupervisedSandboxesMixin` implements a sandbox getter.
        mock_account = patch_it(self, 'swiftclient.client.get_account')
        mock_account.return_value = (
            'ignored', [{'name': 'new', 'bytes': 0, 'count': 0}])
        # Fetch an existing sandbox.
        sandbox = self.mixin.get_sandbox('new')
        mock_account.assert_called_once_with(
            self.mixin.url, self.mixin.token, prefix='sandbox-new')
        self.assertIsInstance(sandbox, SandboxContainer)
        self.assertEquals('new', sandbox.name)
        self.assertEquals(0, sandbox.size)
        self.assertEquals(0, sandbox.count)

    def test_get_sandbox_missing(self):
        # `SupervisedSandboxesMixin.get_sandbox` raises `SandboxNotFound`
        # on missing sandboxes.
        mock_account = patch_it(self, 'swiftclient.client.get_account')
        mock_account.return_value = ('ignored', [])
        # Try to fetch a missing sandbox.
        self.assertRaises(
            SandboxNotFound, self.mixin.get_sandbox, 'missing')
        mock_account.assert_called_once_with(
            self.mixin.url, self.mixin.token, prefix='sandbox-missing')

    def test_get_sandbox_empty_identifier(self):
        # `SupervisedSandboxesMixin.get_sandbox` raises `SandboxNotFound`
        # when called with empty identifier, swift is not even consulted.
        mock_account = patch_it(self, 'swiftclient.client.get_account')
        self.assertRaises(SandboxNotFound, self.mixin.get_sandbox, '')
        mock_account.assert_not_called()

    def test_create_sandbox(self):
        # `SupervisedSandboxesMixin` implements a sandbox creator.
        a_uuid = '02c5fc18-ca73-11e3-88d9-001c4229f9ac'
        mock_uuid = patch_it(self, 'uuid.uuid1')
        mock_uuid.return_value = a_uuid
        patch_it(self, 'swiftclient.client.put_container')
        mock_account = patch_it(self, 'swiftclient.client.get_account')
        container_name = self.mixin._get_container_name(a_uuid)
        mock_account.return_value = (
            'ignored',
            [{'name': container_name, 'bytes': 0, 'count': 0}])
        # Create a new sandbox.
        sandbox = self.mixin.create_sandbox()
        self.assertIsInstance(sandbox, SandboxContainer)
        self.assertEquals(container_name, sandbox.name)
        self.assertEquals(a_uuid, sandbox.identifier)
        self.assertEquals(0, sandbox.size)
        self.assertEquals(0, sandbox.count)
