# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Swift tempurl provider classes tests."""
from __future__ import unicode_literals

from mock import patch
from testtools import TestCase

from gatekeeper.datastore.tests import FakeOptions
from gatekeeper.datastore.tempurl import (
    OSTempUrlProvider,
    HPTempUrlProvider,
)


class TestOpenStackTempUrlProvider(TestCase):
    """Tests for OpenStack tempurl provider."""

    def setUp(self):
        super(TestOpenStackTempUrlProvider, self).setUp()
        url = 'http://os.com/AUTH_1234'
        token = '4321'
        options = FakeOptions()
        self.provider = OSTempUrlProvider(url, token, options)

    def test_instantiation(self):
        # On instantiation and no tempurl can be generated.
        self.assertIsNone(self.provider.temp_key)
        self.assertRaises(
            AssertionError,
            self.provider.get_temp_url, 'testing', 'test.txt')

    def test_set_temp_key(self):
        # Setting tempurl key uses 'client.post_account' to store
        # the given key in Swift.
        temp_key = b'A_KEY'
        with patch('swiftclient.client.post_account') as mock_post:
            self.provider.set_temp_key(temp_key)
        request_headers = {
            'X-Account-Meta-Temp-URL-Key': temp_key,
        }
        mock_post.assert_called_once_with(
            self.provider.url, self.provider.token, request_headers)
        self.assertEquals(temp_key, self.provider.temp_key)

    def test_set_temp_key_no_unicode(self):
        # tempurl key must be string (or bytes), unicode is not accepted
        self.assertRaises(
            AssertionError, self.provider.set_temp_key, 'unicode')

    def test_get_temp_url(self):
        # Once the temp key is set, we can fetch tempurls for a given
        # (container_name, object_name). Note that the tempurl contents
        # are time-dependent (thus time is mocked).
        self.provider.temp_key = b'A_KEY'
        with patch('time.time') as mock_time:
            mock_time.return_value = 1397702769
            temp_url = self.provider.get_temp_url('testing', 'test.txt')
        self.assertEquals(
            'http://os.com/v1/AUTH_1234/testing/test.txt?'
            'temp_url_sig=d894b91af7148cd9b08f86dba910bedd4aba1257&'
            'temp_url_expires=1397703069',
            temp_url)

    def test_get_temp_url_for_put(self):
        # tempurls can be generared for PUT as well, the signature
        # is differs from the GET one (above)
        self.provider.temp_key = b'A_KEY'
        with patch('time.time') as mock_time:
            mock_time.return_value = 1397702769
            temp_url = self.provider.get_temp_url(
                'testing', 'test.txt', method='PUT')
        self.assertEquals(
            'http://os.com/v1/AUTH_1234/testing/test.txt?'
            'temp_url_sig=368c07dc2f77c70991172d8d26f99885f778e9e8&'
            'temp_url_expires=1397703069',
            temp_url)

    def test_get_temp_url_expires(self):
        # tempurls can be generared with expiring time different than
        # the default one (300 s).
        self.assertEquals('300', self.provider.options.expires)
        self.provider.temp_key = b'A_KEY'
        with patch('time.time') as mock_time:
            mock_time.return_value = 1397702769
            temp_url = self.provider.get_temp_url(
                'testing', 'test.txt', expires=3600)
        self.assertEquals(
            'http://os.com/v1/AUTH_1234/testing/test.txt?'
            'temp_url_sig=c7e30a1d104f1e839955a887f89da78594ed5d2a&'
            'temp_url_expires=1397706369',
            temp_url)

    def test_get_temp_url_public_host(self):
        # tempurls can be generared for a configured 'public_host',
        # which is a different Swift frontend dedicated for external
        # access. This is the case for Canonical Prodstack.
        self.assertEquals('', self.provider.options.public_host)
        self.provider.options.public_host = 'http://public.com'
        self.provider.temp_key = b'A_KEY'
        with patch('time.time') as mock_time:
            mock_time.return_value = 1397702769
            temp_url = self.provider.get_temp_url(
                'testing', 'test.txt', expires=3600)
        self.assertEquals(
            'http://public.com/v1/AUTH_1234/testing/test.txt?'
            'temp_url_sig=c7e30a1d104f1e839955a887f89da78594ed5d2a&'
            'temp_url_expires=1397706369',
            temp_url)


class TestHPTempUrlProvider(TestCase):
    """Tests for HP tempurl provider."""

    def setUp(self):
        super(TestHPTempUrlProvider, self).setUp()
        url = 'http://hp.com/AUTH_1234'
        token = '4321'
        options = FakeOptions()
        self.provider = HPTempUrlProvider(url, token, options)

    def test_set_temp_key(self):
        # Differently than OpenStack spec, HP cloud re-uses the
        # a single 'secret_key' for signing tempurls.
        # set_temp_key() ignores the given 'key'.
        self.provider.set_temp_key(b'A_KEY')
        self.assertEquals(
            self.provider.options.secret_key, self.provider.temp_key)

    def test_get_temp_url(self):
        # HP decides that prepending 'tenant_id' and 'access_key_id'
        # to the tempurl signature would be great!
        self.provider.temp_key = self.provider.options.secret_key
        with patch('time.time') as mock_time:
            mock_time.return_value = 1397702769
            temp_url = self.provider.get_temp_url('testing', 'test.txt')
        self.assertEquals(
            'http://hp.com/v1/AUTH_1234/testing/test.txt?'
            'temp_url_sig=TEN_ID:ACC_KEY:1f663506e3faf5eda0da4a39f9a'
            '4afd5efb03141&'
            'temp_url_expires=1397703069',
            temp_url)
