# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Swift datastore ticket container mixin tests."""
from __future__ import unicode_literals

from mock import patch
import unittest

from gatekeeper.datastore.tests import FakeOptions
from gatekeeper.datastore.ticket import (
    TicketContainer,
    TicketContainerNotFound,
    TicketContainerMixin,
)


class TestTicketContainer(unittest.TestCase):
    """Tests for `TicketContainer`."""

    def setUp(self):
        # Creates a sample `TicketContainer`.
        super(TestTicketContainer, self).setUp()
        contents = {
            'name': 'ticket-3',
            'bytes': 17,
            'count': 1,
        }
        self.url = 'http://fake.com/AUTH_1234'
        self.token = '4321'
        self.options = FakeOptions()
        self.container = TicketContainer(
            self.url, self.token, self.options, **contents)

    def test_representation(self):
        # `TicketContainer` representation and base attributes.
        self.assertEquals('ticket-3', unicode(self.container))
        self.assertEquals('ticket-3', self.container.name)
        self.assertEquals(17, self.container.size)
        self.assertEquals(1, self.container.count)

    @patch('swiftclient.client.post_account')
    @patch('time.time')
    @patch('gatekeeper.datastore.ticket.TicketContainer._get_tempurl_key')
    @patch('swiftclient.client.get_container')
    def test_get_objects_tempurls(self, mocked_container, mocked_key,
                                  mocked_time, mocked_post):
        # `TicketContainer.get_objects_tempurls` returns a dictionary
        # keyed by the ticket objects (artifacts) names and their
        # corresponding tempurls (for GET) as values.
        mocked_container.return_value = (
            'ignored', [{'name': 'one'}, {'name': 'two'}])
        mocked_key.return_value = b'02c5fc18-ca73-11e3-88d9-001c4229f9ac'
        mocked_time.return_value = 1397702769

        self.assertEquals(
            {'one': 'http://fake.com/v1/AUTH_1234/ticket-3/one?'
                    'temp_url_sig=5de01a1853ae94eb824d4c12d2d1dafb0d4f1000&'
                    'temp_url_expires=1397703069',
             'two': 'http://fake.com/v1/AUTH_1234/ticket-3/two?'
                    'temp_url_sig=27a61b0817e51fdd1a64ca2d71d98d6d4a1adebf&'
                    'temp_url_expires=1397703069'},
            self.container.get_objects_tempurls())


class TestTicketContainerMixin(unittest.TestCase):
    """Tests for `TicketContainerMixin`."""

    def setUp(self):
        # Creates a sample `TicketContainerMixin`.
        super(TestTicketContainerMixin, self).setUp()
        self.mixin = TicketContainerMixin()
        self.mixin.url = 'http://a-url'
        self.mixin.token = 'a-token'
        self.mixin.options = FakeOptions()

    @patch('swiftclient.client.get_account')
    def test_get_container(self, mocked_account):
        # `TicketContainerMixin` implements a ticket container getter.
        mocked_account.return_value = (
            'ignored', [{'name': 'ticket-9', 'bytes': 0, 'count': 0}])
        container = self.mixin.get_ticket_container('9')
        mocked_account.assert_called_once_with(
            self.mixin.url, self.mixin.token, prefix='ticket-9')
        self.assertIsInstance(container, TicketContainer)
        self.assertEquals('ticket-9', container.name)
        self.assertEquals(0, container.size)
        self.assertEquals(0, container.count)

    @patch('swiftclient.client.get_account')
    def test_get_container_missing(self, mocked_account):
        # `TicketContainerMixin.get_ticket_container` raises
        # `TicketContainerNotFound` on missing tickets.
        mocked_account.return_value = ('ignored', [])
        with self.assertRaises(TicketContainerNotFound):
            self.mixin.get_ticket_container('10')
        mocked_account.assert_called_once_with(
            self.mixin.url, self.mixin.token, prefix='ticket-10')

    @patch('swiftclient.client.get_account')
    def test_get_container_exact(self, mocked_account):
        # `TicketContainerMixin` getter only returns exact matches.
        mocked_account.return_value = (
            'ignored', [{'name': 'ticket-9999', 'bytes': 0, 'count': 0}])
        with self.assertRaises(TicketContainerNotFound):
            self.mixin.get_ticket_container('9')
        mocked_account.assert_called_once_with(
            self.mixin.url, self.mixin.token, prefix='ticket-9')
