# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Swift ticket container hanlder mixin."""
from __future__ import unicode_literals

import uuid

from swiftclient import client

from ci_utils import unit_config
from tempurl import (
    OSTempUrlProvider,
    HPTempUrlProvider,
)


class TicketContainerNotFound(Exception):
    """Raised when a swift ticket container could not be found."""


class TicketContainer(object):
    """Swift ticket container utility with tempurl support. """

    def __init__(self, url, token, options, **kwargs):
        self.url = url
        self.token = token
        self.options = options
        self.name = kwargs.get('name')
        self.size = kwargs.get('bytes')
        self.count = kwargs.get('count')

    def __unicode__(self):
        return self.name

    def _get_tempurl_key(self):
        """Local wrapper to UUID for facilitating tests."""
        return str(uuid.uuid1())

    def _get_tempurl_provider(self):
        """Return the tempurl provider appropriate for this context."""
        if unit_config.is_hpcloud(self.options.auth_url):
            return HPTempUrlProvider(self.url, self.token, self.options)
        return OSTempUrlProvider(self.url, self.token, self.options)

    def get_objects_tempurls(self, method='GET'):
        """Return a tempurls for objects for the given method."""
        provider = self._get_tempurl_provider()
        provider.set_temp_key(self._get_tempurl_key())
        header, contents = client.get_container(
            self.url, self.token, self.name)
        results = []
        for content in contents:
            object_name = content['name']
            temp_url = provider.get_temp_url(
                self.name, object_name, method=method)
            results.append((object_name, temp_url))
        return dict(results)


class TicketContainerMixin(object):
    """Swift datastore mixin for ticket container utility."""

    def get_ticket_container(self, ticket_id):
        """Return a corresponding `TicketContainer`.

        Raises `TicketContainerNotFound` if it could not be found.
        """
        if not ticket_id:
            raise TicketContainerNotFound()
        container_name = 'ticket-{}'.format(ticket_id)
        headers, contents = client.get_account(
            self.url, self.token, prefix=container_name)
        if len(contents) == 0:
            raise TicketContainerNotFound()
        if contents[0].get('name') != container_name:
            raise TicketContainerNotFound()
        return TicketContainer(
            self.url, self.token, self.options, **contents[0])
