# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Restful API test suite."""
from __future__ import unicode_literals

import os
import json

from mock import patch
from testtools import TestCase
from webtest import TestApp

from ci_utils.testing.fixtures import override_last_revision
from gatekeeper.adapters.sourcepackageupload import (
    SourcePackageUploadFailure,
)
from gatekeeper.datastore.sandbox import SandboxNotFound
from gatekeeper.datastore.ticket import TicketContainerNotFound
from gatekeeper.resources.root import make_app
from gatekeeper.resources.v1 import API


class FakeObject(object):

    def __init__(self, title='A_OBJ'):
        self.title = title


class FakeSandbox(object):

    def __init__(self, identifier='A_UUID', title='A_OBJ'):
        self.identifier = identifier
        self.title = title
        self.objects = [FakeObject('obj1'), FakeObject('obj2')]

    def get_temp_url(self, obj_name, method):
        return b'{}/{}/{}'.format(method, self.identifier, obj_name)

    def drop(self):
        return


class FakeTicketContainer(object):

    def get_objects_tempurls(self):
        return {'A_ARTIFACT': 'http://ARTIFACT_TEMPURL'}


class FakeDataStore(object):

    def create_sandbox(self):
        return FakeSandbox()

    def get_sandbox(self, identifier):
        if identifier in ['', 'UNKNOWN']:
            raise SandboxNotFound()
        return FakeSandbox(identifier)

    def get_ticket_container(self, ticket_id):
        if ticket_id in ['', '1000']:
            raise TicketContainerNotFound()
        return FakeTicketContainer()


def fake_datastore_factory(klass):
    return FakeDataStore()


class BaseTestAPI(TestCase):
    """Base TestCase for testing API using a fake `DataStore`."""

    def setUp(self):
        # Run the restful server.
        super(BaseTestAPI, self).setUp()
        wsgi_app = make_app(fake_datastore_factory)
        self.app = TestApp(wsgi_app, relative_to='.')


class TestAPIv1(BaseTestAPI):
    """v1 Root API tests."""
    def tearDown(self):
        super(TestAPIv1, self).tearDown()
        if os.path.exists(API.CACHE_FILE):
            os.unlink(API.CACHE_FILE)

    def test_get_greeting(self):
        # GET on the root URL returns a greeting string as plain/text.
        resp = self.app.get('/api/v1/')
        self.assertEqual(200, resp.status_code, resp.body)
        self.assertEqual('text/plain', resp.content_type)
        self.assertEqual('CI-Airline gatekeeper.\n', resp.body)

    def test_other_methods_not_allowed(self):
        # PUT and DELETE are not allowed.
        self.app.put('/api/v1/', status=405)
        self.app.delete('/api/v1/', status=405)

    @patch('ci_utils.unit_config.get')
    @patch('ci_utils.json_status.JSONStatus.add_data_store_health')
    @patch('urllib2.urlopen')
    def test_get_status(self, urlopen, dstest, get):
        # GET on 'status' works and can be used for monitoring this
        # service.
        urlopen.side_effect = RuntimeError('foo')
        override_last_revision(self, 'test revision')
        resp = self.app.get('/api/v1/status')
        self.assertEqual(200, resp.status_code, resp.body)
        self.assertEqual('application/json', resp.content_type)
        self.assertEqual(
            [
                {"status": "okay", "value": "test revision",
                 "label": "revision"},
                {'status': 'fail', 'value': 'foo',
                 'label': 'keyserver connectivity'},
            ],
            json.loads(resp.body))
        dstest.assert_called_once()

    @patch('ci_utils.unit_config.get')
    @patch('ci_utils.json_status.JSONStatus.add_data_store_health')
    @patch('ci_utils.json_status.JSONStatus.add_url_test')
    def test_get_status_jsonp(self, urltest, dstest, get):
        # GET on 'status' supports JSONP requests, as used by the
        # Airline WebUI.
        override_last_revision(self, 'test revision')
        resp = self.app.get(
            '/api/v1/status?format=jsonp&callback=cb')
        self.assertEqual(200, resp.status_code, resp.body)
        self.assertEqual('application/json', resp.content_type)
        self.assertEqual(
            ['cb([{"status": "okay", "value": "test revision", '
             '"label": "revision"}])'],
            resp.body.splitlines())

    @patch('ci_utils.unit_config.get')
    @patch('ci_utils.json_status.JSONStatus')
    def test_get_status_cache(self, status, get):
        # ensure responses are cached
        type(status()).results = ['foo']
        resp = self.app.get('/api/v1/status')
        self.assertEqual(['foo'], json.loads(resp.body))

        # should get the cached value still
        type(status()).results = ['foobar']
        resp = self.app.get('/api/v1/status')
        self.assertEqual(['foo'], json.loads(resp.body))

        # should invalidate cache
        resp = self.app.get('/api/v1/status?refresh=1')
        self.assertEqual(['foobar'], json.loads(resp.body))

    def test_status_only_supports_get(self):
        # 'status' only supports GET.
        self.app.put('/api/v1/status', status=405)
        self.app.put('/api/v1/status', status=405)
        self.app.delete('/api/v1/status', status=405)


class TestTicketAPI(BaseTestAPI):
    """v1 Ticket API tests."""

    def test_get_not_found(self):
        # Empty GET on ticket/ or pointing to an absent ID returns 404.
        self.app.get('/api/v1/ticket/', status=404)
        self.app.get('/api/v1/ticket/1000/', status=404)

    def test_get_returns_the_list_of_artifacts(self):
        # GET on a ticket URL returns a list to tempurls for its
        # artifacts as a dictionary (JSON object). Since tempurls
        # expires, this page should not be cached in browsers.
        resp = self.app.get('/api/v1/ticket/1')
        self.assertEqual(200, resp.status_code, resp.body)
        self.assertEqual(
            'max-age=0, no-cache', str(resp.cache_control))
        self.assertEqual('application/json', resp.content_type)
        self.assertEqual(
            {'A_ARTIFACT': 'http://ARTIFACT_TEMPURL'},
            json.loads(resp.body))

    def test_get_returns_the_list_of_artifacts_jsonp(self):
        # GET on a ticket URL supports JSONP requests, as used by the
        # Airline WebUI ('callback(result)').
        resp = self.app.get(
            '/api/v1/ticket/1?format=jsonp&callback=cb')
        self.assertEqual(200, resp.status_code, resp.body)
        self.assertEqual('application/json', resp.content_type)
        self.assertEqual(
            ['cb({"A_ARTIFACT": "http://ARTIFACT_TEMPURL"})'],
            resp.body.splitlines())

    def test_other_methods_not_allowed(self):
        # POST, PUT and DELETE are not allowed.
        self.app.post('/api/v1/ticket/1', status=405)
        self.app.put('/api/v1/ticket/1', status=405)
        self.app.delete('/api/v1/ticket/1', status=405)


class TestSandboxAPI(BaseTestAPI):
    """v1 Sandbox API tests."""

    def test_post_creates_a_sandbox(self):
        # POST on the 'sandbox/' URL creates a new sandbox and returns
        # status 201, 'Location' header and body (as 'text/plain')
        # container the created sandbox path.
        resp = self.app.post('/api/v1/ticket/1/sandbox/')
        self.assertEqual(201, resp.status_code, resp.body)
        self.assertEqual('/api/v1/ticket/1/sandbox/A_UUID', resp.location)
        self.assertEqual('text/plain', resp.content_type)
        self.assertEqual('/api/v1/ticket/1/sandbox/A_UUID', resp.body)

    def test_post_on_a_sandbox_creates_a_ticket(self):
        # POST on a existing sandbox with valid contents creates
        # a ticket.
        with patch('gatekeeper.adapters.sourcepackageupload.'
                   'SourcePackageUpload.process') as mock_process:
            mock_process.return_value = 'A_TICKET_URL'
            resp = self.app.post('/api/v1/ticket/1/sandbox/A_UUID')
        mock_process.assert_called_once()
        self.assertEqual(201, resp.status_code, resp.body)
        self.assertEqual('A_TICKET_URL', resp.location)
        self.assertEqual('text/plain', resp.content_type)
        self.assertEqual('A_TICKET_URL', resp.body)

    def test_post_on_a_sandbox_with_invalid_contents(self):
        # POST on a existing sandbox with invalid contents return
        # the validation errors.
        def process_error(ticket_id):
            raise SourcePackageUploadFailure('Broken content ...')
        with patch('gatekeeper.adapters.sourcepackageupload.'
                   'SourcePackageUpload.process') as mock_process:
            mock_process.side_effect = process_error
            resp = self.app.post('/api/v1/ticket/1/sandbox/FAILS')
        mock_process.assert_called_once()
        self.assertEqual(200, resp.status_code, resp.body)
        self.assertEqual('text/plain', resp.content_type)
        self.assertEqual('Broken content ...', resp.body)

    def test_get_sandbox_lookup_failure(self):
        # GET on missing sandboxes returns 404.
        # See the FakeDataStore implementation.
        self.app.get('/api/v1/ticket/1/sandbox/', status=404)
        self.app.get('/api/v1/ticket/1/sandbox/UNKNOWN', status=404)

    def test_get_sandbox_contents(self):
        # GET on a sandbox URL lists its contents
        resp = self.app.get('/api/v1/ticket/1/sandbox/A_UUID')
        self.assertEqual(200, resp.status_code, resp.body)
        self.assertEqual('text/plain', resp.content_type)
        self.assertEqual('obj1\nobj2', resp.body)

    def test_delete_sandbox(self):
        # DELETE on a sandbox URL deletes the sandbox and its contents.
        resp = self.app.delete('/api/v1/ticket/1/sandbox/A_UUID')
        self.assertEqual(200, resp.status_code, resp.body)
        self.assertEqual('text/plain', resp.content_type)
        self.assertEqual('', resp.body)

    def test_put_object(self):
        # PUT on a sandbox object URL returns a tempurl for
        # uploading it.
        resp = self.app.put('/api/v1/ticket/1/sandbox/A_UUID/A_OBJ')
        self.assertEqual(200, resp.status_code, resp.body)
        self.assertEqual('text/plain', resp.content_type)
        self.assertEqual('PUT/A_UUID/A_OBJ', resp.body)

    def test_get_object(self):
        # GET on a sandbox object URL returns a tempurl for
        # downloading it.
        resp = self.app.get('/api/v1/ticket/1/sandbox/A_UUID/A_OBJ')
        self.assertEqual(200, resp.status_code, resp.body)
        self.assertEqual('text/plain', resp.content_type)
        self.assertEqual('GET/A_UUID/A_OBJ', resp.body)
