# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Testing datastore options factories."""
from __future__ import unicode_literals

import mock
import testtools

from ucitests import fixtures as uci_fixtures

from gatekeeper.wsgi import (
    environment_datastore_factory,
    unit_config_datastore_factory,
)


class TestDatastoreFactories(testtools.TestCase):
    """Test Gatekeeper datastore factories."""

    def isolate_datastore_environment(self, env=None):
        """Isolate and optionally override datastore (OS) environment."""
        ds_env = {
            'OS_AUTH_URL': 'http://test.com',
            'OS_USERNAME': 'username',
            'OS_PASSWORD': 'pwd',
            'OS_TENANT_NAME': 'a_tenant',
            'OS_REGION_NAME': 'a_region',
        }
        if env is not None:
            ds_env.update(env)
        uci_fixtures.isolate_from_env(self, ds_env)

    @mock.patch('swiftclient.client.get_auth',
                return_value=('A_URL', 'A_TOKEN'))
    def test_environment_factory_default(self, mock_swift):
        # `environment_datastore_factory` creates a datastore based on
        # environment variables (OS_* and others).
        self.isolate_datastore_environment(
            env={'OS_AUTH_URL': 'http://overriden.com'})
        ds = environment_datastore_factory(None)
        self.assertEquals('http://overriden.com', ds.options.auth_url)
        self.assertEquals('username', ds.options.username)
        self.assertEquals('pwd', ds.options.password)
        self.assertEquals('a_tenant', ds.options.tenant)
        self.assertEquals('a_region', ds.options.region)
        self.assertEquals('2', ds.options.api_version)
        self.assertEquals('', ds.options.access_key_id)
        self.assertEquals('', ds.options.tenant_id)
        self.assertEquals('', ds.options.secret_key)
        self.assertEquals('300', ds.options.expires)
        self.assertEquals('', ds.options.public_host)

    @mock.patch('ci_utils.unit_config.get_auth_config')
    @mock.patch('swiftclient.client.get_auth',
                return_value=('A_URL', 'A_TOKEN'))
    def test_unit_config_factory(self, mock_swift, mock_config):
        # `unit_config_datastore_factory` creates a datastore based on
        # unit_config variables. This is the default production factory.
        mock_config.return_value = {
            'auth_url': 'http://config.com',
            'auth_user': 'user',
            'auth_password': 'pass',
            'auth_tenant_name': 'ten_name',
            'auth_region': 're_gion',
            'auth_access_key_id': 'a_key',
            'auth_tenant_id': 'ten_id',
            'tempurl_signing_key': 'secret',
            'tempurl_expires': '666',
            'tempurl_public_host': 'http://public-config.com'
        }
        ds = unit_config_datastore_factory(None)
        self.assertEquals('http://config.com', ds.options.auth_url)
        self.assertEquals('user', ds.options.username)
        self.assertEquals('pass', ds.options.password)
        self.assertEquals('ten_name', ds.options.tenant)
        self.assertEquals('re_gion', ds.options.region)
        self.assertEquals('2', ds.options.api_version)
        self.assertEquals('a_key', ds.options.access_key_id)
        self.assertEquals('ten_id', ds.options.tenant_id)
        self.assertEquals('secret', ds.options.secret_key)
        self.assertEquals('666', ds.options.expires)
        self.assertEquals('http://public-config.com', ds.options.public_host)
