# -*- coding: utf-8 -*-
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Restish v1 API."""
from __future__ import unicode_literals

import contextlib
import fcntl
import json
import os
import time

from restish import (
    http,
    resource,
)

from ci_utils import json_status, unit_config
from gatekeeper.adapters.sourcepackageupload import (
    SourcePackageUpload,
    SourcePackageUploadFailure,
)
from gatekeeper.datastore.sandbox import SandboxNotFound
from gatekeeper.datastore.ticket import TicketContainerNotFound
from gatekeeper.resources import GatekeeperResource


class SandboxAPI(GatekeeperResource):
    """Sandbox API

    - POST '/' -> creates a sandbox
    - POST '/{sandbox_id}' -> creates a ticket
    - GET '/{sandbox_id}' -> list sandbox objects
    - DELETE '{sandbox_id}' -> delete sandbox and its contents
    - GET '/{sandbox_id}/{object_name}' -> returns a tempurl for
      downloading the object
    - PUT '/{sandbox_id}/{object_name}' -> returns a tempurl for
      uploading the object
    """

    def __init__(self, ticket_id, identifier):
        super(SandboxAPI, self).__init__()
        self.ticket_id = ticket_id
        self.identifier = identifier

    @property
    def path(self):
        return '{}/ticket/{}'.format(API.PATH, self.ticket_id)

    @resource.POST()
    def create(self, request):
        # POSTs with no identifier result in new sandboxes.
        if self.identifier is '':
            sandbox = self.datastore.create_sandbox()
            location = b'{}/sandbox/{}'.format(self.path, sandbox.identifier)
            self.logger.info('Sandbox {} created.'.format(location))
            return http.created(location, list(API.headers), location)
        # POST on an existing sandbox creates and validates a new
        # SourcePackageUpload.
        try:
            sandbox = self.datastore.get_sandbox(self.identifier)
        except SandboxNotFound:
            return http.not_found()
        spu = SourcePackageUpload(sandbox)
        try:
            ticket_url = spu.process(self.ticket_id)
        except SourcePackageUploadFailure as failure:
            self.logger.info(
                'Sandbox {}: {}'.format(self.identifier, str(failure)))
            return http.ok(list(API.headers), str(failure))
        self.logger.info(
            'SPU for ticket {} created.'.format(self.ticket_id))
        return http.created(
            str(ticket_url), list(API.headers), str(ticket_url))

    def _render_response(self, payload):
        return http.ok(list(API.headers), str(payload))

    @resource.GET()
    def contents(self, request):
        try:
            sandbox = self.datastore.get_sandbox(self.identifier)
        except SandboxNotFound:
            return http.not_found()
        objects = [o.title for o in sandbox.objects]
        return self._render_response('\n'.join(objects))

    @resource.DELETE()
    def drop(self, request):
        try:
            sandbox = self.datastore.get_sandbox(self.identifier)
        except SandboxNotFound:
            return http.not_found()
        sandbox.drop()
        self.logger.info(
            'Sandbox {} deleted.'.format(sandbox.identifier))
        return self._render_response('')

    def _get_temp_url(self, object_name, method):
        try:
            sandbox = self.datastore.get_sandbox(self.identifier)
        except SandboxNotFound:
            return http.not_found()
        temp_url = sandbox.get_temp_url(object_name, method)
        return self._render_response(temp_url)

    @resource.child('{object_name}')
    def object_wrapper(self, request, segments, object_name):
        allow = [b'GET', b'PUT']
        if request.method not in allow:
            return http.method_not_allowed(allow)
        return self._get_temp_url(object_name, request.method)


class TicketAPI(GatekeeperResource):
    """Ticket API

    - GET '/{ticket_id}' -> returns a list of tempurls for all `Ticket`
      artifacts.
    - 'sandbox/...' -> traverse to `SandboxAPI`
    """

    def __init__(self, ticket_id):
        super(TicketAPI, self).__init__()
        self.ticket_id = ticket_id

    @resource.GET()
    def contents(self, request):
        try:
            container = self.datastore.get_ticket_container(self.ticket_id)
        except TicketContainerNotFound:
            return http.not_found()
        payload = json.dumps(container.get_objects_tempurls())
        if request.GET.get('format') == 'jsonp':
            callback = request.GET.get('callback')
            payload = b'{}({})\n'.format(callback, payload)
        else:
            payload = b'{}\n'.format(payload)
        headers = [
            (b'Cache-Control', 'max-age=0, no-cache'),
            (b'Content-Type', b'application/json')
        ]
        return http.ok(headers, payload)

    @resource.child('sandbox/{sandbox_identifier}')
    def sandbox_wrapper(self, request, segments, sandbox_identifier):
        return SandboxAPI(self.ticket_id, sandbox_identifier)


class API(GatekeeperResource):
    """Base Gatekeeper API provider.

    - GET '' -> greeting message
    - GET 'status' -> JSON/JSONP status monitoring
    - 'ticket/...' -> traverse to `TicketAPI`
    """

    PATH = '/api/v1'
    GREETINGS = b'CI-Airline gatekeeper.\n'
    CACHE_FILE = '/tmp/gk-status.cache'

    headers = (
        (b'Content-Type', b'text/plain'),
    )

    def __init__(self):
        super(API, self).__init__()

    @resource.child('')
    def root(self, request, segments):
        allow = [b'GET']
        if request.method not in allow:
            return http.method_not_allowed(allow)
        return http.ok(list(self.headers), self.GREETINGS)

    @contextlib.contextmanager
    def _status_cache(self, request):
        with open(API.CACHE_FILE, 'a+') as f:
            try:
                fcntl.lockf(f, fcntl.LOCK_EX)

                data = []
                age = time.time() - os.fstat(f.fileno()).st_mtime
                timed_out = (age > json_status.CACHE_TIMEOUT)
                if not timed_out and request.GET.get('refresh') != '1':
                    # don't try and load a newly created file
                    f.seek(0, os.SEEK_END)
                    if f.tell() > 0:
                        f.seek(0)
                        data = json.load(f)
                yield data

                f.seek(0)
                f.truncate()
                json.dump(data, f)
            finally:
                fcntl.lockf(f, fcntl.LOCK_UN)

    @resource.child()
    def status(self, request, segments):
        allow = [b'GET']
        if request.method not in allow:
            return http.method_not_allowed(allow)

        with self._status_cache(request) as payload:
            if not payload:
                status = json_status.JSONStatus()
                status.add_deployed_revision()
                # find the key for our foobar@example.com user
                url = 'http://' + unit_config.get('keyserver_url')
                url += '/pks/lookup?op=get&search=0xD08F8E66B60D42CE'
                status.add_url_test('keyserver connectivity', url)
                # Check access to the data store.
                ticket = 'gk-test-%d' % time.time()
                ds = status.add_data_store_health(ticket)
                if ds:
                    ds.delete(True)
                payload.extend(status.results)

        payload = json.dumps(payload)
        if request.GET.get('format') == 'jsonp':
            callback = request.GET.get('callback')
            payload = b'{}({})\n'.format(callback, payload)
        else:
            payload = b'{}\n'.format(payload)
        return http.ok([('Content-Type', 'application/json')], payload)

    @resource.child('ticket/{ticket_id}')
    def ticket_wrapper(self, request, segments, ticket_id):
        try:
            self.datastore.get_ticket_container(ticket_id)
        except TicketContainerNotFound:
            return http.not_found()
        return TicketAPI(ticket_id)
