#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import fcntl
import hashlib
import logging
import os
import re
import shutil
import subprocess
import textwrap
import time
import requests
import urlparse
import uuid

import ci_utils

from imagebuilder import ImageException
from ci_utils.image_store import ImageStore
from ci_utils import dump_stack, data_store, launchpad


log = logging.getLogger('Imagebuilder')


def _get_args():
    parser = argparse.ArgumentParser(description='Modify a cloud image')
    parser.add_argument(
        "-i", "--image", required=True,
        help="URL for the cloud image to modify"),
    parser.add_argument(
        "-r", "--repository", required=True, action='append',
        help="List of PPAs to add")
    parser.add_argument(
        "-p", "--package", required=True, action='append',
        help="List of extra packages to install")
    parser.add_argument(
        "-s", "--series", required=True,
        help="Series of Ubuntu the image is based on")

    args = parser.parse_args()
    return args


def _parse_ppas(lp, ppalist, series):
    aptlines = []
    keys = {}
    for ppa in ppalist:
        ppamatch = re.match("ppa:(.+)/(.+)", ppa)
        # This will throw an exception for malformed ppas, intentional for now
        (team, repo) = ppamatch.groups()
        ppa = lp.people[team].getPPAByName(name=repo)
        keys[ppa.name] = ppa.signing_key_fingerprint
        if ppa.private:
            # lp.me won't work, you have to get yourself by person
            user = lp.people[launchpad.get_lp_user_name()]
            url = user.getArchiveSubscriptionURL(archive=ppa)
        else:
            url = 'http://ppa.launchpad.net/{}/{}/ubuntu'.format(team, repo)
        aptlines.append('deb {} {} main'.format(url, series))
        aptlines.append('deb-src {} {} main'.format(url, series))
    return aptlines, keys


def _chrooted(chroot, cmd):
    log.info('Running command in image: {0}'.format(cmd))
    if type(cmd) is not list:
        cmd = cmd.split(' ')
    output = subprocess.check_output(
        ['chroot', chroot] + cmd, stderr=subprocess.STDOUT)
    if output:
        log.info(output)


def _apt_get_update(chroot):
    # apt-get update isn't always reliable, give it a few tries
    retries = 5
    for x in range(retries):
        try:
            _chrooted(chroot, "/usr/bin/apt-get update")
            break
        except subprocess.CalledProcessError as e:
            x += 1
            if x == retries:
                raise e
            else:
                # retry delays will be 1, 16, 81, 256, 625
                # ie about 1s, 16s, 1.5m, 4m, 10m
                timeout = x * x * x * x
                fmt = 'apt-get update failed, wait %ds and try %d more times'
                log.info(fmt, timeout, retries - x)
                time.sleep(timeout)


def _setup_apt_mirror(status_cb, fname):
    try:
        mirror = ci_utils.unit_config.get('apt_mirror')
    except KeyError:
        status_cb('no apt mirror configured')
        return False
    status_cb('updating apt mirror to: {}'.format(mirror))
    with open(fname) as f:
        lines = f.readlines()
    with open(fname, 'w') as f:
        for line in lines:
            f.write(line.replace('archive.ubuntu.com', mirror))
    return True


def _setup_ppas(chroot, repos, series, status_cb):
    status_cb('importing ppa signing keys...')
    lp = launchpad.lp_login()

    status_cb('grabbing ppa subscription urls and keys...')
    deblines, keys = _parse_ppas(lp, repos, series)

    for ppa, key in keys.iteritems():
        status_cb('importing signing key for: %s' % ppa)
        cmd_base = 'apt-key adv --keyserver keyserver.ubuntu.com --recv-keys '
        _chrooted(chroot, cmd_base + key)

    status_cb('adding ppas to apt list...')
    with open('%s/etc/apt/sources.list.d/ci.list' % chroot, 'a+') as f:
        for debline in deblines:
            f.write('\n%s' % debline)

    status_cb('running apt-get-update...')
    _apt_get_update(chroot)


def _run_chroot_cmds(chroot, pkglist, status_cb):
    '''Run commands in the chroot to do update the sources and add packages
    '''
    status_cb('Adding requested packages to the image...')
    os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
    _chrooted(chroot, 'apt-get install -y {}'.format(' '.join(pkglist)))
    _chrooted(chroot, "/usr/bin/apt-get clean")

    # create a big null data file and remove it to make it more compressable
    status_cb('updating chroot to be more compressible...')
    zerofile = '%s/zero.tmp' % chroot
    filldata = '\0' * 4096
    with open(zerofile, 'wb') as f:
        while True:
            try:
                f.write(filldata)
            except IOError:
                break
    os.unlink(zerofile)


def _setup_grub(mountpoint, status_cb):
    '''ensure grub uses the proper rootfs

    When installing kernels grub will reconfigure but set the wrong root
    device to boot from. This ensures its probed correctly
    '''
    status_cb('ensuring grub uses proper rootfs')
    path = os.path.join(mountpoint, 'etc/default/grub.d/99-ci-eng.cfg')
    if not os.path.exists(path):
        # FIXME: On precise, this all appears to have changed, but I
        # don't know yet whether this will break kernel installs
        return
    with open(path, 'w') as f:
        f.write('LINUX_ROOT_DEVICE="LABEL=cloudimg-rootfs"\n')
        f.write('GRUB_DEVICE="LABEL=cloudimg-rootfs"\n')


def _md5_for_url(url):
    md5url = os.path.join(os.path.dirname(url), 'MD5SUMS')
    response = requests.get(md5url)
    filename = os.path.basename(url)
    md5_pattern = '([a-f0-9]+) \*' + filename
    md5_search = re.search(md5_pattern, response.text)
    if not md5_search:
        log.warn('No md5sum found for image at {}'.format(url))
        return ''
    return md5_search.group(1)


def _download(url, path, retry=3):
    ''' Download from a url, retry up to 3 times by default '''
    urlpath = urlparse.urlsplit(url).path
    filename = os.path.basename(urlpath)
    filename = os.path.join(path, filename)
    try:
        response = requests.get(url, stream=True)
        response.raise_for_status()
        with open(filename, 'wb') as f:
            for chunk in response.iter_content(chunk_size=64 * 1024):
                # Ignore keep-alives
                if not chunk:
                    continue
                f.write(chunk)
                f.flush()
    except (requests.ConnectionError, requests.HTTPError,
            ImageException) as e:
        if retry == 1:
            raise(e)
        log.warn('Download failed, retrying in 30 sec')
        time.sleep(30)
        return _download(url, path, retry - 1)
    return filename


def _md5_of_file(filename):
    ''' return the md5 of a file '''
    md5 = hashlib.md5()
    chunksize = 1024 * 1024
    with open(filename) as f:
        while True:
            chunk = f.read(chunksize)
            if not chunk:
                break
            md5.update(chunk)
    return md5.hexdigest()


def _nbd_loaded():
    log.info('Checking to see if nbd is loaded')
    with open('/proc/modules') as f:
        if '\nnbd ' in f.read():
            return True
        else:
            return False


def _load_nbd():
    '''load the nbd kernel module unless it's loaded already'''
    if not _nbd_loaded():
        runcmd('modprobe nbd')
    if not _nbd_loaded():
        raise ImageException(
            'nbd module is unavailable and could not be loaded')


def runcmd(cmd):
    log.info('Running command: {0}'.format(cmd))
    output = subprocess.check_output(cmd.split(' '))
    if output:
        log.info(output)


class CloudImage(object):
    def __init__(self, image_path, tmpdir):
        image_size = ci_utils.unit_config.get('image_size')

        # we need to ensure only one of these runs on the system
        log.info('waiting for cloud image lock')
        self.lockfile = open('/tmp/cloudimage.lck', 'a+')
        fcntl.lockf(self.lockfile, fcntl.LOCK_EX)

        log.info('resizing image to be %s', image_size)
        subprocess.check_call(['qemu-img', 'resize', image_path, image_size])
        self.mountpoint = self._mount_image(image_path, tmpdir)

    def __enter__(self):
        return self.mountpoint

    def __exit__(self, exc, value, tb):
        self._umount_image()
        fcntl.lockf(self.lockfile, fcntl.LOCK_UN)
        self.lockfile.close()

    def _mount_image(self, image_path, tmpdir):
        # Find all the unused network block devices
        mountpoint = '%s/mountpoint' % tmpdir
        os.mkdir(mountpoint)
        nbd_devs = [
            dev for dev in os.listdir('/sys/block') if dev.startswith('nbd')
            and not os.path.exists('/sys/block/%s/pid' % dev)
        ]
        self.nbd_dev = nbd_devs[0]
        runcmd('qemu-nbd -c /dev/%s %s' % (self.nbd_dev, image_path))
        for x in xrange(10):
            if os.path.exists('/sys/block/%s/pid' % self.nbd_dev):
                break
            time.sleep(1)
        else:
            raise ImageException('Block device %s never became available' %
                                 self.nbd_dev)
        runcmd('udevadm settle')
        runcmd('mount /dev/%sp1 %s' % (self.nbd_dev, mountpoint))

        # Mount /proc /sys and /dev in the image
        for bmount in ['/proc', '/sys', '/dev']:
            runcmd('mount --bind %s %s/%s' % (bmount, mountpoint, bmount))

        # Creating /run/lock
        run_lock = os.path.join(mountpoint, 'run', 'lock')
        os.mkdir(run_lock)
        runcmd(
            'mount -t tmpfs -o nodev,noexec,nosuid,size=5242880 none '
            '{}'.format(run_lock))
        # Backup resolv.conf and copy in the one from the local system
        if os.path.lexists('%s/etc/resolv.conf' % mountpoint):
            shutil.move('%s/etc/resolv.conf' % mountpoint,
                        '%s/etc/resolv.conf.orig' % mountpoint)
        shutil.copy('/etc/resolv.conf', '%s/etc/resolv.conf' % mountpoint)
        policyfile = textwrap.dedent("""\
            #!/bin/sh
            while true; do
            case "$1" in
                -*) shift ;;
                makedev) exit 0;;
                x11-common) exit 0;;
                *) exit 101;;
            esac
            done""")
        with open('%s/usr/sbin/policy-rc.d' % mountpoint, 'w') as f:
            f.write(policyfile)
        os.chmod('%s/usr/sbin/policy-rc.d' % mountpoint, 0o755)

        return mountpoint

    def _umount_image(self):
        for bmount in ['/proc', '/sys', '/dev']:
            runcmd('umount %s/%s' % (self.mountpoint, bmount))
        # Removing /run/lock
        run_lock = os.path.join(self.mountpoint, 'run', 'lock')
        runcmd('umount {}'.format(run_lock))
        shutil.rmtree(run_lock)
        if os.path.lexists('%s/etc/resolv.conf.orig' % self.mountpoint):
            shutil.move('%s/etc/resolv.conf.orig' % self.mountpoint,
                        '%s/etc/resolv.conf' % self.mountpoint)
        os.unlink('%s/usr/sbin/policy-rc.d' % self.mountpoint)
        runcmd('umount %s' % self.mountpoint)
        runcmd('qemu-nbd -d /dev/%s' % self.nbd_dev)


def _download_needed(glance_filename):
    ''' Check to see if we already have the image in glance'''
    glance = ImageStore(ci_utils.unit_config.get_auth_config())
    image_list = [x.name for x in glance.glance.images.list()]
    if glance_filename in image_list:
        log.info(
            '{} cached in glance, using cached copy'.format(glance_filename))
        return False  # No, we don't need to download it
    else:
        log.info(
            '{} not cached in glance, downloading...'.format(glance_filename))
        return True  # Yes, download is needed


def get_image(image, status_cb=None):
    '''Copy the requested cloud image to glance'''
    # API Version 20140721
    if not status_cb:
        status_cb = lambda x: 0  # just provide a no-op

    image_md5 = _md5_for_url(image)
    upload_filename = image_md5 + '.img'
    if image_md5 and not _download_needed(upload_filename):
        status_cb('Image {} already in glance, no download '
                  'needed'.format(upload_filename))
        return upload_filename
    else:
        status_cb('MD5 not detected for image at {}, downloading from '
                  'scratch'.format(image))

    with ci_utils.TmpDir() as tmpdir:
        status_cb('downloading %s...' % image)
        image_path = _download(image, tmpdir)
        real_md5 = _md5_of_file(image_path)
        if image_md5 and image_md5 != real_md5:
            # The md5sum was specified, but doesn't match
            raise ImageException('md5sum expected was {} but md5sum found'
                                 'was {}'.format(image_md5, real_md5))
        # Just in case we couldn't get the md5 from the site
        upload_filename = real_md5 + '.img'
        # We need to check if it's already in glance again so we don't
        # end up with a duplicate upload!
        if _download_needed(upload_filename):
            glance = ImageStore(ci_utils.unit_config.get_auth_config())
            status_cb(
                'Uploading image to glance as {}'.format(upload_filename))
            glance.image_create(upload_filename, image_path)
        else:
            status_cb('Image {} already found in glance, not '
                      'redownloading'.format(upload_filename))
        return upload_filename


def modify_image(image, repos, series, packages, ds, status_cb=None):
    '''Add packges and PPAs to a cloud image mounted in a chroot'''
    if not status_cb:
        status_cb = lambda x: 0   # just provide a no-op

    status_cb('loading nbd...')
    _load_nbd()
    with ci_utils.TmpDir() as tmpdir:
        status_cb('downloading %s...' % image)
        image_path = _download(image, tmpdir)
        with CloudImage(image_path, tmpdir) as mountpoint:
            status_cb('updating image %s...' % image)

            _setup_grub(mountpoint, status_cb)

            _setup_apt_mirror(
                status_cb, os.path.join(mountpoint, 'etc/apt/sources.list'))
            _setup_ppas(mountpoint, repos, series, status_cb)

            _run_chroot_cmds(mountpoint, packages, status_cb)

        glance = ImageStore(ci_utils.unit_config.get_auth_config())
        converted_image_path = os.path.join(tmpdir, 'new.img')
        status_cb('Shrinking the converted image...')
        runcmd('qemu-img convert -O qcow2 -c %s %s' % (
            image_path, converted_image_path))
        upload_filename = str(uuid.uuid4()) + '.img'
        status_cb(
            'Uploading converted image to glance as %s' % upload_filename)
        glance.image_create(upload_filename, converted_image_path)
        try:
            ds.change_visibility()
            with open(converted_image_path, 'r') as fp:
                location = ds.put_file(upload_filename, fp)
        except data_store.DataStoreException:
            status_cb('ERROR: unable to upload image to swift')
        return upload_filename, location


def main():
    import mock
    ds = mock.Mock()
    dump_stack.install_stack_dump_signal()
    args = _get_args()

    def cb(msg):
        print(msg)

    try:
        return modify_image(args.image, args.repository, args.series,
                            args.package, ds, cb)
    except subprocess.CalledProcessError as e:
        print(e.output)

if __name__ == '__main__':
    main()
