# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import tempfile
import textwrap
import unittest

import mock

from imagebuilder import cloud_image


class TestCloudImage(unittest.TestCase):
    def setUp(self):
        super(TestCloudImage, self).setUp()

    @mock.patch('ci_utils.unit_config')
    def test_apt_mirror_not_configured(self, unit_config):
        unit_config.get.side_effect = KeyError('apt_mirror')
        status_cb = mock.Mock()
        self.assertFalse(cloud_image._setup_apt_mirror(status_cb, '/ignored'))

    @mock.patch('ci_utils.unit_config')
    def test_apt_mirror_configured(self, unit_config):
        unit_config.get.return_value = 'foo.archive.ubuntu.com'
        status_cb = mock.Mock()

        orig = textwrap.dedent('''\
        deb http://archive.ubuntu.com/ubuntu saucy main restricted universe
        deb http://archive.ubuntu.com/ubuntu saucy-updates main restricted
        deb http://security.ubuntu.com/ubuntu saucy-security main restricted
        ''')
        expected = textwrap.dedent('''\
        deb http://foo.archive.ubuntu.com/ubuntu saucy main restricted universe
        deb http://foo.archive.ubuntu.com/ubuntu saucy-updates main restricted
        deb http://security.ubuntu.com/ubuntu saucy-security main restricted
        ''')

        with tempfile.NamedTemporaryFile() as f:
            f.write(orig)
            f.flush()
            self.assertTrue(cloud_image._setup_apt_mirror(status_cb, f.name))
            with open(f.name) as f:
                self.assertEqual(expected, f.read())

    @mock.patch('imagebuilder.cloud_image.launchpad')
    def test_parse_ppas(self, launchpad):
        launchpad.get_lp_user_name.return_value = 'foo'

        lp = mock.MagicMock()
        ppa = mock.Mock()
        ppa.name = 'foo'
        ppa.signing_key_fingerprint = 'foofp'
        ppa.private = True
        lp.people['foo'].getPPAByName.return_value = ppa
        lp.people['foo'].getArchiveSubscriptionURL.return_value = 'foo_url'

        series = 'precise'
        deblines, keys = cloud_image._parse_ppas(lp, ['ppa:/foo/bar'], series)
        self.assertEqual({'foo': 'foofp'}, keys)
        expected = ['deb foo_url precise main', 'deb-src foo_url precise main']
        self.assertListEqual(expected, deblines)

    @mock.patch('time.sleep')
    @mock.patch('imagebuilder.cloud_image._chrooted')
    def test_apt_get_update_fails(self, chrooted, sleep):
        chrooted.side_effect = subprocess.CalledProcessError(1, 'cmd', 'out')
        with self.assertRaises(subprocess.CalledProcessError):
            cloud_image._apt_get_update('/blah')

    @mock.patch('time.sleep')
    @mock.patch('imagebuilder.cloud_image._chrooted')
    def test_apt_get_update_retries(self, chrooted, sleep):
        '''ensure it can apt-get update works after retrying'''
        chrooted.side_effect = [
            subprocess.CalledProcessError(1, 'cmd', 'out'),
            subprocess.CalledProcessError(1, 'cmd', 'out'),
            None,
        ]
        cloud_image._apt_get_update('/blah')
        self.assertEqual(3, chrooted.call_count)
        self.assertEqual(2, sleep.call_count)

    @mock.patch('imagebuilder.cloud_image.launchpad')
    @mock.patch('imagebuilder.cloud_image._chrooted')
    @mock.patch('__builtin__.open')
    def test_setup_ppas(self, mopen, chroot, lp):
        status_cb = mock.Mock()
        series = 'saucy'
        ppas = ['ppa:user/foo', 'ppa:team/bar']
        mopen.return_value = mock.MagicMock(spec=file)
        fd = mopen.return_value.__enter__.return_value
        cloud_image._setup_ppas('/chroot', ppas, series, status_cb)
        self.assertTrue(fd.write.call_count)
        # todo look at chroot calls


class TestGetImage(unittest.TestCase):
    def setUp(self):
        super(TestGetImage, self).setUp()
        self.fake_image_url = 'http://somewhere.fake/current/image.img'

    @mock.patch('ci_utils.unit_config.get_auth_config')
    def test_get_image_ok(self, get_auth_config):
        ''' Image needs to be downloaded, and download is good '''
        cloud_image._md5_for_url = mock.Mock(return_value='deadbeef')
        cloud_image._md5_of_file = mock.Mock(return_value='deadbeef')
        cloud_image._download_needed = mock.Mock(return_value=True)
        cloud_image.ImageStore = mock.Mock()
        upload_filename = cloud_image.get_image(self.fake_image_url)
        self.assertEquals(upload_filename, 'deadbeef.img')

    def test_get_image_bad_md5(self):
        ''' Image needs download, downloaded md5 mismatch '''
        cloud_image._md5_for_url = mock.Mock(return_value='deadbeef')
        cloud_image._md5_of_file = mock.Mock(return_value='aaaaaaaa')
        cloud_image._download_needed = mock.Mock(return_value=True)
        cloud_image._download = mock.Mock()
        cloud_image.ImageStore = mock.Mock()
        self.assertRaises(cloud_image.ImageException, cloud_image.get_image,
                          self.fake_image_url)

    def test_get_image_no_download(self):
        ''' Download is already cached, doesn't need redownload '''
        cloud_image._md5_for_url = mock.Mock(return_value='deadbeef')
        cloud_image._md5_of_file = mock.Mock(return_value='deadbeef')
        cloud_image._download_needed = mock.MagicMock(return_value=False)
        cloud_image._download = mock.Mock()
        cloud_image.ImageStore = mock.Mock()
        upload_filename = cloud_image.get_image(self.fake_image_url)
        self.assertEquals(upload_filename, 'deadbeef.img')
        assert not cloud_image._download.called
