#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import atexit
import base64
from distutils import (
    dir_util,
    spawn,
)
import glob
import gzip
import hashlib
import importlib
import itertools
import logging
import os
import random
import shutil
import stat
import string
import subprocess
import sys
import tarfile
import tempfile
import textwrap
import time
import yaml


import bzrlib
from bzrlib import (
    builtins,
    bzrdir,
    cmd_version_info,
    plugin,
    workingtree,
)
from novaclient.v1_1 import client as nova_client
import novaclient.exceptions
import swiftclient


log = logging.getLogger(__name__)
HERE = os.path.abspath(os.path.dirname(__file__))
SSH_OPTS = ['-o', 'UserKnownHostsFile=/dev/null',
            '-o', 'StrictHostKeyChecking=no']

# ci-utils probably isn't in our pythonpath
sys.path.append(os.path.abspath(os.path.join(HERE, '..', 'ci-utils')))
from ci_utils import (
    data_store,
    dump_stack,
    juju,
    sourcecode,
    unit_config,
)


# Initialize bzrlib and its plugins
bzrlib.initialize()
plugin.load_plugins()


def gen_random_key(length=40):
    """Returns a random ASCII key of the specified length.

    Uses `random.SystemRandom` (true-random) on a `ascii_letters + digits`
    universe for getting maximum entropy with no seeds.

    :param length: key string length (defaults to 40, HMAC-SHA1).
    """
    choices = string.ascii_letters + string.digits
    urandom = random.SystemRandom()
    return "".join([urandom.choice(choices) for i in range(length)])


def _get_net_id():
    """Returns the appropriate network id required when creating testbeds."""

    def internal_networks(networks):
        for network in networks:
            if network.get('router:external') is False:
                yield network['id']

    try:
        kw = {'auth_url': os.environ['GLANCE_OS_AUTH_URL'],
              'username': os.environ['GLANCE_OS_USERNAME'],
              'password': os.environ['GLANCE_OS_PASSWORD'],
              'tenant_name': os.environ['GLANCE_OS_TENANT_NAME'],
              'region_name': os.environ['GLANCE_OS_REGION_NAME']}
    except KeyError as e:
        raise EnvironmentError(
            'Invalid nova environment, missing {}'.format(e.message))
    # FIXME: fginther 20140925 - Defer the import of neutronclient.
    # This is causing problems with production deployments of NFSS due to a
    # missing python dependency (despite python-neutronclient being present
    # in the dependencies branch).
    # The solution thus far is to specify a dummy value for
    # EPHEMERAL_CLOUD_NET_ID in the environment and avoid execution of this
    # method.
    from neutronclient.neutron import client as neutron_client
    neutron = neutron_client.Client('2.0', **kw)
    # FIXME: This can fail on transient errors and needs to be retried (or at
    # least caught to not introduce weird error recovery modes in the callers
    # -- vila 2014-09-12
    networks = neutron.list_networks()['networks']
    results = [x for x in internal_networks(networks)]
    if len(results) > 1:
        raise EnvironmentError('More than one network to choose from')
    elif len(results) < 1:
        raise EnvironmentError('No internal networks configured')

    return results[0]


def check_environment():
    '''Cheetah requires environment variables for populating our templates.

    Some default values are also set here as a side-effect.
    '''
    cheetah_vars = {
        'OS_USERNAME': None,
        'GLANCE_OS_USERNAME': None,
        'GLANCE_OS_AUTH_URL': None,
        'GLANCE_OS_PASSWORD': None,
        'GLANCE_OS_TENANT_NAME': None,
        'GLANCE_OS_REGION_NAME': None,
        'EPHEMERAL_CLOUD_NET_ID': None,
        'CI_CODE_SOURCE': 'branch',
        'CI_BRANCH': 'lp:uci-engine',
        'CI_PAYLOAD_URL': 'n/a',
        'CI_STORAGE_PROVIDER': None,
        'CI_KEYS_BRANCH': '',
        'CI_LAUNCHPAD_USER': None,
        'CI_LAUNCHPAD_PPA_OWNER': None,
        'CI_PRIVATE_PPAS_ONLY': '0',
        'CI_OAUTH_CONSUMER_KEY': None,
        'CI_OAUTH_TOKEN': None,
        'CI_OAUTH_TOKEN_SECRET': None,
        'CI_PPA': 'ppa:canonical-ci-engineering/ci-airline-phase-0',
        'CI_MASTER_PPA': 'ppa:ci-engineering-airline/ci-archive',
        'CI_TEMPURL_EXPIRES': '300',
        'CI_TEMPURL_PUBLIC_HOST': '',
        'CI_TEMPURL_SIGNING_KEY': None,
        'CI_KEYRING_URI': '',
        'CI_KEYSERVER_URL': 'keyserver.ubuntu.com',
        'HP_ACCESS_KEY_ID': None,
        'HP_TENANT_ID': None,
    }

    # The CI_STORAGE_PROVIDER is used for postgresql, for 'local' (lxc
    # containers, local files are used), for openstack deployments ('openstack'
    # provider), nova volumes named after the units are used.
    env_name = juju.environment_name()
    env = juju.environment(env_name)
    if env is not None:
        env_type = env.get('type', None)
    else:
        # Provide a safe default value when no juju env exist
        env_type = 'local'
    if env_type == 'openstack':
        os.environ['CI_STORAGE_PROVIDER'] = 'block-storage-broker'
    elif env_type == 'local':
        os.environ['CI_STORAGE_PROVIDER'] = 'local'
    else:
        msg = 'ERROR: {} is not a known juju deployment type for {}'
        sys.exit(msg.format(env_type, env_name))

    is_hpcloud = unit_config.is_hpcloud()
    missing_required = []
    for key in cheetah_vars:
        if key not in os.environ or len(os.environ[key]) == 0:
            default = cheetah_vars[key]
            if default is not None:
                os.environ[key] = default
                continue

            # Attempt to determine EPHEMERAL_CLOUD_NET_ID if it's not supplied.
            if key == 'EPHEMERAL_CLOUD_NET_ID':
                if unit_config.is_canonistack():
                    os.environ[key] = ''
                else:
                    try:
                        os.environ[key] = _get_net_id()
                    except EnvironmentError as e:
                        sys.exit(
                            'EPHEMERAL_CLOUD_NET_ID was not specified and can '
                            'not be determined ({}). Check for the '
                            'appropriate value via \'neutron net-list\''
                            '.'.format(e.message))
                continue

            # CI_TEMPURL_SIGNING_KEY is required for generating TempUrls.
            # HP requires it to be defined with as its internally
            # generated 'secret_key'.
            if key == 'CI_TEMPURL_SIGNING_KEY':
                if is_hpcloud:
                    sys.exit(
                        'CI_TEMPURL_SIGNING_KEY has to be defined as your '
                        'HP access \'secret_key\', which can be found in '
                        'https://horizon.hpcloud.com/.')
                # For non-hp deployments the signing_key is overwriten
                # on every deployment for better safety.
                os.environ[key] = gen_random_key()
                continue

            # Define default (empty) variables for missing HP_* keys
            # if it is not targeted to HPCloud.
            if not is_hpcloud and key.startswith('HP'):
                os.environ[key] = ''
                continue
            missing_required.append(key)

    if len(missing_required) > 0:
        template = textwrap.dedent('''
        ERROR: Missing required environment variables:
          {}
        Please ensure the novarc file has been sourced.''')
        sys.exit(template.format('\n  '.join(missing_required)))


def bootstrap():
    """Bootstrap juju."""

    args = ['juju', 'bootstrap', '--constraints=mem=1024M']
    try:
        subprocess.check_output(args)
    except subprocess.CalledProcessError as e:
        # FIXME: if check_output raise an exception, we don't get a proper
        # output to check for the text below -- vila 2014-02-07
        if 'environment is already bootstrapped' not in e.output:
            raise


def juju_status(environment=None):
    '''Return the deployment status as a dictionary, parsed from yaml.
    The default environment will be used if one is not provided.'''

    cmd = ['juju', 'status', '--format', 'yaml']
    if environment:
        cmd += ['-e', environment]
    return yaml.safe_load(subprocess.check_output(cmd))


def _get_control_bucket():
    """The swift bucket name used by juju for state information."""
    env = juju.environment(juju.environment_name())
    return env['control-bucket']


def _delete_control_bucket():
    """Remove the juju control bucket and all of its files."""

    bucket = _get_control_bucket()
    conn = _connect_to_swift()
    try:
        files = [x['name'] for x in conn.get_container(bucket)[1]]
        for f in files:
            conn.delete_object(bucket, f)
        conn.delete_container(bucket)
    except swiftclient.exceptions.ClientException:
        # Assume the bucket is not present
        pass


def _connect_to_swift():
    opts = {'tenant_name': os.environ['OS_TENANT_NAME'],
            'region_name': os.environ['OS_REGION_NAME'],
            }
    conn = swiftclient.client.Connection(
        os.environ['OS_AUTH_URL'],
        os.environ['OS_USERNAME'],
        os.environ['OS_PASSWORD'],
        os_options=opts,
        auth_version='2.0')
    return conn


def _get_from_swift(bucket, path):
    """Download the file path from swift at bucket."""

    conn = _connect_to_swift()
    # get_object return a tuple of (metadata, data)
    return conn.get_object(bucket, path)[1]


def get_bootstrap_ip(instance, max_tries=120):
    """Wait for the juju bootstrap node to boot and obtain an IP address.

    :raises: NotFound if the instance has not been spawned.

    :raises: Exception if no IP can be found after max_tries*3 seconds.
    """

    log.info("Getting the bootstrap IP address.")

    # Special case here: while we use GLANCE_* vars for access to glance/nova,
    # we know we have access to nova where we bootstrap juju with the same
    # credentials used for swift. -- vila 2014-07-01
    args = [os.environ['OS_USERNAME'], os.environ['OS_PASSWORD'],
            os.environ['OS_TENANT_NAME'], os.environ['OS_AUTH_URL']]
    kwargs = {'region_name': os.environ['OS_REGION_NAME'],
              'service_type': 'compute'}
    nt = nova_client.Client(*args, **kwargs)

    tries = 0
    while tries < max_tries:
        # Raises NotFound if the instance does not exist.
        server = nt.servers.get(instance)
        networks = server.networks.values()
        if networks:
            # We use a single network and don't care how it's named
            net = networks[0]
            if net:
                # Either we have a single address or the last one has been
                # added to make the instance reachable (floating or public).
                net = net[-1]
                # Don't sleep anymore. We have our IP, let's move on
                return net

        tries += 1
        if tries != max_tries:
            time.sleep(3)
    raise Exception('Timed out waiting for bootstrap IP.')


def get_bootstrap_nova_instance():
    """Get the Nova instance identifier for the juju bootstrap node.

    Raises swiftclient.exceptions.ClientException if the instance cannot be
    found."""

    bucket = _get_control_bucket()

    state = yaml.safe_load(_get_from_swift(bucket, 'provider-state'))
    if not state:
        # Cause unknown (yet) but this has been seen in real life. I can't
        # reproduce anymore but I suspect 'juju bootstrap' failed mid-way so
        # returning None and letting the caller delete the bucket is the Right
        # Thing to do -- vila 2014-02-07
        log.warning("Empty 'provider-state' in the juju bucket")
        return None
    return state['state-instances'][0]


def wait_for_node(ip):
    """Wait for the bootstrap node to become available on ssh. We need to wait
    for this state before attempting to set up a sshuttle."""

    log.info("Waiting for the bootstrap node to respond to ssh.")

    cmd = ['ssh'] + SSH_OPTS + ['ubuntu@%s' % ip, 'exit']
    tries = 0
    while tries < 120:
        try:
            subprocess.check_call(cmd)
            return
        except subprocess.CalledProcessError:
            tries += 1
            time.sleep(5)

    raise Exception('Timed out waiting for bootstrap node.')


def needs_bootstrap():
    """Check to see if there's a Swift bucket for juju. If not, we need to
    boostrap. Check to see if the IP of the bootstrap node is owned by a
    running Nova instance. If not, we need to bootstrap."""

    log.info("Checking to see if a bootstrap is needed.")

    def inconsistent_state(msg):
        log.warning(msg)
        _delete_control_bucket()

    try:
        instance = get_bootstrap_nova_instance()
    except swiftclient.exceptions.ClientException:
        instance = None
    if instance is None:
        # We don't have an instance. We need to bootstrap.
        inconsistent_state('No bootstrap instance in the swift bucket')
        return True

    try:
        get_bootstrap_ip(instance, 1)
    except novaclient.exceptions.NotFound:
        # Juju claims we have an instance that does not exist. We're in an
        # inconsistent state. Delete the bucket and bootstrap.
        inconsistent_state('No IP for the bootstrap instance')
        return True

    return False


def check_bootstrapped():
    juju_env_name = juju.environment_name()
    if juju_env_name is None:
        log.error('No juju environment specified.')
        sys.exit(1)
    try:
        # XXX: Added a print to give user feedback in case sshuttle isn't
        # running and it "hangs", maybe correct behavior is to add a timeout
        # here, but too much work for the benefit, so I'm only providing a
        # hint now -- Ursinha (2014-02-17)
        print 'Checking juju status'
        juju_status()
    except subprocess.CalledProcessError:
        # There's no need to print an error message as juju status already
        # does that.
        sys.exit(1)


def check_dependencies(needed_commands=None, needed_imports=None):
    '''Exit the application if we are missing any dependency programs'''

    err = "%s is missing. Please install it before continuing."
    if needed_commands:
        for cmd in needed_commands:
            if not spawn.find_executable(cmd):
                err = err % cmd
                print >>sys.stderr, err
                sys.exit(1)
    if needed_imports:
        for impt in needed_imports:
            try:
                importlib.import_module(impt)
            except:
                err = err % impt
                print >>sys.stderr, err
                sys.exit(1)


def populate_all_templates(working_dir=None):
    deployer_dir = HERE
    if working_dir is None:
        temp = tempfile.mkdtemp()
        atexit.register(lambda: shutil.rmtree(temp))
        # A place to put our generated .yaml files.
        working_dir = os.path.join(temp, 'deployer')
    else:
        # Use the absolute path of the given working dir to avoid
        # misleading paths when calling `juju-deployer`.
        working_dir = os.path.abspath(working_dir)
    # Copy all existing files (including the .yaml and .yaml.tmpl ones),
    # they don't need to be populated but juju-deployer expect them in
    # 'working_dir'.
    dir_util.copy_tree(deployer_dir, working_dir)
    # Create .yaml files from .yaml.tmpl files.
    populate_templates(deployer_dir, working_dir)
    return working_dir


def populate_templates(deployer_dir, working_dir):
    # The command to populate the template.
    cmd_base = ['cheetah', 'fill', '--env', '-p']

    for tmpl in os.listdir(deployer_dir):
        path = os.path.join(deployer_dir, tmpl)
        if os.path.isdir(path):
            # When called with a directory, recurse and process all templates
            # found there.
            src = os.path.join(deployer_dir, tmpl)
            dst = os.path.join(working_dir, tmpl)
            if not os.path.isdir(dst):
                os.mkdir(dst)
            populate_templates(src, dst)
            continue
        if not tmpl.endswith('.yaml.tmpl'):
            continue
        # Strip the .tmpl off .yaml.tmpl and write a novarc-populated version
        # of the yaml to our working directory.
        with open(os.path.join(working_dir, tmpl[:-5]), 'w') as out:
            args = [os.path.join(deployer_dir, tmpl)]
            subprocess.check_call(cmd_base + args, stdout=out)


def generate_argument_list(d, specific=None):
    for f in sorted(os.listdir(d)):
        if not f.endswith('.yaml'):
            continue
        # If we're looking for a specific set of components to deploy and this
        # isn't one of them, continue.
        if specific and f[:-5] not in specific:
            continue
        yield ('-c', os.path.join(d, f))


def prepend_path(env_var, path):
    """Put our dependencies paths first so they don't get hidden."""
    val = os.environ.get(env_var, None)
    paths = (val and val.split(os.path.pathsep)) or []
    paths.insert(0, path)
    os.environ[env_var] = os.path.pathsep.join(paths)


def install_websocket_client(where=None):
    """Install python-websocket-client from bzr, as jujuclient needs it."""
    if where is None:
        where = os.path.join(HERE, '..', 'branches')
    dest = os.path.abspath(os.path.join(where, 'websocket-client'))
    prepend_path('PYTHONPATH', dest)
    return dest


def install_jujuclient(where=None):
    """Install python-jujuclient from bzr. r28 fixes bugs we were seeing around
    resetting the websocket connection."""
    if where is None:
        where = os.path.join(HERE, '..', 'branches')
    dest = os.path.abspath(os.path.join(where, 'python-jujuclient'))
    prepend_path('PYTHONPATH', dest)
    return dest


def install_deployer(where=None):
    """Install juju-deployer from bzr."""
    if where is None:
        where = os.path.join(HERE, '..', 'branches')
    dest = os.path.abspath(os.path.join(where, 'juju-deployer'))
    prepend_path('PYTHONPATH', dest)
    # FIXME: This is a symptom that our setup story needs to receive some more
    # love -- vila 2014-04-10
    sys.path.insert(0, dest)
    # FIXME: This duplicates setup.py and shouldn't be needed. This specific
    # piece (juju-deployer) is a weird beast, it installs from source (not pip
    # nor package), it's used both in unit tests that are executed in a venv
    # and in ./tests/run.py which is not executed in venv. My head hurts --
    # vila 2014-02-12
    deployer_bin = os.path.join(dest, 'juju-deployer')
    with open(deployer_bin, 'w') as fp:
        tmpl = ('#!/bin/sh\nexport PYTHONPATH={}\n'
                'exec python {}/deployer/cli.py $*')
        fp.write(tmpl.format(os.environ['PYTHONPATH'], dest))
    st = os.stat(deployer_bin)
    mode = st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    os.chmod(deployer_bin, mode)
    prepend_path('PATH', dest)
    return dest


def install_unit_config(working_dir):
    """Copy populated unit config to the tree.

    Only used in tests, simulates production environment.
    """
    unit_config_path = os.path.join(
        working_dir, 'configs', 'unit_config.yaml')
    dest_path = os.path.join(HERE, '..', 'unit_config')
    shutil.copy2(unit_config_path, dest_path)


def _do_bzr_checkout(uri, path):
    """Create a checkout of the given branch uri on the given path."""
    acc, source = bzrdir.BzrDir.open_tree_or_branch(uri)
    source.create_checkout(path, None, True, acc)


def install_keys(working_dir):
    """Install system keys from the specified branch.

    The keys branch is specified via 'CI_KEYS_BRANCH' and installed in
    '<working_dir>/keys'.

    If '<working_dir>/keys' already exist, this function does nothing
    (pre-existing working_dir or static keys).

    If 'CI_KEYS_BRANCH' is not defined or empty and the keys directory is
    not available, we display a warning message about the lack of access.
    """
    key_names = ['id_rsa', 'id_rsa.pub', 'gpg.sec', 'gpg.pub']
    keys_path = os.path.join(working_dir, 'keys')
    if os.path.isdir(keys_path):
        for key_name in key_names:
            if not os.path.exists(os.path.join(keys_path, key_name)):
                break
        else:
            return
    keys_branch = os.environ.get('CI_KEYS_BRANCH')
    if not keys_branch:
        print(
            'Deployment keys are not available (CI_KEYS_BRANCH is not '
            'defined and {} path does not exist). This deployment will not '
            'be able to access LP resources that depend on it.'.format(
                keys_path))
        os.mkdir(keys_path)
        for key_name in key_names:
            with open(os.path.join(keys_path, key_name), 'w') as fd:
                fd.write('# Stub content, CI_KEYS_BRANCH was not specified.')
        return
    print('Installing keys from {}'.format(keys_branch))
    _do_bzr_checkout(keys_branch, keys_path)


def setup_config_manager():
    '''Use a single copy of config-manager and point it at a single working
    directory under branches/.'''

    where = os.path.join(HERE, '..', 'branches')
    sourcebuild_dir = os.path.abspath(os.path.join(where, 'charm-helpers'))
    os.environ['SOURCEBUILD_DIR'] = sourcebuild_dir


def build_sourcedeps(into):
    '''Build out conf/sourcedeps.conf.'''

    conf = os.path.join(HERE, '..', 'conf')
    config_filename = os.path.join(conf, 'sourcedeps.conf')
    cache_filename = os.path.join(conf, 'sourcedeps.cache')
    args = (into, config_filename, cache_filename, False, False, False)
    kwargs = {'quiet': True, 'use_http': False}
    sourcecode.update_sourcecode(*args, **kwargs)


def build_charms(working_dir, charms_dir=None):
    """Build local charms and copy them to the local juju repository."""
    if charms_dir is None:
        charms_dir = os.path.abspath(os.path.join(HERE, '../charms'))
    charms_makefiles = os.path.join(charms_dir, 'precise/*/Makefile')
    for f in glob.glob(charms_makefiles):
        if os.path.isfile(f):
            charm_path = os.path.dirname(f)
            print('Building charm: {}'.format(os.path.basename(charm_path)))
            # Subprocess sets PATH unless specified. We need the venv PATH for
            # running the charmhelpers sync script.
            subprocess.check_call(['make'], cwd=charm_path, env=os.environ)
    repository = os.path.join(working_dir, 'charms')
    dir_util.copy_tree(charms_dir, repository)
    os.environ['JUJU_REPOSITORY'] = repository


def generate_version_info(path):
    """Create file with version info from the current working tree."""
    with open(path, 'w') as f:
        cmd = cmd_version_info.cmd_version_info()
        # Shame that bzr doesn't provide a better way

        def redirect():
            cmd.outf = f
        cmd._setup_outf = redirect
        tmpl = '{branch_nick}@{revno} - {date}'
        cmd.run_argv_aliases(['--custom', '--template', tmpl])


def build_payload(path):
    """Build the payload as a compressed tarball at the given path.

    :returns: A fingerprint uniquely identifying the payload.
    """
    payload_path = '.'
    cmd = builtins.cmd_export()
    with tempfile.NamedTemporaryFile() as ball:
        # We want uncommitted changes. We take file timestamps as they carry
        # useful information (dir timestamps are broken though).
        cmd.run_argv_aliases([ball.name, '--root', payload_path,
                              '--per-file-timestamps',
                              '--format=tar', '--uncommitted'])
        # Build the fingerprint from everything included by bazaar in the
        # tarball, dirs have no fingerprint (but paths are included so we don't
        # miss renames).
        sha1 = hashlib.sha1()
        wt, _ = workingtree.WorkingTree.open_containing(payload_path)
        wt.lock_read()
        try:
            ids = wt.all_file_ids()
            for i in ids:
                i_sha = wt.get_file_sha1(i)
                if i_sha is not None:  # dirs have no sha
                    sha1.update(i_sha)
                sha1.update(wt.id2path(i))
        finally:
            wt.unlock()

        # We add some more files not under version control. As such, we don't
        # include them in the sha1 (if they matter for the sha1, they should be
        # under version control ;)
        uploaded_version_info_path = './.uploaded_version_info'
        generate_version_info(uploaded_version_info_path)

        def reset_ids(tarinfo):
            # We don't want to get the uid/gid from the uploader
            tarinfo.uid = tarinfo.gid = 0
            tarinfo.uname = tarinfo.gname = ''
            return tarinfo

        with tarfile.open(ball.name, 'a') as tarball:
            # Always use relpaths below to mimick bzr's behavior building a
            # tarball that can be extracted anywhere.
            for added_path in [uploaded_version_info_path]:
                tarball.add(added_path, filter=reset_ids)
        # Finally, we compress the resulting tarball
        with gzip.GzipFile(path, 'w') as gz, open(ball.name) as tarball:
            gz.write(tarball.read())
    return sha1.hexdigest()


def upload_payload_to_swift():
    """Upload the versions controlled code tree to swift.

    Also set the relevant environment variables used to populate the deployer
    scripts and charm configurations.
    """
    print('Preparing local branch upload...')

    config = {
        'auth_url': os.environ['OS_AUTH_URL'],
        'auth_user': os.environ['OS_USERNAME'],
        'auth_password': os.environ['OS_PASSWORD'],
        'auth_tenant_name': os.environ['OS_TENANT_NAME'],
        'auth_region': os.environ['OS_REGION_NAME']
    }
    try:
        ds = data_store.DataStore('ci-payload', config, public=True)
    except data_store.DataStoreException as exc:
        error_msg = ("ERROR: data-store: {}")
        sys.exit(error_msg.format(exc))

    with tempfile.NamedTemporaryFile() as ball:
        fingerprint = build_payload(ball.name)
        # The sha-1 fingerprint is used as the name of the file in swift so we
        # can avoid uploading if it's already there.
        if fingerprint + '.tgz' not in ds.list_files():
            with open(ball.name) as f:
                print('Uploading local branch,'
                      ' fingerprint {}'.format(fingerprint))
                payload_url = ds.put_file(fingerprint + '.tgz', f)
        else:
            print('Reusing local branch, fingerprint {}'.format(fingerprint))
            payload_url = ds.file_path(fingerprint + '.tgz')
    # the django charm uses branch for both meanings of payload
    os.environ['CI_CODE_SOURCE'] = 'tarball'
    os.environ['CI_PAYLOAD_URL'] = payload_url
    os.environ['CI_BRANCH'] = payload_url


def print_juju_unit_file(unit, path):
    """Dump a file content from a juju unit.

    :param unit: The unit name as a string.

    :param path: The path on the unit.
    """
    cmd = ['juju', 'ssh', unit, 'cat', path]
    try:
        print '-' * len(path)
        print path
        print '-' * len(path)
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError:
        # Don't fail so callers can collect more file
        msg = 'Failed to capture juju file {} from unit {}'.format(path, unit)
        print >>sys.stderr, msg


def print_juju_unit_log(unit):
    path = '/var/log/juju/unit-{}.log'.format(unit.replace('/', '-'))
    print_juju_unit_file(unit, path)


def units_in_error(status):
    """Iterate units in a error state.

    :param status: juju_status() output (parsed from yaml).
    """
    services = status['services']
    for service, service_data in services.iteritems():
        units = service_data.get('units')
        if not units:
            continue
        for unit, unit_data in units.iteritems():
            if unit_data.get('agent-state') == 'error':
                yield unit, unit_data
            # Tricky: If the unit has subordinates, the error state will be
            # reported there
            subordinates = unit_data.get('subordinates', {})
            for sub_unit, sub_data in subordinates.iteritems():
                if sub_data.get('agent-state') == 'error':
                    yield sub_unit, sub_data


def dump_juju_logs(status):
    """Dump the juju log of any service in an error state."""
    all_machines = False
    for unit, unit_data in units_in_error(status):
        print_juju_unit_log(unit)
        all_machines = True

    if all_machines:
        # If at least one unit was in error, also collect the all-machines log
        # file as it may help diagnose issues involving multiple units.
        print_juju_unit_file('0', '/var/log/juju/all-machines.log')


def _install(name, deployer_dir, generated, build=False):
    args = ['juju-deployer', '-v']

    if build:
        args.append('-b')

    args.extend(generated + [name])

    print 'Running {}'.format(' '.join(args))
    try:
        # Subprocess sets PATH unless specified and we need the venv PATH.
        subprocess.check_call(args, env=os.environ)
        return 0
    except Exception as exc:
        print 'Problem deploying "{}": {}'.format(name, str(exc))
        try:
            status = juju_status()
            dump_juju_logs(status)
        except subprocess.CalledProcessError:
            print '`juju status` failed. Unable to get log files.'
        return 1


def _get_deployer(configs, name):
    import deployer.config   # lazy load since this is added by this script
    import deployer.utils

    deployer.utils.setup_logging()
    dep = deployer.config.ConfigStack(configs).get(name)

    # ensure we pull down the charms we need for upgrade
    args = []
    for c in configs:
        args.extend(['-c', c])
    args.append(name)
    # Subprocess sets PATH unless specified and we need the venv PATH.
    subprocess.check_call(['juju-deployer', '-b'] + args, env=os.environ)

    return dep


def _deployer_options(service, deployer_dir):
    keyvals = []
    for key, val in service.get('options', {}).iteritems():
        include_token = 'include-base64://'
        if type(val) is str and val.startswith(include_token):
            val = val.replace(include_token, '')
            p = os.path.join(deployer_dir, val)
            with open(p) as f:
                val = base64.b64encode(f.read())
        keyvals.append('%s=%s' % (key, val))
    return keyvals


def _upgrade(name, operations, deployer_dir, configs):
    print 'Discovering services to upgrade'
    deployment = _get_deployer(configs, name)
    services = juju_status()['services']

    if 'all' in operations:
        # 'code' is already performed via 'config'.
        operations = ['charms', 'config']

    for operation in operations:
        for svc in services:
            dsvc = deployment.data['services'].get(svc)
            if not dsvc:
                print('Skipping unknown service: {}'.format(svc))
                continue
            print('Upgrading %s for %s' % (operation, svc))
            if operation == 'charms':
                subprocess.check_call(
                    ['juju', 'upgrade-charm', '--force', svc])
            elif operation == 'config':
                keyvals = _deployer_options(dsvc, deployer_dir)
                if keyvals:
                    subprocess.check_call(['juju', 'set', svc] + keyvals)
            elif operation == 'code':
                # this is a really a special case of "config". but its
                # special enough to merit its own code path that *only*
                # updates code
                current = dsvc.get('options', {}).get('current_code')
                avail = dsvc.get('options', {}).get('available_code')
                if current and avail:
                    settings = [
                        'available_code=' + avail,
                        'current_code=' + current
                    ]
                    print('   %s' % ', '.join(settings))
                    subprocess.check_call(['juju', 'set', svc] + settings)

    print('Upgrade requests have been issued.')
    print('Check back with "juju status" to track completion')
    return 0


def _get_args(args=None):
    parser = argparse.ArgumentParser(
        description='use juju-deployer to deploy services')
    parser.add_argument('--private-ppas-only', action='store_true',
                        help='''Sets the environment to use only private PPAs
                        when building packages.''')
    parser.add_argument('--upgrade', action='append',
                        choices=('charms', 'code', 'config', 'all'),
                        help='''Upgrade an existing deployment. This argument
                        allows you to choose the order and number of upgrade
                        operations. ie - you could just update the charms,
                        update the configs, or update the code.''')
    parser.add_argument('specific', metavar='<config.yaml>', nargs='*',
                        help='Use a specific list of juju-deployer configs.')
    parser.add_argument('--build-only', action='store_true',
                        help='Just build the deployment.')
    parser.add_argument('--build', action='store_true',
                        help='Build the deployment for --working-dir.')
    parser.add_argument('--name', default='ci-airline',
                        choices=('ci-airline', 'ci-airline-experimental',
                                 'ci-airline-for-britney'),
                        help='''The name of the juju-deployer "stack".
                             Default=%(default)s''')
    parser.add_argument(
        '--working-dir', action='store', default=None,
        help=('Build configuration files and deploy on an existing '
              'directory. Additional configuration files present in '
              'the given path will be populated (.tmpl) and deployed.')
    )

    return parser.parse_args(args)


def setup(for_tests=False, list_only=False, bootstrap_juju=False):
    """Setup the environment for a deployment.

    This includes:
    - installing the source dependencies,
    - check the environment variables,
    - uploading our source payload to swift,
    - populating the templates,
    - starting the juju boootstrap node if needed.
    """
    install_websocket_client()
    install_jujuclient()
    install_deployer()

    check_environment()
    if for_tests:
        if not list_only:
            upload_payload_to_swift()
        # FIXME: The tests that requires the CI_KEYRING_URI should set it when
        # needed -- vila 2014-06-11
        # Include the testing keyring in the Gatekeeper run-time for tests.
        os.environ['CI_KEYRING_URI'] = (
            'file:///srv/ci_airline_gatekeeper_restish/ci-utils/'
            'ci_utils/testing/data/testing_keyring.txt')
        working_dir = populate_all_templates()
        if not list_only:
            install_unit_config(working_dir)
            setup_config_manager()
            build_charms(working_dir)
        # FIXME: We use JUJU_DEPLOYER_DIR to tell deployers.DeployerTest where
        # the populated templates are, there should be a better way than using
        # an env var as a global variable -- vila 2014-04-11
        os.environ['JUJU_DEPLOYER_DIR'] = working_dir
    # FIXME: This should become a command or an option for the deploy.py and
    # test runner[s] scripts -- vila 2014-04-23
    if bootstrap_juju:
        if needs_bootstrap():
            bootstrap()


def main():
    args = _get_args()

    # You can configure the system to use private PPAs only or by providing
    # the option --private-ppas-only or by setting the environment variable
    # CI_PRIVATE_PPAS_ONLY.
    # If provided, --private-ppas-only option overrides the environment
    # variable value.
    if args.private_ppas_only:
        os.environ['CI_PRIVATE_PPAS_ONLY'] = '1'

    check_dependencies(needed_commands=['cheetah'])
    branches = os.path.join(HERE, '..', 'branches')
    build_sourcedeps(branches)
    setup()

    if not args.build_only and not args.build:
        check_bootstrapped()

    priv = 'enabled' if int(os.environ['CI_PRIVATE_PPAS_ONLY']) else 'disabled'
    print('Private PPAs: ' + priv)

    needs_upload =\
        not args.build_only and \
        ((args.working_dir and args.build) or not args.working_dir)
    if needs_upload:
        upload_payload_to_swift()

    if needs_upload or args.build_only:
        working_dir = populate_all_templates(args.working_dir)
        setup_config_manager()
        build_charms(working_dir)
        install_keys(working_dir)
    else:
        working_dir = args.working_dir
        os.environ['JUJU_REPOSITORY'] = os.path.join(working_dir, 'charms')

    if not args.build_only:
        # Generate an argument list of -c FILE for each .yaml file in our
        # working directory.
        i = generate_argument_list(working_dir, args.specific)
        generated = list(itertools.chain.from_iterable(i))

        if args.upgrade:
            for x in generated:
                configs = [x for x in generated if x != '-c']
            return _upgrade(args.name, args.upgrade, working_dir, configs)
        else:
            return _install(args.name, working_dir, generated, args.build)


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    sys.exit(main())
