#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from distutils import dir_util
import os
import mock
import tempfile
import shutil
import StringIO
import subprocess
import unittest
import yaml

from novaclient import exceptions
import swiftclient

from ucitests import (
    assertions,
    fixtures,
)

from ci_utils.testing import features

import deploy


class TestCheckEnvironment(unittest.TestCase):

    DEFAULT_ENV = {
        'OS_USERNAME': 'os-user',
        'GLANCE_OS_USERNAME': 'glance-os-user',
        'GLANCE_OS_AUTH_URL': 'glance-url',
        'GLANCE_OS_REGION_NAME': 'glance-region',
        'GLANCE_OS_TENANT_NAME': 'glance-tenant',
        'GLANCE_OS_PASSWORD': 'glance-password',
        'EPHEMERAL_CLOUD_NET_ID': 'ephemeral-cloud-net-id',
        'OS_AUTH_URL': 'something.stack',
        'CI_OAUTH_CONSUMER_KEY': 'key',
        'CI_OAUTH_TOKEN': 'token',
        'CI_OAUTH_TOKEN_SECRET': 'secret',
        'CI_LAUNCHPAD_USER': 'lp-user',
        'CI_LAUNCHPAD_PPA_OWNER': 'lp-ppa-owner',
    }

    def setUp(self):
        super(TestCheckEnvironment, self).setUp()
        fixtures.set_uniq_cwd(self)
        # The following vars from the host system should be ignored
        # (deleted) in the context of this test.
        ignored_env = {
            'CI_CODE_SOURCE': None,
            'CI_PRIVATE_PPAS_ONLY': None,
            'CI_TEMPURL_EXPIRES': None,
            'CI_TEMPURL_SIGNING_KEY': None,
            'CI_KEYRING_URI': None,
            'CI_KEYSERVER_URL': None,
            'HP_ACCESS_KEY_ID': None,
            'HP_TENANT_ID': None,
        }
        fixtures.isolate_from_env(self, ignored_env)

    @mock.patch('deploy.gen_random_key', return_value='RaMdOm_KeY')
    def test_check_environment_defaults(self, mocked_gen_key):
        # check that some of our values can be defaulted correctly:
        env_required = self.DEFAULT_ENV.copy()
        fixtures.isolate_from_env(self, env_required)
        deploy.check_environment()
        # defaulted by code
        self.assertEqual(os.environ['CI_CODE_SOURCE'], 'branch')
        self.assertEqual(os.environ['CI_PRIVATE_PPAS_ONLY'], '0')
        self.assertEqual(os.environ['CI_TEMPURL_EXPIRES'], '300')
        self.assertEqual(os.environ['CI_KEYRING_URI'], '')
        self.assertEqual(
            os.environ['CI_KEYSERVER_URL'], 'keyserver.ubuntu.com')
        # CI_TEMPURL_SIGNING_KEY used for canonistack/prodstack TempURL
        # generation and `deploy.py` generated a new random HMAC-SHA1
        # (40 chars) key for each deployment.
        self.assertEqual(os.environ['CI_TEMPURL_SIGNING_KEY'], 'RaMdOm_KeY')
        # Other HP_* variables are set empty if the deployment is not
        # targeted to HPCloud.
        self.assertEqual(os.environ['HP_ACCESS_KEY_ID'], '')
        self.assertEqual(os.environ['HP_TENANT_ID'], '')

    def test_check_environment_hp(self):
        # check_environment requires HP_* variables if the deployment
        # is targeted to HPCloud ('hpcloud' in OS_AUTH_URL)
        hp_env = self.DEFAULT_ENV.copy()
        hp_env.update({
            'OS_AUTH_URL': 'something.hpcloud',
            'CI_TEMPURL_SIGNING_KEY': 'FROM_HORIZON',
            'HP_ACCESS_KEY_ID': 'FROM_HORIZON',
            'HP_TENANT_ID': 'FROM_HORIZON',
        })
        fixtures.isolate_from_env(self, hp_env)
        deploy.check_environment()

    def test_check_environment_hp_signing_key(self):
        # check_environment bails out if it's a HPCloud deployment and
        # it can't find the required CI_TEMPURL_SIGNING_KEY.
        broken_hp_env = self.DEFAULT_ENV.copy()
        broken_hp_env.update({
            'OS_AUTH_URL': 'something.hpcloud',
            'HP_ACCESS_KEY_ID': 'FROM_HORIZON',
            'HP_TENANT_ID': 'FROM_HORIZON',
        })
        fixtures.isolate_from_env(self, broken_hp_env)
        with self.assertRaises(SystemExit) as cm:
            deploy.check_environment()
        self.assertEqual(
            'CI_TEMPURL_SIGNING_KEY has to be defined as your HP access '
            '\'secret_key\', which can be found in '
            'https://horizon.hpcloud.com/.',
            cm.exception.message)

    def test_check_environment_hp_fails(self):
        # check_environment bails out if it's a HPCloud deployment and
        # it can't find the required HP_* variables.
        broken_hp_env = self.DEFAULT_ENV.copy()
        broken_hp_env.update({
            'OS_AUTH_URL': 'something.hpcloud',
            'CI_TEMPURL_SIGNING_KEY': 'FROM_HORIZON',
        })
        fixtures.isolate_from_env(self, broken_hp_env)
        with self.assertRaises(SystemExit) as cm:
            deploy.check_environment()
        self.assertEqual(
            ['',
             'ERROR: Missing required environment variables:',
             '  HP_TENANT_ID',
             '  HP_ACCESS_KEY_ID',
             'Please ensure the novarc file has been sourced.'],
            cm.exception.message.splitlines())

    def test_check_environment_fails(self):
        # check_environment bails out if a valid novarc isn't sourced.
        fixtures.isolate_from_env(self, dict(OS_USERNAME=''))
        with mock.patch('sys.stderr'):
            with self.assertRaises(SystemExit):
                deploy.check_environment()

    def test_gen_random_key(self):
        # `gen_random_key` generates random alphanumeric keys of a
        # specified length.
        key_hmac_sha1 = deploy.gen_random_key()
        self.assertTrue(key_hmac_sha1.isalnum())
        self.assertTrue(40, len(key_hmac_sha1))
        key_hmac_md5 = deploy.gen_random_key(length=32)
        self.assertTrue(key_hmac_md5.isalnum())
        self.assertTrue(32, len(key_hmac_md5))


class TestCheckDependencies(unittest.TestCase):

    def setUp(self):
        super(TestCheckDependencies, self).setUp()
        fixtures.set_uniq_cwd(self)

    def test_check_dependencies_needed_commands(self):
        # If we cannot find any of our dependencies, this program should bail
        # out.
        fixtures.isolate_from_env(self, dict(PATH=''))
        with mock.patch('sys.stderr'):
            with self.assertRaises(SystemExit):
                deploy.check_dependencies(needed_commands=['cheetah'])

    def test_check_dependencies_needed_imports(self):
        # If we cannot find any of our import dependencies, this program should
        # bail out.
        with mock.patch('sys.stderr'):
            with self.assertRaises(SystemExit):
                deploy.check_dependencies(needed_imports=['IDONTEXIST'])


class TestBootStrap(unittest.TestCase):

    def test_check_not_bootstrapped(self):
        # If juju environment isn't bootstrapped, this program should bail
        # out.
        with mock.patch('subprocess.check_output',
                        side_effect=subprocess.CalledProcessError(
                            -1, 'juju status')):
            with self.assertRaises(SystemExit):
                deploy.check_bootstrapped()

    @features.requires(features.bootstrapped_juju)
    def test_check_bootstrapped(self):
        with mock.patch('subprocess.check_output', return_value='ok'):
            # If juju status returns ok, this should just work.
            deploy.check_bootstrapped()


class TestInstallDeployer(unittest.TestCase):

    def setUp(self):
        super(TestInstallDeployer, self).setUp()
        fixtures.set_uniq_cwd(self)

    @mock.patch('ci_utils.sourcecode.update_sourcecode')
    def test_install_deployer(self, mocked_us):
        fixtures.isolate_from_env(self, dict(PYTHONPATH=''))

        def local_us(*args, **kwargs):
            os.makedirs('juju-deployer/.bzr')
        mocked_us.side_effect = local_us
        deploy.build_sourcedeps('.')
        deployer_dir = deploy.install_deployer('.')
        self.addCleanup(shutil.rmtree, deployer_dir)
        self.assertTrue(os.path.exists(os.path.join(deployer_dir, '.bzr')))
        self.assertEqual(deployer_dir, os.environ['PYTHONPATH'])
        deployer_script = os.path.join(deployer_dir, 'juju-deployer')
        self.assertTrue(os.access(deployer_script, os.X_OK))
        self.assertTrue(os.path.exists(deployer_script))


class TestInstallKeys(unittest.TestCase):

    def setUp(self):
        super(TestInstallKeys, self).setUp()
        fixtures.set_uniq_cwd(self)

    def create_keys(self, keys_path):
        """Creates a 'keys' directory with the relevant key files."""
        os.mkdir(keys_path)
        key_names = ['id_rsa', 'id_rsa.pub', 'gpg.sec', 'gpg.pub']
        for key_name in key_names:
            with open(os.path.join(keys_path, key_name), 'w') as fd:
                fd.write('')

    @mock.patch('deploy._do_bzr_checkout')
    @mock.patch('sys.stdout', new_callable=StringIO.StringIO)
    def test_local_key_are_preferred(self, mocked_stdout, mocked_co):
        # If there are valid keys in the working dir, the CI_KEYS_BRANCH
        # is ignored and nothing is done. By definition, local keys are
        # maintained manually.
        fixtures.isolate_from_env(self, dict(CI_KEYS_BRANCH='A_BRANCH'))
        working_dir = os.getcwd()
        keys_dir = os.path.join(working_dir, 'keys')
        self.create_keys(keys_dir)

        deploy.install_keys(working_dir)

        self.assertFalse(mocked_co.called)
        self.assertEqual('', mocked_stdout.getvalue())

    @mock.patch('deploy._do_bzr_checkout')
    @mock.patch('sys.stdout', new_callable=StringIO.StringIO)
    def test_keys_branch_provided(self, mocked_stdout, mocked_co):
        # When CI_KEYS_BRANCH is provided and keys do not exist in the
        # working dir, the keys branch is checked out.
        keys_branch = 'A_BRANCH_URI'
        fixtures.isolate_from_env(self, dict(CI_KEYS_BRANCH=keys_branch))
        working_dir = os.getcwd()

        deploy.install_keys(working_dir)

        keys_dir = os.path.join(working_dir, 'keys')
        mocked_co.assert_called_once_with(keys_branch, keys_dir)
        self.assertEqual(
            'Installing keys from {}\n'.format(keys_branch),
            mocked_stdout.getvalue())

    @mock.patch('sys.stdout', new_callable=StringIO.StringIO)
    def test_keys_branch_not_provided(self, mocked_stdout):
        # When CI_KEYS_BRANCH is not provided and keys do not exist in
        # the working dir, empty 'keys' are created to satisfy deployment
        # configurations 'include's and a warning message is issued.
        fixtures.isolate_from_env(self, dict(CI_KEYS_BRANCH=''))
        working_dir = os.getcwd()

        deploy.install_keys(working_dir)

        keys_dir = os.path.join(working_dir, 'keys')
        self.assertEqual(
            ['gpg.pub', 'gpg.sec', 'id_rsa', 'id_rsa.pub'],
            sorted(os.listdir(keys_dir)))
        self.assertEqual(
            'Deployment keys are not available (CI_KEYS_BRANCH is not '
            'defined and {} path does not exist). This deployment '
            'will not be able to access LP resources that depend on it.'
            '\n'.format(keys_dir),
            mocked_stdout.getvalue())


class TestGenerateArgumentList(unittest.TestCase):

    def setUp(self):
        super(TestGenerateArgumentList, self).setUp()
        fixtures.set_uniq_cwd(self)

    def test_generate_empty_list(self):
        # Calling without anything in the directory should produce an empty
        # list.
        self.assertEqual([], list(deploy.generate_argument_list('.')))

    def create_file(self, filename, working_dir='.'):
        """Creates a file in the testing path with a given filename.

        Return the created file path.
        """
        filepath = os.path.join(working_dir, filename)
        with open(filepath, 'w') as fd:
            fd.write('nothing')
        return filepath

    def test_generate_explicit_list(self):
        # Calling with a basic configuration should produce the deployer
        # configuration as [('-c', '<file path>')].
        expected_filename = self.create_file('foo.yaml')
        self.assertEqual(
            [('-c', expected_filename)],
            list(deploy.generate_argument_list('.')))

    def test_generate_ignores_other_files(self):
        # `generate_argument_list` only considers configuration files
        # (*.yaml). Other files in the working directory are ignored.
        self.create_file('zoing.yaml.tpl')
        self.create_file('bazinga.txt')
        self.assertEqual([], list(deploy.generate_argument_list('.')))

    def test_generate_filtered_list(self):
        # `generate_argument_list` filter configurations matching a
        # given filename whitelist and results are ordered (alphabetically)
        # Create noise files.
        self.create_file('zoing.yaml')
        # Create the files mentioned in the testing whitelist.
        whitelist = ['foo', 'baz']
        expected_results = [
            ('-c', self.create_file(filename + '.yaml'))
            for filename in sorted(whitelist)]
        # Only the whitelisted files are considered.
        self.assertEqual(
            expected_results,
            list(deploy.generate_argument_list('.', whitelist)))


class TestPopulateAllTemplates(unittest.TestCase):

    def setUp(self):
        """Setup common `mock`a for this test suite.

        Mocking `atexit.register` and `deploy.populate_templates`.
        """
        super(TestPopulateAllTemplates, self).setUp()
        atexit_mocker = mock.patch('atexit.register')
        self.mock_atexit = atexit_mocker.start()
        self.addCleanup(atexit_mocker.stop)
        pt_mocker = mock.patch('deploy.populate_templates')
        self.mocked_populate = pt_mocker.start()
        self.addCleanup(pt_mocker.stop)

    def test_no_working_dir(self):
        # `populate_all_templates` recursively copies the contents of
        # 'juju-deployer' directory (*.yaml and templates) to a
        # temporary directory and populates them.

        working_dir = deploy.populate_all_templates()

        # Since `atexit` is mocked, we have to cleanup the tempdir.
        self.addCleanup(shutil.rmtree, os.path.dirname(working_dir))

        # The 'working_dir' is a temporary directory and contains the
        # copy of all 'deployer_dir' files, which were populated in the
        # destination.
        # A `atexit` hook is created to purge this temporary directory.
        self.assertTrue(
            working_dir.startswith(tempfile.gettempdir()))
        self.mocked_populate.assert_called_with(deploy.HERE, working_dir)
        self.assertEquals(
            sorted(os.listdir(deploy.HERE)),
            sorted(os.listdir(working_dir)))
        self.mock_atexit.assert_called()

    def makeWorkingDir(self, filename=None):
        """Creates a temporary directory to be used as a 'working_dir'.

        Setup the corresponding filesystem cleanup and optionally creates
        a testing file with the given 'filename'.
        """
        working_dir = tempfile.mkdtemp()
        if filename is not None:
            with open(os.path.join(working_dir, filename), 'w') as fd:
                fd.write('user content')
        self.addCleanup(shutil.rmtree, working_dir)
        return working_dir

    def test_with_working_dir(self):
        # `populate_all_templates` recursively copies the contents of
        # 'juju-deployer' directory (*.yaml and templates) to a
        # existing (given) directory and populates old and new files.
        filename = 'foo.yaml'
        user_dir = self.makeWorkingDir(filename)
        working_dir = deploy.populate_all_templates(user_dir)
        self.mocked_populate.assert_called_with(deploy.HERE, working_dir)
        self.assertEquals(user_dir, working_dir)
        self.assertEquals(
            sorted(os.listdir(deploy.HERE) + [filename]),
            sorted(os.listdir(working_dir)))
        self.mock_atexit.assert_not_called()

    def test_with_working_dir_relative_path(self):
        # `populate_all_templates` returns the absolute path
        # of the given 'working_dir' to avoid confusions when
        # calling `juju-deployer`.
        user_dir = self.makeWorkingDir()
        rel_path = os.path.relpath(user_dir)
        working_dir = deploy.populate_all_templates(rel_path)
        self.assertTrue(os.path.isabs(working_dir))
        self.assertEquals(user_dir, working_dir)

    def test_with_working_dir_overwrite(self):
        # `populate_all_templates` silently overwrites files in
        # the given 'working_dir' conflicting with the ones from
        # 'juju-deployer'.
        filename = 'gatekeeper.yaml.tmpl'
        user_dir = self.makeWorkingDir(filename)
        working_dir = deploy.populate_all_templates(user_dir)
        self.assertEquals(
            open(os.path.join(deploy.HERE, filename)).read(),
            open(os.path.join(user_dir, filename)).read(),
            'WORKING_DIR file was not overwritten.'
        )
        self.assertEquals(user_dir, working_dir)


class TestPopulateTemplates(unittest.TestCase):

    def setUp(self):
        super(TestPopulateTemplates, self).setUp()
        fixtures.set_uniq_cwd(self)
        self.target = os.path.join('.', 'target')
        os.mkdir(self.target)
        self.source = os.path.join('.', 'source')
        os.mkdir(self.source)

    def test_populate_templates_empty_list(self):
        deploy.populate_templates(self.source, self.target)
        # Nothing happened
        self.assertEqual([], os.listdir(self.source))

    def test_populate_templates_explicit_list(self):
        fixtures.override_env(self, 'USER', 'foo')
        stem = 'config.yaml'
        with open(os.path.join(self.source, stem + '.tmpl'), 'w') as f:
            f.write('me: ${USER}')
        deploy.populate_templates(self.source, self.target)
        with open(os.path.join(self.target, stem)) as f:
            content = f.read()
        self.assertEqual('me: {}'.format(os.environ['USER']), content)

    def test_populate_recurse(self):
        fixtures.override_env(self, 'USER', 'foo')
        configs_dir = os.path.join(self.source, 'configs')
        os.mkdir(configs_dir)
        stem = 'config.yaml'
        with open(os.path.join(configs_dir, stem + '.tmpl'), 'w') as f:
            f.write('me: ${USER}')
        # The only existing template is one level below self.source so
        # populate_templates should recurse there.
        deploy.populate_templates(self.source, self.target)
        # Accordingly, the results should one level below self.target
        with open(os.path.join(self.target, 'configs', stem)) as f:
            content = f.read()
        self.assertEqual('me: {}'.format(os.environ['USER']), content)


class TestBuildCharms(unittest.TestCase):

    def setUp(self):
        super(TestBuildCharms, self).setUp()
        fixtures.set_uniq_cwd(self)

    def test_build_charms(self):
        # `build_charms` builds tree charms, copies them to the working-dir
        # and sets JUJU_REPOSITORY.
        charms_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, charms_dir)
        working_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, working_dir)
        # Creates 'tree' fake-charms directories.
        # - precise
        # | - charm1 (empty)
        # | - charm2 (Makefile for creating 'make_called')
        p = os.path.join(charms_dir, 'precise')
        os.mkdir(p)
        charm1 = os.path.join(p, 'charm1')
        os.mkdir(charm1)
        charm2 = os.path.join(p, 'charm2')
        os.mkdir(charm2)
        with open(os.path.join(charm2, 'Makefile'), 'w') as f:
            f.write('all:\n')
            f.write('\t@touch make_called\n')
        # Copy the unbuilt tree for simulating a pre-existing working_dir.
        charms_copy = os.path.join(working_dir, 'charms')
        dir_util.copy_tree(charms_dir, charms_copy)
        # Track charm built indicators. Both tree and working-dir are unbuilt.
        built_path = 'precise/charm2/make_called'
        tree_built = os.path.join(charms_dir, built_path)
        copy_built = os.path.join(charms_copy, built_path)
        self.assertFalse(os.path.exists(tree_built))
        self.assertFalse(os.path.exists(copy_built))

        deploy.build_charms(working_dir, charms_dir)

        self.assertTrue(os.path.exists(tree_built))
        self.assertTrue(os.path.exists(copy_built))
        self.assertItemsEqual(['charms'], os.listdir(working_dir))
        self.assertItemsEqual(os.listdir(charms_dir), os.listdir(charms_copy))
        self.assertEquals(
            os.path.join(charms_copy), os.environ['JUJU_REPOSITORY'])


class TestNeedsBotstrap(unittest.TestCase):

    @mock.patch('deploy._get_control_bucket',
                return_value='juju-d911b43e427c204cfcfd8c7276c4b2aa')
    def test_needs_bootstrap(self, mock_get_control_bucket):
        provider_state = '{ state-instances: ["abcdefg"] }'
        with mock.patch('deploy._get_from_swift', return_value=provider_state):
            # Bootstrap needed - the bucket exists but the instance does not.
            e = exceptions.NotFound('Four oh four')
            with mock.patch('deploy.get_bootstrap_ip', side_effect=e):
                with mock.patch('deploy._delete_control_bucket') as dcb:
                    self.assertTrue(deploy.needs_bootstrap())
                    dcb.assert_called_with()

            # Bootstrap not needed.
            ip_addr = '127.0.0.1'
            with mock.patch('deploy.get_bootstrap_ip', return_value=ip_addr):
                self.assertFalse(deploy.needs_bootstrap())

        # Bootstrap needed - the bucket does not exist.
        e = swiftclient.exceptions.ClientException('not found')
        with mock.patch('deploy._get_from_swift', side_effect=e):
            with mock.patch('deploy._delete_control_bucket') as dcb:
                self.assertTrue(deploy.needs_bootstrap())
                dcb.assert_called_with()


class TestUtilityFunctions(unittest.TestCase):

    def test_juju_status(self):
        mocked = {'environment': 'local', 'machines': {}, 'services': {}}
        mocked = yaml.safe_dump(mocked)
        with mock.patch('subprocess.check_output', return_value=mocked):
            status = deploy.juju_status()
            self.assertEqual('local', status['environment'])

    @mock.patch('sys.stdout', new_callable=StringIO.StringIO)
    def test_print_juju_unit_file(self, mocked_out):
        with mock.patch('subprocess.check_call'):
            log_path = '/var/log/juju/unit-foo-service-1.log'
            deploy.print_juju_unit_file('foo-service/1', log_path)
            actual = mocked_out.getvalue().split('\n')
            self.assertIn(log_path, actual)

    @mock.patch('sys.stderr', new_callable=StringIO.StringIO)
    def test_print_juju_unit_file_fails(self, mocked_err):
        e = subprocess.CalledProcessError(-1, 'juju ssh')
        with mock.patch('subprocess.check_call', side_effect=e):
            log_path = '/var/log/juju/unit-foo-service-1.log'
            deploy.print_juju_unit_file('foo-service/1', log_path)
            actual = mocked_err.getvalue()
            self.assertIn('Failed to capture juju file', actual)

    def test_units_in_error(self):
        units = {'foo-service/0': {},
                 'foo-service/1': {'agent-state': 'error'}}
        srv = {'foo-service': {'charm': 'foo', 'units': units},
               'bar-service': {'charm': 'bar'}}
        status = {'environment': 'local', 'machines': {}, 'services': srv}
        errors = list(deploy.units_in_error(status))
        assertions.assertLength(self, 1, errors)
        self.assertEquals('foo-service/1', errors[0][0])

    def test_subordinate_in_error(self):
        subordinates = {'bar-service/0': {},
                        'bar-service/1': {'agent-state': 'error'}}
        units = {'foo-service/0': {},
                 'foo-service/1': {'subordinates': subordinates}}
        srv = {'foo-service': {'charm': 'foo', 'units': units},
               'bar-service': {'charm': 'bar'}}
        status = {'environment': 'local', 'machines': {}, 'services': srv}
        errors = list(deploy.units_in_error(status))
        assertions.assertLength(self, 1, errors)
        self.assertEquals('bar-service/1', errors[0][0])


class TestUpgrade(unittest.TestCase):

    def setUp(self):
        super(TestUpgrade, self).setUp()

        class FakeDeployment(object):
            pass

        self.fake_deployment = FakeDeployment()
        self.fake_deployment.data = {
            'services': {
                'foo-service': {
                    'options': {
                        'a': 'b',
                        'c': 'd',
                    },
                },
            },
        }
        _p = mock.patch('deploy._get_deployer')
        _p.start().return_value = self.fake_deployment
        self.addCleanup(_p.stop)

        self.fake_status = {
            'services': self.fake_deployment.data['services'].keys()}
        _p = mock.patch('deploy.juju_status')
        _p.start().return_value = self.fake_status
        self.addCleanup(_p.stop)

    @mock.patch('sys.stdout')
    @mock.patch('subprocess.check_call')
    def test_upgrade_charms(self, mocked_call, mocked_out):
        # '--upgrade charms' invokes juju to upgrade the charm on the
        # selected services.
        name = 'ci-airline'
        operations = ['charms']
        deployer_dir = 'IGNORED'
        deploy._upgrade(name, operations, deployer_dir, [])
        mocked_call.assert_called_once_with(
            ['juju', 'upgrade-charm', '--force', 'foo-service'])

    @mock.patch('sys.stdout')
    @mock.patch('subprocess.check_call')
    def test_upgrade_config(self, mocked_call, mocked_out):
        # '--upgrade config' invokes juju to update all configuration
        # options for the selected services.
        name = 'ci-airline'
        operations = ['config']
        deployer_dir = 'IGNORED'
        deploy._upgrade(name, operations, deployer_dir, [])
        mocked_call.assert_called_once_with(
            ['juju', 'set', 'foo-service', 'a=b', 'c=d'])

    @mock.patch('sys.stdout')
    @mock.patch('subprocess.check_call')
    def test_upgrade_code(self, mocked_call, mocked_out):
        # '--upgrade code' is a special case of 'config' and invokes
        # juju to update the 'current_code' and 'available_code'
        # configuration options only for the selected services.
        options = (
            self.fake_deployment.data['services']['foo-service']['options'])
        options.update({
            'current_code': 'B',
            'available_code': 'A B C',
        })
        name = 'ci-airline'
        operations = ['code']
        deployer_dir = 'IGNORED'
        deploy._upgrade(name, operations, deployer_dir, [])
        mocked_call.assert_called_once_with(
            ['juju', 'set', 'foo-service', 'available_code=A B C',
             'current_code=B'])

    @mock.patch('sys.stdout')
    @mock.patch('subprocess.check_call')
    def test_upgrade_all(self, mocked_call, mocked_out):
        # '--upgrade all' invokes juju twice to upgrade the charm and
        # its configuration options for the selected services.
        name = 'ci-airline'
        operations = ['all']
        deployer_dir = 'IGNORED'
        deploy._upgrade(name, operations, deployer_dir, [])
        self.assertEqual(
            [['juju', 'upgrade-charm', '--force', 'foo-service'],
             ['juju', 'set', 'foo-service', 'a=b', 'c=d']],
            [kall[0][0] for kall in mocked_call.call_args_list])

    @mock.patch('sys.stdout')
    @mock.patch('subprocess.check_call')
    def test_upgrade_only_selected_services(self, mocked_call, mocked_out):
        # '--upgrade' only acts on selected services (deploy.py positional
        # arguments, configuration names). Services not mentioned are not
        # affected.
        # Simulate an aditional deployed services.
        self.fake_status = ['foo-service', 'bar-service']
        name = 'ci-airline'
        operations = ['charms']
        deployer_dir = 'IGNORED'
        deploy._upgrade(name, operations, deployer_dir, [])
        # the upgrade requested only affected the 'selected' services.
        mocked_call.assert_called_once_with(
            ['juju', 'upgrade-charm', '--force', 'foo-service'])


if __name__ == '__main__':
    unittest.main()
