#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import os
import subprocess
import sys

import lander
from lander import workflow
from ci_utils import (
    dump_stack,
)


HERE = os.path.abspath(os.path.dirname(__file__))
logger = lander.get_logger('lander')


def _do_step(step, results_file):
    cmd = os.path.join(HERE, 'lander_service_wrapper.py')
    rc = subprocess.call(
        [cmd, '--service', step['name'], '--results-file', results_file])
    logger.debug('Step completed with rc=%d', rc)
    return rc


def _get_parser():
    parser = argparse.ArgumentParser(
        description='Process a ticket through the various lander steps.')
    parser.add_argument('-u', '--ticket-system-url', required=True,
                        help='URL of the ticket system')
    parser.add_argument('-r', '--results', required=True,
                        help='File to store JSON results in')
    parser.add_argument('--keep-ppa', action='store_true',
                        help='Do not free up PPA after processing')
    ex = parser.add_mutually_exclusive_group(required=True)
    ex.add_argument('-t', '--ticket-data', required=False)
    ex.add_argument('--resume', action='store_true', required=False)

    return parser


def _create_workflow(args):
    if args.resume:
        return workflow.load(args.results)

    with open(args.results, 'w') as f:
        ticket_data = json.loads(args.ticket_data)
        wflow = workflow.create(
            args.ticket_system_url, ticket_data)
        wflow.save(f)
        # lander-service-wrapper calls in _do_step need this file
        # so ensure its flushed to disk
        f.flush()
        return wflow


def _resume_from_steps(steps):
    '''Find the first incomplete step and return everything from there.'''
    for i in range(len(steps)):
        if steps[i].get('state') != 'COMPLETED':
            break
    return steps[i:]


def _handle_pause(results, wflow, step):
    if 'pause' in step and not step['pause'].get('handled'):
        logger.info('pausing ticket at step(%s)', step['name'])
        step['pause']['handled'] = True
        with open(results, 'w') as f:
            wflow.save(f)
        wflow.ticket_api().pause_ticket()
        sys.exit(0)


def main(args):
    rc = 1
    wflow = _create_workflow(args)

    ticket_id = wflow.ticket_id

    steps = wflow.steps()
    if args.resume:
        steps = _resume_from_steps(steps)
    step_names = [x['name'] for x in steps]
    logger.info('Steps to run ticket: %s', ', '.join(step_names))

    for step in steps:
        lander.log_status("running step: %s" % step)
        lander.add_status(ticket_id,
                          {'status_type': 'workflow', 'step': step})
        _handle_pause(args.results, wflow, step)
        rc = _do_step(step, args.results)
        if rc != 0:
            break

    return rc


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    args = _get_parser().parse_args()
    sys.exit(main(args))
