# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import json
import logging
import os
import sys

import ci_utils


LOG_STATUS_FILENAME = 'log_status.json'
STATUS_FILENAME = 'status.log'


def _setup_log_path(filename, directory=None):
    log_dir = ci_utils.get_logs_dir()

    if directory is not None:
        log_dir = os.path.join(log_dir, directory)

    try:
        # Try to create needed dirs
        os.makedirs(log_dir)
    except OSError as e:
        # They are already there, no worries
        if e.errno != errno.EEXIST:
            raise

    path = os.path.join(log_dir, filename)

    return path


def _get_status_logfile():
    return _setup_log_path(STATUS_FILENAME)


def _get_status_datafile(ticket_id):
    return _setup_log_path(LOG_STATUS_FILENAME, directory=str(ticket_id))


def _init_logging():
    # set up a root logger for everything
    handler = logging.StreamHandler(stream=sys.stderr)
    formatter = logging.Formatter(
        '%(asctime)s %(levelname)-5s %(name)s: %(message)s',
        datefmt='%H:%M:%S')
    handler.setFormatter(formatter)
    l = logging.getLogger('')
    l.addHandler(handler)
    l.setLevel(logging.DEBUG)

    # set up a status logger for prodution testing
    lander_status_logfile = _get_status_logfile()
    if (lander_status_logfile and
            os.path.exists(os.path.dirname(lander_status_logfile))):
        l.info("Setting up status logger")
        handler = logging.FileHandler(lander_status_logfile)
        handler.setFormatter(formatter)
        l = logging.getLogger('lander_status')
        l.addHandler(handler)
        l.setLevel(logging.DEBUG)
    else:
        l.info("Unable to set up status logger: %s", lander_status_logfile)

    # the keystoneclient.httpclient debug logging leaks passwords. It enables
    # this based on the setting of the "swiftclient" level
    logging.getLogger('swiftclient').setLevel(logging.INFO)
    return True

_init_logging.called = False


def get_logger(name):
    if not _init_logging.called:
        _init_logging.called = _init_logging()
    return logging.getLogger(name)


def log_status(msg):
    status_logger = get_logger('lander_status')
    status_logger.info(msg)


def add_status(ticket_id, data):
    data_file = _get_status_datafile(ticket_id)
    with open(data_file, 'a') as fp:
        fp.write(json.dumps(data))
        fp.write('\n')
        fp.flush()
