#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
import json
import os
import re
import subprocess
import sys
import time

import lander
from lander.ticket_api import TicketApi

from ci_utils import amqp_utils, data_store, dump_stack, unit_config

HERE = os.path.abspath(os.path.dirname(__file__))
SERVICE_PATH = os.path.join(HERE, '../../../../juju-relations')
PROCESS_TICKET = os.path.join(HERE, '../bin/lander_process_ticket.py')

logger = lander.get_logger('lander_next_ticket')
status_logger = lander.get_logger('lander_status')


def _get_parser():
    logdir = os.path.join(HERE, '../../../../lander-logs')
    parser = argparse.ArgumentParser(
        description='Wraps the REST API calls with a progress queue monitor.')
    parser.add_argument('--service-name',
                        required=True,
                        help='The name of the service to connect.')
    parser.add_argument('--service-port',
                        required=True,
                        help='The port number of the service to connect.')
    parser.add_argument('--delay',
                        required=True,
                        type=int,
                        help='The delay in seconds between poll attempts.')
    parser.add_argument('--log-dir',
                        required=False, default=logdir,
                        help='''Directory to store lander logs for a ticket.
                        default=%(default)s''')
    return parser


def _store_artifacts(ts_url, ticket_id, fname, artifact_type, resume):
    try:
        auth = unit_config.get_auth_config()
        ds = data_store.create_for_ticket(ticket_id, auth)
        name = os.path.basename(fname)
        with open(fname) as f:
            url = ds.put_file(name, f.read(), 'text/plain')
            if not resume:
                # only create ticket objects the first time
                ticket = TicketApi(ts_url, ticket_id)
                ticket.add_artifact(name, url, artifact_type)
    except:
        logging.exception('unable to store artifact for ticket %s', ticket_id)


def _update_local_result(ticket_id, path):
    '''Grab the results file from swift and save it locally'''
    ds = data_store.create_for_ticket(ticket_id, unit_config.get_auth_config())
    with open(path, 'w') as f:
        f.write(ds.get_file(os.path.basename(path)))


def _process_ticket(ts_url, ticket_data, log_dir):
    ticket_id = ticket_data['id']
    logging.info('ticket_id: %s', ticket_id)
    logging.info('request_parameters: %s', ticket_data)

    results = os.path.join(log_dir, 'ticket-%s_results.json' % ticket_id)
    args = [PROCESS_TICKET, '-u', ts_url, '-r', results]

    if ticket_data.get('resume'):
        logging.info('resuming ticket')
        _update_local_result(ticket_id, results)
        args.append('--resume')
        mode = 'a'
    else:
        mode = 'w'
        args.extend(['-t', json.dumps(ticket_data)])

    with open(os.path.join(log_dir, 'ticket-%s.log' % ticket_id), mode) as f:
        logging.info('ticket processing log: %s', f.name)
        rc = subprocess.call(args, stdout=f, stderr=f)
        logging.info('process ticket exited with: %d', rc)

    resume = ticket_data.get('resume', False)
    _store_artifacts(ts_url, ticket_id, f.name, 'LOGS', resume)
    _store_artifacts(ts_url, ticket_id, results, 'RESULTS', resume)


class Lander(object):
    def __init__(self, ts_url, logdir):
        self.ts_url = ts_url
        self.logdir = logdir

    def on_message(self, msg):
        lander.log_status("worker start message: %s" % msg.__dict__)

        ticket = None
        try:
            ticket = json.loads(msg.body)
        except Exception as e:
            logging.error(e)
            pass

        if ticket is None:
            logging.error('Invalid message: %s', msg)
            logging.error('acking message so its expunged from queue')
            msg.channel.basic_ack(msg.delivery_tag)
            return

        lander.add_status(ticket.get('id'), {
            'status_type': 'run_worker',
            'type': 'start',
            'data': ticket,
        })

        try:
            msg.channel.basic_ack(msg.delivery_tag)
            _process_ticket(self.ts_url, ticket, self.logdir)
        except:
            logging.error('Failed processing ticket: %s', msg.body)
            raise
        finally:
            lander.log_status("worker end message: %s" % msg.__dict__)
            lander.add_status(ticket.get('id'), {
                'status_type': 'run_worker',
                'type': 'end',
                'data': msg.body,
            })


def get_service_url(service_name, service_port):
    if not os.path.exists(SERVICE_PATH):
        return None
    for name in os.listdir(SERVICE_PATH):
        with open(os.path.join(SERVICE_PATH, name)) as f:
            buff = f.read().strip()
            if re.match('%s:' % service_name, buff):
                (junk, url) = buff.split(':')
                return 'http://{}:{}'.format(url.strip(), service_port)


def _wait_for_service(service, port):
    while True:
        logging.info('Trying to acquire url for %s', service)
        url = get_service_url(service, port)
        if not url:
            logging.error('Unable to acquire service url, retrying')
            time.sleep(args.delay)
        else:
            return url


def _wait_for_config(args):
    while True:
        config = amqp_utils.get_config()
        if config:
            return config
        logging.error('No rabbit config availble, retrying')
        time.sleep(args.delay)


def main(args):
    logging.info('Starting lander worker.')
    if not os.path.exists(args.log_dir):
        os.makedirs(args.log_dir)

    ts_url = _wait_for_service(args.service_name, args.service_port)
    lander = Lander(ts_url, args.log_dir)
    cfg = _wait_for_config(args)
    amqp_utils.process_queue(cfg, amqp_utils.LANDER_QUEUE, lander.on_message)


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    logging.basicConfig(level=logging.INFO,
                        format="%(asctime)s %(levelname)s %(message)s")
    args = _get_parser().parse_args()
    sys.exit(main(args))
