# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import sys
import tempfile
import unittest

import mock

from lander import workflow

HERE = os.path.abspath(os.path.dirname(__file__))
sys.path.append(os.path.join(HERE, '../../bin'))
import lander_process_ticket


class TestProcessTicket(unittest.TestCase):
    '''Sanity tests of the lander_process_ticket.py script.'''
    def setUp(self):
        super(TestProcessTicket, self).setUp()
        self.tmpfile = tempfile.mkstemp()[1]
        self.addCleanup(os.unlink, self.tmpfile)
        tdata = json.dumps({
            'id': 1234,
            'steps': [
                {'name': 'ppa_creator'},
                {'name': 'bsbuilder'},
                {'name': 'publisher'},
            ],
        })
        self.base_args = ['-u', 'ts-url', '-t', tdata,
                          '-r', self.tmpfile]

    @mock.patch('lander_process_ticket._do_step')
    def test_run_succeeds(self, do_step):
        do_step.return_value = 0
        args = lander_process_ticket._get_parser().parse_args(self.base_args)
        self.assertEqual(0, lander_process_ticket.main(args))

    @mock.patch('lander_process_ticket._do_step')
    def test_run_fails(self, do_step):
        do_step.return_value = 123
        args = lander_process_ticket._get_parser().parse_args(self.base_args)
        self.assertEqual(123, lander_process_ticket.main(args))
        self.assertEqual(1, do_step.call_count)

    @mock.patch('lander_process_ticket._do_step')
    def test_run_resumes(self, do_step):
        do_step.return_value = 0
        data = {
            'ticket_id': 10,
            'steps': [
                {'name': 'step1'},
                {'name': 'step2'},
                {'name': 'step3'},
            ],
        }
        wflow = workflow._Workflow(data)

        steps = wflow.steps()
        steps[0]['state'] = 'COMPLETED'
        steps[1]['state'] = 'COMPLETED'
        with open(self.tmpfile, 'w') as f:
            wflow.save(f)

        args = ['-u', 'ts', '-r', self.tmpfile, '--resume']
        args = lander_process_ticket._get_parser().parse_args(args)
        self.assertEqual(0, lander_process_ticket.main(args))
        self.assertEqual(len(steps) - 2, do_step.call_count)
        first_called = do_step.call_args_list[0][0][0]['name']
        self.assertEqual(steps[2]['name'], first_called)

    @mock.patch('lander.ticket_api')
    @mock.patch('lander_process_ticket._do_step')
    def test_run_pause(self, do_step, ticket_api):
        do_step.return_value = 0
        data = {
            'id': 10,
            'steps': [
                {'name': 'step1'},
                {'name': 'step2', 'pause': {}},
            ],
        }
        wflow = workflow._Workflow(data)
        with open(self.tmpfile, 'w') as f:
            wflow.save(f)
        args = ['-u', 'ts', '-r', self.tmpfile,
                '-t', json.dumps(data)]
        args = lander_process_ticket._get_parser().parse_args(args)
        with self.assertRaises(SystemExit):
            lander_process_ticket.main(args)
        self.assertEqual(1, do_step.call_count)
        self.assertEqual(1, ticket_api.TicketApi().pause_ticket.call_count)
