# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import sys
import tempfile
import unittest

import mock

import lander.workflow


HERE = os.path.abspath(os.path.dirname(__file__))
sys.path.append(os.path.join(HERE, '../../bin'))
import lander_service_wrapper


class TestWrapper(unittest.TestCase):
    """Sanity tests of the lander_service_wrapper.py script."""

    def setUp(self):
        """Creates a default lander workflow/context for all tests."""
        super(TestWrapper, self).setUp()
        patch = mock.patch('ci_utils.amqp_utils.get_config')
        patch.start()
        self.addCleanup(patch.stop)
        patch_uuid4 = mock.patch('uuid.uuid4', return_value='1234567890')
        patch_uuid4.start()
        self.addCleanup(patch_uuid4.stop)

        self.results = tempfile.mkstemp(prefix='lander-test-')[1]
        self.addCleanup(os.unlink, self.results)

        tsdata = {
            'id': 1,
            'steps': [
                {'name': 'ppa_creator'},
                {'name': 'bsbuilder'},
                {'name': 'validator'},
                {'name': 'image_builder'},
                {'name': 'test_runner'},
                {'name': 'publisher'},
            ]
        }
        workflow = lander.workflow.create('ts_url', tsdata)
        workflow.set_results('ppa_creator', {'ppa_name': 'new-ppa',
                                             'unique_arches': 'arches'})
        workflow.set_results('image_builder', {'image_id': 'new-image'})
        with open(self.results, 'w') as f:
            workflow.save(f)

        self.base_args = ['--results-file', self.results]

    def mock_lander_loop(self, response):
        """Mock the lander loop to execute a set of defined callbacks.

        Emulates the response of worker in normal operation.

        Wires a WAITING response (and callback) and another one with
        the given response.
        """
        def get_mocked_server():
            def run_forever(channel, queue, callback):
                m = mock.Mock()
                m.body = json.dumps({'state': 'WAITING'})
                callback(m)
                m.body = json.dumps(response)
                callback(m)
            return run_forever

        _m = mock.patch('ci_utils.amqp_utils._run_forever',
                        new_callable=get_mocked_server)
        _m.start()
        self.addCleanup(_m.stop)

    def testBadService(self):
        # Make sure we exit gracefully for a bad service name.
        args = self.base_args + ['--service', 'bad-name']
        with self.assertRaises(SystemExit):
            lander_service_wrapper._get_parser().parse_args(args)

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testPPACreatorSucceeds(self, connection, send, api):
        # Ensure we generate the proper response if things succeed.
        api = api()
        # Mock the required information from the fullticket API.
        api.get_full_ticket.return_value = {
            'subticket': [{
                'id': 1,
                'sourcepackage': {'name': 'foobar'},
                'source_package_uploads': [{
                    'architecture': 'fake-arch',
                }],
            }],
            'uuid': 'AUUID',
            'series': 'a_series',
        }
        # Mock the lander loop with the expected (successful) ppa-creator
        # response.
        response = {
            'name': 'ppa_creator',
            'ppa_name': 'ppa:foobar/ephemeral-ppa',
            'state': 'COMPLETED',
            'result': 'PASSED',
            'exit': True,
        }
        self.mock_lander_loop(response)
        # Lander run succeeds.
        args = self.base_args + ['--service', 'ppa_creator']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(0, lander_service_wrapper.main(args))
        # The context ticket is updated with the just-created ppa name.
        api.set_ppacreation_waiting.assert_called_once()
        api.set_ppacreation_complete.assert_called_once_with(True, response)
        # The lander workflow is updated as well.
        self.assertEqual(
            response,
            lander.workflow.load(self.results).get_step('ppa_creator'))

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testPPACreatorMultiSucceeds(self, connection, send, api):
        api = api()
        api.get_full_ticket.return_value = {
            'subticket': [{
                'id': 1,
                'sourcepackage': {'name': 'foobar1'},
                'source_package_uploads': [{
                    'architecture': 'fake-arch1',
                }],
            }, {
                'id': 2,
                'sourcepackage': {'name': 'foobar2'},
                'source_package_uploads': [{
                    'architecture': 'fake-arch2',
                }],
            }, {
                'id': 3,
                'sourcepackage': {'name': 'foobar3'},
                'source_package_uploads': [{
                    'architecture': 'fake-arch2',
                }],
            }],
            'uuid': 'AUUID',
            'series': 'a_series',
        }

        response = {
            'name': 'ppa_creator',
            'ppa_name': 'ppa:foobar/ephemeral-ppa',
            'state': 'COMPLETED',
            'result': 'PASSED',
            'exit': True,
        }
        self.mock_lander_loop(response)
        params = {
            'ticket_id': 1,
            'name': 'AUUID',
            'displayname': 'AUUID',
            'description': 'AUUID',
            'architectures': ('fake-arch2', 'fake-arch1'),
            'series': 'a_series',
            'progress_trigger': 'lander-1234567890-ppacreator',
        }
        # Lander run succeeds.
        args = self.base_args + ['--service', 'ppa_creator']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(0, lander_service_wrapper.main(args))

        send.assert_called_once_with('ppacreator',
                                     json.dumps(params),
                                     True)

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    def testBSBuilderUnexpected(self, send, api):
        # Ensure BSBuilder unexpected errors still generate a response.
        err_msg = 'foo bar'
        send.side_effect = RuntimeError(err_msg)
        api = api()
        api.get_full_ticket.return_value = {'subticket': 1}

        args = self.base_args + ['--service', 'bsbuilder']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(1, lander_service_wrapper.main(args))

        api.set_pkg_waiting.assert_called_once()
        api.set_pkg_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('bsbuilder')
        self.assertEqual('FAILED', data['result'])
        self.assertEqual(err_msg, data['error'])

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testBSBuilderSucceeds(self, connection, send, api):
        # Ensure BSBuilder generates the proper response if things succeed.
        api = api()
        api.get_full_ticket.return_value = {
            'subticket': 1
        }
        response = {
            'name': 'bsbuilder',
            'state': 'COMPLETED',
            'result': 'PASSED',
            'foo': 'bar',
            'exit': True,
        }
        self.mock_lander_loop(response)

        args = self.base_args + ['--service', 'bsbuilder']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(0, lander_service_wrapper.main(args))

        api.set_pkg_waiting.assert_called_once()
        api.set_pkg_complete.assert_called_once_with(True, response)

        self.assertEqual(
            response,
            lander.workflow.load(self.results).get_step('bsbuilder'))

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    def testValidatorUnexpected(self, send, api):
        # Unexpected validator errors are stored, ticket is updated and
        # the lander is not blocked.
        err_msg = 'foo bar'
        send.side_effect = RuntimeError(err_msg)
        api = api()
        api.get_full_ticket.return_value = {
            'ticket_id': 1,
            'progress_trigger': 'A_TRIGGER',
            'subticket': [{
                'id': 1,
                'sourcepackage': {'name': 'foobar'},
                'source_package_upload': []
            }],
            'source_archive': 'A_PPA',
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'current_workflow_step': 'Queued',
        }

        args = self.base_args + ['--service', 'validator']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(1, lander_service_wrapper.main(args))

        api.set_validation_waiting.assert_called_once()
        api.set_validation_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('validator')
        self.assertEqual('FAILED', data['result'])
        self.assertEqual(err_msg, data['error'])

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testValidatorSucceeds(self, connection, send, api):
        # Validator calls succeeds.
        api = api()
        # Provide ticket information (fullticket) to be fed into the
        # validator by the lander.
        api.get_full_ticket.return_value = {
            'ticket_id': 1,
            'progress_trigger': 'A_TRIGGER',
            'subticket': [{
                'id': 1,
                'sourcepackage': {'name': 'foobar'},
                'source_package_upload': []
            }],
            'source_archive': 'A_PPA',
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'current_workflow_step': 'Queued',
        }
        # Mock the lander loop with a successfull validador response.
        response = {
            'name': 'validator',
            'state': 'COMPLETED',
            'result': 'PASSED',
            'foo': 'bar',
            'exit': True,
        }
        self.mock_lander_loop(response)
        # Invoke the lander for the validator service and confirms
        # it's return code is ZERO (successs).
        args = self.base_args + ['--service', 'validator']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(0, lander_service_wrapper.main(args))
        # The ticket status is updated accordingly (WAITING then COMPLETED).
        api.set_validation_waiting.assert_called_once()
        api.set_validation_complete.assert_called_once_with(True, response)
        # The internal lander status registry is updated accordingly
        # (with the validator response).
        self.assertEqual(
            response,
            lander.workflow.load(self.results).get_step('validator'))

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    def testImageBuilderUnexpected(self, send, api):
        # Ensure ImageBuilder unexpected errors still generate a response.
        err_msg = 'foo bar'
        send.side_effect = RuntimeError(err_msg)
        api = api()
        api.get_full_ticket.return_value = {
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'base_image': 'A_IMAGE',
        }

        args = self.base_args + ['--service', 'image_builder']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(1, lander_service_wrapper.main(args))

        api.set_image_waiting.assert_called_once()
        api.set_image_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('image_builder')
        self.assertEqual('FAILED', data['result'])
        self.assertEqual(err_msg, data['error'])

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testImageBuilderSucceeds(self, connection, send, api):
        # Ensure ImageBuilder succeeds.
        api = api()
        api.get_full_ticket.return_value = {
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'base_image': 'A_IMAGE',
        }

        response = {
            'name': 'image_builder',
            'state': 'COMPLETED',
            'result': 'PASSED',
            'exit': True,
        }
        self.mock_lander_loop(response)

        args = self.base_args + ['--service', 'image_builder']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(0, lander_service_wrapper.main(args))

        api.set_image_waiting.assert_called_once()
        api.set_image_complete.assert_called_once_with(True, response)

        self.assertEqual(
            response,
            lander.workflow.load(self.results).get_step('image_builder'))

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    def testPublisherUnexpected(self, send, api):
        # Ensure Publisher unexpected errors still generate a response.
        err_msg = 'foo bar'
        send.side_effect = RuntimeError(err_msg)
        api = api()
        api.get_full_ticket.return_value = {
            'ticket_id': 1,
            'progress_trigger': 'A_TRIGGER',
            'subticket': [{
                'id': 1,
                'sourcepackage': {'name': 'foobar'},
                'source_package_uploads': []
            }],
            'source_archive': 'A_PPA',
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'current_workflow_step': 'Queued',
        }

        args = self.base_args + ['--service', 'publisher']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(1, lander_service_wrapper.main(args))

        api.set_publishing_waiting.assert_called_once()
        api.set_publishing_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('publisher')
        self.assertEqual('FAILED', data['result'])
        self.assertEqual(err_msg, data['error'])

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testPublisherSucceeds(self, connection, send, api):
        # Ensure Publisher generates the proper response if things succeed.
        api = api()
        api.get_full_ticket.return_value = {
            'ticket_id': 1,
            'progress_trigger': 'A_TRIGGER',
            'subticket': [{
                'id': 1,
                'sourcepackage': {'name': 'foobar'},
                'source_package_upload': []
            }],
            'source_archive': 'A_PPA',
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'current_workflow_step': 'Queued',
        }

        response = {
            'name': 'publisher',
            'state': 'COMPLETED',
            'result': 'PASSED',
            'exit': True,
        }
        self.mock_lander_loop(response)

        args = self.base_args + ['--service', 'publisher']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(0, lander_service_wrapper.main(args))

        api.set_publishing_waiting.assert_called_once()
        api.set_publishing_complete.assert_called_once_with(True, response)

        self.assertEqual(
            response,
            lander.workflow.load(self.results).get_step('publisher'))

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testRunnerSucceeds(self, connection, send, api):
        # Ensure TestRunner generates the proper response if things succeed.
        api = api()
        api.get_full_ticket.return_value = {
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'subticket': [{
                'id': 1,
                'sourcepackage': {'name': 'foobar'},
                'source_package_upload': []
            }],
        }

        response = {
            'name': 'test_runner',
            'state': 'COMPLETED',
            'result': 'PASSED',
            'exit': True,
        }
        self.mock_lander_loop(response)

        args = self.base_args + ['--service', 'test_runner']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(0, lander_service_wrapper.main(args))

        api.set_testing_waiting.assert_called_once()
        api.set_testing_complete.assert_called_once_with(True, response)

        self.assertEqual(
            response,
            lander.workflow.load(self.results).get_step('test_runner'))

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    def testRunnerUnexpected(self, send, api):
        # Ensure TestRunner unexpected errors still generate a response.
        err_msg = 'foo bar'
        send.side_effect = RuntimeError(err_msg)
        api = api()
        api.get_full_ticket.return_value = {
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'subticket': [{
                'id': 1,
                'sourcepackage': {'name': 'foobar'},
                'source_package_upload': []
            }],
        }

        args = self.base_args + ['--service', 'test_runner']
        args = lander_service_wrapper._get_parser().parse_args(args)
        self.assertEqual(1, lander_service_wrapper.main(args))

        api.set_testing_waiting.assert_called_once()
        api.set_testing_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('test_runner')
        self.assertEqual('FAILED', data['result'])
        self.assertEqual(err_msg, data['error'])
