# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import mock
import unittest

from ci_utils.ticket_states import (
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)
from lander import ticket_api


class TestTicketApi(unittest.TestCase):
    '''Tests for the ticket_api.py module.'''

    ticket_data = {
        'artifact': [{
            'id': 123,
            'name': 'artifact_name',
            'reference': 'artifact_reference',
            'resource_uri': '/api/v1/fullticketartifact/123',
            'type': 'LOGS',
        }],
        'subticket': [{
            'id': 12,
            'artifact': [{
                'id': 321,
                'name': 'subticket_artifact_name',
                'reference': 'subticket_artifact_reference',
                'resource_uri': '/api/v1/fullticketartifact/12',
                'type': 'LOGS'
            }],
        }]
    }

    def setUp(self):
        super(TestTicketApi, self).setUp()
        self.ticket = ticket_api.TicketApi('http://ts-url', '123')

    def test_add_artifact_default(self):
        # `Ticket` artifacts can be created and uses the context
        # 'ticket_id' (or UUID) to fetch the actual ticket 'resource_uri'
        # before posting.
        self.ticket._post = mock.Mock()
        self.ticket.add_artifact('name', 'url')
        self.ticket._post.assert_called_with(
            'http://ts-url/api/v1/ticketartifact/',
            {'type': 'LOGS',
             'name': 'name',
             'reference': 'url',
             'ticket': '/api/v1/ticket/123/'})

    def test_add_artifact_subticket(self):
        self.ticket._post = mock.Mock()
        post_url = 'http://ts-url/api/v1/subticketartifact/'
        self.ticket.add_artifact('name', 'url', artifact_type='RESULT',
                                 subticket_id=12)
        data = {
            'type': 'RESULT',
            'name': 'name',
            'reference': 'url',
            'subticket': '/api/v1/subticket/12/'
        }
        self.ticket._post.assert_called_with(post_url, data)

    def test_get_artifact_list_from_ticket(self):
        expected = ['artifact_name']
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        actual = self.ticket.get_artifact_list()
        self.assertItemsEqual(expected, actual)

    def test_get_artifact_list_from_subticket(self):
        expected = ['subticket_artifact_name']
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        actual = self.ticket.get_artifact_list(subticket_id=12)
        self.assertItemsEqual(expected, actual)

    def test_get_artifact_list_from_subticket_assert(self):
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        with self.assertRaises(ValueError):
            self.ticket.get_artifact_list(subticket_id=10)

    def test_process_artifacts(self):
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        self.ticket.add_artifact = mock.Mock()
        data = {
            'artifacts': [{
                'name': 'new_artifact_name',
                'reference': 'new_artifact_reference',
                'type': 'LOGS'}],
        }
        self.ticket.process_artifacts(data)
        self.ticket.add_artifact.assert_called_with(
            'new_artifact_name',
            'new_artifact_reference',
            'LOGS',
            None)

    def test_process_artifacts_subticket(self):
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        self.ticket.add_artifact = mock.Mock()
        data = {
            'artifacts': [{
                'name': 'new_artifact_name',
                'reference': 'new_artifact_reference',
                'type': 'LOGS'}],
        }
        self.ticket.process_artifacts(data, subticket_id=12)
        self.ticket.add_artifact.assert_called_with(
            'new_artifact_name',
            'new_artifact_reference',
            'LOGS',
            12)

    def test_process_artifacts_none(self):
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        self.ticket.add_artifact = mock.Mock()
        data = {}
        self.ticket.process_artifacts(data)
        self.assertEqual(0, self.ticket.add_artifact.called)

    def _test_status(self, method, step, status, extra=None):
        api = ticket_api.TicketApi('foo', 3)
        api._patch_ticket = mock.Mock()
        if extra is not None:
            method(api, extra)
        else:
            method(api)
        api._patch_ticket.assert_called_once_with(step, status)

    def test_status_pkg_waiting(self):
        self._test_status(ticket_api.TicketApi.set_pkg_waiting,
                          ticket_api.TicketWorkflowStep.PKG_BUILDING,
                          ticket_api.TicketWorkflowStepStatus.QUEUED)

    def test_status_pkg_inprogress(self):
        self._test_status(ticket_api.TicketApi.set_pkg_inprogress,
                          ticket_api.TicketWorkflowStep.PKG_BUILDING,
                          ticket_api.TicketWorkflowStepStatus.INPROGRESS)

    def test_status_pkg_complete(self):
        self._test_status(ticket_api.TicketApi.set_pkg_complete,
                          ticket_api.TicketWorkflowStep.PKG_BUILDING,
                          ticket_api.TicketWorkflowStepStatus.COMPLETED)
        self._test_status(ticket_api.TicketApi.set_pkg_complete,
                          ticket_api.TicketWorkflowStep.PKG_BUILDING,
                          ticket_api.TicketWorkflowStepStatus.FAILED,
                          False)

    def test_status_image_waiting(self):
        self._test_status(ticket_api.TicketApi.set_image_waiting,
                          ticket_api.TicketWorkflowStep.IMAGE_BUILDING,
                          ticket_api.TicketWorkflowStepStatus.QUEUED)

    def test_status_image_inprogress(self):
        self._test_status(ticket_api.TicketApi.set_image_inprogress,
                          ticket_api.TicketWorkflowStep.IMAGE_BUILDING,
                          ticket_api.TicketWorkflowStepStatus.INPROGRESS)

    def test_status_image_complete(self):
        self._test_status(ticket_api.TicketApi.set_image_complete,
                          ticket_api.TicketWorkflowStep.IMAGE_BUILDING,
                          ticket_api.TicketWorkflowStepStatus.FAILED,
                          False)

    def test_status_testing_waiting(self):
        self._test_status(ticket_api.TicketApi.set_testing_waiting,
                          ticket_api.TicketWorkflowStep.IMAGE_TESTING,
                          ticket_api.TicketWorkflowStepStatus.QUEUED)

    def test_status_testing_inprogress(self):
        self._test_status(ticket_api.TicketApi.set_testing_inprogress,
                          ticket_api.TicketWorkflowStep.IMAGE_TESTING,
                          ticket_api.TicketWorkflowStepStatus.INPROGRESS)

    def test_status_testing_complete(self):
        self._test_status(ticket_api.TicketApi.set_testing_complete,
                          ticket_api.TicketWorkflowStep.IMAGE_TESTING,
                          ticket_api.TicketWorkflowStepStatus.COMPLETED)

    def test_status_publish_waiting(self):
        self._test_status(ticket_api.TicketApi.set_publishing_waiting,
                          ticket_api.TicketWorkflowStep.PKG_PUBLISHING,
                          ticket_api.TicketWorkflowStepStatus.QUEUED)

    def test_status_publish_inprogress(self):
        self._test_status(ticket_api.TicketApi.set_publishing_inprogress,
                          ticket_api.TicketWorkflowStep.PKG_PUBLISHING,
                          ticket_api.TicketWorkflowStepStatus.INPROGRESS)

    def test_status_publish_complete(self):
        self._test_status(ticket_api.TicketApi.set_publishing_complete,
                          ticket_api.TicketWorkflowStep.COMPLETED,
                          ticket_api.TicketWorkflowStepStatus.COMPLETED)
        self._test_status(ticket_api.TicketApi.set_publishing_complete,
                          ticket_api.TicketWorkflowStep.PKG_PUBLISHING,
                          ticket_api.TicketWorkflowStepStatus.FAILED,
                          False)


class TestCleanRequestData(unittest.TestCase):
    """Verifies utility for request data cleanup."""

    def test_simple(self):
        # `clean_request_data` transforms given DBItems (lazr.enum)
        # into their 'value' (integer).
        data = {
            'int': 1,
            'string': 'foo',
            'dbitem': TicketWorkflowStepStatus.FAILED,
        }
        self.assertEqual(
            {'dbitem': 300, 'int': 1, 'string': 'foo'},
            ticket_api.clean_request_data(data))


class TestTicketApiStatusHandlers(unittest.TestCase):
    """Verify and document `TicketAPI` step/status handlers."""

    def setUp(self):
        """Mock the `TicketAPI._patch` for this context."""
        super(TestTicketApiStatusHandlers, self).setUp()
        _m = mock.patch('lander.ticket_api.TicketApi._patch')
        self.mocked_patch = _m.start()
        self.addCleanup(_m.stop)
        self.api = ticket_api.TicketApi('http://ts-server:8080', '1')

    def assertPatched(self, details):
        """Assert the api patch method was invoked with the given details."""
        # 'Clean' expected details before checking.
        data = ticket_api.clean_request_data(details)
        # Reproduce the keys implicitly injected.
        data.update({
            'lander_unit': 0,
        })
        # Assert patch method was invoked properly.
        self.mocked_patch.assert_called_once_with(
            'http://ts-server:8080/api/v1/ticket/1/', data)

    def test_set_ppacreation_waiting(self):
        # `set_ppacreation_waiting` patches ticket to:
        # PPA_CREATING + QUEUED
        self.api.set_ppacreation_waiting()
        self.assertPatched({
            'current_workflow_step': TicketWorkflowStep.PPA_CREATING,
            'status': TicketWorkflowStepStatus.QUEUED,
        })

    def test_set_ppacreation_inprogress(self):
        # `set_ppacreation_inprogress` patches ticket to:
        # PPA_CREATING + INPROGRESS
        self.api.set_ppacreation_inprogress()
        self.assertPatched({
            'current_workflow_step': TicketWorkflowStep.PPA_CREATING,
            'status': TicketWorkflowStepStatus.INPROGRESS,
        })

    def test_set_ppacreation_complete_success(self):
        # `set_ppacreation_complete` patches ticket to:
        # PPA_CREATING + COMPLETED + 'build_ppa'
        result = {'ppa_name': 'ppa:foobar/A_UUID'}
        self.api.set_ppacreation_complete(result=result)
        self.assertPatched({
            'current_workflow_step': TicketWorkflowStep.PPA_CREATING,
            'status': TicketWorkflowStepStatus.COMPLETED,
            'build_ppa': 'ppa:foobar/A_UUID',
        })

    def test_set_ppacreation_complete_fails_if_no_ppa(self):
        # `set_ppacreation_complete` requires a 'ppa_name' result
        # on success.
        with self.assertRaises(AssertionError) as cm:
            self.api.set_ppacreation_complete()
        self.assertEqual(
            'ppa_name must be given on success.',
            str(cm.exception))

    def test_set_ppacreation_complete_failure(self):
        # `set_ppacreation_complete(False)` patches ticket to:
        # PPA_CREATING + FAILED
        self.api.set_ppacreation_complete(False)
        self.assertPatched({
            'current_workflow_step': TicketWorkflowStep.PPA_CREATING,
            'status': TicketWorkflowStepStatus.FAILED,
        })

    def test_status_validation_waiting(self):
        # `set_validation_waiting` patches the ticket to:
        #   PKG_VALIDATION + QUEUED
        self.api.set_validation_waiting()
        self.assertPatched({
            'current_workflow_step': TicketWorkflowStep.PKG_VALIDATION,
            'status': TicketWorkflowStepStatus.QUEUED,
        })

    def test_status_validation_inprogress(self):
        # `set_validation_inprogress` patches the ticket to:
        #   PKG_VALIDATION + INPROGRESS
        self.api.set_validation_inprogress()
        self.assertPatched({
            'current_workflow_step': TicketWorkflowStep.PKG_VALIDATION,
            'status': TicketWorkflowStepStatus.INPROGRESS,
        })

    def test_status_validation_complete_success(self):
        # `set_validation_complete` patches the ticket to:
        #   PKG_VALIDATION + COMPLETED
        self.api.set_validation_complete()
        self.assertPatched({
            'current_workflow_step': TicketWorkflowStep.PKG_VALIDATION,
            'status': TicketWorkflowStepStatus.COMPLETED,
        })

    def test_status_validation_complete_failure(self):
        # `set_validation_complete(False)` patches the ticket to:
        #   PKG_VALIDATION + FAILED
        self.api.set_validation_complete(False)
        self.assertPatched({
            'current_workflow_step': TicketWorkflowStep.PKG_VALIDATION,
            'status': TicketWorkflowStepStatus.FAILED,
        })
