# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import urllib2
import urlparse

from ci_utils.ticket_states import (
    Item as TicketStateItem,
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


def clean_request_data(data):
    """Cleans request data before issuing POST|PATCH requests.

    Transforms DBItems into their identifiers, as expected by the API.

    Retuns a new clean dictionary.
    """
    cleaned = {}
    for k, v in data.iteritems():
        if isinstance(v, TicketStateItem):
            cleaned[k] = v.value
        else:
            cleaned[k] = v
    return cleaned


class TicketApi(object):

    def __init__(self, url, ticket_id):
        self._url = url
        self._api = '/api/v1'
        self._ticket_id = str(ticket_id)

    def _patch(self, resource, data):
        req = urllib2.Request(resource)
        req.add_header('Content-Type', 'application/json')
        req.get_method = lambda: 'PATCH'
        urllib2.urlopen(req, json.dumps(data))

    def _post(self, url, data):
        req = urllib2.Request(url)
        req.add_header('Content-Type', 'application/json')
        resp = urllib2.urlopen(req, json.dumps(data))
        return resp.headers['location']

    def _get(self, url):
        resp = urllib2.urlopen(url)
        resp = resp.read()
        return json.loads(resp)

    @property
    def ticket_url(self):
        path = os.path.join(self._api, 'ticket', self._ticket_id, '')
        return urlparse.urljoin(self._url, path)

    def pause_ticket(self):
        params = {
            'status': TicketWorkflowStepStatus.PAUSED.value,
            'lander_unit': int(os.environ.get('JUJU_UNIT_IDX', '0')),
        }
        self._patch(self.ticket_url, params)

    def _patch_ticket(self, step, status, **kwargs):
        params = {
            'current_workflow_step': step,
            'status': status,
            'lander_unit': int(os.environ.get('JUJU_UNIT_IDX', '0')),
        }
        # Update and clean the request data with other (arbitrary) ticket
        # parameters to be patched.
        params.update(kwargs)
        params = clean_request_data(params)
        self._patch(self.ticket_url, params)

    def _patch_pkg_status(self, status):
        self._patch_ticket(TicketWorkflowStep.PKG_BUILDING, status)

    def set_pkg_waiting(self):
        self._patch_pkg_status(TicketWorkflowStepStatus.QUEUED)

    def set_pkg_inprogress(self):
        self._patch_pkg_status(TicketWorkflowStepStatus.INPROGRESS)

    def set_pkg_complete(self, success=True, result=None):
        state = TicketWorkflowStepStatus.COMPLETED
        if not success:
            state = TicketWorkflowStepStatus.FAILED
        self._patch_pkg_status(state)

    def set_validation_waiting(self):
        self._patch_ticket(
            TicketWorkflowStep.PKG_VALIDATION,
            TicketWorkflowStepStatus.QUEUED)

    def set_validation_inprogress(self):
        self._patch_ticket(
            TicketWorkflowStep.PKG_VALIDATION,
            TicketWorkflowStepStatus.INPROGRESS)

    def set_validation_complete(self, success=True, result=None):
        if success:
            state = TicketWorkflowStepStatus.COMPLETED
        else:
            state = TicketWorkflowStepStatus.FAILED
        self._patch_ticket(
            TicketWorkflowStep.PKG_VALIDATION, state)

    def _patch_subticket_status(self, subticket_id, step, status):
        path = os.path.join(
            self._api, 'updatesubticketstatus', str(subticket_id), '')
        url = urlparse.urljoin(self._url, path)
        self._patch(url, {'status': status, 'current_workflow_step': step})

    def set_subticket_progress(self, data):
        '''Processes the content of messages with subticket_status.

        Not all messages will have this content.'''
        for subticket in data.get('subticket_status', []):
            self._patch_subticket_status(subticket['id'], subticket['step'],
                                         subticket['status'])
            self.process_artifacts(subticket, subticket['id'])

    def _patch_image_status(self, status):
        self._patch_ticket(TicketWorkflowStep.IMAGE_BUILDING, status)

    def set_image_waiting(self):
        self._patch_image_status(TicketWorkflowStepStatus.QUEUED)

    def set_image_inprogress(self):
        self._patch_image_status(TicketWorkflowStepStatus.INPROGRESS)

    def set_image_complete(self, success=True, result=None):
        state = TicketWorkflowStepStatus.COMPLETED
        if not success:
            state = TicketWorkflowStepStatus.FAILED
        self._patch_image_status(state)

    def _patch_testing_status(self, status):
        self._patch_ticket(TicketWorkflowStep.IMAGE_TESTING, status)

    def set_testing_waiting(self):
        self._patch_testing_status(TicketWorkflowStepStatus.QUEUED)

    def set_testing_inprogress(self):
        self._patch_testing_status(TicketWorkflowStepStatus.INPROGRESS)

    def set_testing_complete(self, success=True, result=None):
        state = TicketWorkflowStepStatus.COMPLETED
        if not success:
            state = TicketWorkflowStepStatus.FAILED
        self._patch_testing_status(state)

    def _patch_publishing_status(self, status):
        step = TicketWorkflowStep.PKG_PUBLISHING
        if status == TicketWorkflowStepStatus.COMPLETED:
            logging.info('publishing is final step, marking ticket completed')
            step = TicketWorkflowStep.COMPLETED
        self._patch_ticket(step, status)

    def set_publishing_waiting(self):
        self._patch_publishing_status(TicketWorkflowStepStatus.QUEUED)

    def set_publishing_inprogress(self):
        self._patch_publishing_status(TicketWorkflowStepStatus.INPROGRESS)

    def set_publishing_complete(self, success=True, result=None):
        state = TicketWorkflowStepStatus.COMPLETED
        if not success:
            state = TicketWorkflowStepStatus.FAILED
        self._patch_publishing_status(state)

    def set_ppacreation_waiting(self):
        self._patch_ticket(
            TicketWorkflowStep.PPA_CREATING,
            TicketWorkflowStepStatus.QUEUED)

    def set_ppacreation_inprogress(self):
        self._patch_ticket(
            TicketWorkflowStep.PPA_CREATING,
            TicketWorkflowStepStatus.INPROGRESS)

    def set_ppacreation_complete(self, success=True, result=None):
        step = TicketWorkflowStep.PPA_CREATING
        if not success:
            self._patch_ticket(step, TicketWorkflowStepStatus.FAILED)
            return

        assert result is not None and result.get('ppa_name'), (
            'ppa_name must be given on success.')
        self._patch_ticket(
            step,
            TicketWorkflowStepStatus.COMPLETED,
            build_ppa=result['ppa_name'])

    def add_artifact(self, name, artifact_url, artifact_type='LOGS',
                     subticket_id=None):
        '''Adds an artifact to the ticket or subticket if defined.

        It is the caller's responsibility to not create duplicates.'''
        data = {
            'type': artifact_type,
            'name': name,
            'reference': artifact_url,
        }
        if subticket_id:
            url = urlparse.urljoin(
                self._url, os.path.join(self._api, 'subticketartifact', ''))
            data['subticket'] = os.path.join(
                self._api, 'subticket', str(subticket_id), '')
        else:
            url = urlparse.urljoin(
                self._url, os.path.join(self._api, 'ticketartifact', ''))
            data['ticket'] = os.path.join(
                self._api, 'ticket', self._ticket_id, '')
        try:
            return self._post(url, data)
        except:
            logging.exception('Unable to add %s to ticket-system' % name)

    def get_artifact_list(self, subticket_id=None):
        '''Fetches the list of artifact names for the ticket or subticket.'''
        full_ticket = self.get_full_ticket()
        current_artifacts = None
        if subticket_id:
            for subticket in full_ticket.get('subticket', []):
                if str(subticket['id']) == str(subticket_id):
                    current_artifacts = subticket['artifact']
                    break
            if not current_artifacts:
                raise ValueError('subticket id not found: '
                                 '{}'.format(subticket_id))
        else:
            current_artifacts = full_ticket['artifact']
        return [artifact['name'] for artifact in current_artifacts]

    def process_artifacts(self, data, subticket_id=None):
        '''Process all of the artifacts enclosed in the data dictionary.

        This will register the artifact to the ticket or subticket (if
        defined).  Not all data input may have artifacts.
        Duplicate artifacts names are not registered, this allows a progress
        update routine to continue to send back the same artifact data over
        multiple updates. Attempts to create duplicate artifacts should be
        caught when the artifact is placed in the data store by the creator.'''
        current_artifacts = self.get_artifact_list(subticket_id)
        artifact_list = data.get('artifacts', None)
        if artifact_list:
            for artifact in artifact_list:
                if artifact['name'] not in current_artifacts:
                    self.add_artifact(artifact['name'], artifact['reference'],
                                      artifact['type'], subticket_id)

    def get_full_ticket(self):
        path = os.path.join(self._api, 'fullticket', self._ticket_id, '')
        url = urlparse.urljoin(self._url, path)
        return self._get(url)

    def get_binaries(self):
        url = urlparse.urljoin(
            self._url, os.path.join(self._api, 'binarypackage', ''))
        return self._get(url)['objects']
