#!/usr/bin/env python3
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from oauthlib.common import (
    generate_client_id,
    generate_token,
)
from argparse import ArgumentParser
from configparser import ConfigParser
import string
from textwrap import dedent
import sys

import nfss


def main(args=None):
    parser = ArgumentParser('nfss')
    subparsers = parser.add_subparsers(help='Commands', dest="command")

    subparsers.add_parser(
        'keys-add',
        help="Generate a new set of access keys.",
    )
    parser_keys_del = subparsers.add_parser(
        'keys-del',
        help="Remove an existing set of access keys.",
    )
    parser_keys_del.add_argument(
        'client_key',
        help="The client access key you want to invalidate."
    )
    subparsers.add_parser(
        'keys-list',
        help="List current client access keys."
    )
    subparsers.add_parser(
        'database-migrate',
        help="Migrate the database.",
    )
    subparsers.add_parser(
        'database-clean',
        help="Run database maintenance tasks."
    )
    arguments = parser.parse_args(args or sys.argv[1:])
    if arguments.command is None:
        parser.error("Missing command string.")
        return

    if arguments.command == 'keys-add':
        keys_add()
    elif arguments.command == 'keys-del':
        keys_del(arguments.client_key)
    elif arguments.command == 'keys-list':
        keys_list()
    elif arguments.command == 'database-migrate':
        nfss.db_migrate.main()
    elif arguments.command == 'database-clean':
        database_clean()


def keys_add():
    repeat = True
    name = description = poc = ""
    while repeat:
        name = input("Client name: ")
        poc = input("Point of contact (name & email address): ")
        description = input("Description of client: ")
        print(
            dedent(
                """
                Client Summary:

                Client Name: {}
                Point of Contact: {}
                Description: {}

                """.format(name, poc, description)
            )
        )
        correct = input("Is this information correct (y/n)? ")
        repeat = correct.lower().strip() == "n"

    access_key = generate_token()
    owner_key = generate_token()
    owner_secret = generate_client_id()
    config = ConfigParser()
    config['nfss'] = {
        'client_access_key': access_key,
        'resource_owner_key': owner_key,
        'resource_owner_secret': owner_secret,
        'backend': 'http://nfss.ubuntu.com/api/v1',
    }

    with nfss.database.get_scoped_connection() as database_connection:
        nfss.database.auth_add_client_details(
            database_connection,
            name,
            description,
            poc,
            access_key,
            owner_key,
            owner_secret
        )

    valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
    filename = ''.join(c for c in name if c in valid_chars)
    filename += '.conf'
    while True:
        try:
            with open(filename, 'w') as f:
                config.write(f)
        except IOError:
            print("Cannot write to the default location (%s)" % filename)
            filename = input("Where should the new script be written to? ")
        else:
            print(
                "%r has been written. Use this to provide credentials for"
                "nfss_insert.py" % filename
            )
            break


def keys_del(client_key):
    with nfss.database.get_scoped_connection() as db_connection:
        if not nfss.database.get_auth_client_key_exists(
            db_connection,
            client_key
        ):
            print("Error: the client key '%s' does not exist. Use 'keys-list'")
            print(" to get a list of all the client access keys.")
            sys.exit(1)
        nfss.database.invalidate_client_key(db_connection, client_key)
        if not nfss.database.get_auth_client_key_exists(
            db_connection,
            client_key
        ):
            print("Client key has been successfully invalidated.")
        else:
            print("Client key has NOT been invalidated.")


def keys_list():
    with nfss.database.get_scoped_connection() as db_connection:
        data = nfss.database.get_auth_client_key_list(db_connection)
        if data:
            keylen = max([len(r[0]) for r in data]) + 2
            namelen = max([len(r[1]) for r in data]) + 2
            desclen = max([len(r[2]) for r in data]) + 2
            format_str = "%%%ds %%%ds %%%ds %%s" % (keylen, namelen, desclen)
            print(format_str % ("Key:", "Name:", "Description:", "Contact:"))
            for key, name, desc, poc in data:
                print(format_str % (key, name, desc, poc))
        print("Total: %d keys" % len(data))


def database_clean():
    with nfss.database.get_scoped_connection() as db_connection:
        nfss.database.clean_old_nonces(db_connection)


if __name__ == '__main__':
    main()
