# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""V1 of the REST API.

This file contains the views for the v1 API, as well as the URL routing table
for the v1 API.

"""
from datetime import (
    datetime,
    timezone,
)
import json
from pyramid.httpexceptions import HTTPBadRequest

from nfss import database
from nfss.auth import oauth_protected


def configure_routes(config):
    """Add url routes to 'config' object."""
    config.add_route('v1.root', '/')
    config.add_route('v1.project', '/{project}')
    config.add_route('v1.test', '/{project}/{test}')


def configure_views(config):
    """Add views to 'config' object."""
    config.add_view(root, route_name='v1.root', renderer='json',
                    request_method='GET')
    config.add_view(project, route_name='v1.project', renderer='json',
                    request_method='GET')
    config.add_view(test, route_name='v1.test', renderer='json',
                    request_method='GET')
    config.add_view(test_add, route_name='v1.test', renderer='json',
                    request_method='POST')


def root(request):
    """Return summary data about the projects and tests."""
    data = database.get_details_for_all_projects(request.database())
    return dict(
        projects={
            r[0]: {
                'tests': json.loads(r[1]),
                'last_updated': r[2].isoformat(),
                'path': request.route_path('v1.project', project=r[0])
            } for r in data},
    )


def project(request):
    """Return detailed information about this project's tests."""
    project_name = request.matchdict['project']
    project_data = database.get_details_for_project(
        request.database(),
        project_name
    )

    for test in project_data:
        project_data[test]['path'] = request.route_path(
            'v1.test',
            project=project_name,
            test=test
        )
    return dict(project_name=project_name, tests=project_data)


def test(request):
    """Return data points for this particular test.

    Check the Range HTTP header, or return the last 30 days worth of data if
    it's not present.

    """
    # import pudb; pudb.set_trace()
    project_name = request.matchdict['project']
    test_name = request.matchdict['test']
    start = try_parse_date(request.params.get('start_date', None))
    end = try_parse_date(request.params.get('end_date', None))
    data, f, l, df, dl = database.get_details_for_test(
        request.database(),
        project_name,
        test_name,
        start,
        end,
    )
    return dict(
        project_name=project_name,
        test_name=test_name,
        first_data=f.isoformat(),
        last_data=l.isoformat(),
        data_range_first=df.isoformat(),
        data_range_last=dl.isoformat(),
        data=json.loads(data) if data else [],
    )


def try_parse_date(date_param):
    if date_param is not None:
        try:
            return datetime.fromtimestamp(float(date_param), timezone.utc)
        except:
            return None
    return None


def get_timestamp_from_request_params(parameters):
    timestamp = parameters.get('timestamp', None)
    if timestamp:
        try:
            return datetime.fromtimestamp(timestamp, timezone.utc)
        except:
            raise HTTPBadRequest(
                'Invalid time zone format; '
                'Should be epoch seconds in UTC, microseconds optional')
    return datetime.now(timezone.utc)


def raise_if_timestamp_invalid(timestamp):
    time_delta = (timestamp - datetime.now(tz=timezone.utc))
    if time_delta.days > 0 or (time_delta.days == 0 and
                               time_delta.seconds > 1800):
        raise HTTPBadRequest(
            'Timestamp %s is more than 30 minutes in the future; '
            'Please check your system clock' % timestamp)
    if timestamp.year < 2004:
        raise HTTPBadRequest(
            'Timestamp %s is older than 2004; '
            'Please make sure this is a valid Ubuntu benchmark' % timestamp)


@oauth_protected()
def test_add(request, oauth_request):
    """Add data points to this test.

    Return the created id of the data point added to the DB if it was
    successful, or an error if it wasn't.

    """
    try:
        project_name = request.matchdict['project']
        test_name = request.matchdict['test']
        data = request.params['data']
    except KeyError as e:
        raise HTTPBadRequest("Missing data in POST request: %s" % e)
    else:
        timestamp = get_timestamp_from_request_params(request.params)
        raise_if_timestamp_invalid(timestamp)
        created_id = database.insert_test_data(
            request.database(),
            project_name,
            test_name,
            data,
            oauth_request.client_key,
            timestamp,
        )
        return dict(created_id=created_id)
