#!/usr/bin/env python3
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from oauthlib.oauth1 import RequestValidator, WebApplicationServer
from pyramid.httpexceptions import HTTPForbidden
import functools

from nfss import database


class NFSSRequestValidator(RequestValidator):

    def __init__(self, database_connection):
        super().__init__()
        self._database = database_connection

    @property
    def enforce_ssl(self):
        return False

    @property
    def dummy_client(self):
        return "dummy-client-access-token"

    def validate_client_key(self, client_key, request):
        return database.get_auth_client_key_exists(self._database, client_key)

    def validate_access_token(self, client_key, resource_owner_key, request):
        return database.get_auth_resource_owner_key_for_client_key(
            self._database,
            client_key
        )

    def validate_timestamp_and_nonce(self, client_key, timestamp, nonce,
                                     request, request_token=None,
                                     access_token=None):
        token = request_token or access_token
        if not database.get_auth_nonce_already_used(
            self._database,
            client_key,
            timestamp,
            nonce,
            token
        ):
            database.store_nonce(
                self._database,
                client_key,
                timestamp,
                nonce,
                token
            )
            return True
        return False

    def validate_realms(self, client_key, token, request, uri=None,
                        realms=None):
        # Realms are used so we can split the app into several sections with
        # different authentications in each. We're not using realms, so we
        # just return True straight away:
        return True

    def get_client_secret(self, client_key, request):
        # We don't use the client secret, so this will always return the
        # empty string.
        return ''

    def get_access_token_secret(self, client_key, owner_key, request):
        return database.get_auth_resource_owner_secret_for_client_key(
            self._database,
            client_key
        )


def oauth_protected(realms=None):
    def wrapper(f):
        @functools.wraps(f)
        def verify_oauth(request, *args, **kwargs):
            provider = WebApplicationServer(
                NFSSRequestValidator(
                    request.database()
                )
            )
            v, r = provider.validate_protected_resource_request(
                request.url,
                http_method=request.method,
                body=request.body,
                headers=request.headers,
                realms=realms or []
            )
            if v:
                return f(request, r, *args, **kwargs)
            else:
                raise HTTPForbidden()
        return verify_oauth
    return wrapper
