# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Functions that access the database."""

import contextlib
import os.path
import functools
import json
import logging

logger = logging.getLogger(__name__)

import psycopg2.pool


def get_settings_path():
    return os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '..',
            '..',
            '..',
            '..',
            'pgsql.json'
        )
    )


@functools.lru_cache()
def get_db_settings(settings_path):
    """Return a dictionary containing the database settings, as provided by
    juju.

    If the settings file does not exist an IOError is returned.
    If the settings file exists, but is not valid JSON, a ValueError is
        returned.

    """

    with open(settings_path, 'r') as settings_file:
        return json.load(settings_file)


def translate_juju_db_settings(db_settings):
    """Given a dictionary of juju database settings, return a dictionary of
    psycopg2 database connection settings.

    """
    return dict(
        database=db_settings['database'],
        user=db_settings['schema_user'],
        password=db_settings['schema_password'],
        host=db_settings['host'],
        port=db_settings['port'],
    )


def get_connection_for_request(request=None):
    """Get a database connection.

    This is bound to every request object we recieve as 'request.database'.

    Note: If this argument is not called with the 'request' parameter then the
        caller *must* call '.close()' on the returned connection object.

    """
    juju_db_settings = get_db_settings(get_settings_path())
    db_kwargs = translate_juju_db_settings(juju_db_settings)
    connection = psycopg2.connect(**db_kwargs)
    if request is not None:
        def cleanup(request):
            connection.close()
        request.add_finished_callback(cleanup)
    return connection


@contextlib.contextmanager
def get_scoped_connection():
    """Get a database connection.

    This context manager will return a database connection, and will
    automatically close the connection when the context manager exits.

    Use like so::

        with nfss.database.get_scoped_connection() as db:
            cursor = db.cursor()
            cursor.execute('SELECT 1+1;')
            db.commit()

    """
    connection = get_connection_for_request()
    try:
        yield connection
    finally:
        connection.close()


def autocommit(fn):
    """A simple utility decorator that wraps a method that gets content from
    the database. Using this will ensure that:

    1) If the wrapped function returns without raising an exception, the
       connection's 'commit()' method is called.
    2) If the wrapped function raises an exception, the connection's
       'rollback()' method will be called.

    This is important since by default, any operation (including simple
    SELECT queries) starts a new transaction. Without calling 'commit' or
    'rollback' these transactions gradually use up more and more resources.

    """
    def worker(connection, *args, **kwargs):
        try:
            result = fn(connection, *args, **kwargs)
            connection.commit()
            return result
        except Exception:
            connection.rollback()
            raise
    return worker


@autocommit
def get_details_for_all_projects(connection):
    """Get high level details for all projects in the system.

    The return value will be a list of lists. Each inner list will contain:

        [project_name, test_array, max_date]

    * project_name is obvious - it's the name of the project.
    * test_array is a json formatted array of test names that have had data
      entered for this project.
    * max_date is the date of the last data point for this project in any test.

    """
    cursor = connection.cursor()
    cursor.execute('''
        SELECT
            project.name,
            json_agg(distinct test.name),
            max(data.date_entered)
        FROM project
        INNER JOIN test ON project.id = test.project_id
        INNER JOIN data ON test.id = data.test_id
        GROUP BY project.name;
    ''')
    return cursor.fetchall()


@autocommit
def get_details_for_project(connection, project_name):
    """Get high-level details for a single project.

    The return value will be a json string containing the test name, first and
    last data point dates, and a count of data points, in an array, for each
    test.

    """
    cursor = connection.cursor()
    cursor.execute(
        '''
        SELECT
            test.name as test_name,
            min(data.date_entered) as first_data,
            max(data.date_entered) as last_data,
            count(data.id) as data_points
        FROM test
        INNER JOIN data on data.test_id = test.id
        INNER JOIN project on test.project_id = project.id
        WHERE project.name = %s
        GROUP BY test.name;
        ''',
        (project_name,),
    )
    data = {}
    for row in cursor:
        data[row[0]] = dict(
            first_data=row[1].isoformat(),
            last_data=row[2].isoformat(),
            data_points=row[3]
        )
    return data


@autocommit
def get_details_for_test(connection, project_name, test_name, start_time=None,
                         end_time=None):
    """Get data points for a particular project / test combo.

    If start_time is specified, end time must also be specified. If either are
    specified, they must be datetime objects. If neither are specified, then
    the data range will be 30 days from the last data point.

    This functions returns a tuple that contins, in order:

     * A json string containing the test data points.
     * A datetime object representing the point at which the returned data
       range starts.
     * A datetime object representing the point at which the returned data
       range ends.
     * A datetime object representing the first data point in the test data.
     * A datetime object represending the last data point in the test data.

    """
    cursor = connection.cursor()

    if all((start_time, end_time)):
        cursor.execute(
            '''
            SELECT
                min(data.date_entered),
                max(data.date_entered)
            FROM data
            INNER JOIN test on data.test_id = test.id
            INNER JOIN project on test.project_id = project.id
            WHERE project.name = %s
            AND test.name = %s;
            ''',
            (project_name, test_name)
        )
        first, last = cursor.fetchone()
        cursor.execute(
            '''
            SELECT json_agg(d)
            FROM (
                SELECT
                    data.date_entered AS date_entered,
                    data.id AS id,
                    data.data AS data
                FROM data
                INNER JOIN test on data.test_id = test.id
                INNER JOIN project on test.project_id = project.id
                WHERE project.name = %s
                AND test.name = %s
                AND data.date_entered >= %s
                AND data.date_entered <= %s
            ) as d;
            ''',
            (project_name, test_name, start_time, end_time)
        )
        return (
            cursor.fetchone()[0],
            first,
            last,
            start_time,
            end_time,
        )
    else:
        # user didn't specify dates.
        cursor.execute(
            '''
            SELECT
                min(data.date_entered),
                max(data.date_entered) - interval '30 days',
                max(data.date_entered)
            FROM data
            INNER JOIN test on data.test_id = test.id
            INNER JOIN project on test.project_id = project.id
            WHERE project.name = %s
            AND test.name = %s;
            ''',
            (project_name, test_name)
        )
        first, data_first, last = cursor.fetchone()
        cursor.execute(
            '''
            SELECT json_agg(d)
            FROM (
                SELECT
                    data.date_entered AS date_entered,
                    data.id AS id,
                    data.data AS data
                FROM data
                INNER JOIN test on data.test_id = test.id
                INNER JOIN project on test.project_id = project.id
                WHERE project.name = %s
                AND test.name = %s
                AND data.date_entered >= %s
                AND data.date_entered <= %s
            ) as d;
            ''',
            (project_name, test_name, data_first, last)
        )
        return (
            cursor.fetchone()[0],
            first,
            last,
            data_first,
            last,
        )


@autocommit
def insert_test_data(connection, project_name, test_name, data, client_key,
                     timestamp):
    project_id = _maybe_create_project(connection, project_name)
    test_id = _maybe_create_test(connection, project_id, test_name)
    cursor = connection.cursor()
    cursor.execute(
        "INSERT INTO data (date_entered, data, test_id, client_id) "
        "VALUES (TIMESTAMP WITH TIME ZONE %s, %s, %s, "
        "(SELECT id FROM client where access_key = %s)) RETURNING id;",
        (timestamp, data, test_id, client_key)
    )
    return cursor.fetchone()[0]


def _maybe_create_project(connection, project_name):
    """If the project exists, return it's id. If it doesn't, create it and
    return it's id.

    """
    cursor = connection.cursor()
    cursor.execute('SELECT id FROM project WHERE name = %s;', (project_name,))
    if cursor.rowcount > 0:
        return cursor.fetchone()[0]
    cursor.execute(
        'INSERT INTO project (name) VALUES (%s) RETURNING id;',
        (project_name,)
    )
    return cursor.fetchone()[0]


def _maybe_create_test(connection, project_id, test_name):
    cursor = connection.cursor()
    cursor.execute(
        'SELECT id FROM test WHERE name = %s AND project_id = %s;',
        (test_name, project_id)
    )
    if cursor.rowcount > 0:
        return cursor.fetchone()[0]
    cursor.execute(
        'INSERT INTO test (name, project_id) VALUES (%s, %s) RETURNING id;',
        (test_name, project_id)
    )
    return cursor.fetchone()[0]


@autocommit
def get_auth_client_key_exists(connection, client_key):
    cursor = connection.cursor()
    cursor.execute(
        'SELECT id FROM client WHERE access_key = %s AND valid = true;',
        (client_key,)
    )
    return cursor.rowcount != 0


@autocommit
def get_auth_client_key_list(connection):
    """Get a list of valid client keys."""
    cursor = connection.cursor()
    cursor.execute(
        'SELECT access_key, name, description, point_of_contact '
        'FROM client WHERE valid = true;'
    )
    return cursor.fetchall()


@autocommit
def invalidate_client_key(connection, client_key):
    cursor = connection.cursor()
    cursor.execute(
        'UPDATE client SET valid = false WHERE access_key = %s;',
        (client_key,)
    )


@autocommit
def get_auth_resource_owner_key_for_client_key(connection, client_key):
    cursor = connection.cursor()
    cursor.execute(
        'SELECT resource_owner_key FROM client WHERE access_key = %s;',
        (client_key,)
    )
    return cursor.fetchone()[0]


@autocommit
def auth_add_client_details(connection, name, description, poc, access_key,
                            owner_key, owner_secret):
    cursor = connection.cursor()
    cursor.execute(
        '''
        INSERT INTO client (access_key, name, description, point_of_contact,
            resource_owner_key, resource_owner_secret)
        VALUES (%s, %s, %s, %s, %s, %s);
        ''',
        (access_key, name, description, poc, owner_key, owner_secret)
    )


@autocommit
def get_auth_resource_owner_secret_for_client_key(connection, client_key):
    cursor = connection.cursor()
    cursor.execute(
        'SELECT resource_owner_secret FROM client where access_key = %s;',
        (client_key,)
    )
    return cursor.fetchone()[0]


@autocommit
def get_auth_nonce_already_used(conn, client_key, timestamp, nonce, owner_key):
    cursor = conn.cursor()
    cursor.execute(
        '''
        SELECT * FROM nonce
        WHERE access_key = %s
        AND timestamp = to_timestamp(%s)
        AND nonce = %s
        AND resource_owner_key = %s;
        ''',
        (client_key, timestamp, nonce, owner_key)
    )
    return cursor.rowcount != 0


@autocommit
def store_nonce(connection, client_key, timestamp, nonce, owner_key):
    cursor = connection.cursor()
    cursor.execute(
        '''
        INSERT INTO nonce (access_key, timestamp, nonce, resource_owner_key)
        VALUES (%s, to_timestamp(%s), %s, %s);
        ''',
        (client_key, timestamp, nonce, owner_key)
    )
    return


@autocommit
def clean_old_nonces(connection):
    cursor = connection.cursor()
    cursor.execute(
        '''
        DELETE FROM nonce
        WHERE timestamp <
        CURRENT_TIMESTAMP AT TIME ZONE 'UTC' - CAST('1 day' AS interval);
        '''
    )
