#!/usr/bin/env python3
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Migration script for the Non Functional Stats Service.

This script is called when the service is installed, and any time the service
is upgraded. It follows a few simple steps to manage the database version:

1) Connecxt to the database, and read the version from the db_version table. If
    it doesn't exist, assume the version is -1.

2) Scan the db_patches directory for files that end in '.sql'. The files must
    be named 'NNN-{name}.sql'. The NNN is the numeric order of the patches,
    with leading '0's.

3) Apply patches in order. Each patch is applied in a DB transaction, and after
    each patch we incriment the database revision number in the db_version
    table. As a special case, if we just upgraded to version '0', we create
    the table.

"""

import glob
import os.path
import psycopg2
import re
import sys

from nfss.database import get_scoped_connection


def main():
    with get_scoped_connection() as connection:
        current_version = get_current_database_version(connection)
        maximum_version = get_maximum_version()
        print("Current database version is: %d" % current_version)
        if current_version != maximum_version:
            print(
                "Database schema can be updated to version: %d" %
                maximum_version
            )
            for new_version in range(current_version + 1, maximum_version + 1):
                upgrade_database_to_version(connection, new_version)


def get_current_database_version(connection):
    cursor = connection.cursor()
    try:
        cursor.execute('SELECT version FROM db_version;')
    except psycopg2.ProgrammingError:
        # table doesn't exist, we have a blank database, and need to start from
        # scratch.
        connection.rollback()
        return -1

    data = cursor.fetchone()
    if data is None:
        # table exists, but contains no data - not sure what this means
        # as we should never get here. Let's assume version -1.
        return -1
    else:
        return int(data[0])


def report_error(message):
    # TODO: replace with proper logging!
    sys.stderr.write(message + '\n')


def get_maximum_version():
    """Return the maximum version we can patch the database to."""
    patches = get_all_patch_paths()
    return int(patches[-1][:3])


def get_all_patch_paths():
    patches_path = get_database_patch_dir()
    return sorted(
        filter(
            lambda f: re.match(r'\d{3}-.*.sql', f),
            os.listdir(patches_path)
        )
    )


def get_patch_file_path_for_version(version_number):
    """Given a database version number, get the path to a patch that upgrades
    the schema to that version.

    raise a RuntimeError if more than one file matches.
    """
    matches = glob.glob(
        os.path.join(
            get_database_patch_dir(),
            "%03d-*.sql" % version_number
        )
    )
    if len(matches) > 1:
        raise RuntimeError(
            "More than one file matched for version %d: %r" % (
                version_number, matches)
        )
    if not matches:
        raise RuntimeError(
            "No patch file found for version %d" % version_number
        )
    return matches[0]


def get_database_patch_dir():
    return os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '..',
            'db_patches'
        )
    )


def upgrade_database_to_version(connection, new_version):
    """Do the work to upgrade the database to version 'new_version'.

    This function starts a transaction on 'connection', and will either
    commit or rollback the transaction before the function ends.

    On success the function returns None, on failure the function raises an
    exception.

    Note: Only call this function if the database is at 'new_version' - 1.

    """
    patch_file = get_patch_file_path_for_version(new_version)
    print("Upgrading to version %s ..." % new_version)
    cursor = connection.cursor()
    try:
        with open(patch_file, 'r') as patch:
            cursor.execute(patch.read())
            if new_version == 0:
                cursor.execute('CREATE TABLE db_version (version integer);')
                cursor.execute('INSERT INTO db_version (version) VALUES (0);')
            else:
                cursor.execute('UPDATE db_version SET version=%d'
                               % new_version)
    except Exception as err:
        print("Error: %s" % err)
        connection.rollback()
        raise
    print("Done!")
    connection.commit()
