#!/usr/bin/env python3

# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import datetime
import json
import logging
import urllib.request
import sys


logger = logging.getLogger(__name__)


def _main():
    arguments = _parse_arguments()
    logger.info(
        'Checking if the last data point was submitted less than {} hours ago.'
        '\nserver: {}, project: {}, test: {}'.format(
            arguments.hours, arguments.server, arguments.project,
            arguments.test))
    project_data = _get_project_data(arguments.server, arguments.project)
    if project_data:
        last_data_datetime = _get_last_data_datetime(
            project_data, arguments.test)
        _check_last_data_delta(last_data_datetime, arguments.hours)
    else:
        # FIXME The server sends empty tests data when the project
        # doesn't exist. See http://pad.lv/1391033 --elopio - 2014-11-09
        sys.exit('Error: The project {} does not exist.'.format(
            arguments.project))


def _get_project_data(server, project):
    url = server.rstrip('/') + '/api/v1/' + project
    try:
        response = urllib.request.urlopen(url)
    except ValueError:
        sys.exit('Error: wrong URL {}.'.format(server))
    try:
        read_response = response.read()
        data = json.loads(read_response.decode('utf-8'))
    except:
        sys.exit(
            'Error: Failed to decode the message returned by the server: '
            '{}'.format(read_response))
    return data['tests']


def _get_last_data_datetime(project_data, test):
    test_data = project_data.get(test)
    if test_data:
        last_data = project_data[test]['last_data']
        return datetime.datetime.strptime(
            last_data.replace(':', ''), '%Y-%m-%dT%H%M%S.%f%z')
    else:
        sys.exit('Error: The test {} does not exist.'.format(test))


def _check_last_data_delta(last_data_datetime, maximum_hours):
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    delta = now - last_data_datetime
    maximum_seconds = maximum_hours * 60 * 60
    last_update_message = 'The last update was on {}.'.format(
        last_data_datetime)
    if delta.total_seconds() <= maximum_seconds:
        sys.stdout.write(
            'OK: {}\nLess than {} hours ago.\n'.format(
                last_update_message, maximum_hours))
        sys.exit(0)
    else:
        error = (
            'Error: {}\nMore than {} hours ago.'.format(
                last_update_message, maximum_hours))
        sys.exit(error)


def _parse_arguments():
    parser = argparse.ArgumentParser(
        description='Check the last data point submission to an NFSS project.')
    parser.add_argument('server', help='The URL of the NFSS server.')
    parser.add_argument('project', help='The name of the project.')
    parser.add_argument('test', help='The name of the test.')
    parser.add_argument(
        'hours', type=int,
        help='The maximum number of hours since the last data point.')
    return parser.parse_args()


if __name__ == '__main__':
    _main()
