# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import http.server
import json
import logging
import urllib.parse


logger = logging.getLogger(__name__)


_database = {}


class FakeNFSSServer(http.server.HTTPServer):

    def __init__(self, server_address):
        super().__init__(server_address, _FakeNFSSRequestHandler)
        global _database
        _database = {}

    def add_fake_data(self, project, data):
        global _database
        _database[project] = data

    def add_fake_test_data(self, project, test, last_data):
        data = {
            'project_name': project,
            'tests': {
                test: {
                    'last_data': last_data
                }
            }
        }
        self.add_fake_data(project, data)


class _FakeNFSSRequestHandler(http.server.BaseHTTPRequestHandler):

    _API_PATH = '/api/v1/'
    _UNEXISTING_PROJECT_DATA = {'project_name': 'dummy', 'tests': {}}

    def log_message(self, format, *args):
        logger.debug(format, args)

    def do_GET(self):
        parsed_path = urllib.parse.urlparse(self.path)
        if parsed_path.path.startswith(self._API_PATH):
            project = parsed_path.path[len(self._API_PATH):]
            self._send_project_data(project)
        else:
            logger.error('Not implemented path in fake server: {}'.format(
                self.path))
            raise NotImplementedError(self.path)

    def _send_project_data(self, project):
        global _database
        project_data = _database.get(project)
        if project_data:
            if isinstance(project_data, bytes):
                self._send_reply(200, project_data)
            else:
                self._send_json_reply(200, json.dumps(project_data))
        else:
            # FIXME The real server sends empty tests data when the project
            # doesn't exist. See http://pad.lv/1391033 --elopio - 2014-11-09
            unexisting_data = copy.deepcopy(self._UNEXISTING_PROJECT_DATA)
            unexisting_data['project_name'] = project
            self._send_json_reply(200, json.dumps(unexisting_data))

    def _send_reply(self, code, reply):
        self.send_response(code)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(reply)

    def _send_json_reply(self, code, reply_json):
        self._send_reply(code, reply_json.encode())
