# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import subprocess

import testscenarios
import testtools

from nfss.scripts import nfss_check
from nfss.testing import fixture_setup


logger = logging.getLogger(__name__)


class NFSSCheckBaseTestCase(testtools.TestCase):

    """Base test case for NFSS Check script tests."""

    def run_nfss_check(self, *args):
        """Run the NFSS check script.

        :args: The list of arguments to pass to the script.
        :return: A tuple with the return code and the output of the script.

        """
        script = nfss_check.__file__
        command = [script]
        if args:
            command.extend(args)
        logger.info('Running the command: {}'.format(command))
        try:
            output = subprocess.check_output(
                command, stderr=subprocess.STDOUT, universal_newlines=True)
            return_code = 0
        except subprocess.CalledProcessError as error:
            output = error.output
            return_code = error.returncode
        return return_code, output


class NFSSCheckArgumentsTestCase(
        testscenarios.TestWithScenarios, NFSSCheckBaseTestCase):

    USAGE = 'usage: nfss_check.py [-h] server project test hours'
    HELP = USAGE + """

Check the last data point submission to an NFSS project.

positional arguments:
  server      The URL of the NFSS server.
  project     The name of the project.
  test        The name of the test.
  hours       The maximum number of hours since the last data point.

optional arguments:
  -h, --help  show this help message and exit
"""
    ERROR = '\nnfss_check.py: error: '
    MISSING_ARGUMENTS_FORMAT = (
        ERROR + 'the following arguments are required: {}\n')
    UNRECOGNIZED_ARGUMENTS_FORMAT = (
        ERROR + 'unrecognized arguments: {}\n')

    scenarios = [
        ('Calling with --help must succeed and print help', {
            'arguments': ('--help',),
            'expected_return_code': 0,
            'expected_output': HELP
        }),
        ('Calling with -h must succeed and print help', {
            'arguments': ('-h',),
            'expected_return_code': 0,
            'expected_output': HELP
        }),
        ('Calling without arguments must fail and print usage and error', {
            'arguments': (),
            'expected_return_code': 2,
            'expected_output': (
                USAGE + MISSING_ARGUMENTS_FORMAT.format(
                    'server, project, test, hours'))
        }),
        ('Calling with less arguments must fail and print usage and error', {
            'arguments': ('dummy',),
            'expected_return_code': 2,
            'expected_output': (
                USAGE + MISSING_ARGUMENTS_FORMAT.format(
                    'project, test, hours'))
        }),
        ('Calling with more arguments must fail and print usage and error', {
            'arguments': (
                'dummy1', 'dummy2', 'dummy3', '0', 'test_unrecognized_arg'),
            'expected_return_code': 2,
            'expected_output': (
                USAGE + UNRECOGNIZED_ARGUMENTS_FORMAT.format(
                    'test_unrecognized_arg'))
        }),
        ('Calling with non-int hours must fail and print usage and error', {
            'arguments': (
                'dummy1', 'dummy2', 'dummy3',  'non-int'),
            'expected_return_code': 2,
            'expected_output': (
                USAGE + ERROR +
                "argument hours: invalid int value: 'non-int'\n")
        }),
        ('Calling with malformed URL must fail and print error', {
            'arguments': (
                'notaurl', 'dummy2', 'dummy3',  '0'),
            'expected_return_code': 1,
            'expected_output': 'Error: wrong URL notaurl.\n'
        }),
    ]

    def test_nfss_check_arguments(self):
        return_code, output = self.run_nfss_check(*self.arguments)
        self.assertEqual(
            self.expected_return_code, return_code, 'Wrong return code')
        self.assertEqual(self.expected_output, output, 'Wrong output')


class NFSSCheckLastDateTestCase(NFSSCheckBaseTestCase):

    def setUp(self):
        super().setUp()
        # XXX We should use a test local server with in-memory storage instead
        # of a fake. See http://pad.lv/1391039 --elopio - 2014-10-09
        self.fake_server_fixture = fixture_setup.FakeNFSSServerRunning()
        self.useFixture(self.fake_server_fixture)
        self.url = self.fake_server_fixture.url

    def test_check_unexisting_project_must_fail_and_print_error(self):
        return_code, output = self.run_nfss_check(
            self.url, 'unexisting_project', 'dummy', '0')
        self.assertEqual(1, return_code, 'Wrong return code')
        self.assertEqual(
            'Error: The project unexisting_project does not exist.\n', output)

    def test_check_unexisting_test_must_fail_and_print_error(self):
        self.fake_server_fixture.server.add_fake_test_data(
            'test_project', 'test_test', 'test_last_data')

        return_code, output = self.run_nfss_check(
            self.url, 'test_project', 'unexisting_test', '0')
        self.assertEqual(1, return_code, 'Wrong return code')
        self.assertEqual(
            'Error: The test unexisting_test does not exist.\n', output)

    def test_check_old_last_data_must_fail_and_print_error(self):
        self.fake_server_fixture.server.add_fake_test_data(
            'test_project', 'test_test', '2014-01-01T00:00:01.000000+00:00')

        return_code, output = self.run_nfss_check(
            self.url, 'test_project', 'test_test', '24')
        self.assertEqual(1, return_code, 'Wrong return code')
        self.assertEqual(
            'Error: The last update was on 2014-01-01 00:00:01+00:00.\n'
            'More than 24 hours ago.\n',
            output)

    def test_check_recent_data_must_succeed_and_print_info(self):
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        self.fake_server_fixture.server.add_fake_test_data(
            'test_project', 'test_test', now.isoformat())

        return_code, output = self.run_nfss_check(
            self.url, 'test_project', 'test_test', '24')
        self.assertEqual(0, return_code, 'Wrong return code')
        self.assertEqual(
            'OK: The last update was on {}.\n'
            'Less than 24 hours ago.\n'.format(now),
            output)

    def test_check_does_not_return_json_must_fail_and_print_error(self):
        self.fake_server_fixture.server.add_fake_data(
            'test_project', b'nojson')

        return_code, output = self.run_nfss_check(
            self.url, 'test_project', 'dummy', '0')
        self.assertEqual(1, return_code, 'Wrong return code')
        self.assertEqual(
            "Error: Failed to decode the message returned by the server: "
            "b'nojson'\n",
            output)
