# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest

import datetime

from pyramid.httpexceptions import HTTPBadRequest

try:
    import mock
    have_mock = True
except ImportError:
    have_mock = False

from nfss import api


@unittest.skipUnless(have_mock, "No Mock available")
class TimestampTests(unittest.TestCase):

    def test_get_timestamp_from_request_params_gets_parameter(self):
        parameters = mock.MagicMock()
        api.v1.get_timestamp_from_request_params(parameters)
        parameters.get.assert_called_once_with('timestamp', None)

    def test_get_timestamp_from_request_params_returns_now_for_none(self):
        parameters = mock.MagicMock()
        parameters.get.return_value = None
        with mock.patch('nfss.api.v1.datetime') as dt:
            self.assertEqual(
                api.v1.get_timestamp_from_request_params(parameters),
                dt.now.return_value)

    def test_get_timestamp_from_request_params_calls_datetime(self):
        parameters = mock.MagicMock()
        with mock.patch('nfss.api.v1.datetime') as dt:
            api.v1.get_timestamp_from_request_params(parameters)
            dt.fromtimestamp.assert_called_once_with(
                parameters.get.return_value, datetime.timezone.utc)

    def test_get_timestamp_from_request_params_returns_datetime_object(self):
        parameters = mock.MagicMock()
        with mock.patch('nfss.api.v1.datetime') as dt:
            self.assertEqual(
                api.v1.get_timestamp_from_request_params(parameters),
                dt.fromtimestamp.return_value)

    def test_get_timestamp_from_raises_if_datetime_raises(self):
        parameters = mock.MagicMock()
        with mock.patch('nfss.api.v1.datetime') as dt:
            dt.fromtimestamp.side_effect = Exception
            self.assertRaises(
                HTTPBadRequest,
                lambda: api.v1.get_timestamp_from_request_params(parameters))

    def test_get_timestamp_from_raises_if_passed_string(self):
        parameters = mock.MagicMock()
        parameters.get.return_value = '1410289926'
        self.assertRaises(
            HTTPBadRequest,
            lambda: api.v1.get_timestamp_from_request_params(parameters))

    def test_raise_if_timestamp_invalid_raises_on_near_future(self):
        timestamp = datetime.datetime.now(datetime.timezone.utc)
        timestamp += datetime.timedelta(minutes=31)
        self.assertRaises(
            HTTPBadRequest,
            lambda: api.v1.raise_if_timestamp_invalid(timestamp))

    def test_raise_if_timestamp_invalid_raises_on_far_future(self):
        timestamp = datetime.datetime.now(datetime.timezone.utc)
        timestamp += datetime.timedelta(days=1)
        self.assertRaises(
            HTTPBadRequest,
            lambda: api.v1.raise_if_timestamp_invalid(timestamp))

    def test_raise_if_timestamp_invalid_raises_on_far_past(self):
        timestamp = datetime.datetime(year=2003, month=1, day=1,
                                      tzinfo=datetime.timezone.utc)
        self.assertRaises(
            HTTPBadRequest,
            lambda: api.v1.raise_if_timestamp_invalid(timestamp))
