# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import tempfile
import unittest

try:
    from mock import patch
    have_mock = True
except ImportError:
    have_mock = False

from nfss import db_migrate as dbm
from nfss import database


class DBMigrationTests(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.addCleanup(database.get_db_settings.cache_clear)

    def test_get_settings_returns_dict_on_no_file(self):
        with self.assertRaises(IOError):
            database.get_db_settings('/this/path/does/not/exist')

    def test_get_settings_returns_dict_on_bad_json(self):
        with tempfile.NamedTemporaryFile(mode='w') as f:
            f.write('sdfsdf')
            f.flush()

            with self.assertRaises(ValueError):
                    database.get_db_settings(f.name)

    def test_get_settings_works_on_example_json(self):
        with tempfile.NamedTemporaryFile(mode='w') as f:
            f.write(
                '{"database": "ci-airline-nfss-restish", '
                '"allowed-units": "ci-airline-nfss-restish/0", '
                '"state": "standalone", '
                '"schema_user": "db_2_ci_airline_nfss_restish_schema", '
                '"schema_password": "secret", '
                '"private-address": "10.0.3.236", '
                '"host": "10.0.3.236", '
                '"user": "db_2_ci_airline_nfss_restish", '
                '"password": "secret", '
                '"port": "5432"}'
            )
            f.flush()

            expected = {
                "database": "ci-airline-nfss-restish",
                "allowed-units": "ci-airline-nfss-restish/0",
                "state": "standalone",
                "schema_user": "db_2_ci_airline_nfss_restish_schema",
                "schema_password": "secret",
                "private-address": "10.0.3.236",
                "host": "10.0.3.236",
                "user": "db_2_ci_airline_nfss_restish",
                "password": "secret",
                "port": "5432"
            }
            self.assertEqual(
                expected,
                database.get_db_settings(f.name)
            )

    @unittest.skipUnless(have_mock, "No Mock available")
    def test_get_maximum_version_works_with_single_patch(self):
        with patch.object(dbm, 'get_all_patch_paths') as p:
            p.return_value = ['000-foo.sql']

            self.assertEqual(0, dbm.get_maximum_version())

    @unittest.skipUnless(have_mock, "No Mock available")
    def test_get_maximum_version_works_with_many_patches(self):
        with patch.object(dbm, 'get_all_patch_paths') as p:
            p.return_value = ['000-foo.sql', '001-bar.sql', '002-baz.sql']

            self.assertEqual(2, dbm.get_maximum_version())
