#!/usr/bin/env python3

from configparser import ConfigParser
from datetime import (
    datetime,
    timezone,
)
import json
import sys
from requests_oauthlib import OAuth1Session


def raise_if_timestamp_invalid(timestamp):
    time_delta = (timestamp - datetime.now(tz=timezone.utc))
    if time_delta.days > 0 or (time_delta.days == 0 and
                               time_delta.seconds > 1800):
        raise ValueError(
            'Timestamp %s is more than 30 minutes in the future; '
            'Please check your system clock' % timestamp)
    if timestamp.year < 2004:
        raise ValueError(
            'Timestamp %s is older than 2004; '
            'Please make sure this is a valid Ubuntu benchmark' % timestamp)


def main():
    if len(sys.argv) not in (4, 5):
        sys.stderr.write("Usage: %s <config file> <projectname> <testname> "
                         "<timestamp> (optional)\n" % sys.argv[0])
        sys.stderr.write(
            "Pipe json test data to this script to insert it into the database"
            ".\n")
        sys.exit(1)

    config_file = sys.argv[1]
    project = sys.argv[2]
    test = sys.argv[3]
    if len(sys.argv) == 5:
        timestamp = sys.argv[4]
        raise_if_timestamp_invalid(
            datetime.fromtimestamp(float(timestamp), timezone.utc))
    else:
        timestamp = None

    if sys.stdin.isatty():
        sys.stderr.write(
            "Error: Pipe json test data through this script, like so:\n")
        sys.stderr.write("$ cat test_data | %s\n" % ' '.join(sys.argv))
        sys.exit(2)

    config = ConfigParser()
    config.read(config_file)
    if 'nfss' not in config:
        sys.stderr.write('No nfss section in config file\n')
        sys.stderr.write('Please verify you have a valid config file, '
                         'obtained by running keys_add on the instance.\n')
        sys.exit(3)
    for config_item in (
        'client_access_key',
        'resource_owner_key',
        'resource_owner_secret',
        'backend'
    ):
        if config_item not in config['nfss']:
            sys.stderr.write('%s not present in config file\n' % config_item)
            sys.stderr.write('Please verify you have a valid config file, '
                             'obtained by running keys_add on the instance.\n')
            sys.exit(3)

    data = sys.stdin.read()
    try:
        json.loads(data)
    except ValueError as e:
        sys.stderr.write(
            "Error: Data does not appear to be valid json: %s\n" % e)
        sys.exit(3)

    test_session = OAuth1Session(
        config['nfss']['client_access_key'],
        resource_owner_key=config['nfss']['resource_owner_key'],
        resource_owner_secret=config['nfss']['resource_owner_secret']
    )
    url = '/'.join((config['nfss']['backend'], project, test))
    r = test_session.post(url, dict(data=data.encode(), timestamp=timestamp))
    print(r.text)
    if r.status_code != 200:
        sys.stderr.write('Unexpected HTTP response code: %i\n' % r.status_code)
        sys.exit(r.status_code)

if __name__ == '__main__':
    main()
