var app = angular.module(
    'NonFunctional',
    ['ngAnimate', 'nvd3', 'ngQuickDate', 'NonFunctional.graphing', 'NonFunctional.support']
);

var DOMAIN = '';
var API_PREFIX = "/api/v1/";
var GRAPH_DIR = "graphs/";

if (window.location.origin == 'file://') {
    DOMAIN = 'http://nfss.staging.ubuntu.com';
}

// This allows reversing the display order of lists in angular templates
// by doing something like ng-repeat="foo in bar | reverse"
app.filter('reverse', function() {
    return function(items) {
        return items.slice().reverse();
    };
});


// Produce a date range for the REST API if appropriate
function dateRange(url) {
    var start = url.start_date;
    var end =   url.end_date;
    if (start && end) {
        return '?start_date=' + start + '&end_date=' + end;
    }
    return '';
}

// Generate a URL query string that contains the appropriate date range
// defined by the arguments start and end (which should be date objects).
function makeMonthLink(url, start, end) {
    return ('#?project=' + url.project + '&test=' + url.test +
            '&start_date=' + (start.getTime() / 1000) + '&end_date=' + (end.getTime() / 1000));
}


function appController($scope, $location, $http, $rootScope, dates) {
    $scope.loading = false;
    $scope.graphAvailable = false;

    $scope.endDate = null;
    $scope.startDate = null;

    $scope.dateFormatter = dates.dateFormatter;

    $scope.setLoading = function(state) {
        $scope.loading = state;
    };

    $scope.setGraphAvailable = function(state) {
        $scope.graphAvailable = state;
    };

    function dateWatcher() {
        if ($scope.startDate && $scope.endDate) {
            $scope.url.start_date = $scope.startDate.getTime() / 1000;
            $scope.url.end_date   = $scope.endDate.getTime()   / 1000;
            $location.search($scope.url);
            $scope.setLoading(true);
            $scope.getGraphData();
        }
    }

    $scope.$watch('endDate', dateWatcher);
    $scope.$watch('startDate', dateWatcher);

    // Parse URL query paramters into a dict and return it.
    function URLParams() {
        return $location.search();
    }

    // Watch for changes in the URL and apply that data to the scope.
    // Values are stored as eg $scope.url.project
    function URLWatcher(queryParams) {
        $scope.url = queryParams;
    }

    $scope.$watch(URLParams, URLWatcher);

    function setChosenProjectData(data) {
        $scope.chosenProjectData = data.tests;
    }

    // When a project is chosen, get a list of its tests
    function projectWatcher(chosenProject) {
        $scope.tests = [];
        if (chosenProject in $scope.allProjects) {
            $scope.tests = $scope.allProjects[chosenProject].tests.sort();
            $http.get(DOMAIN + $scope.allProjects[chosenProject].path)
                 .success(setChosenProjectData);
        }
    }

    // Handle results from initial rest request: Show the list of projects.
    function projectInitializer(data) {
        $scope.projects = Object.keys(data.projects).sort();
        $scope.allProjects = data.projects;
        $scope.$watch('url.project', projectWatcher);
    }

    // Make the initial rest request.
    $http.get(DOMAIN + API_PREFIX).success(projectInitializer);

    // Set the values for the next/previous month links after the graph data has loaded.
    function setMonthLinks(blob) {
        $scope.hidePrevLink = false;
        $scope.hideNextLink = false;
        $scope.veryFirst = dates.dateParser(blob.first_data);
        $scope.veryLast = dates.dateParser(blob.last_data);
        var first = $scope.url.start_date ? new Date($scope.url.start_date * 1000) : dates.dateParser(blob.data_range_first);
        var last  = $scope.url.end_date   ? new Date($scope.url.end_date   * 1000) : dates.dateParser(blob.data_range_last);
        if (first.getTime() <= $scope.veryFirst.getTime()) {
            $scope.hidePrevLink = true;
        }
        if (last.getTime() >= $scope.veryLast.getTime()) {
            $scope.hideNextLink = true;
        }
        $scope.prevMonthLink = makeMonthLink($scope.url, dates.deltaMonth(first, -1), first);
        $scope.nextMonthLink = makeMonthLink($scope.url, last, dates.deltaMonth(last, 1));
    }
    $scope.setMonthLinks = setMonthLinks;

    // Fetch the actual graph data from the db.
    $scope.getGraphData = function() {
        $http.get(DOMAIN + API_PREFIX + $scope.url.project + '/' + $scope.url.test + dateRange($scope.url))
        .success(setMonthLinks)
        .success(function(data){
            $scope.$emit("newdata", data);
        });
    };

    $scope.$on('newdata', function(e, data) {
        var data_available = data.data.length > 0;
        $scope.setLoading(!data_available);
        $scope.setGraphAvailable(data_available);
    });
}

app.directive('customGraph', function($compile, $http, $templateCache, $q, $rootScope, definitions) {
    // This function handles the switching of graph definitions and loading of
    // graph data when the user navigates between pages. The functions defined
    // within this function will make more sense if you read them bottom to top.
    function graphLinker(scope, element, attrs) {

        // Take a string of raw HTML data, insert it into the DOM, and
        // tell angular that the DOM has changed so it can make it go.
        function graphChanger(htmlData) {
            if(htmlData === undefined) return;
            element.html(htmlData);
            $compile(element.contents())(scope);

            // Now that the graph definition is ready, let's load the graph data.
            scope.getGraphData();
        }

        // return a promise object that will load the requested file_url.
        function resolveController(file_url) {
            var deferred = $q.defer();
            var script = document.createElement('script');
            script.src = file_url;
            script.onload = function() {
                $rootScope.$apply(deferred.resolve());
            };
            document.body.appendChild(script);
            return deferred.promise;
        }

        // Make the AJAX request to load the raw HTML data and javascript file., calling graphChanger
        function loadGraphDef(graph_files) {
            scope.setLoading(true);

            resolveController(graph_files.deps).then(
                function(data) {
                    $http({
                        method: 'GET',
                        url: graph_files.templateUrl,
                        cache: $templateCache
                    }).success(graphChanger);
                },
                function(data) {
                    console.log("Unable to load graph definition javascript.");
                }
            );
        }

        // Load the most appropriate graph definition as defined in
        // graphs/graphing-module.js. First a project+test specific graph is tried, then
        // a general project-wide definition, then a general test-wide one.
        scope.graphWatcher = function(urlParams) {
            scope.setGraphAvailable(false);
            scope.startDate = null;
            scope.endDate = null;

            // getGraphDef will return a default graph if no other options are available.
            var graph_def_files = definitions.getGraphDef(scope.url);
            if (scope.url.test) {
                loadGraphDef(graph_def_files);
            }
        };

        scope.$watch('url', scope.graphWatcher);
    }

    return {
        replace: true,
        link: graphLinker
    };
});
