describe('The dateRange function', function() {
    it('should return URL query parameters when appropriate.', function() {
        expect(dateRange({start_date: 1, end_date: 100}))
        .toBe('?start_date=1&end_date=100');
    });
    it('should not return URL query parameters otherwise.', function() {
        expect(dateRange({})).toBe('');
        expect(dateRange({start_date: 1000})).toBe('');
        expect(dateRange({end_date: 42})).toBe('');
    });
});

describe('The makeMonthLink function', function() {
    it('should return URL query parameters.', function() {
        var url = { project: 'foo', test: 'bar' };
        var start = new Date(Date.UTC(1970, 0, 1));
        var end   = new Date(Date.UTC(1970, 0, 2));
        expect(makeMonthLink(url, start, end))
        .toBe('#?project=foo&test=bar&start_date=0&end_date=86400');
    });
});

describe('The appController function data retrieval', function() {
    var scope, ctrl, $httpBackend, basic_response

    beforeEach(module('NonFunctional'));

    beforeEach(inject(function(_$httpBackend_, $rootScope, $controller) {
        basic_response = {
            data_range_first: "2014-01-01T12:59:59+00:00",
            first_data: "2014-01-01T12:59:59+00:00",
            last_data: "2014-02-02T15:59:59+00:00",
            data_range_last: "2014-02-02T15:59:59+00:00",
            data: [],
        };

        $httpBackend = _$httpBackend_;

        scope = $rootScope.$new();
        ctrl = $controller('appController', {$scope: scope});
    }));

    it('must have a getGraphData function.', function() {
        expect(scope.getGraphData).toBeDefined();
    });

    describe('getGraphData data requests', function() {
        beforeEach(function() {
            $httpBackend.expect('GET', DOMAIN + API_PREFIX).respond({projects: {}});
            $httpBackend.when('GET', DOMAIN + API_PREFIX + 'foo/bar').respond(basic_response);
            $httpBackend.when('GET', DOMAIN + API_PREFIX + 'foo/bar?start_date=1&end_date=69').respond(
                basic_response
            );
        });

        // Fail tests if there is any un-handled get requests.
        afterEach(function() {
            $httpBackend.verifyNoOutstandingExpectation();
            $httpBackend.verifyNoOutstandingRequest();
        });

        it('getGraphData function must request data for project + test in url.', function() {
            expect(scope.getGraphData).toBeDefined();
            scope.url = {
                project: 'foo',
                test: 'bar',
            };
            expect(scope.loading).toBe(false);
            $httpBackend.expectGET(DOMAIN + API_PREFIX + 'foo/bar');
            scope.getGraphData();
            $httpBackend.flush();
            expect(scope.loading).toBe(true);
        });

        it('getGraphData function must include start and end dates defined in url.', function() {
            $httpBackend.expectGET(DOMAIN + API_PREFIX + 'foo/bar?start_date=1&end_date=69');
            scope.url = {
                project: 'foo',
                test: 'bar',
                start_date: 1,
                end_date: 69,
            };
            scope.getGraphData();
            $httpBackend.flush();
        });
    });

    it('must show next & last month links when data is available', function() {
        scope.url = {
            project: 'flib',
            test: 'jib',
        };
        scope.setMonthLinks(basic_response);

        expect(scope.veryFirst instanceof Date).toBe(true);
        expect(scope.veryLast instanceof Date).toBe(true);
        expect(scope.hidePrevLink).toBe(true);
        expect(scope.hideNextLink).toBe(true);
        expect(scope.prevMonthLink).toBe(
            '#?project=flib&test=jib&start_date=1385902799&end_date=1388581199');
        expect(scope.nextMonthLink).toBe(
            '#?project=flib&test=jib&start_date=1391356799&end_date=1393775999');
    });

    it('must show next & last month links when data is available', function() {
        scope.url = {
            project: 'flib',
            test: 'jib',
        };
        basic_response.first_data = '1970-10-20T23:11:30Z';
        scope.setMonthLinks(basic_response);
        expect(scope.hidePrevLink).toBe(false);
        expect(scope.hideNextLink).toBe(true);

        basic_response.last_data = '2015-12-25 12:13:14';
        scope.setMonthLinks(basic_response);
        expect(scope.hidePrevLink).toBe(false);
        expect(scope.hideNextLink).toBe(false);
    });
});

describe('The appController function', function() {
    var scopeMock, locationMock, httpMock, datesMock;
    beforeEach(function() {
        scopeMock = {
            'url': {},
            '$watch': NOP,
            '$on': function(){},
        };
        locationMock = {
            'search': NOP,
        };
        datesMock = {
            'dateFormatter': NOP,
        };
        httpMock = {
            'get': function () {
                return {
                    success: function() { return this },
                    error: function() { return this }
                }
            }
        };
    });
    it('should set some basic state.', function() {
        appController(scopeMock, locationMock, httpMock, {}, datesMock);
        expect(scopeMock.loading).toBe(false);
        expect(scopeMock.endDate).toBeNull();
        expect(scopeMock.startDate).toBeNull();
    });
    it('should watch for changes.', function() {
        spyOn(scopeMock, '$watch');
        appController(scopeMock, locationMock, httpMock, {}, datesMock);
        expect(scopeMock.$watch).toHaveBeenCalled();
        var watchCalls = scopeMock.$watch.calls.all();
        expect(watchCalls[0].args[0]).toBe('endDate');
        expect(watchCalls[0].args[1].name).toBe('dateWatcher');
        expect(watchCalls[1].args[0]).toBe('startDate');
        expect(watchCalls[1].args[1].name).toBe('dateWatcher');
        expect(watchCalls[2].args[0].name).toBe('URLParams');
        expect(watchCalls[2].args[1].name).toBe('URLWatcher');
    });
    it('should respond to the date picker.', function() {
        spyOn(scopeMock, '$watch');
        spyOn(locationMock, 'search');
        appController(scopeMock, locationMock, httpMock, {}, datesMock);
        spyOn(scopeMock, 'getGraphData');
        var dateWatcher = scopeMock.$watch.calls.first().args[1];
        dateWatcher();
        expect(scopeMock.url.start_date).toBeUndefined();
        expect(scopeMock.url.end_date).toBeUndefined();
        scopeMock.startDate = new Date();
        dateWatcher();
        expect(scopeMock.url.start_date).toBeUndefined();
        expect(scopeMock.url.end_date).toBeUndefined();
        scopeMock.endDate = new Date();
        dateWatcher();
        expect(scopeMock.url.start_date).toBe(scopeMock.startDate.getTime() / 1000);
        expect(scopeMock.url.end_date).toBe(scopeMock.endDate.getTime() / 1000);
        expect(locationMock.search).toHaveBeenCalledWith(scopeMock.url);
        expect(scopeMock.getGraphData).toHaveBeenCalledWith();
    });
    it('should have a URLParams function.', function() {
        spyOn(scopeMock, '$watch');
        spyOn(locationMock, 'search').and.returnValue('sentinel');
        appController(scopeMock, locationMock, httpMock, {}, datesMock);
        expect(scopeMock.$watch).toHaveBeenCalled();
        var URLParams = scopeMock.$watch.calls.all()[2].args[0];
        var result = URLParams();
        expect(locationMock.search).toHaveBeenCalled();
        expect(result).toBe('sentinel');
    });
    it('should have a URLWatcher function.', function() {
        spyOn(scopeMock, '$watch');
        spyOn(locationMock, 'search').and.returnValue('sentinel');
        appController(scopeMock, locationMock, httpMock, {}, datesMock);
        expect(scopeMock.$watch).toHaveBeenCalled();
        var URLWatcher = scopeMock.$watch.calls.all()[2].args[1];
        URLWatcher('hiya!');
        expect(scopeMock.url).toBe('hiya!');
    });
    it('should have a projectInitializer and projectWatcher.', function() {
        var projectInitializer;
        spyOn(scopeMock, '$watch');
        spyOn(httpMock, 'get').and.returnValue({
            success: function(f) {
                projectInitializer = f;
            }
        });
        appController(scopeMock, locationMock, httpMock, {}, datesMock);
        expect(httpMock.get).toHaveBeenCalled();
        var data = { projects: { foo: 1, bar: 2, baz: 3 } };
        projectInitializer(data);
        expect(scopeMock.projects).toEqual([ 'bar', 'baz', 'foo' ]);
        expect(scopeMock.allProjects).toBe(data.projects);
        expect(scopeMock.$watch).toHaveBeenCalled();
        watchCall = scopeMock.$watch.calls.mostRecent();
        expect(watchCall.args[0]).toBe('url.project');
        var projectWatcher = watchCall.args[1];
        expect(projectWatcher.name).toBe('projectWatcher');
        scopeMock.allProjects = {
            foo: {
                tests: [ 'one', 'two', 'three' ],
                path: '/api/path/here',
            },
        };
        scopeMock.tests = 'wrong';
        expect(httpMock.get.calls.count()).toBe(1);

        projectWatcher('foo');
        expect(scopeMock.tests).toEqual([ 'one', 'three', 'two' ]);
        expect(httpMock.get.calls.mostRecent().args[0]).toBe(DOMAIN + '/api/path/here');
        expect(httpMock.get.calls.count()).toBe(2);

        projectWatcher('nonexistent');
        expect(scopeMock.tests).toEqual([]);
        expect(httpMock.get.calls.count()).toBe(2);
    });
});
