'use strict';

describe("Support module", function() {
    beforeEach(module("NonFunctional.support"));

    describe("dates factory", function() {
        var dates;
        beforeEach(inject(function(_dates_) {
            dates = _dates_;
        }));

        describe('The dateParser function', function() {
            it('should parse dates properly.', function() {
                var result = dates.dateParser('2014-01-02T03:04:05Z');
                expect(result instanceof Date).toBe(true);
                expect(result.getUTCFullYear()).toBe(2014);
                expect(result.getUTCMonth()).toBe(0);
                expect(result.getUTCDate()).toBe(2);
                expect(result.getUTCHours()).toBe(3);
                expect(result.getUTCMinutes()).toBe(4);
                expect(result.getUTCSeconds()).toBe(5);

                var result = dates.dateParser('2008-08-30 12:30:45');
                expect(result.toUTCString()).toBe('Sat, 30 Aug 2008 12:30:45 GMT');
            });

            it('should parse dates with timezones properly.', function() {
                var result = dates.dateParser('2008-08-30 12:30:45+0000');
                expect(result.toUTCString()).toBe('Sat, 30 Aug 2008 12:30:45 GMT');
            });
        });

        describe('The deltaMonth function', function() {
            it('should be able to offset months.', function() {
                var result = dates.deltaMonth(new Date(2014, 0, 15), -1);
                expect(result.getMonth()).toBe(11);
                expect(result.getYear()).toBe(113);

                var result = dates.deltaMonth(new Date(2014, 0, 15), 1);
                expect(result.getMonth()).toBe(1);
                expect(result.getYear()).toBe(114);
            });
        });

        describe('The dateFormatter function', function() {
            it('should use a sensible default format.', function() {
                var now = new Date();
                spyOn(d3.time, 'format').and.returnValue(
                    function() { return arguments });
                var result = dates.dateFormatter()(now);
                expect(d3.time.format).toHaveBeenCalledWith('%Y-%m-%d %H:%M:%S %Z');
                expect(result[0]).toEqual(now);
            });

            it('should accept a custom format.', function() {
                var now = new Date();
                spyOn(d3.time, 'format').and.returnValue(
                    function() { return arguments });
                var result = dates.dateFormatter('%Y-%m-%d')(now);
                expect(d3.time.format).toHaveBeenCalledWith('%Y-%m-%d');
                expect(result[0]).toEqual(now);
            });
        });

    });

    describe("modifiers factory", function() {
        var modifiers;
        beforeEach(inject(function(_modifiers_) {
            modifiers = _modifiers_;
        }));

        describe('The accessor function', function() {
            it('should return the correct values.', function() {
                var data = { foo: 42, bar: 27 };
                expect(modifiers.accessor('foo')(data)).toBe(42);
                expect(modifiers.accessor('bar')(data)).toBe(27);
                expect(modifiers.accessor('qux')(data)).toBeUndefined();
            });
        });

        describe('The numberFormatter function', function() {
            it('should use a sensible default format.', function() {
                var n = 49.3;
                spyOn(d3, 'format').and.returnValue(
                    function() { return arguments[0] });
                var result = modifiers.numberFormatter(null, ' parsecs')(n);
                expect(d3.format).toHaveBeenCalledWith(',.2f');
                // Unformatted because d3 is mocked out.
                expect(result).toEqual('49.3 parsecs');
            });
            it('should accept a custom format.', function() {
                var n = 24.7;
                spyOn(d3, 'format').and.returnValue(
                    function(n) { return n });
                var result = modifiers.numberFormatter('.6f', ' seconds')(n);
                expect(d3.format).toHaveBeenCalledWith('.6f');
                // Unformatted because d3 is mocked out.
                expect(result).toEqual('24.7 seconds');
            });
            it('should not require a unit.', function() {
                var n = 24.123;
                spyOn(d3, 'format').and.returnValue(
                    function(n) { return n });
                expect(modifiers.numberFormatter('.2f')(n)).toBe('24.123');
            });
        });
    });

});
