#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from lazr.restfulclient.errors import (
    BadRequest,
    ClientError,
    Unauthorized,
)
import logging

from ci_utils import (
    launchpad,
    unit_config,
)

log = logging.getLogger('ppacreator')


class PPACreationError(Exception):
    """Raised if there was an error creating a new PPA."""


class PPADeVirtualizationError(Exception):
    """Raised if there was an error while de-virtualizing PPA."""


def create_ppa(name,
               displayname,
               architectures=None,
               description=None,
               series=None,
               private=False):
    """Creates a LP PPA with the given information.

    Creates a PPA to the configured 'launchpad_owner_name' LP user logged
    in as the configured 'launchpad_user' & 'launchpad_token'.

    In case of failure, it raises `PPACreationError` with the reason
    why the PPA could not be created.

    Return the just-created PPA reference (ppa:<owner>/<name>) on success.
    """
    user_name = unit_config.get('launchpad_user')
    owner_name = unit_config.get('launchpad_ppa_owner')
    ppa_name = "ppa:{}/{}".format(owner_name, name)

    log.info('Attempting to create: {}'.format(ppa_name))

    try:
        lp = launchpad.lp_login()
    except ClientError:
        raise PPACreationError(
            'Could not log in LP using the configured \'launchpad_user\' '
            '({}).'.format(user_name))

    try:
        owner = lp.people[owner_name]
    except (ClientError, KeyError):
        raise PPACreationError(
            'Could not find the configured \'launchpad_ppa_owner\' '
            '({}).'.format(owner_name))

    try:
        # XXX cprov 20140911: At some point we will have to support PPA
        # creation for other distributions (e.g. RTM).
        ppa = owner.createPPA(
            name=name,
            displayname=displayname,
            description=description,
            private=private)
    except BadRequest as exc:
        ppa = owner.getPPAByName(name=name)
        if ppa is not None:
            raise PPACreationError('PPA {} already exists.'.format(ppa_name))
        raise PPACreationError(
            'The LP request to create the PPA failed:\n{}'.format(
                str(exc)))
    except Unauthorized as exc:
        raise PPACreationError(
            'Configured \'launchpad_user\' ({}) has no permission to create '
            'PPAs in the context of the configured \'launchpad_ppa_owner\' '
            '({}).'.format(user_name, owner_name))
    log.info('PPA {} created.'.format(ppa_name))

    # no need to attempt to de-virt the ppa if and only if all is stated
    if not ("all" in architectures and len(architectures) == 1):
        modify_ppa(lp, name, series, architectures, ppa)

    return ppa_name


def modify_ppa(lp,
               name,
               series,
               architectures,
               ppa):
    log.info('Attempting to modify: {} to enable '
             'restricted processors'.format(ppa.name))
    ppa_admins = 'launchpad-ppa-self-admins'
    user_name = unit_config.get('launchpad_user')

    try:
        # XXX psivaa 20140919 At some point we will have to support
        # de-virtualizing ppa's for other distributions (e.g. RTM).
        ubuntu = lp.distributions['ubuntu']
    except (ClientError, KeyError):
        raise PPACreationError(
            'Could not find Ubuntu distribution')

    log.info('Distribution {} is found'.format(ubuntu))

    try:
        lp_series = ubuntu.getSeries(name_or_version=series)
    except (ClientError, KeyError):
        raise PPACreationError(
            'Could not find the series ({}).'.format(series))

    log.info('The series found to be: {}'.format(lp_series))

    # processor.restricted is only available in lp.net/devel/processors
    restricted_processors = [
        das.processor for das in lp_series.architectures
        if das.processor.restricted]

    # Here we calculate the required restricted processors for a given
    # set of architecture_tags (identical to processor names)
    if 'any' in architectures:
        needed_processors = restricted_processors
    else:
        needed_processors = [
            proc for proc in restricted_processors
            if proc.name in architectures]

    # If no restricted procs are needed, dismiss
    if not needed_processors:
        return

    if not lp.people[user_name] in lp.people[ppa_admins].members_details:
        raise PPADeVirtualizationError(
            'The system was deployed with the credential of ({}) who '
            'is not a member of the ({}) team. Modifying the ppa enabling '
            'restricted processors failed.'.format(user_name, ppa_admins))

    try:
        for proc in needed_processors:
            ppa.enableRestrictedProcessor(processor=proc)
    except BadRequest:
        raise PPACreationError(
            'Enabling restricted processors for the created ppa ({}) failed'
            'when attempted with the credentials of \'launchpad_user\''
            ' ({}) '.format(user_name, ppa.name))
    ppa.require_virtualized = False
    ppa.save()
