#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging


from ci_utils import (
    amqp_utils,
    amqp_worker,
    unit_config,
)
from ppacreator.create_ppa import (
    create_ppa,
    PPACreationError,
    PPADeVirtualizationError,
)


class PPACreatorWorker(amqp_worker.AMQPWorker):
    def __init__(self):
        super(PPACreatorWorker, self).__init__('ppacreator')

    def engine_health(self):
        status = super(PPACreatorWorker, self).engine_health()
        status.add_de_virt_ppa_capabilities()
        return status

    def handle_request(self, params, log):
        name = params['name']
        displayname = params['displayname']
        description = params['description']
        trigger = params['progress_trigger']
        ticket_id = params['ticket_id']
        architectures = params['architectures']
        series = params['series']

        def status_cb(msg):
            log.info(msg)
            amqp_utils.progress_update(trigger, {'message': msg})

        status_cb('Creating PPA {} for ticket, {}'.format(name, ticket_id))
        retval = {}
        try:
            ppa_name = create_ppa(name,
                                  displayname,
                                  architectures,
                                  description,
                                  series)
        except PPACreationError as exc:
            retval['message'] = exc.message
            return amqp_utils.progress_failed, retval

        except PPADeVirtualizationError as exc:
            # XXX psivaa 20140921: in production deployments this should
            # be considered an error, but currently we have no precise
            # way to identify whether the current deployment is
            # 'production' or not (read development, trial, etc).
            log.warning(str(exc))
            owner_name = unit_config.get('launchpad_ppa_owner')
            ppa_name = "ppa:{}/{}".format(owner_name, name)
            architectures = []

        retval = {'ppa_name': ppa_name, 'unique_arches': architectures}
        return amqp_utils.progress_completed, retval


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO,
                        format="%(asctime)s %(levelname)s %(message)s")
    PPACreatorWorker().main(amqp_utils.PPA_CREATOR_QUEUE)
