# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from lazr.restfulclient.errors import (
    BadRequest,
    ClientError,
    Unauthorized,
)
import mock
import unittest

from ppacreator.create_ppa import (
    create_ppa,
    modify_ppa,
    PPACreationError,
    PPADeVirtualizationError,
)


class FakeResponse(object):
    """Fake HTTP response implementation."""

    status = 'X'
    reason = 'Y'

    def items(self):
        return []


class FakeLaunchpad(object):
    """Fake Launchpad API client implementation."""

    def __init__(self):
        self.people = {}
        self.team = {}
        self.series = {}
        self.architectures = {}


class FakeLaunchpadPPA(object):
    """Fake Launchpad PPA API object."""

    def __init__(self, owner, name):
        self.owner = owner
        self.name = name
        self.processor = FakeLaunchpadProcessor('armhf')

    def enableRestrictedProcessor(self, processor):
        return processor

    def save(self):
        return


class FakeLaunchpadPerson(object):
    """Fake Launchpad API person."""

    def __init__(self, name, failure=None):
        self.name = name
        self.failure = failure
        self._ppas = {}
        self.members_details = {}

    def createPPA(self, **kwargs):
        if self.failure is not None:
            raise self.failure
        name = kwargs['name']
        ppa = FakeLaunchpadPPA(self, name)
        self._ppas[name] = ppa
        return ppa

    def getPPAByName(self, name):
        return self._ppas.get(name)


class FakeLaunchpadTeam(object):

    def __init__(self, name, members=[FakeLaunchpadPerson,
                                      FakeLaunchpadPerson]):
        self.name = name
        self.members_details = members


class FakeLaunchpadDistro(object):

    def __init__(self, name, series=None):
        self.name = name
        self.series = FakeLaunchpadSeries(series)

    def getSeries(self, name_or_version=None):
        return FakeLaunchpadSeries(name_or_version)


class FakeLaunchpadSeries(object):
    def __init__(self, name):
        self.architectures = [FakeLaunchpadDistroArchSeries(name)]


class FakeLaunchpadDistroArchSeries(object):
    def __init__(self, name):
        self.name = name
        self.processor = FakeLaunchpadProcessor('a_proc')


class FakeLaunchpadProcessor(object):
    def __init__(self, name, restricted=None):
        self.name = name
        self.restricted = True


class TestCreatePPA(unittest.TestCase):

    def setUp(self):
        """Mocks `unit_config.get` with testing configuration."""
        super(TestCreatePPA, self).setUp()

        _p = mock.patch('ci_utils.unit_config.get')
        mocked_config_get = _p.start()
        self.addCleanup(_p.stop)
        self.test_config = {
            'oauth_consumer_key': 'A_KEY',
            'oauth_token': 'A_TOKEN',
            'oauth_token_secret': 'A_SECRET',
            'launchpad_user': 'A_USER',
            'launchpad_ppa_owner': 'A_OWNER',
        }
        mocked_config_get.side_effect = lambda k: self.test_config.get(k)

    @mock.patch('ci_utils.launchpad.lp_login')
    def test_login_fails(self, mocked_lp):
        # `create_ppa` raises `PPACreationError` if it is not possible to
        # log on LP with the configured credentials.
        mocked_lp.side_effect = [
            ClientError(None, None)
        ]
        with self.assertRaises(PPACreationError) as error:
            create_ppa('A_PPA', 'A_DISPLAYNAME')

        self.assertEqual(
            "Could not log in LP using the configured 'launchpad_user' "
            "(A_USER).",
            str(error.exception))

    @mock.patch('ci_utils.launchpad.lp_login')
    def test_owner_not_found(self, mocked_lp):
        # `create_ppa` raises `PPACreationError` if the configured PPA
        # owner could not be found.
        fake_lp = FakeLaunchpad()
        mocked_lp.return_value = fake_lp

        with self.assertRaises(PPACreationError) as error:
            create_ppa('A_PPA', 'A_DISPLAYNAME')

        self.assertEqual(
            "Could not find the configured 'launchpad_ppa_owner' (A_OWNER).",
            str(error.exception))

    @mock.patch('ci_utils.launchpad.lp_login')
    def test_broken_request(self, mocked_lp):
        # `create_ppa` raises `PPACreationError` if the creation request
        # returns HTTP 400 and verbose request information is included in
        # the exception content (logged).
        fake_lp = FakeLaunchpad()
        fake_owner = FakeLaunchpadPerson(
            self.test_config['launchpad_ppa_owner'],
            failure=BadRequest(FakeResponse(), 'ERROR'))
        fake_lp.people = {
            'A_OWNER': fake_owner,
        }
        mocked_lp.return_value = fake_lp

        with self.assertRaises(PPACreationError) as error:
            create_ppa('A_PPA', 'A_DISPLAYNAME')

        self.assertEqual(
            ['The LP request to create the PPA failed:',
             'HTTP Error X: Y',
             'Response headers:',
             '---',
             '',
             '---',
             'Response body:',
             '---',
             'ERROR',
             '---'],
            str(error.exception).splitlines())

    @mock.patch('ci_utils.launchpad.lp_login')
    def test_duplicated(self, mocked_lp):
        # `create_ppa` raises `PPACreationError` if there is already a PPA
        # with the same name.
        fake_lp = FakeLaunchpad()
        fake_owner = FakeLaunchpadPerson(
            self.test_config['launchpad_ppa_owner'])
        fake_owner.createPPA(name='A_PPA')
        fake_owner.failure = BadRequest(None, None)
        fake_lp.people = {
            'A_OWNER': fake_owner,
        }
        mocked_lp.return_value = fake_lp

        with self.assertRaises(PPACreationError) as error:
            create_ppa('A_PPA', 'A_DISPLAYNAME')

        self.assertEqual(
            'PPA ppa:A_OWNER/A_PPA already exists.',
            str(error.exception))

    @mock.patch('ci_utils.launchpad.lp_login')
    def test_unauthorized(self, mocked_lp):
        # `create_ppa` raises `PPACreationError` if the configured
        # credentials are not authorized to do so.
        fake_lp = FakeLaunchpad()
        fake_lp.people = {
            'A_OWNER': FakeLaunchpadPerson(
                self.test_config['launchpad_ppa_owner'],
                failure=Unauthorized(None, None)),
        }
        mocked_lp.return_value = fake_lp

        with self.assertRaises(PPACreationError) as error:
            create_ppa('A_PPA', 'A_DISPLAYNAME')

        self.assertEqual(
            "Configured 'launchpad_user' (A_USER) has no permission to "
            "create PPAs in the context of the configured "
            "'launchpad_ppa_owner' (A_OWNER).",
            str(error.exception))

    @mock.patch('ci_utils.launchpad.lp_login')
    def test_creates(self, mocked_lp):
        # `create_ppa` returns the just-created PPA reference on success.
        fake_lp = FakeLaunchpad()
        fake_lp.people = {
            'A_OWNER': FakeLaunchpadPerson(
                self.test_config['launchpad_ppa_owner']),
            'A_USER': FakeLaunchpadPerson(
                self.test_config['launchpad_user']),
            'launchpad-ppa-self-admins': FakeLaunchpadPerson(
                'admins')
        }
        mocked_lp.return_value = fake_lp

        # all special case
        self.assertEqual(
            'ppa:A_OWNER/A_PPA',
            create_ppa('A_PPA', 'A_DISPLAYNAME', {'all'}))

    @mock.patch('ppacreator.create_ppa.modify_ppa')
    @mock.patch('ci_utils.launchpad.lp_login')
    def test_modifies(self, mocked_lp, modify):
        # `modify_ppa` is called for cases where arch is not `all` and
        # the user is in lp-ppa-self-admins
        fake_lp = FakeLaunchpad()
        a_user = FakeLaunchpadPerson('A_USER')
        b_user = FakeLaunchpadPerson('B_USER')
        fake_lp.people = {
            'A_OWNER': FakeLaunchpadPerson(
                self.test_config['launchpad_ppa_owner']),
            'A_USER': a_user,
            'launchpad-ppa-self-admins': FakeLaunchpadTeam(
                'launchpad-ppa-self-admins',
                [a_user, b_user])
        }

        mocked_lp.return_value = fake_lp

        self.assertEqual(
            'ppa:A_OWNER/A_PPA',
            create_ppa('A_PPA', 'A_DISPLAYNAME', {'arch1', 'arch2'}))
        modify.assert_called_once()

    @mock.patch('ci_utils.launchpad.lp_login')
    def test_not_found(self, mocked_lp):
        # `modify_ppa` raises `PPACreationError` if `ubuntu`
        # distribution is not found.
        fake_lp = FakeLaunchpad()
        a_user = FakeLaunchpadPerson('A_USER')
        b_user = FakeLaunchpadPerson('B_USER')
        fake_lp.people = {
            'A_OWNER': FakeLaunchpadPerson(
                self.test_config['launchpad_ppa_owner']),
            'A_USER': a_user,
            'launchpad-ppa-self-admins': FakeLaunchpadTeam(
                'launchpad-ppa-self-admins',
                [a_user, b_user])
        }

        fake_lp.distributions = {
            'a_distro': FakeLaunchpadDistro('a_distro')
        }
        ppa = FakeLaunchpadPPA('A_OWNER', 'A_PPA')
        fake_name = 'A_PPA'
        fake_arch = {'A_ARCH1', 'A_ARCH2'}
        fake_lp_series = 'A_SERIES'
        mocked_lp.return_value = fake_lp

        with self.assertRaises(PPACreationError) as error:
            modify_ppa(fake_lp, fake_name, fake_lp_series, fake_arch, ppa)

        self.assertEqual(
            "Could not find Ubuntu distribution",
            str(error.exception))

    @mock.patch('ci_utils.launchpad.lp_login')
    def test_creates_modifies(self, mocked_lp):
        # `modify_ppa` returns the just-created PPA
        # with de-virt capabilities
        fake_lp = FakeLaunchpad()
        a_user = FakeLaunchpadPerson('A_USER')
        b_user = FakeLaunchpadPerson('B_USER')
        fake_lp.people = {
            'A_OWNER': FakeLaunchpadPerson(
                self.test_config['launchpad_ppa_owner']),
            'A_USER': a_user,
            'launchpad-ppa-self-admins': FakeLaunchpadTeam(
                'launchpad-ppa-self-admins',
                [a_user, b_user])
        }
        fake_lp.distributions = {
            'ubuntu': FakeLaunchpadDistro('ubuntu')
        }
        mocked_lp.return_value = fake_lp

        self.assertEqual(
            'ppa:A_OWNER/A_PPA',
            create_ppa('A_PPA', 'A_DISPLAYNAME', {'any', 'arch2'}))

    @mock.patch('ci_utils.launchpad.lp_login')
    def test_unauth_modification(self, mocked_lp):
        # `modify_ppa` raises `PPADeVirtualizationError` if the configured
        # user is not a member of launchpad-ppa-self-admins
        # and a restricted processor is requested
        fake_lp = FakeLaunchpad()
        a_user = FakeLaunchpadPerson('A_USER')
        b_user = FakeLaunchpadPerson('B_USER')
        c_user = FakeLaunchpadPerson('C_USER')
        fake_lp.people = {
            'A_OWNER': FakeLaunchpadPerson(
                self.test_config['launchpad_ppa_owner']),
            'A_USER': a_user,
            'launchpad-ppa-self-admins': FakeLaunchpadTeam(
                'launchpad-ppa-self-admins',
                [b_user, c_user])
        }

        fake_lp.distributions = {
            'ubuntu': FakeLaunchpadDistro('ubuntu')
        }
        ppa = FakeLaunchpadPPA('A_USER', 'A_PPA')
        fake_name = 'A_PPA'
        fake_arch = {'any', 'A_ARCH2'}
        fake_lp_series = 'A_SERIES'

        mocked_lp.return_value = fake_lp

        with self.assertRaises(PPADeVirtualizationError) as error:
            modify_ppa(fake_lp,
                       fake_name, fake_lp_series, fake_arch, ppa)

        self.assertEqual(
            "The system was deployed with the credential of (A_USER) who "
            "is not a member of the (launchpad-ppa-self-admins) team."
            " Modifying the ppa enabling restricted processors failed.",
            str(error.exception))

    @mock.patch('ci_utils.launchpad.lp_login')
    def test_creates_no_need_to_modify(self, mocked_lp):
        # `modify_ppa` returns the just-created PPA when restricted
        # processors are not required despite the user not being a
        # member of `launchpad-ppa-self-admins` team
        fake_lp = FakeLaunchpad()
        a_user = FakeLaunchpadPerson('A_USER')
        b_user = FakeLaunchpadPerson('B_USER')
        c_user = FakeLaunchpadPerson('C_USER')
        fake_lp.people = {
            'A_OWNER': FakeLaunchpadPerson(
                self.test_config['launchpad_ppa_owner']),
            'A_USER': a_user,
            'launchpad-ppa-self-admins': FakeLaunchpadTeam(
                'launchpad-ppa-self-admins',
                [b_user, c_user])
        }
        fake_lp.distributions = {
            'ubuntu': FakeLaunchpadDistro('ubuntu')
        }
        mocked_lp.return_value = fake_lp

        self.assertEqual(
            'ppa:A_OWNER/A_PPA',
            create_ppa('A_PPA', 'A_DISPLAYNAME', {'arch1', 'arch2'}))
