# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import mock
import unittest

from ppacreator import run_worker
from ci_utils.json_status import JSONStatus


class TestPPACreatorWorker(unittest.TestCase):

    def setUp(self):
        super(TestPPACreatorWorker, self).setUp()
        self.worker = run_worker.PPACreatorWorker()

    @mock.patch('ppacreator.run_worker.create_ppa')
    @mock.patch('ci_utils.amqp_utils.progress_completed')
    @mock.patch('ci_utils.amqp_utils.progress_update')
    def test_handle_request(self, mocked_progress_update,
                            mocked_progress_completed,
                            mocked_create_ppa):

        mocked_log = mock.MagicMock()
        mocked_create_ppa.return_value = 'ppa:A_OWNER/A_PPA'
        params = {
            'name': 'fake-ppa',
            'displayname': 'fake-ppa-displayname',
            'description': 'fake-description',
            'private': True,
            'architectures': 'fake-architecture',
            'progress_trigger': 'ticket_1_trigger',
            'ticket_id': 'fake-ticket-id',
            'series': 'a_series'
        }
        retcode, results = self.worker.handle_request(params, mocked_log)

        # A message is passed to 'log.info' and to 'progress_update'
        # announcing the processing period.
        message = 'Creating PPA fake-ppa for ticket, fake-ticket-id'
        mocked_log.info.assert_called_once_with(message)

        self.assertEquals(mocked_progress_completed, retcode)
        self.assertEquals({'ppa_name': 'ppa:A_OWNER/A_PPA',
                           'unique_arches': 'fake-architecture'}, results)

    @mock.patch('ppacreator.run_worker.create_ppa')
    @mock.patch('ci_utils.amqp_utils.progress_failed')
    @mock.patch('ci_utils.amqp_utils.progress_update')
    def test_processing_error_handling(self,
                                       mocked_progress_update,
                                       mocked_progress_failed,
                                       mocked_create_ppa):

        mocked_log = mock.MagicMock()

        mocked_create_ppa.side_effect = [
            run_worker.PPACreationError('Processing Error!')
        ]

        params = {
            'name': 'fake-ppa',
            'displayname': 'fake-ppa-displayname',
            'description': 'fake-description',
            'private': True,
            'architectures': 'fake-architecture',
            'progress_trigger': 'ticket_1_trigger',
            'ticket_id': 'fake-ticket-id',
            'series': 'a_series'
        }
        retcode, results = self.worker.handle_request(params, mocked_log)

        self.assertEquals(mocked_progress_failed, retcode)
        self.assertEquals(
            'Processing Error!',
            results['message'])

    @mock.patch('ci_utils.json_status.JSONStatus.add_de_virt_ppa_capabilities')
    def test_engine_health(self, mocked_cap):
        self.assertIsInstance(self.worker.engine_health(), JSONStatus)
        mocked_cap.assert_called_once()
