#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import signal
import time

from ci_utils import (
    amqp_utils,
    health_check,
    ticket_states,
)


class Timeout(Exception):
    pass


TIMEOUT = 60 * 3  # 3 minutes.


def timeout_handler(sig_num, frame):
    raise Timeout('No worker responding after {} seconds'.format(TIMEOUT))


signal.signal(signal.SIGALRM, timeout_handler)
signal.alarm(TIMEOUT)


class MessageSync(object):

    progress = logging.getLogger('PROGRESS_TRIGGER')

    def __init__(self):
        self.data = None

    def on_message(self, msg):
        data = json.loads(msg.body)
        state = data['state']

        # print "state" as a label for the rest of the items in the list
        params = {x: y for (x, y) in data.iteritems() if x != 'state'}
        if params:
            self.progress.info('%s: %s', state, params)
        else:
            self.progress.info(state)

        msg.channel.basic_ack(msg.delivery_tag)
        if data.get('exit'):
            self.data = data
            # this will kick us out of the run-forever worker loop
            msg.channel.basic_cancel(msg.consumer_tag)

    @staticmethod
    def drain(queue):
        sync = MessageSync()
        cfg = amqp_utils.get_config()
        amqp_utils.process_queue(cfg, queue, sync.on_message, delete=True)
        return sync.data


def check_worker():
    queue_name = 'publisher-in-production-test-%d' % time.time()
    step = str(ticket_states.TicketWorkflowStep.PKG_PUBLISHING)
    input_ppa = 'ppa:canonical-ci-engineering/ci-airline-phase-0'
    output_ppa = 'ppa:canonical-ci-engineering/publisher-test-output'
    params = {
        'ticket_id': 'publisher-in-production-test',
        'progress_trigger': queue_name,
        'subtickets': [
            {'id': 1,
             'source_package_upload': {
                 'sourcepackage': {'name': 'uci-tests'}}}],
        'source_archive': input_ppa,
        'destination_archive': output_ppa,
        'series': 'precise',
        'current_workflow_step': step,
    }
    amqp_utils.send(
        amqp_utils.PUBLISHER_QUEUE, json.dumps(params), True)
    try:
        data = MessageSync.drain(queue_name)
        output = 'data returned: {}'.format(data)
        result = 'COMPLETED' == data['state'] and 'PASSED' == data['result']
    except Timeout as e:
        result = False
        output = unicode(e)
    return result, output


def main():
    with health_check.HealthCheck('publisher') as check:
        result, output = check_worker()
        if result:
            check.write_result_ok()
        else:
            check.write_result_failed(output)


if __name__ == '__main__':
    main()
