# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ci_utils.ticket_states import TicketWorkflowStep
from publisher import ppa_sync


class PublisherProcessError(Exception):
    """Raised when there is an error processing publisher request."""


class PublisherValidationError(Exception):
    """Raised when on publisher validation errors."""


class PublisherCopyError(Exception):
    """Raised when on publisher result copying errors."""


def _is_publish_time(current_workflow_step):
    """Whether the workflow step is PKG_PUBLISHING or not."""
    return (current_workflow_step == str(TicketWorkflowStep.PKG_PUBLISHING))


def copy_ticket_packages(
        source_archive, destination_archive, packages, series, status_cb=None):
    try:
        ppa_sync.copy(
            source_archive, destination_archive, packages, series, status_cb)
    except Exception as exc:
        raise PublisherCopyError(
            "Error copying packages: {}".format(exc.message))


def validate_subticket(subticket):
    """Validates `Subticket` information."""
    # Until the consistency checks are added, validate should always return
    # success.
    return True


def process_request(params, data_store, status_cb=None):
    """Process a publisher request."""
    try:
        # Required variables
        subtickets = params['subtickets']
        series = params['series']
        source_archive = params['source_archive']
        destination_archive = params['destination_archive']
        current_workflow_step = params['current_workflow_step']
    except KeyError as exc:
        raise PublisherProcessError(
            "Missing required variables: '{}'.".format(exc.message))

    # Register a no-op status-callback if it wasn't provided.
    if status_cb is None:
        status_cb = lambda msg: None

    packages = []
    out_data = {
        'checks': [],
        'copy': None,
    }
    # Validade all subtickets.
    for subticket in subtickets:
        package_name = (
            subticket['sourcepackage']['name'])
        subticket_results = {
            'subticket_id': subticket['id'],
            'package': package_name,
            'result': 'FAILED',
            'info': '',
        }
        try:
            validate_subticket(subticket)
        except PublisherValidationError as exc:
            subticket_results['result'] = 'FAILED'
            subticket_results['info'] = exc.message
            status_cb(
                'Validating source \'{}\': FAILED\nERROR: {}'.format(
                    package_name, exc.message))
        else:
            subticket_results['result'] = 'PASSED'
            status_cb(
                'Validating source \'{}\': PASSED'.format(package_name))

        packages.append(package_name)
        out_data['checks'].append(subticket_results)

    # Dismiss if the ticket cannot be published yet
    # XXX: This logic is flawed, every time the lander calls the publisher it
    # will be through _handle_publisher, and there we set the ticket to
    # publishing step. We need another handler to call when we're only
    # validating, removing the need of this, or we pass a variable directly
    # setting publish=True -- Ursinha (2014-06-23).
    if not _is_publish_time(current_workflow_step):
        return out_data

    # Dismiss if there are validation failures.
    if set([c['result'] for c in out_data['checks']]) != set(['PASSED']):
        status_cb('Cannot copy results due to validation errors.')
        return out_data

    # Try to copy ticket results to the destination archive.
    copy_results = {
        'result': 'FAILED',
        'packages': packages,
        'info': '',
    }
    try:
        copy_ticket_packages(
            source_archive, destination_archive, packages, series, status_cb)
    except PublisherCopyError as exc:
        copy_results['result'] = 'FAILED'
        copy_results['info'] = exc.message
        status_cb('Copying results: FAILED\nERROR: {}'.format(exc.message))
    else:
        copy_results['result'] = 'OK'
        status_cb('Copying results: OK')

    out_data['copy'] = copy_results

    return out_data
