#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# adapted from: http://code.launchpad.net/ppa-dev-tools

import logging
import subprocess

from lazr.restfulclient import errors

from ci_utils.launchpad import lp_login


def get_package_data(ppa, series, package):
    source = ppa.getPublishedSources(exact_match=True,
                                     source_name=package,
                                     distro_series=series)[0]
    publishing_id = str(source).rpartition('/')[-1]
    source_data = dict()
    source_data['package'] = package
    source_data['binaries'] = source.getPublishedBinaries()
    source_data['version'] = source.source_package_version
    source_data['records'] = ppa.getBuildRecords(source_name=package)
    source_data['build_status'] = ppa.getBuildSummariesForSourceIds(
        source_ids=publishing_id)[publishing_id]
    return source_data


def is_built(source_data):
    try:
        status = source_data['build_status']['status']
        if status not in 'FULLYBUILT':
            logging.info(
                'SKIP %s has status %s', source_data['package'], status)
            return False
        else:
            return True
    except KeyError:
        logging.info('SKIP %s not present in PPA', source_data['package'])
        return False


def records_built(records, package, version):
    arches = set()
    for record in records:
        title = record.title.split()
        # title[0]:arch
        # title[3] package name
        # title[4] version
        if package == title[3] and version == title[4]:
            if record.buildstate != 'Successfully built':
                raise EnvironmentError('%s - %s has not built yet' %
                                       (package, version))
            arches.add(title[0])
    if not arches:
        raise EnvironmentError('%s - %s seems to be a binary copy' %
                               (package, version))
    return arches


def records_published(binaries):
    built_arches = set()
    for build in binaries:
        if build.status == 'Pending':
            raise EnvironmentError('%s has not publised yet' %
                                   (build.display_name))
        if not build.architecture_specific:
            return None
        built_arches.add(build.display_name.split()[4])
    return built_arches


def _get_team_ppa(lp, ppaurl):
    # convert ppa:team/name into a tuple of (team, name)
    team, ppa = ppaurl.split(':')[1].split('/')
    team = lp.people[team]
    ppa = team.getPPAByName(name=ppa)
    return team, ppa


def _should_update(ppa, package, series, source_version):
    try:
        target = ppa.getPublishedSources(
            exact_match=True, source_name=package, distro_series=series)[0]
        if target.date_made_pending:
            return True
        tver = target.source_package_version
        rc = subprocess.call(
            ['dpkg', '--compare-versions', source_version, 'gt', tver])
        return rc == 0
    except:
        return True
    return False


def copy(source_ppa_url, target_ppa_url, packages, series, status_cb):
    if not status_cb:
        status_cb = lambda x: 0

    launchpad = lp_login()

    result = {}

    ppa_team, ppa = _get_team_ppa(launchpad, source_ppa_url)
    target_ppa_team, target_ppa = _get_team_ppa(launchpad, target_ppa_url)

    ubuntu = launchpad.distributions['ubuntu']
    series = ubuntu.getSeries(name_or_version=series)

    for package in packages:
        source_data = get_package_data(ppa, series, package)
        if not is_built(source_data):
            continue
        try:
            arches = records_built(source_data['records'], package,
                                   source_data['version'])
            # Arches built None means binary independent
            arches_built = records_published(source_data['binaries'])
        except EnvironmentError as exc:
            status_cb('Skipped {}: {}'.format(package, exc.message))
            continue

        if arches_built is not None and arches != arches_built:
            status_cb(
                '{} is not fully published in the source PPA'.format(package))
            continue

        status_cb('Copying {} - {} for arches {}'.format(
            package, source_data['version'], arches_built))
        try:
            # XXX: Change to copySources to do this atomically, if something
            # happens in the middle of this loop we'll have half packages
            # copied.
            target_ppa.syncSource(from_archive=ppa,
                                  include_binaries=True,
                                  source_name=package,
                                  to_pocket='Release',
                                  version=source_data['version'],
                                  to_series=series.name)
        except errors.Unauthorized as exc:
            # When a kernel package is being copied, launchpad can
            # fail with Unauthorized trying to copy d-is, that doesn't work
            # for PPAs. So here we check if the relevant source was copied
            # succesfully, and report success if so.
            #
            # I won't check if publishing record is PENDING, because we don't
            # know if publisher ran in the meantime and set it to PUBLISHED,
            # that would cause the query to return empty and give a false
            # negative. Also it's not possible that more than one record
            # per version exists on a PPA.
            copied_source = target_ppa.getPublishedSources(
                source_name=package, version=source_data['version'],
                distro_series=series)
            if not copied_source:
                status_cb(
                    'Failed to copy packages from {} to {}. User ~{} is not '
                    'allowed to copy to target PPA.'.format(
                        source_ppa_url, target_ppa_url, launchpad.me.name))
        except Exception as exc:
            # Record something wrong happened with a package but keeps trying
            # to copy others.
            result += "{}: {}\n".format(package, exc.message)

    return result
