# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import mock
import unittest

from ci_utils.ticket_states import TicketWorkflowStep
from publisher import (
    copy_ticket_packages,
    process_request,
    validate_subticket,
    PublisherCopyError,
    PublisherProcessError,
    PublisherValidationError,
)


class TestValidateSubticket(unittest.TestCase):

    def test_validate_subticket_always_returns_success(self):
        self.assertTrue(validate_subticket('whatever'))


class TestCopyTicketPackages(unittest.TestCase):

    def setUp(self):
        """Setup testing parameters."""
        super(TestCopyTicketPackages, self).setUp()
        self.source_archive = 'SRC_PPA'
        self.destination_archive = 'DST_PPA'
        self.packages = ['foobar', 'barfoo']
        self.series = 'A_SERIES'

    @mock.patch('publisher.ppa_sync.copy')
    def test_copy_ticket_packages(self, mocked_copy):
        mocked_copy.return_value = {}
        mocked_cb = mock.MagicMock()
        copy_ticket_packages(
            self.source_archive, self.destination_archive, self.packages,
            self.series, mocked_cb)
        mocked_copy.assert_called_once_with(
            self.source_archive, self.destination_archive, self.packages,
            self.series, mocked_cb)

    @mock.patch('publisher.ppa_sync.copy')
    def test_copy_ticket_packages_fails(self, mocked_copy):
        mocked_cb = mock.MagicMock()
        mocked_copy.side_effect = [
            Exception("Can't connect to Launchpad")
        ]
        with self.assertRaises(PublisherCopyError) as cm:
            copy_ticket_packages(
                self.source_archive, self.destination_archive, self.packages,
                self.series, mocked_cb)

        self.assertEquals(
            "Error copying packages: Can't connect to Launchpad",
            cm.exception.message)


class TestProcessRequest(unittest.TestCase):

    def setUp(self):
        """Setup testing parameters."""
        super(TestProcessRequest, self).setUp()
        self.params = {
            'ticket_id': 1,
            'progress_trigger': 'A_TRIGGER',
            'subtickets': [{
                'id': 1,
                'sourcepackage': {'name': 'foobar'},
                'source_package_uploads': [
                    {'version': '1.0'},
                ]
            }, {
                'id': 2,
                'sourcepackage': {'name': 'barfoo'},
                'source_package_uploads': [
                    {'version': '1.0'},
                ],
            }],
            'source_archive': 'SRC_PPA',
            'destination_archive': 'DST_PPA',
            'series': 'A_SERIES',
            'current_workflow_step': 'Queued',
        }

    def test_process_request_missing_variables(self):
        # 'process_request' will raise `PublisherProcessError` on
        # missing parameters.
        missing_params = self.params.copy()
        del missing_params['series']
        with self.assertRaises(PublisherProcessError) as cm:
            process_request(missing_params, None, None)
        self.assertEquals(
            "Missing required variables: 'series'.",
            cm.exception.message)

    @mock.patch('publisher.copy_ticket_packages')
    @mock.patch('publisher.validate_subticket')
    def test_process_request_check_only(self, mocked_validate_subticket,
                                        mocked_copy):
        # 'process_request' validates all given `Subtickets`.
        mocked_ds = mock.MagicMock()
        mocked_cb = mock.MagicMock()
        self.assertEquals({
            'checks': [
                {'subticket_id': 1,
                 'package': 'foobar',
                 'result': 'PASSED',
                 'info': ''},
                {'subticket_id': 2,
                 'package': 'barfoo',
                 'result': 'PASSED',
                 'info': ''},
            ],
            'copy': None,
        }, process_request(self.params, mocked_ds, mocked_cb))
        # Validation and status_cb are called for each subticket and 'copy'
        # is not called.
        self.assertEquals(2, mocked_validate_subticket.call_count)
        self.assertEquals(2, mocked_cb.call_count)
        self.assertEquals(0, mocked_copy.call_count)
        # XXX cprov 2014-06-02: data_store is currently unused.
        self.assertEquals(0, mocked_ds.call_count)

    @mock.patch('publisher.copy_ticket_packages')
    @mock.patch('publisher.validate_subticket')
    def test_process_request_publishing(self, mocked_validate_subticket,
                                        mocked_copy):
        # `process_request` will publish (copy) ticket results if the
        # ticket is in 'PKG_PUBLISHING' status and if validation passes.
        mocked_ds = mock.MagicMock()
        mocked_cb = mock.MagicMock()
        # Modify testing parameters to trigger publication ('copy').
        pub_params = self.params.copy()
        pub_params['current_workflow_step'] = str(
            TicketWorkflowStep.PKG_PUBLISHING)
        # Return data contains copy information.
        self.assertEquals({
            'checks': [
                {'subticket_id': 1,
                 'package': 'foobar',
                 'result': 'PASSED',
                 'info': ''},
                {'subticket_id': 2,
                 'package': 'barfoo',
                 'result': 'PASSED',
                 'info': ''},
            ],
            'copy': {'info': '',
                     'packages': ['foobar', 'barfoo'],
                     'result': 'OK'},
        }, process_request(pub_params, mocked_ds, mocked_cb))
        self.assertEquals(2, mocked_validate_subticket.call_count)
        self.assertEquals(3, mocked_cb.call_count)
        mocked_copy.assert_called_once_with(
            pub_params['source_archive'],
            pub_params['destination_archive'],
            [s['sourcepackage']['name'] for s in pub_params['subtickets']],
            pub_params['series'], mocked_cb)

    @mock.patch('publisher.copy_ticket_packages')
    @mock.patch('publisher.validate_subticket')
    def test_validation_error(self, mocked_validate_subticket, mocked_copy):
        # `process_request` copes with validation errors.
        mocked_ds = mock.MagicMock()
        mocked_cb = mock.MagicMock()
        mocked_validate_subticket.side_effect = [
            PublisherValidationError('Cannot Pass!')
        ]
        failed_params = self.params.copy()
        failed_params['subtickets'] = [
            {'id': 1,
             'sourcepackage': {'name': 'foobar'},
             'source_package_upload': [{'version': '1.0'}]},
        ]
        self.assertEquals({
            'checks': [
                {'subticket_id': 1,
                 'package': 'foobar',
                 'result': 'FAILED',
                 'info': 'Cannot Pass!'},
            ],
            'copy': None,
        }, process_request(failed_params, mocked_ds, mocked_cb))
        mocked_validate_subticket.assert_called_once_with(
            failed_params['subtickets'][0])
        mocked_cb.assert_called_once_with(
            "Validating source 'foobar': FAILED\nERROR: Cannot Pass!")

    @mock.patch('publisher.copy_ticket_packages')
    @mock.patch('publisher.validate_subticket')
    def test_publishing_error(self, mocked_validate_subticket, mocked_copy):
        # `process_request` copes with copying errors.
        mocked_ds = mock.MagicMock()
        mocked_cb = mock.MagicMock()
        mocked_copy.side_effect = [
            PublisherCopyError('Cannot Copy!')
        ]
        failed_params = self.params.copy()
        failed_params['subtickets'] = [
            {'id': 1,
             'sourcepackage': {'name': 'foobar'},
             'source_package_upload': [{'version': '1.0'}]},
        ]
        failed_params['current_workflow_step'] = str(
            TicketWorkflowStep.PKG_PUBLISHING)
        self.assertEquals({
            'checks': [
                {'subticket_id': 1,
                 'package': 'foobar',
                 'result': 'PASSED',
                 'info': ''},
            ],
            'copy': {'packages': ['foobar'],
                     'result': 'FAILED',
                     'info': 'Cannot Copy!'},
        }, process_request(failed_params, mocked_ds, mocked_cb))
        self.assertEquals(2, mocked_cb.call_count)
        mocked_cb.assert_called_with(
            'Copying results: FAILED\nERROR: Cannot Copy!')

    @mock.patch('publisher.copy_ticket_packages')
    @mock.patch('publisher.validate_subticket')
    def test_no_publishing_invalid(self, mocked_validate_subticket,
                                   mocked_copy):
        # `process_request` does not publish after validation errors.
        mocked_ds = mock.MagicMock()
        mocked_cb = mock.MagicMock()
        mocked_validate_subticket.side_effect = [
            PublisherValidationError('Cannot Publish!')
        ]
        failed_params = self.params.copy()
        failed_params['subtickets'] = [
            {'id': 1,
             'sourcepackage': {'name': 'foobar'},
             'source_package_upload': [{'version': '1.0'}]},
        ]
        failed_params['current_workflow_step'] = str(
            TicketWorkflowStep.PKG_PUBLISHING)
        self.assertEquals({
            'checks': [
                {'subticket_id': 1,
                 'package': 'foobar',
                 'result': 'FAILED',
                 'info': 'Cannot Publish!'},
            ],
            'copy': None,
        }, process_request(failed_params, mocked_ds, mocked_cb))
        self.assertEquals(2, mocked_cb.call_count)
        mocked_cb.assert_called_with(
            'Cannot copy results due to validation errors.')
