# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import mock
import unittest

from publisher import run_worker


class TestPublisherWorker(unittest.TestCase):

    def setUp(self):
        super(TestPublisherWorker, self).setUp()
        self.worker = run_worker.PublisherWorker('publisher')

    @mock.patch('publisher.run_worker.process_request')
    @mock.patch('ci_utils.amqp_utils.progress_completed')
    @mock.patch('ci_utils.amqp_utils.progress_update')
    @mock.patch('ci_utils.amqp_worker.AMQPWorker._create_data_store')
    def test_handle_request(self, mocked_create_data_store,
                            mocked_progress_update,
                            mocked_progress_completed,
                            mocked_process_request):
        # `PublisherWorker.handle_request` validates initial data, creates
        # a ticket data store and dispatches `publisher.process_request`.
        mocked_log = mock.MagicMock()
        mocked_ds = mock.MagicMock()
        mocked_create_data_store.return_value = mocked_ds
        mocked_process_request.return_value = {
            'checks': ['results'], 'copy': None
        }
        params = {
            'ticket_id': 1, 'progress_trigger': 'A_TRIGGER'
        }
        result, output = self.worker.handle_request(params, mocked_log)
        # `process_request` is called.
        # XXX cprov 2014-06-02: cannot assert passed arguments because the
        # local 'status_cb' is unreachable.
        mocked_process_request.assert_called_once()
        # A ticket data_store acessor is created.
        mocked_create_data_store.assert_called_once_with(
            'ticket-{}'.format(params['ticket_id']))
        # A message is passed to 'log.info' and to 'progress_update'
        # announcing the processing period.
        message = 'Processing ticket 1:'
        mocked_log.info.assert_called_once_with(message)
        mocked_progress_update.assert_called_once_with(
            params['progress_trigger'], {'message': message})
        # The result contains the 'completed' progress indicator and the
        # mocked request output.
        self.assertEquals(mocked_progress_completed, result)
        self.assertEquals(
            {'checks': ['results'], 'copy': None}, output)

    @mock.patch('ci_utils.amqp_utils.progress_failed')
    def test_handle_request_invalid_input(self, mocked_progress_failed):
        # `PublisherWorker.handle_request` requires 'ticket_id' and
        # 'progress_trigger' parameters
        params = {
            'ticket_id': 1,
        }
        result, output = self.worker.handle_request(params, None)
        # The result contains the 'failed' progress indicator and
        # the failure information.
        self.assertEquals(mocked_progress_failed, result)
        self.assertEquals(
            {'message': 'Invalid input: progress_trigger',
             'status': 'FAILED'},
            output)

    @mock.patch('publisher.run_worker.process_request')
    @mock.patch('ci_utils.amqp_utils.progress_failed')
    @mock.patch('ci_utils.amqp_utils.progress_update')
    @mock.patch('ci_utils.amqp_worker.AMQPWorker._create_data_store')
    def test_processing_error_handling(self, mocked_create_data_store,
                                       mocked_progress_update,
                                       mocked_progress_failed,
                                       mocked_process_request):
        # `PublisherWorker.handle_request` handles `process_request`
        # errors.
        mocked_log = mock.MagicMock()
        mocked_ds = mock.MagicMock()
        mocked_create_data_store.return_value = mocked_ds
        mocked_process_request.side_effect = [
            run_worker.PublisherProcessError('Processing Error!')
        ]

        params = {
            'ticket_id': 1, 'progress_trigger': 'A_TRIGGER',
        }
        result, output = self.worker.handle_request(params, mocked_log)
        # The result contains the 'failed' progress indicator and
        # the failure information.
        self.assertEquals(mocked_progress_failed, result)
        self.assertEquals(
            {'message': 'Processing Error!', 'status': 'FAILED'},
            output)
