#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import signal
import uuid


from ci_utils import (
    amqp_utils,
    health_check,
    image_store,
)


class Timeout(Exception):
    pass


# 7 minutes should be enough to process libpng on trusty
TIMEOUT = 60 * 7  # 420 seconds


# FIXME: This is wrong by design: it could happen (and it did in real life)
# that no worker is available for 7 minutes. This leads to a timeout and the
# progress queue staying alive without anybody subscribing to it. When the
# request is processed later, it left the expected 6 messages hanging around in
# the queue... -- vila 2014-10-31
def timeout_handler(sig_num, frame):
    raise Timeout('No worker responding after {} seconds'.format(TIMEOUT))


signal.signal(signal.SIGALRM, timeout_handler)
signal.alarm(TIMEOUT)


class MessageSync(object):
    progress = logging.getLogger('PROGRESS_TRIGGER')

    def __init__(self):
        self.data = None

    def on_message(self, msg):
        data = json.loads(msg.body)
        state = data['state']

        # print "state" as a label for the rest of the items in the list
        params = {x: y for (x, y) in data.iteritems() if x != 'state'}
        if params:
            self.progress.info('%s: %s', state, params)
        else:
            self.progress.info(state)

        msg.channel.basic_ack(msg.delivery_tag)
        if data.get('exit'):
            self.data = data
            # this will kick us out of the run-forever worker loop
            msg.channel.basic_cancel(msg.consumer_tag)

    @staticmethod
    def drain(queue):
        sync = MessageSync()
        cfg = amqp_utils.get_config()
        amqp_utils.process_queue(cfg, queue, sync.on_message, delete=True)
        return sync.data


def check_worker():
    queue = 'testrun-test-{}'.format(uuid.uuid4())
    series = 'trusty'
    architecture = 'amd64'
    image_id = image_store.uci_image_name('cloudimg', series, architecture)
    params = {
        'ticket_id': 'test-runner-test',
        'image_id': image_id,
        'series': series,
        'architecture': architecture,
        'package_list': ['libpng'],
        'ppa_list': [],
        'progress_trigger': queue,
    }
    amqp_utils.send(
        amqp_utils.TEST_RUNNER_QUEUE, json.dumps(params), True)
    try:
        data = MessageSync.drain(queue)
        output = 'data returned: {}'.format(data)
        result = 'COMPLETED' == data['state'] and 'PASSED' == data['result']
    except Timeout as e:
        logging.exception('Run took too long')
        result = False
        output = unicode(e)
    return result, output


def main():
    logging.basicConfig(level=logging.INFO,
                        format="%(asctime)s %(levelname)s %(message)s")
    with health_check.HealthCheck('test_runner') as check:
        result, output = check_worker()
        if result:
            check.write_result_ok()
        else:
            check.write_result_failed(output)


if __name__ == '__main__':
    main()
