#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import os
import subprocess
import subunit
from testtools import content
import unittest


def run(*args, **kwargs):
    """Run a command in a subprocess.

    :param args: The command and its positional arguments.

    :param kwargs: The named arguments for the command.

    The stdout and stderr outputs are captured and returned to the caller.
    """
    check_rc = kwargs.pop('check_rc', True)
    cwd = kwargs.pop('cwd', None)
    proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, cwd=cwd)
    out, err = proc.communicate()
    if check_rc and proc.returncode:
        raise Exception(
            '{} failed with return code: {}'.format(' '.join(args),
                                                    proc.returncode))
    return proc, out, err


def parse_summary_line(line):
    try:
        name, status = line.split(None, 1)
    except ValueError:
        # Irregular line, set an arbitrary status not conflicting with the real
        # ones and report the whole line as wrong.
        return (line, 'error')
    if name == '*' and status.startswith('SKIP '):
        # Tweak a bit to normalize the status
        name = '*' + status[4:]
        status = 'SKIP'
    elif status.startswith('FAIL'):
        # Remove spurious info, more info is available in the details
        # (stdout/stderr for the corresponding test)
        status = 'FAIL'
    return (name, status)


def parse_summary(path):
    tests = []
    with open(path) as f:
        for line in f:
            line = line.strip()
            tests.append(parse_summary_line(line))
    return tests


def get_test_case(name):

    class Test(unittest.TestCase):

        def id(self):
            return name

    # We wont run that test, we only use it as a way to interact with subunit
    # in a simple and supported way.
    return Test('id')


def get_details(name):
    details = {}
    out_path = os.path.join('results', '{}-stdout'.format(name))
    if os.path.exists(out_path):
        with open(out_path) as f:
            details['stdout'] = content.text_content(f.read())
    err_path = os.path.join('results', '{}-stderr'.format(name))
    if os.path.exists(err_path):
        with open(err_path) as f:
            details['stderr'] = content.text_content(f.read())
    return details


def process_test_result(package, result, name, status):
    if status == 'SKIP':
        # adt-run doesn't give a name to the test here (there are none...) we
        # fake one.
        t = get_test_case('{}.skipped'.format(package))
    else:
        t = get_test_case(name)
    result.startTest(t)
    if status == 'PASS':
        result.addSuccess(t, details=get_details(name))
    elif status == 'FAIL':
        result.addFailure(t, details=get_details(name))
    elif status == 'SKIP':
        result.addSkip(t, reason=name)
    else:
        # We are catching unexpected lines in 'summary.log' here, some of them
        # may be valid (but yet unknown) cases. We'll fix them as they are
        # discovered. By turning them into an erroring test, we know the user
        # will see them and can report accordingly.
        error = 'adt-run produced an unexpected result: "{}"'.format(name)
        details = dict(error=content.text_content(error))
        result.addError(t, details=details)
    result.stopTest(t)


def aggregate_versions(dir_path):
    """Aggregate all versions used for the tests.

    :param dir_path: The results directory produced by adt-run.

    :returns: A list of (package, version) sorted by package name without
         duplicates.

    This relies on adt-run producing a 'testbed-package' and '<test>-packages'
    files describing the (package, versions) used for each test.

    """
    # Use a dict to remove duplicates
    versions = {}
    for file_path in os.listdir(dir_path):
        if not file_path.endswith('-packages'):
            continue
        with open(os.path.join(dir_path, file_path)) as f:
            for line in f:
                package, version = line.split()
                versions[package] = version
    return [(k, versions[k]) for k in sorted(versions.keys())]


def package_version(dir_path):
    """Extract the package version.

    :param dir_path: The results directory produced by adt-run.

    :retun: The package version or None if it can't be found.
    """
    try:
        with open(os.path.join(dir_path, 'testpkg-version')) as f:
            content = f.read()
    except IOError as e:
        if e.errno != errno.ENOENT:
            raise
        # No file ? No version.
        return None
    try:
        version = content.split()[1]
    except IndexError:
        # Bad content ? No version
        return None
    return version


def run_test(ip, login, ssh_key_path, package, adt_opts):
    subunit_path = '{}.subunit'.format(package)
    output = open(subunit_path, 'wb')
    result = subunit.TestProtocolClient(output)
    result.startTestRun()
    pkg_version = None
    deps_versions = []
    try:
        summary_path = '{}-summary.log'.format(package)
        log_path = '{}.log'.format(package)
        cmd = ['adt-run',
               '--apt-source', package,
               '--summary', summary_path,
               '--log-file', log_path,
               '--output-dir', 'results']
        if adt_opts:
            cmd.extend(adt_opts)
        cmd.extend(['---', 'adt-virt-ssh',
                    '-H', ip, '-l', login, '-i', ssh_key_path])
        proc, _, _ = run(*cmd, check_rc=False)
        rc = proc.returncode
        pkg_version = package_version('results')
        deps_versions = aggregate_versions('results')
        for name, status in parse_summary(summary_path):
            process_test_result(package, result, name, status)
    finally:
        result.stopTestRun()
        output.close()
    return rc, pkg_version, deps_versions
