#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals


import errno
import logging
import os


from ci_utils import (
    amqp_utils,
    amqp_worker,
    image_store,
    json_status,
)
from ci_utils import unit_config
from tstrun import (
    run_test,
    testbed,
)


class TestRequestFailed(Exception):

    fmt = '{reason} failed for {ticket_id}'

    def __init__(self, reason, ticket_id):
        self.reason = reason
        self.ticket_id = ticket_id

    def __unicode__(self):
        return self.fmt.format([], **self.__dict__)

    __repr__ = __unicode__


def format_britney_result(series, arch, package, retcode,
                          pkg_version, deps_versions):
    content = [series, arch]
    if pkg_version is not None and deps_versions:
        if retcode in (0, 2):
            result = 'PASS'
        else:
            result = 'FAIL'
        content.extend([package, pkg_version])
        content.append(result)
        # Flatten the (package, version) tuples
        for p, v in deps_versions:
            content.extend([p, v])
    else:
        content.append(package)
    return ' '.join(content)


class TestRequestHandler(object):

    def __init__(self, logger, progress_queue_name,
                 data_store, ticket_id, artifact_prefix):
        self.logger = logger
        self.progress_queue_name = progress_queue_name
        self.data_store = data_store
        self.ticket_id = ticket_id
        self.artifact_prefix = artifact_prefix
        self.test_bed = None
        self.artifacts = []

    def status_cb(self, msg):
        self.logger.info(msg)
        amqp_utils.progress_update(self.progress_queue_name, {'message': msg})

    def save_artifact(self, name, value, description=None,
                      kind='LOGS'):
        """Save an artifact catching and reporting exceptions.

        This should only be used to upload artifacts after the pass/fail status
        for the test run has been established so it's safe to fail the upload.

        :param name: The artifact name.

        :param value: The artifact value as a string.

        :param description: For reporting purposes.

        :param kind: The kind of artifact (defaults to 'LOGS').
        """
        if description is None:
            description = name
        try:
            self.logger.info('Saving {}'.format(description))
            url = self.data_store.put_file(name, value, 'text/plain')
            artifact_name = '{}.{}'.format(self.artifact_prefix, name)
            self.artifacts.append(
                dict(name=artifact_name, reference=url, type=kind))
        except:
            self.logger.exception('Unable to upload {}'.format(description))

    def save_testbed_console(self, package):
        self.save_artifact('{}-testbed-cloud-init.log'.format(package),
                           self.test_bed.get_cloud_init_console(),
                           '{} testbed console'.format(package))

    def save_britney_result(self, package, return_code, pkg_version,
                            deps_versions):
        # FIXME: We generate the britney file here as we don't copy all
        # files from the testbed to the britney results service via the
        # swift containers and renames. We need a better ticket container
        # layout and/or a way to trigger specific actions. In the mean
        # time, that will do. --vila 2014-10-17
        cfg = self.test_bed.conf.get
        content = format_britney_result(
            cfg('vm.release'), cfg('vm.architecture'), package,
            return_code, pkg_version, deps_versions)
        self.save_artifact(
            '{}.britney.results'.format(package), content,
            'britney results for {}'.format(package),
            'RESULTS')

    def save_testbed_artifacts(self, package):
        subunit_path = '{}.subunit'.format(package)
        self.save_artifact(
            subunit_path, open(subunit_path).read(),
            'subunit results for {}'.format(package),
            'RESULTS')
        package_log_path = '{}.log'.format(package)
        self.save_artifact(
            package_log_path, open(package_log_path).read(),
            'adt-run log for {}'.format(package))
        summary_log_path = '{}-summary.log'.format(package)
        self.save_artifact(
            summary_log_path, open(summary_log_path).read(),
            'adt-run summary for {}'.format(package))
        syslog_log_path = '{}-syslog.log'.format(package)
        self.save_artifact(
            syslog_log_path,
            self.test_bed.get_remote_content('/var/log/syslog'),
            'syslog for {}'.format(package))

    def setup_testbed(self, series, architecture, image_id, ppa_list, package):
        try:
            self.status_cb('Setting up the testbed for ticket {}'.format(
                self.ticket_id))
            auth_conf = unit_config.get_auth_config()
            tb_name = '{}-testbed-{}'.format(package, self.ticket_id)
            conf = testbed.vms_config_from_auth_config(tb_name, auth_conf)
            conf.set('vm.image', image_id)
            conf.set('vm.release', series)
            # This is for documentation purposes as the glance image is where
            # the architecture is really defined.
            conf.set('vm.architecture', architecture)
            conf.set('vm.ppas', ','.join(ppa_list))
            base_dir = conf.get('vm.vms_dir')
            try:
                # Try to create needed dirs
                os.makedirs(base_dir)
            except OSError as e:
                # They are already there, no worries
                if e.errno != errno.EEXIST:
                    raise
            conf.store.save_changes()
            self.test_bed = testbed.TestBed(conf, self.logger)
        except:
            msg = 'The testbed creation for ticket: {}, package: {} failed'
            self.logger.exception(msg.format(self.ticket_id, package))
            raise TestRequestFailed('The testbed creation', self.ticket_id)

        try:
            self.test_bed.setup()
            # Save the changes that can happen during the setup (mostly for
            # debug)
            conf.store.save_changes()
            self.save_testbed_console(package)
        except:
            self.logger.exception(
                'The testbed setup for {} failed'.format(self.ticket_id))
            if self.test_bed.instance is not None:
                self.save_testbed_console(package)
            raise TestRequestFailed('The testbed setup', self.ticket_id)

        self.status_cb('The test bed is ready')

    def run_tests(self, package, adt_opts):
        try:
            self.status_cb('Testing {}'.format(package))
            # uci-vms shell adt-run ...  --- adt-virt-null
            test_result = run_test.run_test(self.test_bed.ip, 'ubuntu',
                                            self.test_bed.test_bed_key_path,
                                            package, adt_opts)
            self.save_testbed_artifacts(package)
            (return_code, pkg_version, deps_versions) = test_result
            self.save_britney_result(package, return_code,
                                     pkg_version, deps_versions)
            self.status_cb('Test completed for ticket {}'.format(
                self.ticket_id))
            return return_code
        except:
            self.logger.exception(
                'Exception for ticket: {}, package: {}'.format(
                    self.ticket_id, package))
            raise TestRequestFailed('Testing {}'.format(package),
                                    self.ticket_id)

    def teardown_testbed(self):
        """Tearing down the testbed.

        This should never fail or even a passing test would be reported as
        failing for a wrong reason (all known cases are caused by nova
        transient failures).
        """
        if self.test_bed is not None:
            try:
                self.test_bed.teardown()
            except:
                msg = 'Tear down testbed failed, nova id: {}, floating ip: {}'
                if self.test_bed.instance is not None:
                    tb_id = self.test_bed.instance.id
                else:
                    tb_id = '<unknown>'
                fl_ip = self.test_bed.floating_ip or '<unknown>'
                self.logger.exception(msg.format(tb_id, fl_ip))
            self.test_bed = None

    def handle(self, params, package):
        # There are fields in the message that are not relevant to the test
        # runner ('state' for one) so we extract only the ones we need.
        series = params['series']
        architecture = params['architecture']
        image_id = params['image_id']
        ppa_list = params.get('ppa_list', [])
        adt_opts = params.get('adt_opts', [])

        try:
            self.setup_testbed(series, architecture, image_id,
                               ppa_list, package)
            retcode = self.run_tests(package, adt_opts)
            return retcode, self.artifacts
        except TestRequestFailed:
            return 'FAILED', self.artifacts
        finally:
            self.teardown_testbed()


# FIXME: This could/should become a nagios check (in-production test) when we
# figure out the best place to call it -- vila 2014-10-15
class TestRunnerJSONStatus(json_status.JSONStatus):

    def add_testbed_health(self):
        """Check that a testbed can be created/deleted.

        This requires a glance image that can be used to boot a nova instance
        and will issue an apt-get update via ssh.

        With that, glance, nova, archive reachability and connection via ssh
        between the worker and the nova instance are validated.
        """
        try:
            auth_conf = unit_config.get_auth_config()
            tb_name = 'testbed-health-check'
            conf = testbed.vms_config_from_auth_config(tb_name, auth_conf)
            series = 'precise'
            architecture = 'amd64'
            image_id = image_store.uci_image_name('cloudimg',
                                                  series, architecture)
            conf.set('vm.release', series)
            conf.set('vm.architecture', architecture)
            conf.set('vm.image', image_id)
            base_dir = conf.get('vm.vms_dir')
            if not os.path.exists(base_dir):
                os.makedirs(base_dir)
            test_bed = testbed.TestBed(conf)
            test_bed.setup()
            test_bed.teardown()
            conf.store.save_changes()
            self.add_okay('nova provisioning', 'functional')
        except Exception as e:
            self.add_fail('nova provisioning', unicode(e))


class TestRunnerWorker(amqp_worker.AMQPWorker):

    def __init__(self, _data_store_factory=None):
        """Test runner worker.

        :param _data_store_factory: A callable returning a ticket specific data
            store (intended to be used by tests).
        """
        super(TestRunnerWorker, self).__init__('image_test')
        self.data_store_factory = _data_store_factory
        self.data_store = None

    def _create_data_store(self, ticket_id):
        if self.data_store_factory is None:
            self.data_store = super(TestRunnerWorker,
                                    self)._create_data_store(ticket_id)
        else:
            self.data_store = self.data_store_factory(ticket_id)
        return self.data_store

    def engine_health(self):
        status = TestRunnerJSONStatus()
        status.add_deployed_revision()
        status.add_data_store_health()
        status.add_image_store_health()
        status.add_testbed_health()
        return status

    def handle_request(self, params, logger):
        logger.info('Running in {} with HOME: {}'.format(os.getcwd(),
                                                         os.environ['HOME']))
        ticket_id = params['ticket_id']
        progress_queue_name = params['progress_trigger']
        package_list = params['package_list']
        data_store = self._create_data_store(params['ticket_id'])

        # The request parameters are forwarded to the rabbit worker processing
        # the status/result queue to enable unidirectional communication
        # between the requester and the ultimate receiver.
        results = dict(artifacts=[], request=params)

        # The tests will succeed unless they fail ;)
        notify = amqp_utils.progress_completed
        for package in package_list:
            if params.get('cancelled'):
                self.logger.error('The request for {} has been cancelled,'
                                  ' exiting'.format(ticket_id))
                return amqp_utils.progress_failed, results
            req_handler = TestRequestHandler(logger, progress_queue_name,
                                             data_store, ticket_id,
                                             self.logger_name)
            return_code, artifacts = req_handler.handle(params, package)
            # 0 is success, 2 means "at least one test skipped" and is
            # considered a success
            if return_code not in (0, 2):
                # At least one test failed
                notify = amqp_utils.progress_failed
            results['artifacts'].extend(artifacts)
        return notify, results


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO,
                        format="%(asctime)s %(levelname)s %(message)s")
    # We're started in the charm code directory, let's go to a safer place
    home = os.environ.get('HOME', None)
    if home is None:
        logging.warning('$HOME not set ! Going to /tmp')
        home = '/tmp'
    os.chdir(home)
    TestRunnerWorker().main(amqp_utils.TEST_RUNNER_QUEUE)
