#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import errno
import logging
import os
import subprocess
import time

import requests

from novaclient import (
    client,
    exceptions,
)
from uciconfig import options
from ucivms import (
    config,
    vms,
)
from ci_utils import unit_config


HERE = os.path.abspath(os.path.dirname(__file__))


def register(option):
    options.option_registry.register(option)


register(options.Option('os.username', default_from_env='GLANCE_OS_USERNAME',
                        default=options.MANDATORY,
                        help_string='''The Open Stack user name.

This is generally set via GLANCE_OS_USERNAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('os.password', default_from_env='GLANCE_OS_PASSWORD',
                        default=options.MANDATORY,
                        help_string='''The Open Stack password.

This is generally set via GLANCE_OS_PASSWORD, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('os.region_name',
                        default_from_env='GLANCE_OS_REGION_NAME',
                        default=options.MANDATORY,
                        help_string='''The Open Stack region name.

This is generally set via GLANCE_OS_REGION_NAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('os.tenant_name',
                        default_from_env='GLANCE_OS_TENANT_NAME',
                        default=options.MANDATORY,
                        help_string='''The Open Stack tenant name.

This is generally set via GLANCE_OS_TENANT_NAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('os.auth_url', default_from_env='GLANCE_OS_AUTH_URL',
                        default=options.MANDATORY,
                        help_string='''The Open Stack keystone url.

This is generally set via GLANCE_OS_AUTH_URL, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.ListOption('vm.flavors', default=None,
                            help_string='''\
A list of flavors for all supported clouds.

The first known one is used.
'''))
register(options.Option('vm.image', default=options.MANDATORY,
                        help_string='''The base image to run the tests.'''))
register(options.Option('vm.net_id', default=None,
                        help_string='''The network id for the testbed.'''))
register(config.PathOption('vm.ssh_key_path', default='~/.ssh/id_rsa',
                           help_string='''The ssh key for the testbed.'''))
register(options.ListOption('vm.apt_get.update.timeouts',
                            default='15.0, 90.0, 240.0',
                            help_string='''apt-get update timeouts in seconds.

When apt-get update fails on hash sum mismatches, retry after the specified
timeouts. More values mean more retries.
'''))
register(options.ListOption('vm.ppas',
                            default=None,
                            help_string='''PPAs to be added to the testbed.

This works around cloud-init not activating the deb-src line and not providing
a way to do so. This is intended to be fixed in uci-vms so vm.apt_sources can
be used again.
'''))
register(options.Option('vm.nova.boot_timeout', default='300',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time to boot a nova instance (in seconds).'''))
register(options.Option('vm.nova.set_ip_timeout', default='300',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time for a nova instance to get an IP (in seconds).'''))
register(options.Option('vm.nova.cloud_init_timeout', default='1200',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time for cloud-init to fisnish (in seconds).'''))


class TestBedException(Exception):
    pass


def vms_config_from_auth_config(testbed_name, auth_conf):
    """Populate a ucivms config from an auth config.

    :param testbed_name: The name of the test bed in the ucivms config file.

    :param auth_conf: A dict with the openstack credentials.
    """
    conf = config.VmStack(testbed_name)
    conf.set('vm.name', testbed_name)
    conf.set('os.username', auth_conf.get('glance_auth_user'))
    conf.set('os.password', auth_conf.get('glance_auth_password'))
    conf.set('os.tenant_name', auth_conf.get('glance_auth_tenant_name'))
    conf.set('os.region_name', auth_conf.get('glance_auth_region'))
    conf.set('os.auth_url', auth_conf.get('glance_auth_url'))
    net_id = auth_conf.get('ephemeral_cloud_net_id', None)
    if net_id:
        # Only set this if defined
        conf.set('vm.net_id', net_id)
    conf.set('vm.flavors', ','.join(auth_conf.get('tr_flavors')))
    # User can override locally in unit_config
    conf.set('vm.ssh_key_path',
             unicode(auth_conf.get('auth_testbed_sshkey',
                                   '~/.ssh/id_rsa')))
    return conf


class NovaClient(object):
    """A nova client re-trying requests on known transient failures."""

    def __init__(self, *args, **kwargs):
        self.logger = kwargs.pop('logger')
        debug = kwargs.pop('debug', False)
        # Activating debug will output the http requests issued by nova and the
        # corresponding responses (/!\ including credentials).
        if debug:
            self.logger.setLevel(logging.DEBUG)
        self.nova = client.Client('1.1', *args, http_log_debug=debug,
                                  **kwargs)

    def retry(self, func, *args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.ConnectionError:
            # Most common transient failure: the API server is unreachable
            nap_time = 1
            self.logger.warn('Received connection error for {},'
                             ' retrying)'.format(func.__name__))
        except exceptions.OverLimit:
                # This happens rarely but breaks badly if not caught. elmo
                # recommended a 30 seconds nap in that case.
                nap_time = 30
                msg = ('Rate limit reached for {},'
                       ' will sleep for {} seconds')
                self.logger.exception(msg.format(func.__name__, nap_time))
        time.sleep(nap_time)
        return func(*args, **kwargs)  # Retry once

    def flavors_list(self):
        return self.retry(self.nova.flavors.list)

    def images_list(self):
        return self.retry(self.nova.images.list)

    def create_server(self, name, flavor, image, user_data, nics,
                      availability_zone):
        return self.retry(self.nova.servers.create, name=name,
                          flavor=flavor, image=image, userdata=user_data,
                          nics=nics, availability_zone=availability_zone)

    def delete_server(self, server_id):
        return self.retry(self.nova.servers.delete, server_id)

    def create_floating_ip(self):
        return self.retry(self.nova.floating_ips.create)

    def delete_floating_ip(self, floating_ip):
        return self.retry(self.nova.floating_ips.delete, floating_ip)

    def add_floating_ip(self, instance, floating_ip):
        return self.retry(instance.add_floating_ip, floating_ip)

    def get_server_details(self, server_id):
        return self.retry(self.nova.servers.get, server_id)

    def get_server_console(self, server, length=None):
        return self.retry(server.get_console_output, length)


class TestBed(vms.VM):

    nova_client_class = NovaClient

    def __init__(self, conf, logger):
        super(TestBed, self).__init__(conf)
        self.logger = logger
        self.instance = None
        self.floating_ip = None
        self.nova = self.build_nova_client()
        self.test_bed_key_path = None
        self.conf.set('vm.ssh_authorized_keys',
                      self.conf.get('vm.ssh_key_path') + '.pub')
        # No need to reboot a nova instance
        self.conf.set('vm.poweroff', 'False')
        self.conf.set('vm.final_message', 'testbed setup completed.')

    def build_nova_client(self):
        args = [self.conf.get('os.username'),
                self.conf.get('os.password'),
                self.conf.get('os.tenant_name'),
                self.conf.get('os.auth_url')]
        kwargs = {'region_name': self.conf.get('os.region_name'),
                  'service_type': 'compute'}
        nova_client = self.nova_client_class(*args, logger=self.logger,
                                             **kwargs)
        return nova_client

    def ensure_ssh_key_is_available(self):
        self.test_bed_key_path = self.conf.get('vm.ssh_key_path')
        # From the test runner, we need an ssh key that can be used to connect
        # to the test bed. For testing purposes, we rely on self.auth_conf to
        # provide this key.
        if not os.path.exists(self.test_bed_key_path):
            base_dir = os.path.dirname(self.test_bed_key_path)
            try:
                # Try to create needed dirs
                os.makedirs(base_dir)
            except OSError as e:
                # They are already there, no worries
                if e.errno != errno.EEXIST:
                    raise
            # First time the test runner instance needs to create a test bed
            # instance, we need to create the ssh key pair.
            subprocess.call(
                ['ssh-keygen', '-t', 'rsa', '-q',
                 '-f', self.test_bed_key_path, '-N', ''])

    def find_flavor(self):
        flavors = self.conf.get('vm.flavors')
        existing_flavors = self.nova.flavors_list()
        for flavor in flavors:
            for existing in existing_flavors:
                if flavor == existing.name:
                    return existing
        raise TestBedException(
            'None of [{}] can be found'.format(','.join(flavors)))

    def find_nova_image(self):
        image_name = self.conf.get('vm.image')
        existing_images = self.nova.images_list()
        for existing in existing_images:
            if image_name == existing.name:
                return existing
        raise TestBedException('Image "{}" cannot be found'.format(image_name))

    def find_nics(self):
        net_id = self.conf.get('vm.net_id')
        if net_id:
            return [{'net-id': self.conf.get('vm.net_id')}]
        return None

    def setup(self):
        flavor = self.find_flavor()
        image = self.find_nova_image()
        nics = self.find_nics()
        self.ensure_ssh_key_is_available()
        self.create_user_data()
        with open(self._user_data_path) as f:
            user_data = f.read()
        self.instance = self.nova.create_server(
            name=self.conf.get('vm.name'), flavor=flavor, image=image,
            user_data=user_data, nics=nics,
            # FIXME: We probably want at least a vm.az_name option. And get
            # that option from higher levels too -- vila 2014-10-13
            availability_zone=None)
        self.wait_for_active_instance()
        if unit_config.is_hpcloud(self.conf.get('os.auth_url')):
            self.floating_ip = self.nova.create_floating_ip()
            self.nova.add_floating_ip(self.instance, self.floating_ip)
        self.wait_for_ip()
        self.wait_for_cloud_init()
        self.ensure_ssh_works()
        ppas = self.conf.get('vm.ppas')
        if ppas:
            cmd = ['sudo', 'add-apt-repository']
            if self.conf.get('vm.release') > 'precise':
                cmd.append('--enable-source')
            for ppa in ppas:
                self.ssh(*(cmd + [ppa]))
        # Now we can apt-get update (doing it earlier would lead to the wrong
        # source package to be installed).
        self.safe_apt_get_update()

    def apt_get_update(self):
        return self.ssh('sudo', 'apt-get', 'update')

    def safe_apt_get_update(self):
        for timeout in self.conf.get('vm.apt_get.update.timeouts'):
            proc, out, err = self.apt_get_update()
            if proc.returncode == 0:
                # We're done
                return
            else:
                msg = ('apt-get update failed, wait {}s\n'
                       'stdout:\n{}\n'
                       'stderr:\n{}\n')
                self.logger.info(msg.format(timeout, out, err))
                time.sleep(float(timeout))
        raise TestBedException('apt-get update never succeeded')

    def update_instance(self):
        try:
            # Always query nova to get updated data about the instance
            self.instance = self.nova.get_server_details(self.instance.id)
            return True
        except:
            # But catch exceptions if something goes wrong. Higher levels will
            # deal with the instance not replying.
            return False

    def wait_for_active_instance(self):
        timeout_limit = time.time() + self.conf.get('vm.nova.boot_timeout')
        while (time.time() < timeout_limit
               and self.instance.status not in ('ACTIVE', 'ERROR')):
            time.sleep(5)
            self.update_instance()
        if self.instance.status != 'ACTIVE':
            msg = 'Instance {} never came up (last status: {})'.format(
                self.instance.id, self.instance.status)
            raise TestBedException(msg)

    def wait_for_ip(self):
        timeout_limit = time.time() + self.conf.get('vm.nova.set_ip_timeout')
        while time.time() < timeout_limit:
            if not self.update_instance():
                time.sleep(5)
                continue
            networks = self.instance.networks.values()
            if networks:
                # The network name is arbitrary, can vary for different clouds
                # but there should be only one network so we get the first one
                # and avoid the need for a config option for the network name.
                # We take the last IP address so it's either the only one or
                # the floating one. In both cases that gives us a reachable IP.
                self.ip = networks[0][-1]
                self.logger.info('Got IP {} for {}'.format(
                    self.ip, self.instance.id))
                return
            else:
                self.logger.info(
                    'IP not yet available for {}'.format(self.instance.id))
            time.sleep(5)
        msg = 'Instance {} never provided an IP'.format(self.instance.id)
        raise TestBedException(msg)

    def get_cloud_init_console(self, length=None):
        return self.nova.get_server_console(self.instance, length)

    def wait_for_cloud_init(self):
        timeout_limit = (time.time()
                         + self.conf.get('vm.nova.cloud_init_timeout'))
        final_message = self.conf.get('vm.final_message')
        while time.time() < timeout_limit:
            # A relatively cheap way to catch cloud-init completion is to watch
            # the console for the specific message we specified in user-data).
            console = self.get_cloud_init_console(10)
            if final_message in console:
                # We're good to go
                self.logger.info(
                    'cloud-init completed for {}'.format(self.instance.id))
                return
            time.sleep(5)
        raise TestBedException('Instance never completed cloud-init')

    def ensure_ssh_works(self):
        proc, out, err = self.ssh('whoami')
        if proc.returncode:
            msg = ('testbed {} IP {} cannot be reached with ssh, retcode: {}\n'
                   'stdout:\n{}\n'
                   'stderr:\n{}\n')
            self.logger.info(msg.format(self.instance.id, self.ip,
                                        proc.returncode, out, err))
            raise TestBedException('No ssh access to {}, IP: {}'.format(
                self.instance.id, self.ip))

    def teardown(self):
        if self.instance is not None:
            self.nova.delete_server(self.instance.id)
            self.instance = None
        if self.floating_ip is not None:
            self.nova.delete_floating_ip(self.floating_ip)
            self.floating_ip = None
        # FIXME: Now we can remove the testbed key from known_hosts (see
        # ssh()). -- vila 2014-01-30

    def ssh(self, command, *args, **kwargs):
        """Run a command in the testbed via ssh.

        :param args: The command and its positional arguments.

        :param kwargs: The named arguments for the command.

        The stdout and stderr outputs are captured and returned to the caller.
        """
        user = 'ubuntu'
        host = self.ip
        cmd = ['ssh',
               # FIXME: It would be better to ssh-keygen -f
               # "~/.ssh/known_hosts" -R self.ip once we're done with the test
               # bed to avoid polluting ssh commands stdout, but that will do
               # for now (that's what juju is doing after all ;)
               # -- vila 2014-01-29
               '-oStrictHostKeyChecking=no',
               '-i', self.test_bed_key_path,
               '{}@{}'.format(user, host)]
        if command is not None:
            cmd += [command]
        if args:
            cmd += args
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = proc.communicate()
        return proc, out, err

    def scp(self, local_path, remote_path):
        cmd = ['scp',
               # FIXME: It would be better to ssh-keygen -f
               # "~/.ssh/known_hosts" -R self.ip once we're done with the test
               # bed to avoid polluting ssh commands stdout, but that will do
               # for now (that's what juju is doing after all ;)
               # -- vila 2014-01-29
               '-oStrictHostKeyChecking=no',
               '-i', self.test_bed_key_path,
               local_path, remote_path]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = proc.communicate()
        return proc, out, err

    def get_remote_content(self, path):
        proc, content, err = self.ssh('cat', path)
        if proc.returncode:
            # We didn't get a proper content, report it instead
            content = ("{} couldn't be copied from testbed {}:\n"
                       "error: {}\n".format(path, self.ip, err))
        return content
