# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals


import os
import subprocess
import unittest

import requests
from uciconfig import options
from ucitests import (
    assertions,
    fixtures,
    scenarii,
)
from ucivms.tests import fixtures as vms_fixtures


from ci_utils import (
    image_store,
    unit_config,
)
from ci_utils.testing import features
from tstrun import (
    testbed,
    tests,
)


load_tests = scenarii.load_tests_with_scenarios


@features.requires(features.nova_creds)
@features.requires(features.nova_compute)
class TestNovaClient(unittest.TestCase):
    """Check the nova client behavior when it encounters exceptions.

    This is achieved by overriding specific methods from NovaClient and
    exercising it through the TestBed methods.
    """

    def setUp(self):
        super(TestNovaClient, self).setUp()
        vms_fixtures.isolate_from_disk(self)
        self.tb_name = 'testing-nova-client'
        # Prepare a suitable config, importing the nova credentials
        conf = testbed.vms_config_from_auth_config(
            self.tb_name, unit_config.get_auth_config())
        # Default to precise
        conf.set('vm.release', 'precise')
        # Avoid triggering the 'atexit' hook as the config files are long gone
        # at that point.
        self.addCleanup(conf.store.save_changes)
        self.conf = conf
        os.makedirs(self.conf.get('vm.vms_dir'))

    def get_image_id(self, series, arch):
        return image_store.uci_image_name('cloudimg', series, arch)

    @tests.log_on_failure()
    def test_retry_is_called(self, logger):
        self.retry_calls = []

        class RetryingNovaClient(testbed.NovaClient):

            def retry(inner, func, *args, **kwargs):
                self.retry_calls.append((func, args, kwargs))
                return super(RetryingNovaClient, inner).retry(
                    func, *args, **kwargs)

        image_id = self.get_image_id('trusty', 'amd64')
        self.conf.set('vm.image', image_id)
        fixtures.patch(self, testbed.TestBed,
                       'nova_client_class', RetryingNovaClient)
        tb = testbed.TestBed(self.conf, logger)
        self.assertEqual(image_id, tb.find_nova_image().name)
        assertions.assertLength(self, 1, self.retry_calls)

    @tests.log_on_failure()
    def test_known_failure_is_retried(self, logger):
        self.nb_calls = 0

        class FailingOnceNovaClient(testbed.NovaClient):

            def fail_once(inner):
                self.nb_calls += 1
                if self.nb_calls == 1:
                    raise requests.ConnectionError()
                else:
                    return inner.nova.flavors.list()

            def flavors_list(inner):
                return inner.retry(inner.fail_once)

        fixtures.patch(self, testbed.TestBed,
                       'nova_client_class', FailingOnceNovaClient)
        tb = testbed.TestBed(self.conf, logger)
        tb.find_flavor()
        self.assertEqual(2, self.nb_calls)

    @tests.log_on_failure()
    def test_unknown_failure_is_raised(self, logger):

        class FailingNovaClient(testbed.NovaClient):

            def fail(inner):
                raise AssertionError('Boom!')

            def flavors_list(inner):
                return inner.retry(inner.fail)

        fixtures.patch(self, testbed.TestBed,
                       'nova_client_class', FailingNovaClient)
        tb = testbed.TestBed(self.conf, logger)
        # This mimics what will happen when we encounter unknown transient
        # failures we want to catch: an exception will bubble up and we'll have
        # to add it to NovaClient.retry().
        with self.assertRaises(AssertionError) as cm:
            tb.find_flavor()
        self.assertEqual('Boom!', unicode(cm.exception))


@features.requires(features.nova_creds)
@features.requires(features.nova_compute)
class TestTestbed(unittest.TestCase):

    def setUp(self):
        super(TestTestbed, self).setUp()
        vms_fixtures.isolate_from_disk(self)
        self.tb_name = 'testing-testbed'
        # Prepare a suitable config, importing the nova credentials
        conf = testbed.vms_config_from_auth_config(
            self.tb_name, unit_config.get_auth_config())
        # Default to precise
        conf.set('vm.release', 'precise')
        # Avoid triggering the 'atexit' hook as the config files are long gone
        # at that point.
        self.addCleanup(conf.store.save_changes)
        self.conf = conf
        os.makedirs(self.conf.get('vm.vms_dir'))

    def get_image_id(self, series='precise', arch='amd64'):
        return image_store.uci_image_name('cloudimg', series, arch)

    @tests.log_on_failure()
    def test_create_no_image(self, logger):
        tb = testbed.TestBed(self.conf, logger)
        with self.assertRaises(options.errors.OptionMandatoryValueError) as cm:
            tb.setup()
        self.assertEqual('vm.image must be set.', unicode(cm.exception))

    @tests.log_on_failure()
    def test_create_unknown_image(self, logger):
        image_name = "I don't exist and eat kittens"
        self.conf.set('vm.image', image_name)
        tb = testbed.TestBed(self.conf, logger)
        with self.assertRaises(testbed.TestBedException) as cm:
            tb.setup()
        self.assertEqual('Image "{}" cannot be found'.format(image_name),
                         unicode(cm.exception))

    @tests.log_on_failure()
    def test_create_unknown_flavor(self, logger):
        flavors = "I don't exist and eat kittens"
        self.conf.set('vm.flavors', flavors)
        tb = testbed.TestBed(self.conf, logger)
        with self.assertRaises(testbed.TestBedException) as cm:
            tb.setup()
        self.assertEqual('None of [{}] can be found'.format(flavors),
                         unicode(cm.exception))

    @tests.log_on_failure()
    def test_existing_home_ssh(self, logger):
        # The first request for the worker requires creating ~/.ssh if it
        # doesn't exist, but it may happen that this directory already exists
        # (see http://pad.lv/1334146).
        ssh_home = os.path.expanduser('~/sshkeys')
        os.mkdir(ssh_home)
        self.conf.set('vm.ssh_key_path', os.path.join(ssh_home, 'id_rsa'))
        tb = testbed.TestBed(self.conf, logger)
        tb.ensure_ssh_key_is_available()
        self.assertTrue(os.path.exists(ssh_home))
        self.assertTrue(os.path.exists(os.path.join(ssh_home, 'id_rsa')))
        self.assertTrue(os.path.exists(os.path.join(ssh_home, 'id_rsa.pub')))

    @tests.log_on_failure()
    def test_create_new_ssh_key(self, logger):
        self.conf.set('vm.image', self.get_image_id())
        # We use a '~' path to cover proper uci-vms user expansion
        self.conf.set('vm.ssh_key_path', '~/sshkeys/id_rsa')
        tb = testbed.TestBed(self.conf, logger)
        tb.ensure_ssh_key_is_available()
        self.assertTrue(os.path.exists(os.path.expanduser('~/sshkeys/id_rsa')))
        self.assertTrue(
            os.path.exists(os.path.expanduser('~/sshkeys/id_rsa.pub')))

    @tests.log_on_failure()
    def test_ssh_failure(self, logger):
        self.conf.set('vm.release', 'trusty')
        self.conf.set('vm.image', self.get_image_id())
        tb = testbed.TestBed(self.conf, logger)
        self.addCleanup(tb.teardown)
        tb.setup()
        # Sabotage ssh access
        os.remove(self.conf.get('vm.ssh_key_path'))
        # Oh, we can't ssh anymore !
        with self.assertRaises(testbed.TestBedException) as cm:
            tb.ensure_ssh_works()
        msg = 'No ssh access to {}, IP: {}'
        msg = msg.format(tb.instance.id, tb.ip)
        self.assertEqual(msg, unicode(cm.exception))

    @tests.log_on_failure()
    def test_apt_get_update_retries(self, logger):
        self.conf.set('vm.image', self.get_image_id())
        self.conf.set('vm.apt_get.update.timeouts', '0.1, 0.1')
        tb = testbed.TestBed(self.conf, logger)
        self.nb_calls = 0

        class Proc(object):
            returncode = 0

        def failing_update():
            self.nb_calls += 1
            if self.nb_calls > 1:
                return Proc(), 'stdout success', 'stderr success'
            else:
                # Fake a failed apt-get update
                proc = Proc()
                proc.returncode = 1
                return proc, 'stdout error', 'stderr error'

        tb.apt_get_update = failing_update
        tb.safe_apt_get_update()
        self.assertEqual(2, self.nb_calls)

    @tests.log_on_failure()
    def test_apt_get_update_fails(self, logger):
        self.conf.set('vm.image', self.get_image_id())
        self.conf.set('vm.apt_get.update.timeouts', '0.1, 0.1, 0.1')
        tb = testbed.TestBed(self.conf, logger)

        def failing_update():
            class Proc(object):
                pass

            proc = Proc()
            proc.returncode = 1
            return proc, 'stdout', 'stderr'

        tb.apt_get_update = failing_update
        with self.assertRaises(testbed.TestBedException) as cm:
            tb.safe_apt_get_update()
        self.assertEqual('apt-get update never succeeded',
                         unicode(cm.exception))

    @tests.log_on_failure()
    def test_wait_for_instance_fails(self, logger):
        self.conf.set('vm.image', self.get_image_id())
        # Force a 0 timeout so the instance can't finish booting
        self.conf.set('vm.nova.boot_timeout', '0')
        tb = testbed.TestBed(self.conf, logger)
        self.addCleanup(tb.teardown)
        with self.assertRaises(testbed.TestBedException) as cm:
            tb.setup()
        msg = 'Instance {} never came up (last status: BUILD)'
        msg = msg.format(tb.instance.id)
        self.assertEqual(msg, unicode(cm.exception))

    @tests.log_on_failure()
    def test_wait_for_instance_errors(self, logger):
        self.conf.set('vm.image', self.get_image_id())
        tb = testbed.TestBed(self.conf, logger)
        self.addCleanup(tb.teardown)

        def update_instance_to_error():
            # Fake an instance starting in error state
            tb.instance.status = 'ERROR'
            return True
        tb.update_instance = update_instance_to_error
        with self.assertRaises(testbed.TestBedException) as cm:
            tb.setup()
        msg = 'Instance {} never came up (last status: ERROR)'
        msg = msg.format(tb.instance.id)
        self.assertEqual(msg, unicode(cm.exception))

    @tests.log_on_failure()
    def test_wait_for_ip_fails(self, logger):
        self.conf.set('vm.image', self.get_image_id())
        # Force a 0 timeout so the instance never get an IP
        self.conf.set('vm.nova.set_ip_timeout', '0')
        tb = testbed.TestBed(self.conf, logger)
        self.addCleanup(tb.teardown)
        with self.assertRaises(testbed.TestBedException) as cm:
            tb.setup()
        msg = 'Instance {} never provided an IP'.format(tb.instance.id)
        self.assertEqual(msg, unicode(cm.exception))


@features.requires(features.nova_creds)
@features.requires(features.nova_compute)
class TestUsableTestbed(unittest.TestCase):

    scenarios = scenarii.multiply_scenarios(
        # series
        ([('precise', dict(series='precise', result='skip')),
          ('trusty', dict(series='trusty', result='pass')),
          ('utopic', dict(series='utopic', result='pass'))]),
        # architectures
        ([('amd64', dict(arch='amd64')), ('i386', dict(arch='i386'))]))

    def setUp(self):
        super(TestUsableTestbed, self).setUp()
        vms_fixtures.isolate_from_disk(self)
        self.tb_name = 'testing-testbed-{}-{}'.format(self.series, self.arch)
        # Prepare a suitable config, importing the nova credentials
        conf = testbed.vms_config_from_auth_config(
            self.tb_name, unit_config.get_auth_config())
        conf.set('vm.release', self.series)
        # Avoid triggering the 'atexit' hook as the config files are long gone
        # at that point.
        self.addCleanup(conf.store.save_changes)
        self.conf = conf
        os.makedirs(self.conf.get('vm.vms_dir'))

    def get_image_id(self):
        return image_store.uci_image_name('cloudimg', self.series, self.arch)

    @tests.log_on_failure()
    def test_create_usable_testbed(self, logger):
        self.conf.set('vm.release', 'trusty')
        self.conf.set('vm.image', self.get_image_id())
        tb = testbed.TestBed(self.conf, logger)
        self.addCleanup(tb.teardown)
        tb.setup()
        # We should be able to ssh with the right user
        proc, out, err = tb.ssh('whoami',
                                out=subprocess.PIPE, err=subprocess.PIPE)
        self.assertEqual(0, proc.returncode)
        self.assertEqual('ubuntu\n', out)
