# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

from ucitests import (
    assertions,
    fixtures as uci_fixtures,
    scenarii,
)
from ucivms.tests import fixtures as vms_fixtures
from ci_utils import (
    amqp_utils,
    image_store,
    unit_config,
)
from ci_utils.testing import (
    features,
    fixtures,
)
from tstrun import (
    run_worker,
    testbed,
    tests,
)


load_tests = scenarii.load_tests_with_scenarios


@features.requires(features.nova_creds)
@features.requires(features.nova_compute)
class TestWorkerHandleRequest(unittest.TestCase):

    def setUp(self):
        super(TestWorkerHandleRequest, self).setUp()
        # The fake rabbit input queue
        self.queue = fixtures.FakeTaskQueue('test')
        # The data store and uci-vms need disk access
        vms_fixtures.isolate_from_disk(self)
        self.ds_path = os.path.join(self.uniq_dir, 'data_store')
        # We don't need a config for the data store
        self.data_store = fixtures.FakeDataStore(self.ds_path)

        def create_data_store(component):
            return self.data_store

        self.ds_factory = create_data_store
        self.conf = unit_config.get_auth_config()

    def get_ppa_list(self):
        # We're testing only the test runner so there is no ppa assigned to the
        # ticket, only the master ppa is relevant.
        return [self.conf.get('master_ppa')]

    def get_image_id(self, series, arch):
        return image_store.uci_image_name('cloudimg', series, arch)

    def assertTestRun(self, expected_artifacts, logger, params):
        worker = run_worker.TestRunnerWorker(self.ds_factory)
        self.queue.publish(params)
        returns = self.queue.subscribe(worker.handle_request, logger)
        assertions.assertLength(self, 1, returns)
        self.assertTrue(self.queue.is_empty())
        self.assertEquals(expected_artifacts, self.data_store.list_files())
        return returns

    @tests.log_on_failure()
    def test_two_packages_two_testbeds(self, logger):
        # We use precise where no tests are defined for libpng nor adduser to
        # speed things up a bit without reducing coverage. (~244s instead of
        # 560~ on a single sample (YMMV, high variations there)).
        series = 'precise'
        architecture = 'amd64'
        params = dict(ticket_id=series, progress_trigger='progress',
                      series=series,
                      architecture=architecture,
                      image_id=self.get_image_id(series, architecture),
                      ppa_list=self.get_ppa_list(),
                      package_list=['libpng', 'adduser'])
        returns = self.assertTestRun(
            (tests.expected_tr_files('adduser')
             + tests.expected_tr_files('libpng')),
            logger, params)
        (retcode, results) = returns[0]
        self.assertEqual(amqp_utils.progress_failed, retcode)
        self.assertEquals(
            '*                    SKIP no tests in this package\n',
            self.data_store.get_file('libpng-summary.log'))
        self.assertEquals(
            '*                    SKIP no tests in this package\n',
            self.data_store.get_file('adduser-summary.log'))

    @tests.log_on_failure()
    def test_teardown_dont_break_test_run(self, logger):
        series = 'trusty'
        architecture = 'amd64'
        package = 'libpng'
        params = dict(ticket_id=series, progress_trigger='progress',
                      series=series,
                      architecture=architecture,
                      image_id=self.get_image_id(series, architecture),
                      ppa_list=self.get_ppa_list(),
                      package_list=[package])

        def broken_teardown(test_bed):
            self.addCleanup(test_bed.nova.delete_server, test_bed.instance.id)
            if test_bed.floating_ip is not None:
                self.addCleanup(test_bed.nova.delete_floating_ip,
                                test_bed.floating_ip)
            raise AssertionError('Boom !')

        uci_fixtures.patch(self, testbed.TestBed, 'teardown', broken_teardown)
        returns = self.assertTestRun(tests.expected_tr_files(package),
                                     logger, params)
        (retcode, results) = returns[0]
        self.assertEqual(amqp_utils.progress_completed, retcode)


@features.requires(features.nova_creds)
@features.requires(features.nova_compute)
class TestWorkerHandleRequestFor(unittest.TestCase):

    scenarios = scenarii.multiply_scenarios(
        # series
        ([('precise', dict(series='precise', result='skip')),
          ('trusty', dict(series='trusty', result='pass')),
          ('utopic', dict(series='utopic', result='pass'))]),
        # architectures
        ([('amd64', dict(arch='amd64')), ('i386', dict(arch='i386'))]))

    def setUp(self):
        super(TestWorkerHandleRequestFor, self).setUp()
        # The fake rabbit input queue
        self.queue = fixtures.FakeTaskQueue('test')
        # The data store and uci-vms need disk access
        vms_fixtures.isolate_from_disk(self)
        self.ds_path = os.path.join(self.uniq_dir, 'data_store')
        # We don't need a config for the data store
        self.data_store = fixtures.FakeDataStore(self.ds_path)

        def create_data_store(component):
            return self.data_store

        self.ds_factory = create_data_store
        self.conf = unit_config.get_auth_config()

    def get_ppa_list(self):
        # We're testing only the test runner so there is no ppa assigned to the
        # ticket, only the master ppa is relevant.
        return [self.conf.get('master_ppa')]

    def get_image_id(self):
        return image_store.uci_image_name('cloudimg', self.series, self.arch)

    def assertTestRun(self, logger, params):
        worker = run_worker.TestRunnerWorker(self.ds_factory)
        self.queue.publish(params)
        returns = self.queue.subscribe(worker.handle_request, logger)
        assertions.assertLength(self, 1, returns)
        self.assertTrue(self.queue.is_empty())
        return returns

    @tests.log_on_failure()
    def test_handle_request(self, logger):
        ticket_id = '{}-{}'.format(self.series, self.arch)
        package = 'libpng'
        params = dict(ticket_id=ticket_id, progress_trigger='progress',
                      series=self.series,
                      architecture=self.arch,
                      image_id=self.get_image_id(),
                      ppa_list=self.get_ppa_list(),
                      package_list=[package])
        returns = self.assertTestRun(logger, params)
        (retcode, results) = returns[0]
        if self.result == 'skip':
            e_retcode = amqp_utils.progress_failed
            e_summary = '*                    SKIP no tests in this package\n'
        elif self.result == 'pass':
            e_retcode = amqp_utils.progress_completed
            e_summary = 'build                PASS\n'
        self.assertEqual(e_retcode, retcode)
        self.assertEquals(tests.expected_tr_files(package),
                          self.data_store.list_files())
        # We need to check the content here as the worker will put an error
        # message there if it cannot get them from the testbed (so the user get
        # *some* feedback).
        self.assertEquals(e_summary, self.data_store.get_file(
            '{pkg}-summary.log'.format(pkg=package)))


class TestFormatBritneyResult(unittest.TestCase):

    def test_error_no_pkg_version(self):
        self.assertEqual('trusty armhf adduser',
                         run_worker.format_britney_result(
                             'trusty', 'armhf', 'adduser', 1,
                             None, [('foo', '1.0')]))

    def test_error_no_deps_versions(self):
        self.assertEqual('trusty armhf adduser',
                         run_worker.format_britney_result(
                             'trusty', 'armhf', 'adduser', 1,
                             '1.0', []))

    def test_passing(self):
        self.assertEqual('trusty armhf adduser 1.0 PASS foo 1.0',
                         run_worker.format_britney_result(
                             'trusty', 'armhf', 'adduser', 0,
                             '1.0', [('foo', '1.0')]))

    def test_skipping(self):
        self.assertEqual('trusty armhf adduser 1.0 PASS foo 1.0',
                         run_worker.format_britney_result(
                             'trusty', 'armhf', 'adduser', 2,
                             '1.0', [('foo', '1.0')]))
