#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""The CI Engine test runner."""

import logging
import os
import subunit
import sys
import testtools


HERE = os.path.abspath(os.path.dirname(__file__))

# deploy.py isn't in our pythonpath when run as a script
sys.path.append(os.path.join(HERE, '..', 'juju-deployer'))
import deploy
from ucitests import (
    filters,
    loaders,
    matchers,
    runners,
    results,
)


def rename_tests(suite, renamer):
    """Rename all tests in a suite.

    :param suite: The tests to be renamed.

    :param renamer: A callable returning a callable 'id' method for the
       test. It should accept a callable that returns the current test id.

    Note: Care should be taken to ensure that 'renamer' preserve unique ids.
    """
    # We modify the tests in place so it's ok to use iter_flat
    for t in filters.iter_flat(suite):
        # 'renamer' needs to be a two-level lambda or the wrong variable is
        # captured. See tests for examples.
        t.id = renamer(t.id)


def prefix_test_ids(suite, prefix):
    """Change the test ids to embed a prefix for easier matching.

    :param suite: The tests to be renamed.

    :param prefix: A path from the branch root to the root python module (found
       in PYTHONPATH).
    """
    prefix = prefix.replace(os.path.sep, '.')
    # The renamer receives the old_id callable and return a lambda that will be
    # called as test.id().
    renamer = lambda old_id: lambda: prefix + '.' + old_id()
    rename_tests(suite, renamer)


class SysPath(object):
    """Add paths in front of sys.path temporarily."""

    def __init__(self, paths):
        self.paths = paths
        self.orig = sys.path[:]

    def __enter__(self):
        sys.path = self.paths + sys.path

    def __exit__(self, *args):
        sys.path = self.orig


def load_component_tests(loader, component_base_dir):
    """Load tests for a given component.

    A component is organized inside a given base directory containing files
    *and* a python directory with the code and the tests. We care about that
    python directory only.
    """
    # We set the current directory to the component base one and restore it
    # after loading. This could produce duplicate tests if different components
    # define the same python modules but this would be a bug in itself so we
    # just ignore the potential issue.
    suite = loader.suiteClass()
    with SysPath([component_base_dir]):
        sub_loader = loader.SubLoader(root=component_base_dir)
        suite.addTests(sub_loader.loadTestsFromTree('.'))
    return suite


def load_regular_component_tests(include_regexps, exclude_regexps=None):
    """Load tests matching inclusive and exclusive regexps.

    :param include_regexps: A list of regexps describing the tests to include.

    :param exclude_regexps: A list of regexps describing the tests to exclude.

    :return: The test suite for all collected tests.
    """
    components = ['ci-utils',
                  'branch-source-builder',
                  'britney_proxy',
                  'gatekeeper',
                  'image-builder',
                  'lander',
                  'test_runner',
                  'publisher',
                  'ppa-creator',
                  'validator',
                  'coverage-retriever',
                  ]
    loader = loaders.Loader()
    # setuptools tends to leave eggs all over the place
    loader.dir_matcher = matchers.NameMatcher(includes=[r'.*'],
                                              excludes=[r'\.egg$',
                                                        r'\.egg-info$',
                                                        ])
    suite = loader.suiteClass()
    for c in components:
        component_suite = load_component_tests(loader, c)
        prefix_test_ids(component_suite, c)
        suite.addTests(component_suite)
    suite = filters.include_regexps(include_regexps, suite)
    suite = filters.exclude_regexps(exclude_regexps, suite)
    return suite


def load_testing_tests(include_regexps, exclude_regexps=None):
    loader = loaders.Loader()
    suite = loader.suiteClass()
    suite.addTests(loader.loadTestsFromTree('testing'))
    suite = filters.include_regexps(include_regexps, suite)
    suite = filters.exclude_regexps(exclude_regexps, suite)
    return suite


def get_django_runner(component_base_dir):
    """Return a test runner class tailored to django needs.


    The django design forces us to delay the imports so we can prepare the
    environment before django attempt to import modules and set some globals
    there.

    Since the imports needs to be delayed, that DjangoTestRunner class has to
    be defined inside a function so it can inherit from DjangoTestSuiteRunner.
    """
    from django.test import simple

    class DjangoTestRunner(simple.DjangoTestSuiteRunner):

        def __init__(self, component_base_dir, requires_south, out, result,
                     options,
                     **kwargs):
            super(DjangoTestRunner, self).__init__(**kwargs)
            self.component_base_dir = component_base_dir
            self.requires_south = requires_south
            self.out = out
            self.result = result
            self.options = options
            # Used to cache the test suite
            self._cached_test_suite = None

        def build_suite(self, *args, **kwargs):
            # We cache the test suite to ensure it's loaded only once. This is
            # used to workaround DjangoTestSuiteRunner closed design that
            # doesn't allow to hook that more precisely. Concretely,
            # DjangoTestSuiteRunner.run_tests() setup the databases even if
            # there are no tests to run, we take care of that in our own
            # run_tests() by not upcalling when the test suite is empty.
            if self._cached_test_suite is not None:
                return self._cached_test_suite
            loader = loaders.Loader()
            suite = load_component_tests(loader, self.component_base_dir)
            prefix_test_ids(suite, component_base_dir)
            # Filtered as required by the user
            suite = filters.include_regexps(
                self.options.include_regexps, suite)
            suite = filters.exclude_regexps(
                self.options.exclude_regexps, suite)
            self._cached_test_suite = suite
            return suite

        def run_tests(self, test_labels, extra_tests=None, **kwargs):
            # We load the tests early to avoid the costly database setup when
            # it's not needed.
            suite = self.build_suite(test_labels, extra_tests, **kwargs)
            if self.options.list_only:
                ret = runners.list_tests(suite, self.out)
                return ret

            if suite.countTestCases() == 0:
                # No tests to run, we let the caller handle the case where no
                # tests are run at all which is an error. By loading the test
                # suite early we avoid calling
                # DjangoTestSuiteRunner.run_tests() which will setup the
                # databases even if the test suite is empty.
                return 0

            # South needs to patch the way the db is created/synced.
            if self.requires_south:
                from south.management import commands
                commands.patch_for_test_db_setup()

            # Otherwise let the base class handle the setup and the run
            return super(DjangoTestRunner, self).run_tests(
                test_labels, extra_tests=None, **kwargs)

        def run_suite(self, suite, **kwargs):
            suite = parallelize_suite_round_robin(suite,
                                                  self.options.concurrency)
            suite.run(self.result)
            return self.result

    return DjangoTestRunner


class IsolatedImportAndLogging(object):
    """Protect sys.modules, sys.path and logging.root

    This allows running django tests which relies on a lot of global context by
    taking a snapshot of 'sys.path', 'sys.modules' and 'logging.root' when
    entering the context and restoring them afterwards. As such, it creates a
    context where django can freely setup globals in modules all over the place
    without interfering with the outside.

    Several django apps can therefore be tested during the same test run
    without stepping on each other toes.
    """

    def __init__(self):
        # Keep copies of the right objects to be able to restore them
        self.paths = list(sys.path)
        self.modules = sys.modules.copy()
        self.root = logging.root

    def __enter__(self):
        logging.root.handlers = []
        logging.root.setLevel(logging.WARNING)

    def __exit__(self, *args):
        # Restore the copied list
        sys.path = self.paths
        # Remove all added modules
        added = [m for m in sys.modules.keys() if m not in self.modules]
        # Don't mess up with bzrlib modules, they use a lazy import mechanism
        # that doesn't cope with restoring modules this way.
        added = [m for m in added if not m.startswith('bzrlib')]
        if added:
            for m in added:
                del sys.modules[m]
        # Restore deleted or modified modules
        sys.modules.update(self.modules)
        logging.root = self.root


def run_django_tests(component_base_dir, component_name, requires_south,
                     out, result, options):
    """Run tests for a django app.

    :param component_base_dir: Where the component files are.

    :param component_name: The name of the component acting as a base for its
        python name space.

    :param requires_south: Whether the component uses the 'south' module and
        needs to be setup accordingly.

    :param out: The output stream to use.

    :param result: The test result object to use for the run.

    :param options: A dict with the options from the command line.

    :return: 0 on success, 1 on failure.
    """
    with IsolatedImportAndLogging():
        # We don't have to use setdefault here because we don't have (and don't
        # want !) to support overriding. We want to test with the default
        # values defined in the project settings.py
        settings_path = '{}.settings'.format(component_name)
        os.environ["DJANGO_SETTINGS_MODULE"] = settings_path
        kls = get_django_runner(component_base_dir)
        test_runner = kls(component_base_dir, requires_south,
                          out, result, options)
        failures = test_runner.run_tests([])
        return failures


def load_orphan_tests(include_regexps, exclude_regexps=None):
    """Load tests matching inclusive and exclusive regexps.

    :param include_regexps: A list of regexps describing the tests to include.

    :param exclude_regexps: A list of regexps describing the tests to exclude.

    :return: The test suite for all collected tests.
    """
    components = ['juju-deployer']
    loader = loaders.Loader()
    suite = loader.suiteClass()
    for c in components:
        suite.addTests(load_component_tests(loader, c))
        prefix_test_ids(suite, c)
    suite = filters.include_regexps(include_regexps, suite)
    suite = filters.exclude_regexps(exclude_regexps, suite)
    return suite


def load_integration_tests(include_regexps, exclude_regexps=None):
    """Load tests matching inclusive and exclusive regexps.

    :param include_regexps: A list of regexps describing the tests to include.

    :param exclude_regexps: A list of regexps describing the tests to exclude.

    :return: The test suite for all collected tests.
    """
    loader = loaders.Loader()
    suite = loader.suiteClass()
    suite.addTests(loader.loadTestsFromTree('tests'))
    suite = filters.include_regexps(include_regexps, suite)
    suite = filters.exclude_regexps(exclude_regexps, suite)
    return suite


def load_charm_tests(include_regexps, exclude_regexps=None):
    """Load charm unittest from <charm>/unit_test."""
    charms = [
        'charms/precise/lander',
        'charms/precise/rabbitmq-worker',
        'charms/precise/wsgi-app',
        'charms/precise/webui',
        'charms/precise/key-secret-subordinate',
    ]
    loader = loaders.Loader()
    suite = loader.suiteClass()
    for c in charms:
        with SysPath([c, os.path.join(c, 'hooks')]):
            sub_loader = loader.SubLoader(root=c)
            charm_suite = loader.suiteClass()
            # Load only python tests from 'unit_tests' (that's specific
            # for uci-engine charms).
            charm_suite.addTests(sub_loader.loadTestsFromTree('unit_tests'))
            # Change the test ids to embed the leading path for easier matching
            prefix_test_ids(charm_suite, c)
        suite.addTests(charm_suite)
    suite = filters.include_regexps(include_regexps, suite)
    suite = filters.exclude_regexps(exclude_regexps, suite)
    return suite


def run_regular_tests(suite, out, result, options):
    """Run the given test suite under some isolation.

    :param suite: The test suite to run.

    :param out: The output stream to use.

    :param result: The test result object to use for the run.

    :param options: A dict with the options from the command line.

    :return: 0 on success, 1 on failure.
    """
    with IsolatedImportAndLogging():
        if options.list_only:
            # List the tests without running them
            ret = runners.list_tests(suite, out)
        else:
            suite.run(result)
            # TODO: This logic is worth putting into the test result class
            # itself or at least a helper should be provided by uci-tests
            # itself. -- vila 2014-02-08
            ret = int(not (result.wasSuccessful()
                           and result.testsRun > 0))
        return ret


def parallelize_suite_round_robin(suite, concurrency):
    """Parallelize a suite.

    :param suite: The input suite to transform.

    :param concurrency: The level of concurrency.

    :returns: A suite that will spread the execution of the input suite tests
        across ``concurrency`` forked processes. The suite is evenly split
        across the processes in a round-robin fashion. While this doesn't take
        into account the duration of each test, it's doing a good job on
        average as slower tests are most often grouped by class/module in the
        input suite.
    """
    # Nothing to parallelize if there is only one process
    if concurrency > 1:
        suite = testtools.ConcurrentTestSuite(
            suite, runners.fork_for_tests(concurrency))
    return suite


class RunTestsArgParser(runners.RunTestsArgParser):

    def __init__(self):
        super(RunTestsArgParser, self).__init__('run-tests')
        self.add_argument(
            '--no-skip', action='store_true',
            dest='no_skip',
            help='Errors if at lest one test is skipped.')


def main(args, stdout, stderr):
    # Interpret user wishes
    parser = RunTestsArgParser()
    options = parser.parse_args(args)
    if options.list_only:
        # We won't run tests, we won't need a test result
        result = None
    else:
        # Run the tests with the required output
        if options.format == 'text':
            result = results.TextResult(stdout, verbosity=2)
        else:
            result = subunit.TestProtocolClient(stdout)
        # We'll run tests, setup the result accordingly
        result.startTestRun()
    ret = 0
    # Regular components
    suite = load_regular_component_tests(options.include_regexps,
                                         options.exclude_regexps)
    suite = parallelize_suite_round_robin(suite, options.concurrency)
    reg_rc = run_regular_tests(suite, stdout, result, options)
    ret = reg_rc or ret
    # The testing infrastructure itself
    suite = load_testing_tests(options.include_regexps,
                               options.exclude_regexps)
    suite = parallelize_suite_round_robin(suite, options.concurrency)
    tst_rc = run_regular_tests(suite, stdout, result, options)
    ret = tst_rc or ret
    # orphan tests that don't have a proper home yet
    suite = load_orphan_tests(options.include_regexps,
                              options.exclude_regexps)
    suite = parallelize_suite_round_robin(suite, options.concurrency)
    orphan_rc = run_regular_tests(suite, stdout, result, options)
    ret = orphan_rc or ret
    with IsolatedImportAndLogging():
        ts_rc = run_django_tests('ticket_system', 'ticket_system', True,
                                 stdout, result, options)
        ret = ts_rc or ret
    # charms and integration tests have a common setup
    with IsolatedImportAndLogging(), SysPath(['ci-utils']):
        # FIXME: We need some more setup here that tests themselves should do
        # but they aren't ripe for that yet -- vila 2014-04-11
        branches = os.path.join(HERE, '..', 'branches')
        deploy.build_sourcedeps(branches)
        try:
            deploy.setup(for_tests=True, list_only=options.list_only)
        except SystemExit:
            # The setup includes checking the environment and raising
            # SystemExit if one var is missing. Since this is entangled with
            # building the charms, we skip the lot.
            pass
        else:
            # Charm tests
            charm_suite = load_charm_tests(
                options.include_regexps, options.exclude_regexps)
            charm_rc = run_regular_tests(
                charm_suite, stdout, result, options)
            ret = charm_rc or ret
            # Integration tests
            suite = load_integration_tests(
                options.include_regexps, options.exclude_regexps)
            itg_rc = run_regular_tests(suite, stdout, result, options)
            ret = itg_rc or ret

    if not options.list_only:
        # We did run tests, stop the run
        result.stopTestRun()
    # Fails if at least one fail
    if not options.list_only and options.no_skip and result.skip_reasons:
        # Some tests were skipped
        return 1
    return int(ret)
