# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import collections
import errno
import os
import subprocess
import sys
import logging

logging.basicConfig(level=logging.INFO, format='%(msg)s')
log = logging.getLogger('integration-test-runner-deployers')

HERE = os.path.abspath(os.path.dirname(__file__))
# We need access to ci_utils
sys.path.append(os.path.join(HERE, '..', 'ci-utils'))
# We need to find our juju environment in deploy
sys.path.append(os.path.join(HERE, '..', 'juju-deployer'))
import deploy

from ci_utils import (
    amqp_utils,
    testing,
)
from ci_utils.testing import features
from ucitests import fixtures


def _juju_call(juju_args):
    if not isinstance(juju_args, collections.Iterable):
        raise OSError(
            'Given argument is not iterable: {!r}'.format(juju_args))
    args = ['juju']
    args.extend(juju_args)
    try:
        p = subprocess.Popen(
            args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError as e:
        # MISSINGTEST: not covered by any test -- vila 2014-06-20
        if e.errno != errno.ENOENT:
            raise
        raise OSError("juju not found, do you have Juju installed?")
    out, err = p.communicate()
    if p.returncode:
        raise IOError('juju command failed {!r}:\n{}'.format(juju_args, err))
    return out


def juju_run(service, cmd, unit=0):
    """Call a command on a single given unit.

    :param service: A service name.

    :param cmd: The command to execute on a unit. This should be provided as a
        string to avoid evaluation on the originating host (the evaluation will
        happen on the unit itself if needed).

    :param unit: The unit number where the command should be run. It defaults
        to 0 if not provided.

    """
    return _juju_call(['run', '--unit', '{}/{}'.format(service, unit), cmd])


def juju_expose(service, expose=True):
    if expose:
        action = 'expose'
    else:
        action = 'unexpose'
    return _juju_call([action, service])


def juju_get_ip_and_port(service, unit=0):
    status = deploy.juju_status()
    units = status['services'][service]['units']
    ip = units['%s/%d' % (service, unit)]['public-address']
    formatted = '%s/%d' % (service, unit)
    port, _ = units[formatted]['open-ports'][0].split('/')
    return ip, int(port)


# TODO ev 2014-06-16 @features.requires(features.settled_units)
@features.requires(features.bootstrapped_juju)
class DeployerTest(testing.TestCaseWithGnupg):
    '''Base class for building juju deployer based tests.'''

    def get_ip_and_port(self, service, unit=0):
        return juju_get_ip_and_port(service, unit)

    def assert_job_running(self, service, unit=0, upstart_job=None):
        # Ensure the given upstart job is running on a unit
        if upstart_job is None:
            upstart_job = service
        output = juju_run(service, '/sbin/status {}'.format(upstart_job),
                          unit=unit)
        self.assertIn('start/running', output)


RABBIT_SERVICE = 'ci-airline-rabbit'


def run_rabbit_cmd(args):
    juju_run(RABBIT_SERVICE, 'sudo rabbitmqctl {}'.format(' '.join(args)))


def create_rabbit_user(user_name, password=None):
    if password is None:
        password = user_name
    run_rabbit_cmd(['add_user', user_name, password])
    run_rabbit_cmd(['set_permissions', '-p', '/', user_name, '".*" ".*" ".*"'])


def delete_rabbit_user(user_name):
    run_rabbit_cmd(['delete_user', user_name])


def expose_rabbit(test, rabbit_address, user, password):
    """Expose the rabbit server for the test duration.

    This also patches ampq_utils.get_config() to provide the created user
    credentials which will be used to connect to the exposed rabbit.

    :param test: The test that needs an exposed rabbit.

    :param rabbit_address: The rabbit IP address.

    :param user: The rabbit use that will exist for the test duration.

    :param password: The password for the user.

    """
    juju_expose(RABBIT_SERVICE)
    test.addCleanup(juju_expose, RABBIT_SERVICE, False)

    class AmqpConfig(object):

        def __init__(self):
            self.AMQP_HOST = rabbit_address
            self.AMQP_VHOST = '/'
            self.AMQP_USER = user
            self.AMQP_PASSWORD = password

    create_rabbit_user(user, password)
    test.addCleanup(delete_rabbit_user, user)
    # Install the created user as the amqp config for the test duration
    fixtures.patch(test, amqp_utils, 'get_config', AmqpConfig)
