#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import uuid
import unittest
import deployers
import deploy

from ci_utils import amqp_utils
from ucitests import assertions


class TestPPACreator(deployers.DeployerTest):
    """Integration tests for ppacreator service run on a juju deployment"""

    def setUp(self):
        super(TestPPACreator, self).setUp()
        status = deploy.juju_status()
        units = status['services']['ci-airline-rabbit']['units']
        rabbit_ip = units['ci-airline-rabbit/0']['public-address']
        deployers.expose_rabbit(self, rabbit_ip, 'tester', 's3cr3t')

    def drain_queue(self, queue_name):
        self.messages = []

        def on_message(msg):
            body = json.loads(msg.body)
            self.messages.append(body)
            msg.channel.basic_ack(msg.delivery_tag)
            if body.get('exit'):
                # this will kick us out of the run-forever worker loop, relying
                # on the worker last message setting 'exit'
                msg.channel.basic_cancel(msg.consumer_tag)

        conf = amqp_utils.get_config()
        # Process all messages and delete the queue when done
        amqp_utils.process_queue(conf, queue_name, on_message, delete=True)
        return self.messages

    def assertStatusMsg(self, expected, msg):
        self.assertEqual('STATUS', msg['state'])
        without_state = msg.copy()
        del without_state['state']
        self.assertDictEqual(expected, without_state)

    def test_worker_running(self):
        """Ensure the ppacreator worker is deployed and running."""
        self.assert_job_running('ci-airline-ppa-creator-worker')

    def test_create_ppa(self):
        """Ensure that ppa's can be created or their name can
           be retrieved if they are already existing"""
        unique_id = uuid.uuid4()
        progress_queue_name = 'progress-trigger'
        params = dict(
            name='ppacreator-inttest-{}'.format(unique_id),
            displayname='ppacreator-inttest-{}'.format(unique_id),
            description='int-test-ppa-description',
            progress_trigger=progress_queue_name,
            ticket_id='int-test-ppa-ticket-id',
            architectures=('all'),
            series='precise')

        amqp_utils.send(amqp_utils.PPA_CREATOR_QUEUE, json.dumps(params),
                        raise_errors=True)
        msgs = self.drain_queue(progress_queue_name)
        assertions.assertLength(self, 3, msgs)
        self.assertStatusMsg({u'message': u'Creating PPA {} for ticket, '
                              'int-test-ppa-ticket-id'.format(params['name'])},
                             msgs[1])

        final = msgs[2]
        self.assertEqual('COMPLETED', final['state'])
        self.assertEqual('PASSED', final['result'])

if __name__ == "__main__":
    unittest.main()
