# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import unittest

import deployers
from ucitests import fixtures

from ci_utils import amqp_utils
from ci_utils.testing import features


RABBIT_SERVICE = 'ci-airline-rabbit'


# FIXME: We need more tests around prefetch_size=0 to demonstrate workers being
# starved -- vila 2014-10-23
@features.requires(features.bootstrapped_juju)
class TestRabbit(deployers.DeployerTest):

    def setUp(self):
        super(TestRabbit, self).setUp()
        self.create_user('tester', 's3cr3t')
        self.addCleanup(self.delete_user, 'tester')
        deployers.juju_expose(RABBIT_SERVICE)
        self.addCleanup(deployers.juju_expose, RABBIT_SERVICE, False)

        class AmqpConfig(object):

            def __init__(inner):
                ip, _ = self.get_ip_and_port(RABBIT_SERVICE)
                inner.AMQP_HOST = ip
                inner.AMQP_VHOST = '/'
                inner.AMQP_USER = 'tester'
                inner.AMQP_PASSWORD = 's3cr3t'
        fixtures.patch(self, amqp_utils, 'get_config', AmqpConfig)

    def run_ctl_cmd(self, args):
        deployers.juju_run(RABBIT_SERVICE,
                           'sudo rabbitmqctl {}'.format(' '.join(args)))

    def create_user(self, user_name, password=None):
        if password is None:
            password = user_name
        self.run_ctl_cmd(['add_user', user_name, password])
        self.run_ctl_cmd(['set_permissions', '-p', '/',
                          user_name, '".*" ".*" ".*"'])

    def delete_user(self, user_name):
        self.run_ctl_cmd(['delete_user', user_name])

    def assertConsumed(self, channel, queue_name, callback, check):
        tag = channel.basic_consume(callback=callback, queue=queue_name)
        # Calling consume only setup the queue to trigger the callback when a
        # message is received, we need some monitoring to ensure the messages
        # are processed. For the present tests, a single call is expected.
        while not check():
            channel.wait()
        # We're done with the queue, cancel the subscription
        channel.basic_cancel(tag)

    def test_acked_message(self):
        amqp_utils.send('myqueue', 'hello')
        conn, channel = amqp_utils.declare_queue('myqueue')
        self.addCleanup(conn.close)
        self.addCleanup(channel.close)
        self.addCleanup(channel.queue_delete, 'myqueue')
        channel.basic_qos(prefetch_size=0, prefetch_count=1, a_global=False)
        self.ack_called = False

        def ack(msg):
            self.assertEqual('hello', msg.body)
            msg.channel.basic_ack(msg.delivery_tag)
            self.ack_called = True

        self.assertConsumed(channel, 'myqueue', ack, lambda: self.ack_called)

    def test_nacked_message(self):
        amqp_utils.send('myqueue', 'hello')
        conn, channel = amqp_utils.declare_queue('myqueue')
        self.addCleanup(conn.close)
        channel.basic_qos(prefetch_size=0, prefetch_count=1, a_global=False)
        self.nack_called = False

        def nack(msg):
            self.assertEqual('hello', msg.body)
            msg.channel.basic_reject(msg.delivery_tag, requeue=True)
            self.nack_called = True

        self.assertConsumed(channel, 'myqueue', nack, lambda: self.nack_called)
        # Bounce the message by closing the channel otherwise it stays there
        # waiting for another callback to consume it.
        channel.close()

        # Use a different channel to get the message again
        conn, channel = amqp_utils.declare_queue('myqueue')
        self.addCleanup(conn.close)
        self.addCleanup(channel.close)
        self.addCleanup(channel.queue_delete, 'myqueue')
        channel.basic_qos(prefetch_size=0, prefetch_count=1, a_global=False)
        self.ack_called = False

        def ack(msg):
            self.assertEqual('hello', msg.body)
            msg.channel.basic_ack(msg.delivery_tag)
            self.ack_called = True

        self.assertConsumed(channel, 'myqueue', ack, lambda: self.ack_called)

if __name__ == '__main__':
    unittest.main()
