#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import json
import unittest

from sst.actions import (
    assert_title,
    click_button,
    go_to,
    get_element,
    get_element_by_css,
    set_base_url,
    wait_for,
    write_textfield,
)
from sst import (
    browsers,
    cases,
)

import deployers


class WebUITest(cases.SSTTestCase):

    browser_factory = browsers.PhantomJSFactory()

    def setUp(self):
        """Sets browser base_url based on the available deployment."""
        super(WebUITest, self).setUp()
        set_base_url('http://{}/'.format(
            deployers.juju_get_ip_and_port('ci-airline-webui-apache')[0]))

    def test_home(self):
        # Explore the open ticket list (home).
        go_to('/')
        assert_title('Ubuntu CI Engine')

        # 'Tickets' is the active link.
        active_link = get_element_by_css('a.active')
        self.assertEqual('Tickets', active_link.text)

        # And the table headers are ...
        dt_header = wait_for(
            get_element_by_css, 'table.yui3-datatable-table thead')
        self.assertEqual(
            ['Title', 'Landing Progress', 'Review Status'],
            [t.text for t in dt_header.find_elements_by_tag_name('th')])

        # A stock YUI paginator is also present.
        paginator = wait_for(get_element_by_css, '#paginator div')
        self.assertEqual(
            'yui3-pagview-container', paginator.get_attribute("class"))

    def test_create_ticket(self):
        # Create a ticket via WebUI form.
        go_to('/create')
        assert_title('Ubuntu CI Engine')

        wait_for(get_element, tag='form')

        # Submit the form with the required fields.
        title = get_element_by_css('input[name="title"]')
        write_textfield(title, 'Testing WebUI')
        owner = get_element_by_css('input[name="owner"]')
        write_textfield(owner, 'foo@bar.com')
        submit = get_element_by_css('input[type="submit"]')
        click_button(submit, wait=True)

        # Redirects to the ticket page.
        ticket_title = wait_for(get_element, tag='h2')
        self.assertEqual('Testing WebUI', ticket_title.text)

    def test_json_status(self):
        # Ensure the webui charm is providing us a valid status_urls.json.
        go_to('/static/engine-health/status_urls.json')

        # browser automatically wraps JSON (non-html) into a <pre> ...
        pre = get_element_by_css('pre')
        json.loads(pre.text)

    def test_status_page(self):
        # Show we can load engine-health information.

        # XXX cprov 20141105: status page can be slow (> 30s) if the cache
        # is cold, we should probably work on this more agressively to
        # provide quicker information.

        go_to('/status/')
        assert_title('Ubuntu CI Engine')

        # 'Engine health' is the active link.
        active_link = get_element_by_css('a.active')
        self.assertEqual('Engine health', active_link.text)


if __name__ == "__main__":
    unittest.main()
