# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import httplib2
import json
import logging
import socket

from django.conf.urls import url
from django.core.urlresolvers import NoReverseMatch
from tastypie import fields
from tastypie.authorization import Authorization
from tastypie.constants import (
    ALL,
    ALL_WITH_RELATIONS,
)
from tastypie.exceptions import (
    BadRequest,
    NotFound,
)
from tastypie.paginator import Paginator
from tastypie.resources import (
    Bundle,
    ModelResource,
    Resource,
)
from tastypie.utils import trailing_slash

from ci_utils import ticket_states
from ci_utils.json_status import JSONStatus
from ticket.models import (
    get_enum_title,
    get_enum_value,
    MergeProposal,
    Review,
    ReviewStatus,
    SubTicket,
    SourcePackageUpload,
    SubTicketArtifact,
    SubTicketWorkflowStep,
    SubTicketWorkflowStepStatus,
    Ticket,
    TicketArtifact,
    TicketCiTrainOverlay,
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
    Workflow,
)
from project.api import SourcePackageResource


class PageNumberPaginator(Paginator):
    """Copes with YUI 'page'-number based pagination."""

    def get_offset(self):
        """Calculates 'offset' also considering optionally given 'page'.

        If 'offset' is given (different than the default value) it takes
        precedence over 'page'. Otherwise offset is calculated from the
        requested page number ('(page-1) * limit').

        Returns the calculated (or given) offset value.
        """
        offset = super(PageNumberPaginator, self).get_offset()
        # If offset is given, it takes precedence.
        if offset != self.offset:
            return offset
        # If no 'page' is given, default offset is used.
        page = self.request_data.get('page')
        if page is None:
            return offset
        # Validates 'page' as a 1-based integer.
        try:
            page = int(page)
        except ValueError:
            raise BadRequest(
                "Invalid page '{}' provided. Please provide an integer "
                "value.".format(page))
        if page < 1:
            raise BadRequest(
                "Invalid offset '{}' provided. Please provide a "
                "positive integer > 1.".format(page))
        # Calculate offset based on the page number.
        offset = (page - 1) * self.get_limit()
        return offset


class TicketTranslatedResource(ModelResource):

    def dehydrate_current_workflow_step(self, bundle):
        return get_enum_title(bundle.data["current_workflow_step"],
                              TicketWorkflowStep)

    def hydrate_current_workflow_step(self, bundle):
        value = bundle.data.get('current_workflow_step')
        if not value:
            return bundle
        try:
            int(value)
        except ValueError:
            # user passed us "Package building"
            value = get_enum_value(value, TicketWorkflowStep)
            bundle.data['current_workflow_step'] = value
        return bundle

    def dehydrate_status(self, bundle):
        return get_enum_title(bundle.data["status"], TicketWorkflowStepStatus)

    def hydrate_status(self, bundle):
        value = bundle.data.get('status')
        if not value:
            return bundle
        try:
            int(value)
        except ValueError:
            # user passed us "Continue"
            value = get_enum_value(value, TicketWorkflowStepStatus)
            bundle.data['status'] = value
        return bundle

    def hydrate_subtickets(self, bundle):
        """Prevent creation of conflicting (duplicated) `Subticket`s."""
        subtickets = bundle.data.get('subtickets')
        if subtickets is None:
            return bundle

        for subticket in subtickets:
            if isinstance(subticket, basestring):
                continue
            elif not isinstance(subticket, dict):
                sp = subticket.data['sourcepackage']
            else:
                sp = subticket['sourcepackage']

            # Already a resource URI.
            if not isinstance(sp, dict):
                continue

            # Set the 'id' attribute if the subticket already exists,
            # it will prevent tastypie to create a conflict one.
            cls = SubTicketResource._meta.object_class
            try:
                obj = cls.objects.get(
                    ticket=bundle.obj, sourcepackage__name=sp['name'])
            except cls.DoesNotExist:
                continue
            subticket['id'] = obj.id

        return bundle

    def hydrate_workflow(self, bundle):
        """Re-use existing `Workflow` when creating or updating tickets.

        It effectively disables workflow creation and update. It is only
        possible to 'refer' to and existing `Workflow`.
        """
        workflow = bundle.data.get('workflow')
        if not isinstance(workflow, dict):
            return bundle
        cls = WorkflowResource._meta.object_class
        try:
            obj = cls.objects.get(pk=workflow.get('name'))
        except cls.DoesNotExist:
            bundle.data.pop('workflow')
            return bundle

        # If the refered workflow exists, we augment the given specification
        # ('name') with 'steps'. This is the only way tastypie will avoid
        # model creation, because workflows are not exposed in the API and
        # have no 'resource_uri'.
        bundle.data['workflow'].update({'steps': obj.steps})

        return bundle

    def prepend_urls(self):
        """Prepend URL pattern for UUID-based ticket URIs."""
        uuid_regexp = r'[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}'
        return [
            url(r'^(?P<resource_name>{})/(?P<uuid>{}){}$'.format(
                self._meta.resource_name, uuid_regexp, trailing_slash()),
                self.wrap_view('dispatch_detail'),
                name="api_dispatch_detail_uuid"),
        ]

    def get_resource_uri(self, bundle_or_obj=None,
                         url_name='api_dispatch_list'):
        """Use UUID-based URL pattern for building resource URIs."""
        if bundle_or_obj is None:
            return super(TicketTranslatedResource, self).get_resource_uri(
                bundle_or_obj, url_name)

        kwargs = self.resource_uri_kwargs(bundle_or_obj)
        kwargs.pop('pk')
        if isinstance(bundle_or_obj, Bundle):
            kwargs['uuid'] = getattr(bundle_or_obj.obj, 'uuid')
        else:
            kwargs['uuid'] = getattr(bundle_or_obj, 'uuid')

        try:
            return self._build_reverse_url(
                'api_dispatch_detail_uuid', kwargs=kwargs)
        except NoReverseMatch:
            return ''


class SourcePackageUploadResource(ModelResource):

    subticket = fields.ToOneField(
        'ticket.api.SubTicketResource', 'subticket', full=False)

    class Meta:
        queryset = SourcePackageUpload.objects.all()
        allowed_methods = ['get', 'post']
        authorization = Authorization()
        resource_name = 'spu'


class MergeProposalResource(ModelResource):

    subticket = fields.ToOneField(
        'ticket.api.SubTicketResource', 'subticket', full=False)

    class Meta:
        queryset = MergeProposal.objects.all()
        allowed_methods = ['get', 'post']
        authorization = Authorization()
        resource_name = 'mp'


class TicketResource(TicketTranslatedResource):

    # XXX cprov 20140930: the more we understand and polish the way we
    # use tastypie, more this FULL<something> resources makes no sense.
    # We are exposing 'Ticket.subtickets' resource to allow them to be
    # created within the same transaction and it's not 'full' yet just
    # to avoid a bigger test-fixing task.
    subtickets = fields.ToManyField(
        'ticket.api.SubTicketResource', 'subticket_set',
        related_name='ticket', null=True, full=False,
        help_text="Related `SubTicket`s URIs.")

    citrain_overlay = fields.ToOneField(
        'ticket.api.TicketCiTrainOverlayResource', 'citrain_overlay',
        related_name='ticket', null=True, blank=True, unique=True,
        full=True, help_text='Extended fields for CI-Train.')

    reviews = fields.ToManyField(
        'ticket.api.ReviewResource', 'review_set', related_name='ticket',
        null=True, full=True, help_text='Related `Review`s URIs.')

    workflow = fields.ForeignKey(
        'ticket.api.WorkflowResource', 'workflow',
        blank=True, full=True,
        help_text='Ticket `Workflow` information.')

    def build_filters(self, filters=None):
        """Override filters generated by tastypie.

        This is specially target to replace misleading references to
        related m2m objects (in this case 'reviews') which operated
        via 'review_set' but looked up as 'review' (django knows how
        to find the reference).
        """
        filters = super(TicketResource, self).build_filters(filters)
        return {
            k.replace('review_set', 'review'): v
            for k, v in filters.iteritems()
        }

    class Meta:
        # XXX cprov 20140829: 'private' should only be considered via
        # authorization mechanisms, not on the local resource queryset.
        queryset = Ticket.objects.filter(private=False)
        allowed_methods = ['get', 'post', 'patch']
        authorization = Authorization()
        filtering = {
            "status": ALL,
            'reviews': ALL_WITH_RELATIONS,
        }
        ordering = ['id']
        paginator_class = PageNumberPaginator


class TicketCiTrainOverlayResource(ModelResource):
    """`TicketCiTrainOverlay` model resource.

    This resource is only used embedded in a `Ticket` and is not exposed
    in the API.
    """

    class Meta:
        authorization = Authorization()
        list_allowed_methods = []
        detail_allowed_methods = []
        excludes = ['id']
        include_resource_uri = False
        queryset = TicketCiTrainOverlay.objects.all()


class SubTicketTranslatedResource(ModelResource):

    def dehydrate_current_workflow_step(self, bundle):
        return get_enum_title(bundle.data["current_workflow_step"],
                              SubTicketWorkflowStep)

    def dehydrate_status(self, bundle):
        return get_enum_title(bundle.data["status"],
                              SubTicketWorkflowStepStatus)

    conflicts = fields.ToManyField(
        TicketResource, 'conflicts', readonly=True, null=True, full=False)


class SubTicketResource(SubTicketTranslatedResource):

    ticket = fields.ToOneField(TicketResource, 'ticket', full=True)

    sourcepackage = fields.ToOneField(
        SourcePackageResource, 'sourcepackage', full=True)

    source_package_uploads = fields.ToManyField(
        'ticket.api.SourcePackageUploadResource', 'source_package_uploads',
        related_name='subticket', null=True, full=True,
        help_text="Related `SourcePackageUpload`s URIs.")

    merge_proposals = fields.ToManyField(
        'ticket.api.MergeProposalResource', 'merge_proposals',
        related_name='subticket', null=True, full=True,
        help_text="Related `MergeProposal`s URIs.")

    class Meta:
        queryset = SubTicket.objects.all()
        allowed_methods = ['get', 'post']
        authorization = Authorization()

    def hydrate_sourcepackage(self, bundle):
        """Re-use existing `SourcePackage` when creating new SubTickets."""
        sourcepackage = bundle.data.get('sourcepackage')
        # If 'sourcepackage' is not and dictionary, it is already a
        # SourcePackage 'resource_uri' or bogus. Either way, tastypie
        # will handle it properly.
        if not isinstance(sourcepackage, dict):
            return bundle
        # If there is no existing `SourcePackage` matching the given details
        # (name, really), dismiss and let tastypie create a new one.
        cls = SourcePackageResource._meta.object_class
        try:
            obj = cls.objects.get(**sourcepackage)
        except cls.DoesNotExist:
            return bundle

        # Replace the given `SourcePackage` details by the existing
        # object 'resource_uri'.
        bundle.data['sourcepackage'] = self.get_resource_uri(obj)

        return bundle


class TicketArtifactResource(ModelResource):
    ticket = fields.ToOneField(
        TicketResource, 'ticket', full=True)

    class Meta:
        queryset = TicketArtifact.objects.all()
        allowed_methods = ['get', 'post']
        authorization = Authorization()
        filtering = {
            'type': ['exact'],
            'ticket': ['exact'],
        }


class SubTicketArtifactResource(ModelResource):

    subticket = fields.ToOneField(
        SubTicketResource, 'subticket', full=True)

    class Meta:
        queryset = SubTicketArtifact.objects.all()
        allowed_methods = ['get', 'post']
        authorization = Authorization()


class FullTicketArtifactResource(ModelResource):

    class Meta:
        queryset = TicketArtifact.objects.all()
        allowed_methods = ['get']


class FullSubTicketArtifactResource(ModelResource):

    class Meta:
        queryset = SubTicketArtifact.objects.all()
        allowed_methods = ['get']


class FullSubTicketResource(SubTicketTranslatedResource):

    artifact = fields.ToManyField(FullSubTicketArtifactResource,
                                  'subticketartifact_set', full=True)

    sourcepackage = fields.ToOneField(
        SourcePackageResource, 'sourcepackage', full=True)

    source_package_uploads = fields.ToManyField(
        'ticket.api.SourcePackageUploadResource', 'source_package_uploads',
        related_name='subticket', null=True, full=True,
        help_text="Related `SourcePackageUpload`s URIs.")

    merge_proposals = fields.ToManyField(
        'ticket.api.MergeProposalResource', 'merge_proposals',
        related_name='subticket', null=True, full=True,
        help_text="Related `MergeProposal`s URIs.")

    class Meta:
        queryset = SubTicket.objects.all()
        allowed_methods = ['get']


class FullTicketResource(TicketTranslatedResource):

    subticket = fields.ToManyField(FullSubTicketResource, 'subticket_set',
                                   full=True)
    artifact = fields.ToManyField(FullTicketArtifactResource,
                                  'ticketartifact_set', null=True,
                                  full=True)

    citrain_overlay = fields.ToOneField(
        'ticket.api.TicketCiTrainOverlayResource', 'citrain_overlay',
        related_name='ticket', null=True, blank=True, unique=True,
        full=True, help_text='Extended fields for CI-Train.')

    reviews = fields.ToManyField(
        'ticket.api.ReviewResource', 'review_set', null=True, full=True)

    workflow = fields.ForeignKey(
        'ticket.api.WorkflowResource', 'workflow',
        blank=True, full=True,
        help_text='Ticket `Workflow` information.')

    class Meta:
        queryset = Ticket.objects.all()
        allowed_methods = ['get']


class SubTicketUpdateStatusResource(SubTicketTranslatedResource):

    class Meta:
        queryset = SubTicket.objects.all()
        fields = ['id', 'current_workflow_step', 'status']
        allowed_methods = ['get', 'patch']
        authorization = Authorization()
        resource_name = 'updatesubticketstatus'
        filtering = {
            "id": ('exact'),
        }


class SubTicketStatusResource(SubTicketTranslatedResource):

    class Meta:
        queryset = SubTicket.objects.all()
        fields = ['id', 'title', 'current_workflow_step', 'status']
        allowed_methods = ['get']
        filtering = {
            "current_workflow_step": ALL,
        }


class WorkflowStepItem(object):
    def __init__(self, initial=None):
        self.__dict__['_data'] = {}

        if hasattr(initial, 'items'):
            self.__dict__['_data'] = initial
        else:
            self.__dict__['_data']['name'] = 'unknown'
            self.__dict__['_data']['value'] = 0

    def __getattr__(self, name):
        return self._data.get(name, None)

    def __setattr__(self, name, value):
        self.__dict__['_data'][name] = value

    def to_dict(self):
        return self._data


class WorkflowStepResource(Resource):
    name = fields.CharField(attribute='name')
    key = fields.CharField(attribute='key')
    value = fields.IntegerField(attribute='value')

    enum_type = None

    class Meta:
        abstract = True
        allowed_methods = ['get']
        object_class = WorkflowStepItem

    def get_object_list(self, request):
        results = []
        for item in self.enum_type:
            name = item.title
            key = item.token
            try:
                value = getattr(self.enum_type, key).value
            except AttributeError:
                continue

            wfs = WorkflowStepItem(
                initial={
                    'name': name,
                    'value': value,
                    'key': key,
                }
            )
            results.append(wfs)

        return results

    def obj_get_list(self, bundle, **kwargs):
        return self.get_object_list(bundle.request)

    def obj_get(self, bundle, **kwargs):
        key = kwargs['pk']

        result = {}
        try:
            value = getattr(self.enum_type, key).value
            name = getattr(self.enum_type, key).title
            result = WorkflowStepItem(
                initial={
                    'name': name,
                    'value': value,
                    'key': key,
                }
            )
        except AttributeError:
            pass

        return result

    def detail_uri_kwargs(self, bundle_or_obj):
        kwargs = {}

        if isinstance(bundle_or_obj, Bundle):
            kwargs['pk'] = bundle_or_obj.obj.key
        else:
            kwargs['pk'] = bundle_or_obj.key

        return kwargs


class TicketWorkflowStepResource(WorkflowStepResource):

    enum_type = ticket_states.TicketWorkflowStep

    class Meta(WorkflowStepResource.Meta):
        resource_name = 'ticket_workflow_step'


class SubTicketWorkflowStepResource(WorkflowStepResource):

    enum_type = ticket_states.SubTicketWorkflowStep

    class Meta(WorkflowStepResource.Meta):
        resource_name = 'subticket_workflow_step'


class TicketWorkflowStepStatusResource(WorkflowStepResource):

    enum_type = ticket_states.TicketWorkflowStepStatus

    class Meta(WorkflowStepResource.Meta):
        resource_name = 'ticket_workflow_step_status'


class SubTicketWorkflowStepStatusResource(WorkflowStepResource):

    enum_type = ticket_states.SubTicketWorkflowStepStatus

    class Meta(WorkflowStepResource.Meta):
        resource_name = 'subticket_workflow_step_status'


class ReviewResource(ModelResource):

    ticket = fields.ForeignKey(TicketResource, 'ticket')

    def dehydrate_workflow_step(self, bundle):
        return get_enum_title(bundle.data["workflow_step"],
                              TicketWorkflowStep)

    def hydrate_workflow_step(self, bundle):
        value = bundle.data.get('workflow_step')
        if not value:
            return bundle
        try:
            int(value)
        except ValueError:
            # user passed us "Package building"
            value = get_enum_value(value, TicketWorkflowStep)
            bundle.data['workflow_step'] = value
        return bundle

    def dehydrate_status(self, bundle):
        return get_enum_title(bundle.data["status"], ReviewStatus)

    def hydrate_status(self, bundle):
        value = bundle.data.get('status')
        if not value:
            return bundle
        try:
            int(value)
        except ValueError:
            # user passed us "Approved"
            value = get_enum_value(value, ReviewStatus)
            bundle.data['status'] = value
        return bundle

    class Meta:
        queryset = Review.objects.all()
        authorization = Authorization()
        allowed_methods = ['get', 'post', 'patch', 'delete']
        filtering = {
            'ticket': ALL_WITH_RELATIONS,
            'review_type': ALL,
            'status': ALL,
        }


class WorkflowResource(ModelResource):
    """Workflow resource spec. Not exposed via the API."""

    class Meta:
        queryset = Workflow.objects.all()
        authorization = Authorization()
        list_allowed_methods = []
        detail_allowed_methods = []
        include_resource_uri = False

    def dehydrate_steps(self, bundle):
        """Preserve JSON format of this field."""
        return json.loads(bundle.data.get('steps'))


class TSStatus(ModelResource):
    class Meta():
        resource_name = 'status'
        allowed_methods = ['get']
        # return an empty queryset to make tastypie happy

    def get_list(self, request, **kwargs):
        status = JSONStatus()
        status.add_rabbit_configured()
        status.add_deployed_revision()
        create_check = self._check_create_ticket()
        status.add_true_false('ticket creation works', create_check,
                              create_check is True)
        return self.create_response(request, status.results)

    def _get_ip(self):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("launchpad.net", 80))
            ip = s.getsockname()[0]
            s.close()
        except:
            logging.exception('ERROR: unable to determine local IP')
            ip = 'IP lookup error'
        return ip

    def _check_create_ticket(self):
        ip = self._get_ip()
        if ip == 'IP lookup error':
            return ip
        url = 'http://{}:8080/api/v1/ticket/'.format(ip)
        ticket_content = {"owner": "create@example.com",
                          "title": "Created ticket",
                          "description": "This ticket is for creation test",
                          "bug_id": "111111",
                          }
        client = httplib2.Http()
        resp, content = client.request(
            uri=url,
            method='POST',
            headers={'Content-Type': 'application/json'},
            body=json.dumps(ticket_content),
        )
        if resp['status'] != '201':
            return False
        ticket_uuid = resp['location'].split('/')[-2]
        Ticket.objects.get(uuid=ticket_uuid).delete()
        return True


class ResourceWrapper(object):
    """Minimal wrapper class for abstract (non-ORM) resources.

    It optionally copies given 'fields' attributes (supports nested specs
    and callables) from a given 'obj'.

    It allows complete model/field instrospection (tastypie-style) and
    isolatation between the exposed resource and the copied (mirrored)
    model.
    """

    def __init__(self, **kwargs):
        self.__dict__['_info'] = {}
        # Dismiss if 'obj' or 'fields' are not given.
        obj = kwargs.get('obj')
        if obj is None:
            return
        fields = kwargs.get('fields', [])
        # Build local attributes with values mirrored from the given
        # object.
        for name, field in fields.items():
            if not field.attribute:
                continue
            try:
                value = reduce(getattr, field.attribute.split('.'), obj)
            except AttributeError:
                value = ''
            else:
                if callable(value):
                    value = value()
            self.__dict__['_info'][name] = value

    def __getattr__(self, name):
        return self._info.get(name, None)

    def to_dict(self):
        return self._info


class UserInfoResource(Resource):
    """Expose users details for the current requestor."""

    username = fields.CharField(
        'username', readonly=True,
        help_text='Username of the authenticated user')
    email = fields.CharField(
        'email', readonly=True,
        help_text='Email address of the authenticated user.')
    fullname = fields.CharField(
        'get_full_name', readonly=True,
        help_text='Full name (<first> <last>) of the authenticated user.')
    superuser = fields.BooleanField(
        'is_superuser', readonly=True,
        help_text='Whether the authenticated user is a superuser or not.')
    staff = fields.BooleanField(
        'is_staff', readonly=True,
        help_text='Whether the authenticated user is a staff member or not.')
    authenticated = fields.BooleanField(
        'is_authenticated', readonly=True,
        help_text='Whether the user is authenticated or not.')

    class Meta():
        resource_name = 'userinfo'
        object_class = ResourceWrapper
        allowed_methods = ['get']

    def get_list(self, request, **kwargs):
        """Re-implement 'userinfo' resource list.

        It actually returns a single userinfo resource corresponding to the
        current requestor.
        """
        user_info = ResourceWrapper(obj=request.user, fields=self.fields)
        return self.create_response(request, user_info.to_dict())

    def get_object_list(self, request):
        """Satisfy tastypie schema instrospection (/<resource>/schema)."""
        user_info = ResourceWrapper(obj=request.user, fields=self.fields)
        return [user_info]

    def obj_get(self, bundle, **kwargs):
        """Render a 404 for 'userinfo' details lookups."""
        raise NotFound('UserInfo does not allow details lookups')
