# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import uuid
from django.core.exceptions import (
    ObjectDoesNotExist,
    ValidationError,
)
from django.db import models
from django.conf import settings
from project.models import SourcePackage

from ci_utils import amqp_utils, SUPPORTED_SERIES
from ci_utils.data_store import create_for_ticket
from ci_utils.unit_config import get_auth_config
from ci_utils import unit_config

from ci_utils.ticket_states import (
    ReviewStatus,
    STEP_TO_WORKFLOW_MAPPING,
    SubTicketWorkflowStep,
    SubTicketWorkflowStepStatus,
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


def _choices(enum):
    return [
        (int(item.value), item.title) for item in enum.items]


def get_enum_title(value, enumcls):
    return enumcls.items.mapping[value].title


def get_enum_value(title, enumcls):
    for item in enumcls.items:
        if item.title == title:
            return item.value


def get_ticket_uuid():
    """UUIDv1 factory for tickets.

    Returns the textual representation of a new UUIDv1.
    """
    return str(uuid.uuid1())


class Workflow(models.Model):
    '''A ticket is configured with a workflow. This specifies what steps the
    lander needs to handle. We store these in a table so that we can
    add/remove/update workflows in production without having to do an upgrade.
    '''
    name = models.CharField(
        primary_key=True,
        max_length=4096,
        help_text='A human readable label')
    steps = models.CharField(
        max_length=4096,
        help_text='JSON encoded data that specifying the lander steps')

    def __unicode__(self):
        return self.name

    def save(self, *args, **kwargs):
        '''ensure we don't allow invalid json'''
        try:
            json.loads(self.steps)
        except:
            raise ValidationError('Invalid JSON provided for "steps"')
        return super(Workflow, self).save(*args, **kwargs)

    @staticmethod
    def initialize():
        workflows = [{
            'name': 'default',
            'steps': json.dumps([
                {'name': 'ppa_creator'},
                {'name': 'bsbuilder'},
                {'name': 'validator'},
                {'name': 'image_builder'},
                {'name': 'test_runner'},
                {'name': 'publisher'},
            ]),
        }, {
            'name': 'citrain',
            'steps': json.dumps([
                {'name': 'silo_creating'},
                {'name': 'silo_building'},
                {'name': 'silo_testing'},
                {'name': 'silo_publishing'}
            ]),
        }]
        for wf in workflows:
            Workflow.objects.get_or_create(name=wf['name'], defaults=wf)


class Ticket(models.Model):
    class Meta:
        db_table = 'ticket'
        ordering = ['id']

    SERIES_CHOICES = tuple(
        (series, series.upper()) for series in SUPPORTED_SERIES)

    uuid = models.CharField(
        max_length=36, editable=False, unique=True, default=get_ticket_uuid,
        help_text=('UUIDv1 ticket identification represented as '
                   'string (aka \'hyphenated\').'))
    owner = models.EmailField(max_length=254)
    creator = models.EmailField(max_length=254, null=True, blank=True)
    title = models.CharField(max_length=4096)
    description = models.TextField(blank=True)
    bug_id = models.IntegerField(null=True, blank=True)
    current_workflow_step = models.IntegerField(
        choices=_choices(TicketWorkflowStep), max_length=4096,
        default=000)
    status = models.IntegerField(
        choices=_choices(TicketWorkflowStepStatus), max_length=4096,
        default=000)
    base_image = models.CharField(max_length=4096,
                                  default=settings.BASE_IMAGE_DEFAULT)
    series = models.CharField(max_length=4096,
                              choices=SERIES_CHOICES,
                              default=settings.SERIES_DEFAULT)
    master_ppa = models.CharField(max_length=4096,
                                  default=settings.MASTER_PPA_DEFAULT)
    build_ppa = models.CharField(
        max_length=4096, blank=True,
        help_text='Ticket-specific PPA for building context sources.')
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    workflow = models.ForeignKey(
        Workflow, default='default',
        help_text='Ticket `Workflow` information.')

    lander_unit = models.IntegerField(
        default=0,
        help_text='The lander\'s juju unit number processing this ticket')

    private = models.BooleanField(default=False)

    def __unicode__(self):
        return self.title

    def create_container(self):
        ds = create_for_ticket(self.uuid, get_auth_config())
        ds.ensure_swift_client()

    def save(self, *args, **kwargs):
        self.base_image = unit_config.get('image_map').get(self.series)
        if self.base_image is None:
            self.base_image = ''

        is_new = self.pk is None

        if self.creator is None:
            self.creator = self.owner
        super(Ticket, self).save(*args, **kwargs)

        # Create Swift ticket container when the ticket is created.
        # It has to be done after saving because it need the ticket ID.
        if is_new:
            self.create_container()

        # Skip Lander-job dispaching if it's a CI-Train ticket (which is
        # handled externally by pooling, atm).
        try:
            self.citrain_overlay
        except ObjectDoesNotExist:
            self._handle_lander()

    def _handle_lander(self):
        notify = False
        if int(self.current_workflow_step) == int(TicketWorkflowStep.QUEUED):
            notify = (int(self.status) == int(TicketWorkflowStepStatus.QUEUED))

        resume = (int(self.status) == int(TicketWorkflowStepStatus.CONTINUE))

        if notify or resume:
            params = {'id': self.uuid}
            if notify:
                params['steps'] = self._get_workflow_steps()
            if resume:
                params['resume'] = True
            amqp_utils.send(amqp_utils.LANDER_QUEUE, json.dumps(params))

    def _get_workflow_steps(self):
        '''build workflow steps injecting pauses where reviews are needed'''
        steps = json.loads(self.workflow.steps)
        reviews = self.review_set.exclude(
            status=ReviewStatus.APPROVED.value)
        pauses = [x.workflow_step for x in reviews]
        for step in steps:
            step_val = STEP_TO_WORKFLOW_MAPPING.get(step['name'], -1)
            if step_val in pauses:
                step['pause'] = {}
        return steps


class TicketCiTrainOverlay(models.Model):
    """Attaches CI-Train additional fields to the `Ticket` model."""

    ticket = models.OneToOneField(
        Ticket, primary_key=True, related_name='citrain_overlay',
        help_text='CI-Train specific ticket fields.')

    comments = models.TextField(
        help_text='General comments (whiteboard-style) about this ticket.')

    sync_request = models.CharField(
        max_length=256,
        help_text=('Sync ticket request specification '
                   '(e.g. sync:XXX for RTM).'))

    landers = models.TextField(
        help_text=('List of IRC nicknames, separated by space, to be '
                   'notified on ticket changes.'))

    test_notes = models.TextField(
        help_text='General notes and plans for testing this change.')

    job_url = models.URLField(
        help_text='Corresponding Jenkins job URL.')

    class Meta:
        db_table = 'ticketcitrainoverlay'


class SourcePackageUpload(models.Model):
    """Source Package Upload in the context of a subticket.

    This model represents a source package upload directly made by the user
    via CLI (or dput to the build_ppa) or a indirect upload done by the
    system after merging one or more `MergeProposal`s in the same context.
    """

    subticket = models.ForeignKey(
        'SubTicket', related_name='source_package_uploads',
        help_text='SubTicket context of this upload.')

    version = models.CharField(
        max_length=254,
        help_text=('Holds the literal package version of this source '
                   'package upload.'))

    architecture = models.CharField(
        max_length=254,
        help_text=(
            'Holds the literal DSC Architecture content of'
            'this source package upload.'))

    class Meta:
        db_table = 'sourcepackageupload'
        unique_together = ('subticket', 'version')

    def __unicode__(self):
        return "{}_{} ({})".format(
            self.subticket, self.version, self.architecture)


class MergeProposal(models.Model):
    """Merge Proposal in the context of a subticket."""

    subticket = models.ForeignKey(
        'SubTicket', related_name='merge_proposals',
        help_text='SubTicket context of this merge proposal.')

    lp_url = models.CharField(
        max_length=511,
        help_text='Launchpad merge proposal URL (lp:xxx syntax).')

    approved_revno = models.IntegerField(
        null=True, blank=True,
        help_text='Branch revision number approved for this cycle.')

    source_package_upload = models.ForeignKey(
        SourcePackageUpload, null=True, blank=True,
        help_text='Current SPU generated by this merge proposal.')

    class Meta:
        db_table = 'mergeproposal'
        unique_together = ('subticket', 'lp_url')

    def __unicode__(self):
        return '{} {}@{}'.format(
            self.subticket, self.lp_url, self.approved_revno)

    @property
    def proposed_branch(self):
        return "lp:{}".format(self.lp_url.split("/+merge")[0])


class SubTicket(models.Model):
    class Meta:
        db_table = 'subticket'
        ordering = ['id']
        unique_together = ('ticket', 'sourcepackage')

    ticket = models.ForeignKey(Ticket)
    sourcepackage = models.ForeignKey(SourcePackage)
    current_workflow_step = models.IntegerField(
        choices=_choices(SubTicketWorkflowStep), max_length=4096,
        default=000)
    status = models.IntegerField(
        choices=_choices(SubTicketWorkflowStepStatus), max_length=4096,
        default=000)
    assignee = models.EmailField(max_length=254)

    @property
    def conflicts(self):
        """Return `Ticket`s that conflict with this subticket task.

        A ticket is considered a conflict if it contains a `SubTicket`
        targeted to the same `SourcePackage` than this `SubTicket`. I.e.
        conflicts may happen between MPs and a SPUs.

        Only *active* (not NEW|FAILED|COMPLETED) tickets are considered
        for conflicts.
        """
        conflicting_tickets = Ticket.objects.filter(
            models.Q(subticket__sourcepackage=self.sourcepackage),
            ~models.Q(pk=self.ticket.pk),
            ~models.Q(status=TicketWorkflowStepStatus.FAILED),
            current_workflow_step__range=(
                TicketWorkflowStep.SILO_BUILDING,
                TicketWorkflowStep.PKG_PUBLISHING),
        )

        return conflicting_tickets

    def __unicode__(self):
        return "{} {}".format(self.ticket, self.sourcepackage)


class ArtifactBase(models.Model):
    class Meta:
        abstract = True

    # 'reference' provided by the artifact manager
    reference = models.CharField(max_length=4096)
    name = models.CharField(max_length=4096)


class SubTicketArtifact(ArtifactBase):
    SUBTICKET_ARTIFACT_TYPE = (
        ('RESULTS', 'Test Results'),
        ('SPU', 'Source Package Upload'),
        ('LOGS', 'Logs'),
    )

    class Meta:
        db_table = 'subticket_artifact'

    type = models.CharField(choices=SUBTICKET_ARTIFACT_TYPE, max_length=4096)
    subticket = models.ForeignKey(SubTicket)

    def __unicode__(self):
        return "{} {}".format(self.name, self.type)


class TicketArtifact(ArtifactBase):
    TICKET_ARTIFACT_TYPE = (
        ('RESULTS', 'Test Results'),
        ('LOGS', 'Logs'),
        ('IMAGE', 'Image'),
    )

    class Meta:
        db_table = 'ticket_artifact'

    type = models.CharField(choices=TICKET_ARTIFACT_TYPE, max_length=4096)
    ticket = models.ForeignKey(Ticket)

    def __unicode__(self):
        return "{} {}".format(self.name, self.type)


class Review(models.Model):

    ticket = models.ForeignKey(
        Ticket,
        help_text='Context ticket of this review.')

    workflow_step = models.IntegerField(
        choices=_choices(TicketWorkflowStep),
        help_text='Workflow step where the review is required.')

    review_type = models.CharField(
        max_length=254, null=False, blank=False,
        help_text='What type of review is needed.')

    status = models.IntegerField(
        choices=_choices(ReviewStatus),
        default=ReviewStatus.PENDING.value,
        help_text='Current review status.')

    def save(self, *args, **kwargs):
        """If all reviews are approved, resume the ticket processing."""
        super(Review, self).save(*args, **kwargs)
        all_approved = not self.ticket.review_set.exclude(
            status=ReviewStatus.APPROVED.value).exists()
        if all_approved:
            if int(self.ticket.status) == int(TicketWorkflowStepStatus.PAUSED):
                self.ticket.status = TicketWorkflowStepStatus.CONTINUE
                self.ticket.save()

    def __unicode__(self):
        return '{} - {}: {}'.format(
            self.ticket.uuid, self.review_type,
            get_enum_title(self.status, ReviewStatus))
