# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

import mock

from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.test import TestCase
from project.models import SourcePackage
from ticket.models import (
    MergeProposal,
    Review,
    ReviewStatus,
    SourcePackageUpload,
    SubTicket,
    SubTicketArtifact,
    SubTicketWorkflowStep,
    SubTicketWorkflowStepStatus,
    Ticket,
    TicketArtifact,
    TicketCiTrainOverlay,
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
    Workflow,
)


def create_sourcepackage(name='my-package'):
    sourcepackage = SourcePackage()
    sourcepackage.name = name
    sourcepackage.save()
    return sourcepackage


def create_sourcepackageupload(subticket, version='1.1', architecture='all'):
    spu = SourcePackageUpload()
    spu.subticket = subticket
    spu.version = version
    spu.architecture = architecture
    spu.save()
    return spu


def create_merge_proposal(
        subticket, lp_url='~foobar/my-project/my-branch/+merge/123456',
        approved_revno=100):
    merge_proposal = MergeProposal()
    merge_proposal.subticket = subticket
    merge_proposal.lp_url = lp_url
    merge_proposal.approved_revno = approved_revno
    merge_proposal.save()
    return merge_proposal


def create_ticket(title="This is my ticket",
                  description="this is my ticket description",
                  bug_id='12345', owner='test@example.com',
                  current_workflow_step=TicketWorkflowStep.PKG_BUILDING,
                  status=TicketWorkflowStepStatus.INPROGRESS,
                  base_image='17', series='saucy',
                  master_ppa='ci-engine/ppa'):
    ticket = Ticket()
    ticket.owner = owner
    ticket.title = title
    ticket.description = description
    ticket.bug_id = bug_id
    ticket.current_workflow_step = current_workflow_step
    ticket.status = status
    ticket.base_image = base_image
    ticket.series = series
    ticket.master_ppa = master_ppa
    ticket.save()
    return ticket


def create_subticket(
        ticket, sourcepackage,
        current_workflow_step=SubTicketWorkflowStep.PKG_BUILDING,
        status=SubTicketWorkflowStepStatus.PKG_BUILDING_INPROGRESS,
        assignee='test@example.com'):
    subticket = SubTicket()
    subticket.ticket = ticket
    subticket.sourcepackage = sourcepackage
    subticket.current_workflow_step = current_workflow_step
    subticket.status = status
    subticket.assignee = assignee
    subticket.save()
    return subticket


def create_ticketartifact(name="test.log", ticket=None, reference='123abc'):
    artifact = TicketArtifact()
    artifact.name = name
    artifact.ticket = ticket
    artifact.reference = reference
    artifact.save()
    return artifact


def create_subticketartifact(name="test.log", subticket=None,
                             reference='123abc'):
    artifact = SubTicketArtifact()
    artifact.name = name
    artifact.subticket = subticket
    artifact.reference = reference
    artifact.save()
    return artifact


class TicketCreateContainer(TestCase):

    @mock.patch('ticket.models.get_auth_config')
    @mock.patch('ticket.models.create_for_ticket')
    @mock.patch('ci_utils.data_store.DataStore')
    def test_create_container(self, mocked_ds, mocked_creator, mocked_auth):
        # `Ticket` corresponding Swift container is created (using the ticket
        # UUID) only when a new entry is created.
        fake_auth = {}
        mocked_auth.return_value = fake_auth
        fake_ds = mocked_ds('a_component', fake_auth)
        mocked_creator.return_value = fake_ds

        ticket = create_ticket()
        mocked_creator.assert_called_once_with(ticket.uuid, fake_auth)
        fake_ds.ensure_swift_client.assert_called_once()

        ticket.title = 'Something else ...'
        ticket.save()
        mocked_creator.assert_not_called()


class TicketModelsTestCase(TestCase):

    def setUp(self):
        self.sourcepackage = create_sourcepackage()
        # Mock `Ticket.create_container` for all tests in this context as
        # 'self.create_container'.
        _m = mock.patch('ticket.models.Ticket.create_container')
        self.create_container = _m.start()
        self.addCleanup(_m.stop)
        # Creating tickets, independently of its status, creates its
        # container.
        self.ticket = create_ticket(owner='test@example.com')
        self.create_container.assert_called_once()

    def test_creating_an_artifact_subticket(self):
        subticket = create_subticket(
            ticket=self.ticket, sourcepackage=self.sourcepackage)
        self.artifact_1 = create_subticketartifact(subticket=subticket)
        artifact_in_database = SubTicketArtifact.objects.exclude(
            subticket=None)
        self.assertEquals(len(artifact_in_database), 1)
        only_artifact_in_database = artifact_in_database[0]
        self.assertEquals(only_artifact_in_database, self.artifact_1)

        self.assertEquals(only_artifact_in_database.name, "test.log")

    def test_creating_an_artifact_ticket(self):
        self.artifact_2 = create_ticketartifact(ticket=self.ticket)
        artifact_in_database = TicketArtifact.objects.exclude(ticket=None)
        self.assertEquals(len(artifact_in_database), 1)
        only_artifact_in_database = artifact_in_database[0]
        self.assertEquals(only_artifact_in_database, self.artifact_2)

        self.assertEquals(only_artifact_in_database.name, "test.log")

    def test_creating_a_ticket(self):
        ticket_in_database = Ticket.objects.all()
        self.assertEquals(len(ticket_in_database), 1)
        only_ticket_in_database = ticket_in_database[0]
        self.assertEquals(only_ticket_in_database, self.ticket)
        self.assertEquals(only_ticket_in_database.title, "This is my ticket")

    def test_creating_a_ticket_sets_creator(self):
        ticket_in_database = Ticket.objects.all()
        self.assertEquals(len(ticket_in_database), 1)
        only_ticket_in_database = ticket_in_database[0]
        self.assertEquals(only_ticket_in_database, self.ticket)
        self.assertEquals(only_ticket_in_database.title, "This is my ticket")
        self.assertEquals(self.ticket.owner, self.ticket.creator)

    def test_creating_a_subticket_with_spu(self):
        # `SubTicket`s can be associated with one or more
        # `SourcePackageUpload`s.
        subticket = create_subticket(
            ticket=self.ticket, sourcepackage=self.sourcepackage)
        spu = create_sourcepackageupload(subticket=subticket)

        self.assertEqual(
            [subticket], list(self.ticket.subticket_set.all()))
        self.assertEqual('test@example.com', subticket.assignee)
        self.assertEqual(
            [spu], list(subticket.source_package_uploads.all()))
        self.assertEqual('1.1', spu.version)
        self.assertEqual('all', spu.architecture)

    def test_creating_a_subticket_with_mp(self):
        # `SubTicket`s can be associated with one or more
        # `MergeProposal`s.
        subticket = create_subticket(
            ticket=self.ticket, sourcepackage=self.sourcepackage)
        mp = create_merge_proposal(subticket=subticket)

        self.assertEqual(
            [subticket], list(self.ticket.subticket_set.all()))
        self.assertEqual('test@example.com', subticket.assignee)
        self.assertEqual(
            [mp], list(subticket.merge_proposals.all()))
        self.assertEqual(
            '~foobar/my-project/my-branch/+merge/123456', mp.lp_url)
        self.assertEqual(100, mp.approved_revno)

    def test_ticket_updates_do_not_touch_its_container(self):
        # Ticket containers are only created during the ticket creation,
        # they are not affected by ticket updates.
        self.assertEqual(1, self.create_container.call_count)
        self.ticket.current_workflow_step = TicketWorkflowStep.QUEUED
        self.ticket.save()
        self.assertEqual(1, self.create_container.call_count)

    def test_ticket_updates(self):
        # Ticket can be updated independently of their status.
        self.ticket.current_workflow_step = TicketWorkflowStep.NEW
        self.ticket.title = 'New title for a NEW ticket'
        self.ticket.save()
        new_ticket = Ticket.objects.get(pk=self.ticket.id)
        self.assertEqual(
            TicketWorkflowStep.NEW.value, new_ticket.current_workflow_step)
        self.assertEqual(
            'New title for a NEW ticket', new_ticket.title)

    @mock.patch('ci_utils.amqp_utils.send')
    def test_handle_lander_queued(self, send):
        '''ensure a ticket that becomes queued notifies the lander'''
        self.ticket.current_workflow_step = TicketWorkflowStep.QUEUED
        self.ticket.status = TicketWorkflowStepStatus.QUEUED
        self.ticket.save()
        self.assertEqual(1, send.call_count)

    @mock.patch('ci_utils.amqp_utils.send')
    def test_handle_lander_continue(self, send):
        self.ticket.current_workflow_step = TicketWorkflowStep.PKG_PUBLISHING
        self.ticket.status = TicketWorkflowStepStatus.CONTINUE
        self.ticket.save()
        self.assertEqual(1, send.call_count)

    @mock.patch('ci_utils.amqp_utils.send')
    def test_handle_lander_citrain(self, send):
        # A Lander job is not dispatched for CI-Train tickets.
        self.ticket.current_workflow_step = TicketWorkflowStep.QUEUED
        self.ticket.status = TicketWorkflowStepStatus.QUEUED
        overlay = TicketCiTrainOverlay.objects.create(comments='Something')
        self.ticket.citrain_overlay = overlay
        self.ticket.save()
        self.assertEqual(0, send.call_count)

    @mock.patch('ci_utils.amqp_utils.send')
    def test_handle_review(self, send):
        '''ensure a queued ticket with review has pause step'''
        Review.objects.create(ticket=self.ticket, review_type='packaging',
                              workflow_step=TicketWorkflowStep.PKG_PUBLISHING)
        self.ticket.current_workflow_step = TicketWorkflowStep.QUEUED
        self.ticket.status = TicketWorkflowStepStatus.QUEUED
        self.ticket.save()

        self.assertEqual(1, send.call_count)
        data = json.loads(send.call_args[0][1])
        for step in data['steps']:
            if step['name'] == 'publisher':
                self.assertTrue('pause' in step)
                return
        raise ValueError('publisher step not found in: %s' % data)

    @mock.patch('ci_utils.amqp_utils.send')
    def test_handle_review_continue(self, send):
        '''Ensure when reviews are marked approved the ticket continues'''
        self.ticket.current_workflow_step = TicketWorkflowStep.PKG_BUILDING
        self.ticket.status = TicketWorkflowStepStatus.PAUSED
        self.ticket.save()

        r = Review.objects.create(
            ticket=self.ticket, review_type='packaging',
            workflow_step=TicketWorkflowStep.PKG_PUBLISHING)
        self.ticket.save()
        self.assertEqual(0, send.call_count)

        # This will trigger the ticket object to be moved to CONTINUE
        r.status = ReviewStatus.APPROVED.value
        r.save()
        status = Ticket.objects.get(pk=self.ticket.id).status
        self.assertEqual(TicketWorkflowStepStatus.CONTINUE.value, status)
        self.assertEqual(1, send.call_count)


class TestReview(TestCase):

    def setUp(self):
        _m = mock.patch('ticket.models.Ticket.create_container')
        self.create_container = _m.start()
        self.addCleanup(_m.stop)
        self.ticket = create_ticket(owner='test@example.com')

    def test_title(self):
        # Ensure review title include the ticket UUID, review type and
        # its status.
        review = Review.objects.create(
            ticket=self.ticket, review_type='packaging',
            workflow_step=TicketWorkflowStep.PKG_PUBLISHING)

        self.assertEqual(
            '{} - packaging: Pending'.format(self.ticket.uuid),
            unicode(review))


class WorkflowTestCase(TestCase):
    def test_unique_name(self):
        '''make sure the workflow name is unique'''
        Workflow.objects.create(name='foo', steps='[]')
        with self.assertRaises(IntegrityError):
            Workflow.objects.create(name='foo', steps='[]')

    def test_json(self):
        '''make sure we don't allow invalid json'''
        with self.assertRaises(ValidationError):
            Workflow.objects.create(name='foo', steps='this is not json')


class TestTicketConflicts(TestCase):

    def setUp(self):
        """Mock container creation and queue dispatching.

        Also setup a common ticket conflict eco-system.
        """
        _m = mock.patch('ticket.models.Ticket.create_container')
        _m.start()
        self.addCleanup(_m.stop)
        _m = mock.patch('ci_utils.amqp_utils.send')
        _m.start()
        self.addCleanup(_m.stop)

        self.sp_one = create_sourcepackage('ding')
        self.sp_two = create_sourcepackage('dong')
        self.ticket_one = create_ticket(title='One')
        self.ticket_two = create_ticket(title='Two')

    def test_no_conflicts(self):
        # Conflicts are only detected for `SourcePackageUpload`s with
        # the same 'name'.
        subticket_one = create_subticket(
            ticket=self.ticket_one, sourcepackage=self.sp_one)
        subticket_two = create_subticket(
            ticket=self.ticket_two, sourcepackage=self.sp_two)

        self.assertEqual(
            [], list(subticket_one.conflicts))
        self.assertEqual(
            [], list(subticket_two.conflicts))

    def test_has_single_conflict(self):
        # Conflicts are symmetrical across 'SubTicket's on active `Ticket`s
        # (A conflicts with B, thus B conflicts with A) and cover
        # `SourcePackageUpload` and `MergeProposal`s.
        subticket_one = create_subticket(
            ticket=self.ticket_one, sourcepackage=self.sp_one)
        subticket_two = create_subticket(
            ticket=self.ticket_two, sourcepackage=self.sp_one)

        self.assertEqual(
            [self.ticket_two], list(subticket_one.conflicts))
        self.assertEqual(
            [self.ticket_one], list(subticket_two.conflicts))

    def test_no_conflicts_for_inactive(self):
        # Inactive `Ticket`s (NEW, FAILED or COMPLETED) are not considered
        # conflict targets, but still pointing to active conflicts.
        subticket_one = create_subticket(
            ticket=self.ticket_one, sourcepackage=self.sp_one)
        subticket_two = create_subticket(
            ticket=self.ticket_two, sourcepackage=self.sp_one)

        self.ticket_two.current_workflow_step = TicketWorkflowStep.COMPLETED
        self.ticket_two.save()

        self.assertEqual([], list(subticket_one.conflicts))
        self.assertEqual([self.ticket_one], list(subticket_two.conflicts))

    def test_no_conflicts_for_failures(self):
        # `Ticket`s that failed in any workflow steps are not considered
        # conflict targets.
        subticket_one = create_subticket(
            ticket=self.ticket_one, sourcepackage=self.sp_one)
        subticket_two = create_subticket(
            ticket=self.ticket_two, sourcepackage=self.sp_one)

        self.ticket_two.current_workflow_step = (
            TicketWorkflowStep.PKG_BUILDING)
        self.ticket_two.status = TicketWorkflowStepStatus.FAILED
        self.ticket_two.save()

        self.assertEqual([], list(subticket_one.conflicts))
        self.assertEqual([self.ticket_one], list(subticket_two.conflicts))

    def test_has_multiple_conflicts(self):
        # Multiple conflicting tickets are distinct and ordered by id.
        subticket_one = create_subticket(
            ticket=self.ticket_one, sourcepackage=self.sp_one)
        subticket_two = create_subticket(
            ticket=self.ticket_two, sourcepackage=self.sp_one)
        # Create a third conflicting ticket on a MP
        ticket_three = create_ticket(title='Three')
        subticket_three = create_subticket(
            ticket=ticket_three, sourcepackage=self.sp_one)

        self.assertEqual(
            [self.ticket_two, ticket_three], list(subticket_one.conflicts))
        self.assertEqual(
            [self.ticket_one, ticket_three], list(subticket_two.conflicts))
        self.assertEqual(
            [self.ticket_one, self.ticket_two],
            list(subticket_three.conflicts))
