# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import mock
import uuid

from django.conf import settings
from model_mommy import mommy
from model_mommy.recipe import Recipe, seq

from ci_utils import ticket_states
from ci_utils.tastypie.test import TicketTastypieTestCase
from project.models import SourcePackage
from ticket.api import TicketResource
from ticket.models import (
    get_enum_title,
    ReviewStatus,
    SubTicketWorkflowStep,
    SubTicketWorkflowStepStatus,
    TicketCiTrainOverlay,
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


sourcepackage_recipe = Recipe(SourcePackage, name=seq('foobar'))


class TicketLookupTest(TicketTastypieTestCase):

    def setUp(self):
        super(TicketLookupTest, self).setUp('/api/v1')
        self.ticket = mommy.make('Ticket')

    def test_by_id(self):
        # A `Ticket` can be retrieved using its ID (pk, integer).
        obj = self.getResource('ticket/{}/'.format(self.ticket.pk))
        self.assertEqual(self.ticket.pk, obj['id'])

    def test_by_uuid(self):
        # A `Ticket` can be also retrieved by its UUID (uuid, UUIDv1).
        obj = self.getResource('ticket/{}/'.format(self.ticket.uuid))
        self.assertEqual(self.ticket.pk, obj['id'])

    def test_404(self):
        # Failed `Ticket` lookups return an 404 with no contents.
        # Unknown ID.
        response = self.client.get('/api/v1/ticket/1000/')
        self.assertEqual(404, response.status_code)
        self.assertEqual('', response.content)
        # Unknown UUID
        response = self.client.get('/api/v1/ticket/{}/'.format(uuid.uuid1()))
        self.assertEqual(404, response.status_code)
        self.assertEqual('', response.content)

    def test_resource_uris(self):
        # `Ticket` resource URIs are currently based on their UUID.
        r = TicketResource(api_name='v1')
        # Ticket list URIs continues the same.
        self.assertEqual(
            '/api/v1/ticket/', r.get_resource_uri())
        # Individual ticket resources URIs use UUID.
        self.assertEqual(
            '/api/v1/ticket/{}/'.format(self.ticket.uuid),
            r.get_resource_uri(self.ticket))

    def test_get_via_uri_by_uuid(self):
        # `Ticket` current resource URIs (UUID-based) works.
        r = TicketResource(api_name='v1')
        self.assertEqual(
            self.ticket,
            r.get_via_uri('/api/v1/ticket/{}/'.format(self.ticket.uuid)))

    def test_get_via_uri_by_id(self):
        # `Ticket` resource URIs based on ID (old data), still supported.
        r = TicketResource(api_name='v1')
        self.assertEqual(
            self.ticket,
            r.get_via_uri('/api/v1/ticket/{}/'.format(self.ticket.pk)))


class TicketReadAPITest(TicketTastypieTestCase):

    def setUp(self):
        super(TicketReadAPITest, self).setUp('/api/v1')
        self.ticket = mommy.make('Ticket')
        self.sourcepackage = sourcepackage_recipe.make()
        self.subticket = mommy.make(
            'SubTicket', ticket=self.ticket, sourcepackage=self.sourcepackage)
        self.spu = mommy.make(
            'SourcePackageUpload', subticket=self.subticket)
        self.artifact_1 = mommy.make(
            'SubTicketArtifact', subticket=self.subticket)
        self.artifact_2 = mommy.make('TicketArtifact', ticket=self.ticket)
        self.review_1 = mommy.make(
            'Review', ticket=self.ticket, review_type=u'QA', workflow_step=400)
        self.maxDiff = None

    def test_get_artifact_api_subticket(self):
        obj = self.getResource('subticketartifact/')
        self.assertEqual(obj['objects'][0], {
            u'name': unicode(self.artifact_1.name),
            u'reference': unicode(self.artifact_1.reference),
            u'subticket': {
                u'status': unicode(get_enum_title(self.subticket.status,
                                   SubTicketWorkflowStepStatus)),
                u'current_workflow_step': unicode(get_enum_title(
                    self.subticket.current_workflow_step,
                    SubTicketWorkflowStep)),
                u'assignee': unicode(self.subticket.assignee),
                u'sourcepackage': {
                    u'resource_uri': u'/api/v1/sourcepackage/{0}/'.format(
                        self.sourcepackage.pk),
                    u'id': self.sourcepackage.pk,
                    u'name': unicode(self.sourcepackage.name)},
                'merge_proposals': [],
                u'source_package_uploads': [{
                    u'architecture': unicode(self.spu.architecture),
                    u'version': unicode(self.spu.version),
                    u'id': self.spu.pk,
                    'resource_uri': '/api/v1/spu/{0}/'.format(self.spu.pk),
                    'subticket': '/api/v1/subticket/{0}/'.format(
                        self.subticket.pk)}],
                u'ticket': {
                    u'status': unicode(get_enum_title(self.ticket.status,
                                       TicketWorkflowStepStatus)),
                    u'created': unicode(self.ticket.created.strftime(
                        settings.TEST_DATETIME_FORMAT)),
                    u'current_workflow_step': unicode(get_enum_title(
                        self.ticket.current_workflow_step,
                        TicketWorkflowStep)),
                    u'description': unicode(self.ticket.description),
                    u'title': unicode(self.ticket.title),
                    u'bug_id': self.ticket.bug_id,
                    u'owner': unicode(self.ticket.owner),
                    u'creator': unicode(self.ticket.creator),
                    u'base_image': unicode(self.ticket.base_image),
                    u'id': self.ticket.pk,
                    u'lander_unit': 0,
                    u'master_ppa': unicode(self.ticket.master_ppa),
                    u'build_ppa': unicode(self.ticket.build_ppa),
                    u'resource_uri': u'/api/v1/ticket/{0}/'.format(
                        self.ticket.uuid),
                    u'series': unicode(self.ticket.series),
                    u'private': False,
                    u'updated': unicode(self.ticket.updated.strftime(
                        settings.TEST_DATETIME_FORMAT)),
                    u'uuid': self.ticket.uuid,
                    u'subtickets': [
                        '/api/v1/subticket/{}/'.format(s.pk)
                        for s in self.ticket.subticket_set.all()],
                    u'reviews': [{
                        u'id': self.review_1.pk,
                        u'resource_uri': unicode(
                            '/api/v1/review/{0}/'.format(self.review_1.pk)),
                        u'review_type': u'QA',
                        u'status': u'Pending',
                        u'ticket': unicode(
                            '/api/v1/ticket/{}/'.format(self.ticket.uuid)),
                        u'workflow_step': u'Image testing'}],
                    u'citrain_overlay': None,
                    'workflow': {
                        'name': 'default',
                        'steps': [
                            {'name': 'ppa_creator'},
                            {'name': 'bsbuilder'},
                            {'name': 'validator'},
                            {'name': 'image_builder'},
                            {'name': 'test_runner'},
                            {'name': 'publisher'},
                        ]},
                },
                u'id': self.subticket.pk,
                u'conflicts': [],
                u'resource_uri': u'/api/v1/subticket/{0}/'.format(
                    self.subticket.pk)},
            u'type': unicode(self.artifact_1.type),
            u'id': self.artifact_1.pk,
            u'resource_uri': u'/api/v1/subticketartifact/{0}/'.format(
                self.artifact_1.pk),
        })

    def test_get_artifact_api_ticket(self):
        obj = self.getResource('ticketartifact/')
        self.assertEqual(obj['objects'][0], {
            u'name': unicode(self.artifact_2.name),
            u'reference': unicode(self.artifact_2.reference),
            u'ticket': {
                u'status': unicode(get_enum_title(self.ticket.status,
                                   TicketWorkflowStepStatus)),
                u'created': unicode(self.ticket.created.strftime(
                    settings.TEST_DATETIME_FORMAT)),
                u'current_workflow_step': unicode(get_enum_title(
                    self.ticket.current_workflow_step,
                    TicketWorkflowStep)),
                u'description': unicode(self.ticket.description),
                u'title': unicode(self.ticket.title),
                u'bug_id': self.ticket.bug_id,
                u'owner': unicode(self.ticket.owner),
                u'creator': unicode(self.ticket.creator),
                u'base_image': unicode(self.ticket.base_image),
                u'id': self.ticket.pk,
                u'lander_unit': 0,
                u'master_ppa': unicode(self.ticket.master_ppa),
                u'build_ppa': unicode(self.ticket.build_ppa),
                u'resource_uri': u'/api/v1/ticket/{0}/'.format(
                    self.ticket.uuid),
                u'series': unicode(self.ticket.series),
                u'private': False,
                u'updated': unicode(self.ticket.updated.strftime(
                    settings.TEST_DATETIME_FORMAT)),
                u'uuid': self.ticket.uuid,
                'workflow': {
                    'name': 'default',
                    'steps': [
                        {'name': 'ppa_creator'},
                        {'name': 'bsbuilder'},
                        {'name': 'validator'},
                        {'name': 'image_builder'},
                        {'name': 'test_runner'},
                        {'name': 'publisher'},
                    ]},
                u'subtickets': [
                    '/api/v1/subticket/{}/'.format(s.pk)
                    for s in self.ticket.subticket_set.all()],
                u'reviews': [{
                    u'id': self.review_1.pk,
                    u'resource_uri': unicode(
                        '/api/v1/review/{0}/'.format(self.review_1.pk)),
                    u'review_type': u'QA',
                    u'status': u'Pending',
                    u'ticket': unicode(
                        '/api/v1/ticket/{}/'.format(self.ticket.uuid)),
                    u'workflow_step': u'Image testing'}],
                u'citrain_overlay': None},
            u'type': unicode(self.artifact_2.type),
            u'id': self.artifact_2.pk,
            u'resource_uri': u'/api/v1/ticketartifact/{0}/'.format(
                self.artifact_2.pk),
        })

    def test_get_subticket_api(self):
        obj = self.getResource('subticket/')
        self.assertEqual(obj['objects'][0], {
            u'assignee': unicode(self.subticket.assignee),
            u'current_workflow_step': unicode(get_enum_title(
                self.subticket.current_workflow_step,
                SubTicketWorkflowStep)),
            u'id': self.subticket.pk,
            u'resource_uri': u'/api/v1/subticket/{0}/'.format(
                self.subticket.pk),
            u'sourcepackage': {
                u'resource_uri': u'/api/v1/sourcepackage/{0}/'.format(
                    self.sourcepackage.pk),
                u'id': self.sourcepackage.pk,
                u'name': unicode(self.sourcepackage.name)},
            'merge_proposals': [],
            u'source_package_uploads': [{
                u'architecture': unicode(self.spu.architecture),
                u'version': unicode(self.spu.version),
                u'id': self.spu.pk,
                'resource_uri': '/api/v1/spu/{0}/'.format(self.spu.pk),
                'subticket': '/api/v1/subticket/{0}/'.format(
                    self.subticket.pk)}],
            u'status': unicode(get_enum_title(self.subticket.status,
                               SubTicketWorkflowStepStatus)),
            u'ticket': {
                u'base_image': unicode(self.ticket.base_image),
                u'bug_id': self.ticket.bug_id,
                u'created': unicode(self.ticket.created.strftime(
                    settings.TEST_DATETIME_FORMAT)),
                u'current_workflow_step': unicode(get_enum_title(
                    self.ticket.current_workflow_step,
                    TicketWorkflowStep)),
                u'description': unicode(self.ticket.description),
                u'id': self.ticket.pk,
                u'lander_unit': 0,
                u'master_ppa': unicode(self.ticket.master_ppa),
                u'build_ppa': unicode(self.ticket.build_ppa),
                u'owner': unicode(self.ticket.owner),
                u'creator': unicode(self.ticket.creator),
                u'status': unicode(get_enum_title(self.ticket.status,
                                   TicketWorkflowStepStatus)),
                u'title': unicode(self.ticket.title),
                u'resource_uri': u'/api/v1/ticket/{0}/'.format(
                    self.ticket.uuid),
                u'series': unicode(self.ticket.series),
                u'private': False,
                u'updated': unicode(self.ticket.updated.strftime(
                    settings.TEST_DATETIME_FORMAT)),
                u'uuid': self.ticket.uuid,
                'workflow': {
                    'name': 'default',
                    'steps': [
                        {'name': 'ppa_creator'},
                        {'name': 'bsbuilder'},
                        {'name': 'validator'},
                        {'name': 'image_builder'},
                        {'name': 'test_runner'},
                        {'name': 'publisher'},
                    ]},
                u'subtickets': [
                    '/api/v1/subticket/{}/'.format(s.pk)
                    for s in self.ticket.subticket_set.all()],
                u'reviews': [{
                    u'id': self.review_1.pk,
                    u'resource_uri': unicode(
                        '/api/v1/review/{0}/'.format(self.review_1.pk)),
                    u'review_type': u'QA',
                    u'status': u'Pending',
                    u'ticket': unicode(
                        '/api/v1/ticket/{}/'.format(self.ticket.uuid)),
                    u'workflow_step': u'Image testing'}],

                u'citrain_overlay': None},
            u'conflicts': [],
        })

    def test_subticket_api_multiple_spus_and_mps(self):
        # Multiple `SourcePackageUpload`s and `MergeProposals` can be
        # associated with a `Subticket` and they are all listed inline.
        self.spu.delete()
        spu_one = self.subticket.source_package_uploads.create(
            version='1.0', architecture='all')
        spu_two = self.subticket.source_package_uploads.create(
            version='1.1', architecture='any')

        mp_one = self.subticket.merge_proposals.create(
            lp_url='lp:~foo/bar/baz')
        mp_two = self.subticket.merge_proposals.create(
            lp_url='lp:~sheldon/ubuntu/bazinga')

        results = self.getResource('subticket/')
        [obj] = results['objects']

        # Associated SPUs.
        self.assertEqual(
            [{
                'architecture': 'all',
                'id': spu_one.pk,
                'resource_uri': '/api/v1/spu/{}/'.format(spu_one.pk),
                'subticket': '/api/v1/subticket/{}/'.format(
                    self.subticket.pk),
                'version': '1.0',
            }, {
                'architecture': 'any',
                'id': spu_two.pk,
                'resource_uri': '/api/v1/spu/{}/'.format(spu_two.pk),
                'subticket': '/api/v1/subticket/{}/'.format(
                    self.subticket.pk),
                'version': '1.1',
            }],
            obj['source_package_uploads'])

        # Associated MPs.
        self.assertEqual(
            [{
                'approved_revno': None,
                'id': mp_one.pk,
                'lp_url': 'lp:~foo/bar/baz',
                'resource_uri': '/api/v1/mp/{}/'.format(mp_one.pk),
                'subticket': '/api/v1/subticket/{}/'.format(
                    self.subticket.pk),
            }, {
                'approved_revno': None,
                'id': mp_two.pk,
                'lp_url': 'lp:~sheldon/ubuntu/bazinga',
                'resource_uri': '/api/v1/mp/{}/'.format(mp_two.pk),
                'subticket': '/api/v1/subticket/{}/'.format(
                    self.subticket.pk),
            }],
            obj['merge_proposals'])

    def test_subticket_api_conflicts(self):
        # Each `Subticket`s points, individually, to conflicting `Tickets`
        # This way users have a clear indication of which part (subticket)
        # of their task might result in problem when landing.
        another_ticket = mommy.make(
            'Ticket', current_workflow_step=TicketWorkflowStep.QUEUED)
        another_subticket = mommy.make(
            'SubTicket', ticket=another_ticket,
            sourcepackage=self.sourcepackage)
        mommy.make(
            'SourcePackageUpload', subticket=another_subticket,
            version=self.spu.version)

        [old, new] = self.getResource('subticket/')['objects']

        # The 'old' ticket now conflicts with 'new' (just created).
        # Remember, 'conflicts' contains the `Ticket` resource URI.
        self.assertEqual(
            [new['ticket']['resource_uri']], old['conflicts'])

        # The 'new' ticket does not conflict with the 'old' because it's
        # still 'NEW' (not active). See 'test_models' for more descriptive
        # test on this aspect.
        self.assertEqual(
            [], new['conflicts'])

    def test_get_ticket_api(self):
        obj = self.getResource('ticket/')
        self.assertEqual(obj['objects'][0], {
            u'status': unicode(get_enum_title(self.ticket.status,
                               TicketWorkflowStepStatus)),
            u'created': unicode(self.ticket.created.strftime(
                settings.TEST_DATETIME_FORMAT)),
            u'current_workflow_step': unicode(get_enum_title(
                self.ticket.current_workflow_step,
                TicketWorkflowStep)),
            u'description': unicode(self.ticket.description),
            u'title': unicode(self.ticket.title),
            u'bug_id': self.ticket.bug_id,
            u'owner': unicode(self.ticket.owner),
            u'creator': unicode(self.ticket.creator),
            u'base_image': unicode(self.ticket.base_image),
            u'id': self.ticket.pk,
            u'lander_unit': 0,
            u'master_ppa': unicode(self.ticket.master_ppa),
            u'build_ppa': unicode(self.ticket.build_ppa),
            u'resource_uri': u'/api/v1/ticket/{0}/'.format(
                self.ticket.uuid),
            u'series': unicode(self.ticket.series),
            u'updated': unicode(self.ticket.updated.strftime(
                settings.TEST_DATETIME_FORMAT)),
            u'private': False,
            u'uuid': self.ticket.uuid,
            u'subtickets': [
                u'/api/v1/subticket/{}/'.format(s.pk)
                for s in self.ticket.subticket_set.all()],
            u'reviews': [{
                u'id': self.review_1.pk,
                u'resource_uri': unicode(
                    '/api/v1/review/{0}/'.format(self.review_1.pk)),
                u'review_type': u'QA',
                u'status': u'Pending',
                u'ticket': unicode(
                    '/api/v1/ticket/{0}/'.format(self.ticket.uuid)),
                u'workflow_step': u'Image testing'}],
            u'citrain_overlay': None,
            'workflow': {
                'name': 'default',
                'steps': [
                    {'name': 'ppa_creator'},
                    {'name': 'bsbuilder'},
                    {'name': 'validator'},
                    {'name': 'image_builder'},
                    {'name': 'test_runner'},
                    {'name': 'publisher'},
                ]},
        })

    def test_get_ticket_citrain(self):
        # When it exists, the `TicketCiTrainOverlay` record is embeded
        # in the `Ticket` object representation.
        overlay = TicketCiTrainOverlay.objects.create(
            comments='Something',
            job_url='http://foo.com/1',
            landers='bar baz',
            sync_request='sync:100',
            test_notes='Test now!',
        )
        self.ticket.citrain_overlay = overlay
        self.ticket.save()

        [obj] = self.getResource('ticket/')['objects']
        self.assertEqual({
            'comments': 'Something',
            'job_url': 'http://foo.com/1',
            'landers': 'bar baz',
            'sync_request': 'sync:100',
            'test_notes': 'Test now!',
        }, obj['citrain_overlay'])

    def test_get_ticket_api_private(self):
        self.ticket.private = True
        self.ticket.save()

        obj = self.getResource('ticket/')
        self.assertEqual(0, len(obj['objects']))

    def test_get_ticket_status_filtering(self):
        # `Ticket` API supports 'status' filtering.
        # Exact matching.
        results = self.getResource(
            'ticket/?status={}'.format(int(self.ticket.status)))
        [t] = results['objects']
        self.assertEqual(self.ticket.title, t['title'])
        # The django query_spec goodies (exact, lt, lte, gt, gte, etc)
        # are supported.
        results = self.getResource(
            'ticket/?status__lte={}'.format(
                int(TicketWorkflowStepStatus.INPROGRESS)))
        [t] = results['objects']
        self.assertEqual(self.ticket.title, t['title'])
        # Dataset updates take effect immediately.
        self.ticket.status = TicketWorkflowStepStatus.COMPLETED
        self.ticket.save()
        results = self.getResource(
            'ticket/?status={}'.format(int(self.ticket.status)))
        [t] = results['objects']
        self.assertEqual(self.ticket.title, t['title'])
        # Empty queryset.
        results = self.getResource(
            'ticket/?status={}'.format(int(TicketWorkflowStepStatus.FAILED)))
        self.assertEqual(0, len(results['objects']))

    def test_get_ticket_review_filtering(self):
        # `Ticket` API supports 'review' filtering.
        other_ticket = mommy.make('Ticket')
        other_review = other_ticket.review_set.create(
            review_type='DEV',
            workflow_step=TicketWorkflowStep.SILO_TESTING.value,
            status=ReviewStatus.APPROVED.value)

        # Exact matching on status.
        [review] = self.ticket.review_set.all()
        results = self.getResource(
            'ticket/?reviews__status={}'.format(review.status))
        [t] = results['objects']
        self.assertEqual(t['uuid'], review.ticket.uuid)

        results = self.getResource(
            'ticket/?reviews__status={}'.format(other_review.status))
        [t] = results['objects']
        self.assertEqual(t['uuid'], other_review.ticket.uuid)

    def test_get_sourcepackageupload_api(self):
        obj = self.getResource('spu/')
        self.assertEqual(obj['objects'][0], {
            u'architecture': unicode(self.spu.architecture),
            u'version': unicode(self.spu.version),
            'subticket': '/api/v1/subticket/{0}/'.format(
                self.subticket.pk),
            u'id': self.spu.pk,
            u'resource_uri':
            u'/api/v1/spu/{0}/'.format(self.spu.pk),
        })

    def test_get_subticket_update_status_api(self):
        obj = self.getResource('updatesubticketstatus/')
        self.assertEqual(obj['objects'][0], {
            u'id': self.subticket.pk,
            u'resource_uri': u'/api/v1/updatesubticketstatus/{0}/'.format(
                self.subticket.pk),
            u'current_workflow_step': unicode(get_enum_title(
                self.subticket.current_workflow_step,
                SubTicketWorkflowStep)),
            u'status': unicode(get_enum_title(self.subticket.status,
                               SubTicketWorkflowStepStatus)),
        })

    def test_get_single_subticket_update_status_api(self):
        obj = self.getResource(
            'updatesubticketstatus/?id={}'.format(self.subticket.pk))
        self.assertEqual(obj['objects'][0], {
            u'id': self.subticket.pk,
            u'resource_uri': u'/api/v1/updatesubticketstatus/{0}/'.format(
                self.subticket.pk),
            u'current_workflow_step': unicode(get_enum_title(
                self.subticket.current_workflow_step,
                SubTicketWorkflowStep)),
            u'status': unicode(get_enum_title(self.subticket.status,
                               SubTicketWorkflowStepStatus)),
        })

    def test_get_subticket_status_api(self):
        obj = self.getResource('subticketstatus/')
        self.assertEqual(obj['objects'][0], {
            u'id': self.subticket.pk,
            u'resource_uri': u'/api/v1/subticketstatus/{0}/'.format(
                self.subticket.pk),
            u'current_workflow_step': unicode(get_enum_title(
                self.subticket.current_workflow_step,
                SubTicketWorkflowStep)),
            u'status': unicode(get_enum_title(self.subticket.status,
                               SubTicketWorkflowStepStatus)),
        })

    def test_get_ticket_steps(self):
        obj = self.getResource('ticket_workflow_step/')
        self.assertEqual(len(obj['objects']),
                         len(ticket_states.TicketWorkflowStep))

    def test_get_subticket_steps(self):
        obj = self.getResource('subticket_workflow_step/')
        self.assertEqual(len(obj['objects']),
                         len(ticket_states.SubTicketWorkflowStep))

    def test_get_ticket_step_statuses(self):
        obj = self.getResource('ticket_workflow_step_status/')
        self.assertEqual(len(obj['objects']),
                         len(ticket_states.TicketWorkflowStepStatus))

    def test_get_subticket_step_statuses(self):
        obj = self.getResource('subticket_workflow_step_status/')
        self.assertEqual(len(obj['objects']),
                         len(ticket_states.SubTicketWorkflowStepStatus))

    def assert_same(self, uri):
        obj = self.getResource(uri)
        orig = obj['objects'][0]
        resource_uri = obj['objects'][0]['resource_uri']
        item_uri = resource_uri.replace('/api/v1/', '')

        obj = self.getResource(item_uri)
        self.assertEqual(orig, obj)

    def test_get_ticket_steps_detail(self):
        self.assert_same('ticket_workflow_step/')

    def test_get_subticket_steps_detail(self):
        self.assert_same('subticket_workflow_step/')

    def test_get_ticket_step_status_detail(self):
        self.assert_same('ticket_workflow_step_status/')

    def test_get_subticket_step_status_detail(self):
        self.assert_same('subticket_workflow_step_status/')


class TicketOrderingTest(TicketTastypieTestCase):

    @mock.patch('ci_utils.amqp_utils.send')
    def setUp(self, send):
        super(TicketOrderingTest, self).setUp('/api/v1')
        for i in range(5):
            mommy.make('Ticket', title='t-{}'.format(i + 1))

    def test_default_ordering(self):
        # The default order when listing `Ticket`s is ascending 'id'.
        result = self.getResource('ticket/')
        self.assertEqual(
            ['t-1', 't-2', 't-3', 't-4', 't-5'],
            [t['title'] for t in result['objects']])

    def test_ordering_by_id(self):
        # `Ticket`s can optionally be listed by descending 'id'.
        result = self.getResource('ticket/?order_by=-id')
        self.assertEqual(
            ['t-5', 't-4', 't-3', 't-2', 't-1'],
            [t['title'] for t in result['objects']])


class TicketPaginationTest(TicketTastypieTestCase):

    @mock.patch('ci_utils.amqp_utils.send')
    def setUp(self, send):
        super(TicketPaginationTest, self).setUp('/api/v1')
        for i in range(5):
            mommy.make('Ticket', title='ticket-{}'.format(i + 1))

    def test_no_pagination(self):
        # `PageNumberPaginator` honours the default behaviour.
        # Full set (default limit is 20).
        result = self.getResource('ticket/')
        self.assertEqual(5, len(result['objects']))
        # Custom limit.
        result = self.getResource('ticket/?limit=3')
        self.assertEqual(3, len(result['objects']))
        # Custom offset.
        result = self.getResource('ticket/?offset=3')
        self.assertEqual(2, len(result['objects']))
        # Combined offset & limit.
        result = self.getResource('ticket/?offset=1&limit=1')
        [ticket] = result['objects']
        self.assertEqual('ticket-2', ticket['title'])

    def test_page_number(self):
        # `PageNumberPaginator` supports optional 'page' number
        # to cope with YUI datasource requests.
        result = self.getResource('ticket/?page=2&limit=4')
        [ticket] = result['objects']
        self.assertEqual('ticket-5', ticket['title'])

    def test_offset_gets_precedence_over_page_number(self):
        # `PageNumberPaginator` gives precedence to 'offset' if it's
        # provided with 'page' (the latter is ignored.)
        result = self.getResource('ticket/?page=2&offset=1&limit=4')
        self.assertEqual(4, len(result['objects']))
