# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import json

from django.contrib.auth.models import User
from django.test import TestCase

from ci_utils.tastypie.test import TastypieTestCase


class UserInfoTest(TastypieTestCase):

    def setUp(self):
        super(UserInfoTest, self).setUp('/api/v1')

    def test_anonymous_info(self):
        # Anonymous requests to /userinfo/ returns an object with
        # 'authenticated' set as False. The other fields are bogus/empty.
        user_info = self.getResource('userinfo/')
        self.assertEqual({
            'authenticated': False,
            'username': '',
            'email': '',
            'fullname': '',
            'staff': False,
            'superuser': False,
        }, user_info)

    def test_authenticated_info(self):
        # Authenticated requests to /userinfo/ returns the user information.
        password = '1234'
        user = User.objects.create_user(
            username='test', first_name='First', last_name='Last',
            password=password, email='t@t.com')
        self.client.client.login(username=user.username, password=password)

        user_info = self.getResource('userinfo/')
        self.assertEqual({
            'authenticated': True,
            'username': 'test',
            'email': 't@t.com',
            'fullname': 'First Last',
            'staff': False,
            'superuser': False,
        }, user_info)


class UserInfoAuthTest(TestCase):

    def setUp(self):
        """ Creates an `User` for tests."""
        self.test_user = User.objects.create_user(
            username='test', password='1234', first_name='Test')
        self.client.login(username='test', password='1234')

    def _get_current_username(self):
        """Return the current logged-in username.

        By the lack of response 'context' (tastypie thing) we recover the
        logged-in username from the 'userinfo' resource.
        """
        response = self.client.get('/api/v1/userinfo/')
        self.assertEqual(200, response.status_code)
        return json.loads(response.content)['username']

    def assertLoggedIn(self, username):
        """Assert the given username matches the current user."""
        self.assertEqual(username, self._get_current_username())

    def assertNotLoggedIn(self):
        """Assert there is no user currently logged-in."""
        self.assertEqual('', self._get_current_username())

    def test_logout_works(self):
        #'logout' inconditionally logs the user out.
        self.assertLoggedIn('test')

        # By default the 'logout' page redirects to the (webui) site root
        # which is not available in the isolated TS deployment.
        response = self.client.get('/logout/')
        self.assertEqual(302, response.status_code)
        self.assertEqual('http://testserver/', response['location'])

        self.assertNotLoggedIn()

    def test_logout_supports_next(self):
        #'logout' can optionally redirect users to another page, normally
        # where they came from, via 'next' GET parameter.
        self.assertLoggedIn('test')

        response = self.client.get('/logout/?next=/ticket/whatever/')
        self.assertEqual(302, response.status_code)
        self.assertEqual(
            'http://testserver/ticket/whatever/', response['location'])

        self.assertNotLoggedIn()
