# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import mock
import uuid

from model_mommy import mommy
from model_mommy.recipe import Recipe, seq

from ci_utils.tastypie.test import TicketTastypieTestCase
from ci_utils import unit_config
from project.models import (
    SourcePackage,
)
m = mock.patch('ci_utils.unit_config.get',
               return_value={
                   'trusty': 'http://foo/bar'
               })
m.start()
from ticket.models import (
    Review,
    ReviewStatus,
    SourcePackageUpload,
    SubTicket,
    SubTicketWorkflowStep,
    SubTicketWorkflowStepStatus,
    Ticket,
    TicketArtifact,
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
    Workflow,
)


def create_ticket(added=None, removed=None):
    with mock.patch('ci_utils.amqp_utils.send'):
        return mommy.make(
            'Ticket', current_workflow_step=TicketWorkflowStep.QUEUED.value,
            series='trusty')


sourcepackage_recipe = Recipe(SourcePackage, name=seq('foobar'))


class APICreateTicketResourceTest(TicketTastypieTestCase):

    def setUp(self):
        super(APICreateTicketResourceTest, self).setUp('/api/v1')
        self.ticket = create_ticket()
        self.resource = 'ticket/'
        self.detail_url = self.resource + '{0}/'.format(self.ticket.pk)
        self.post_ticket_data = {
            'owner': 'owner@example.com',
            'title': 'My first ticket!',
            'description': 'This if my first ticket. See what it can do',
            'bug_id': '12345',
        }

    def test_post_ticket(self):
        # Check how many are there first.
        self.assertEqual(Ticket.objects.count(), 1)
        self.post(resource=self.resource, params=self.post_ticket_data)
        # Verify a new one has been added.
        self.assertEqual(Ticket.objects.count(), 2)

    def test_base_image_map(self):
        self.post(resource=self.resource, params=self.post_ticket_data)
        image_map = unit_config.get('image_map')
        self.assertEqual(Ticket.objects.get(pk=self.ticket.pk).base_image,
                         image_map.get('trusty'))

    def test_post_ticket_with_uuid(self):
        # `Ticket` can be created with a given UUID via API.
        a_uuid = str(uuid.uuid1())
        params = {
            'owner': 'owner@example.com',
            'title': 'UUID FTW!',
            'description': 'This ticket has UUID',
            'bug_id': '12345',
            'uuid': a_uuid,
        }
        self.post(resource=self.resource, params=params)
        self.assertEqual(2, Ticket.objects.count())
        self.assertEqual(
            'UUID FTW!', unicode(Ticket.objects.get(uuid=a_uuid)))

    def test_post_ticket_atomically_for_sourcepackageuploads(self):
        # `Ticket` can be created with its dependencies in a single
        # transaction (Ticket->[SubTickets->SPU->SP])
        params = self.post_ticket_data.copy()
        params.update({
            'title': 'Atomic Ticket',
            'subtickets': [{
                'sourcepackage': {'name': 'foo'},
                'source_package_uploads': [
                    {'version': '1.0', 'architecture': 'all'}
                ],
                'assignee': 'foo@bar.com',
            }, {
                'sourcepackage': {'name': 'bar'},
                'source_package_uploads': [
                    {'version': '2.0', 'architecture': 'any'},
                ],
                'assignee': 'bar@foo.com',
            }],
        })
        self.post(resource=self.resource, params=params)

        # 'Atomic Ticket' was created.
        [existing, ticket] = list(Ticket.objects.all())
        self.assertEqual('Atomic Ticket', unicode(ticket))
        # and already contain the specified `SubTicket` and corresponding
        # `SourcePackageUpload`s and `SourcePackage`s.
        [sub_foo, sub_bar] = list(ticket.subticket_set.all())
        self.assertEqual('Atomic Ticket foo', unicode(sub_foo))
        self.assertEqual('foo@bar.com', sub_foo.assignee)
        self.assertEqual(
            ['Atomic Ticket foo_1.0 (all)'],
            [unicode(spu) for spu in sub_foo.source_package_uploads.all()])

        self.assertEqual('Atomic Ticket bar', unicode(sub_bar))
        self.assertEqual('bar@foo.com', sub_bar.assignee)
        self.assertEqual(
            ['Atomic Ticket bar_2.0 (any)'],
            [unicode(spu) for spu in sub_bar.source_package_uploads.all()])

    def test_post_ticket_atomically_for_reviews(self):
        # 'Reviews' can be created atomically with the related `Ticket`.

        # We are injecting a review here to check if tastypie it behaving
        # correctly and not updating existing reviews to accomplish the
        # request.
        other_ticket = create_ticket()
        existing_review = other_ticket.review_set.create(
            review_type='QA',
            workflow_step=TicketWorkflowStep.IMAGE_BUILDING)

        # Let's POST to create a new ticket with pre-defined reviews.
        params = self.post_ticket_data.copy()
        params.update({
            'title': 'Atomic Ticket with Review',
            'reviews': [{
                'review_type': 'QA',
                'workflow_step': 'Package validation',
            }, {
                'review_type': 'QA',
                'workflow_step': 'Image building',
            }],
        })
        self.post(resource=self.resource, params=params)

        # 'Atomic Ticket with Review' was created.
        [existing, other, ticket] = list(Ticket.objects.all())

        self.assertEqual('Atomic Ticket with Review', unicode(ticket))
        # and already contain the specified `Review`s
        [review_pkg, review_img] = list(ticket.review_set.all())
        self.assertEqual('QA', review_pkg.review_type)
        self.assertEqual(
            TicketWorkflowStep.PKG_VALIDATION.value, review_pkg.workflow_step)
        self.assertEqual(ReviewStatus.PENDING.value, review_pkg.status)
        self.assertEqual('QA', review_img.review_type)
        self.assertEqual(
            TicketWorkflowStep.IMAGE_BUILDING.value, review_img.workflow_step)
        self.assertFalse(ReviewStatus.PENDING.value, review_img.status)

        # Note that the reviews created via POST are new, they are not
        # mysteriously reused/updated by tastypie.
        self.assertNotIn(existing_review.pk, [review_pkg.pk, review_img.pk])

    def test_patch_ticket_reviews(self):
        # `Ticket` PATCH can update `Review`s.
        review = self.ticket.review_set.create(
            review_type='QA',
            workflow_step=TicketWorkflowStep.IMAGE_BUILDING)

        params = self.post_ticket_data.copy()
        params.update({
            'title': 'Atomic Patched Ticket',
            'reviews': [{
                'id': review.id,
                'status': 'Approved',
                'workflow_step': 'Image building',
            }],
        })
        self.patch(resource=self.detail_url, params=params)

        # The existing `Subticket` was updated and a new one was created.
        [ticket] = list(Ticket.objects.all())
        self.assertEqual('Atomic Patched Ticket', unicode(ticket))
        [new_review] = list(ticket.review_set.all())
        self.assertEqual(review.id, new_review.id)
        self.assertEqual(ReviewStatus.APPROVED.value, new_review.status)

    def test_post_ticket_atomically_reuses_workflows(self):
        # `Ticket` atomic creation re-uses existing `Workflows`s.
        params = self.post_ticket_data.copy()
        params.update({
            'title': 'Atomic Ticket reuses Workflow',
            'workflow': {'name': 'citrain'},
        })
        workflows_before = Workflow.objects.count()
        self.post(resource=self.resource, params=params)

        [existing, ticket] = list(Ticket.objects.all())
        self.assertEqual('Atomic Ticket reuses Workflow', unicode(ticket))
        self.assertEqual('citrain', ticket.workflow.name)
        self.assertEqual(workflows_before, Workflow.objects.count())

    def test_post_ticket_does_not_create_workflow(self):
        # `Ticket` POSTs never create new `Workflow`s.
        params = self.post_ticket_data.copy()
        params.update({
            'title': 'Atomic Ticket with Workflow',
            'workflow': {
                'name': 'new',
                'steps': '[{"name": "bogus"}]',
            },
        })
        workflows_before = Workflow.objects.count()
        self.post(resource=self.resource, params=params)

        [existing, ticket] = list(Ticket.objects.all())
        self.assertEqual('Atomic Ticket with Workflow', unicode(ticket))
        self.assertEqual('default', ticket.workflow.name)
        self.assertEqual(workflows_before, Workflow.objects.count())

    def test_post_ticket_does_not_override_workflows(self):
        # It is not possible to override `Workflows` via ticket POSTing.
        params = self.post_ticket_data.copy()
        params.update({
            'title': 'Atomic Ticket with Workflow',
            'workflow': {
                'name': 'default',
                'steps': '[{"name": "no-way"}]',
            },
        })
        default_steps = Workflow.objects.get(pk='default').steps
        self.post(resource=self.resource, params=params)

        self.assertEqual(
            default_steps, Workflow.objects.get(pk='default').steps)

    def test_post_ticket_atomically_reuses_sourcepackages(self):
        # `Ticket` atomic creation re-uses existing `SourcePackage`s.
        params = self.post_ticket_data.copy()
        params.update({
            'title': 'Atomic Ticket',
            'subtickets': [{
                'sourcepackage': {'name': 'foo'},
                'source_package_uploads': [
                    {'version': '1.0', 'architecture': 'all'},
                ],
                'assignee': 'foo@bar.com',
            }],
        })
        # Create a matching `SourcePackage`.
        existing_sp = SourcePackage.objects.create(name='foo')

        self.post(resource=self.resource, params=params)

        # The existing `SourcePackage` object was used as a reference
        # for the just-created 'subticket.source_package_upload'.
        [existing, ticket] = list(Ticket.objects.all())
        self.assertEqual('Atomic Ticket', unicode(ticket))
        [sub_foo] = list(ticket.subticket_set.all())
        self.assertEqual(existing_sp, sub_foo.sourcepackage)

    def test_patch_ticket_subtickets(self):
        # `Ticket` patch can create new or update existing `Subticket`s.
        params = self.post_ticket_data.copy()
        params.update({
            'title': 'Atomic Patched Ticket',
            'subtickets': [{
                'sourcepackage': {'name': 'foo'},
                'assignee': 'foo@bar.com',
            }, {
                'sourcepackage': {'name': 'bar'},
                'assignee': 'foo@bar.com',
            }],
        })
        # Create a matching `SourcePackage` and `SubTicket`.
        sp = SourcePackage.objects.create(name='foo')
        subticket = self.ticket.subticket_set.create(
            sourcepackage=sp, assignee='boing@boing.com')

        self.patch(resource=self.detail_url, params=params)

        # The existing `Subticket` was updated and a new one was created.
        [ticket] = list(Ticket.objects.all())
        self.assertEqual('Atomic Patched Ticket', unicode(ticket))
        [old, new] = list(ticket.subticket_set.all())
        self.assertEqual(old, subticket)
        self.assertEqual('foo@bar.com', old.assignee)
        self.assertEqual('Atomic Patched Ticket bar', unicode(new))

    def test_post_ticket_citrain_overlay(self):
        # `Ticket` creation supports adding CI-Train extra fields (via
        # 'citrain_overlay' FK).
        params = self.post_ticket_data.copy()
        params.update({
            'title': 'CI-Train Ticket',
            'citrain_overlay': {
                'comments': 'Something',
                'job_url': 'http://foo.com/1',
            },
        })

        self.post(resource=self.resource, params=params)

        [existing, ticket] = list(Ticket.objects.all())
        self.assertEqual('CI-Train Ticket', unicode(ticket))
        self.assertEqual(
            'Something', ticket.citrain_overlay.comments)
        self.assertEqual(
            'http://foo.com/1', ticket.citrain_overlay.job_url)

    def test_patch_detail(self):
        # `Ticket` object can be patched via the API.
        new_data = {'owner': 'ci@example.com'}
        resp = self.client.patch('/api/v1/' + self.detail_url, data=new_data)
        self.assertHttpAccepted(resp)
        self.assertEqual(
            new_data['owner'], Ticket.objects.get(pk=self.ticket.pk).owner)
        self.assertEqual(Ticket.objects.count(), 1)

    def test_patch_detail_citrain(self):
        # `Ticket` citrain_overaly fields can be patched via the API.
        new_data = {
            'citrain_overlay': {
                'comments': 'Something',
            },
        }

        # A new 'citrain_overlay' is implicitly created if it does not
        # exist yet.
        self.patch(resource=self.detail_url, params=new_data)
        ticket = Ticket.objects.get(pk=self.ticket.pk)
        self.assertEqual(
            'Something', ticket.citrain_overlay.comments)

        # The existing 'citrain_overlay' is simply updated if it exists.
        new_data['citrain_overlay']['comments'] = 'Something Else'
        self.patch(resource=self.detail_url, params=new_data)
        ticket = Ticket.objects.get(pk=self.ticket.pk)
        self.assertEqual(
            'Something Else', ticket.citrain_overlay.comments)

    def test_delete_ticket_not_allowed(self):
        # `Ticket` object cannot be deleted via the API.
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class APICreateSPUResourceTest(TicketTastypieTestCase):

    def setUp(self):
        super(APICreateSPUResourceTest, self).setUp('/api/v1')
        self.sourcepackage = sourcepackage_recipe.make()
        self.subticket = mommy.make('SubTicket',
                                    sourcepackage=self.sourcepackage)
        self.spu = mommy.make('SourcePackageUpload',
                              subticket=self.subticket)
        self.resource = 'spu/'
        self.detail_url = self.resource + '{0}/'.format(self.spu.pk)
        self.post_spu_data = {
            'subticket': '/api/v1/subticket/{}/'.format(self.subticket.pk),
            'version': '1.0',
        }

    def test_post_spu(self):
        # Check how many are there first.
        self.assertEqual(SourcePackageUpload.objects.count(), 1)
        self.post(resource=self.resource, params=self.post_spu_data)
        # Verify a new one has been added.
        self.assertEqual(SourcePackageUpload.objects.count(), 2)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['version'] = '2.0'

        self.assertEqual(SourcePackageUpload.objects.count(), 1)
        resp = self.client.patch('/api/v1/' + self.detail_url, data=new_data)
        self.assertHttpMethodNotAllowed(resp)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(SourcePackageUpload.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(SourcePackageUpload.objects.get(
                         pk=self.spu.pk).version, self.spu.version)

    def test_delete_spu_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class APICreateSubTicketResourceTest(TicketTastypieTestCase):

    def setUp(self):
        super(APICreateSubTicketResourceTest, self).setUp('/api/v1')
        self.ticket = mommy.make('Ticket')
        self.sourcepackage = sourcepackage_recipe.make()
        self.subticket = mommy.make('SubTicket', ticket=self.ticket,
                                    sourcepackage=self.sourcepackage)
        self.resource = 'subticket/'
        self.detail_url = self.resource + '{0}/'.format(self.subticket.pk)
        self.post_subticket_data = {
            'sourcepackage': '/api/v1/sourcepackage/{}/'.format(
                self.sourcepackage.pk),
            'ticket': '/api/v1/ticket/{0}/'.format(self.ticket.pk),
            'assignee': 'test@example.com',
        }

    def test_post_creates_new_subticket(self):
        # POSTing on `SubTicket` resource creates a new subticket.
        new_sp = sourcepackage_recipe.make()
        new_data = self.post_subticket_data.copy()
        new_data['sourcepackage'] = '/api/v1/sourcepackage/{}/'.format(
            new_sp.pk)
        self.post(resource=self.resource, params=new_data)
        self.assertEqual(SubTicket.objects.count(), 2)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['assignee'] = 'me@example.com'

        self.assertEqual(SubTicket.objects.count(), 1)
        resp = self.client.patch('/api/v1/' + self.detail_url, data=new_data)
        self.assertHttpMethodNotAllowed(resp)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(SubTicket.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(SubTicket.objects.get(pk=self.subticket.pk).assignee,
                         self.subticket.assignee)

    def test_delete_subticket_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class APICreateTicketArtifactResourceTest(TicketTastypieTestCase):

    def setUp(self):
        super(APICreateTicketArtifactResourceTest, self).setUp('/api/v1')
        self.ticket = mommy.make('Ticket')
        self.artifact = mommy.make('TicketArtifact', ticket=self.ticket)
        self.resource = 'ticketartifact/'
        self.detail_url = self.resource + '{0}/'.format(self.artifact.pk)
        self.ticket_uri = '/api/v1/ticket/{0}/'.format(self.ticket.pk)
        self.post_artifact_data_ticket = {
            'type': 'LOGS',
            'name': 'my_artifact',
            'ticket': self.ticket_uri,
            'reference': 'http://www.example.com/m1CTLg5FHY',
        }

    def test_post_artifact_ticket(self):
        """
        Test creating an artifact for a ticket via the API
        """
        # Check how many are there first.
        self.assertEqual(TicketArtifact.objects.count(), 1)
        self.post(resource=self.resource,
                  params=self.post_artifact_data_ticket)
        # Verify a new one has been added.
        self.assertEqual(TicketArtifact.objects.count(), 2)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['name'] = 'my_upload'

        self.assertEqual(TicketArtifact.objects.count(), 1)
        resp = self.client.patch('/api/v1/' + self.detail_url, data=new_data)
        self.assertHttpMethodNotAllowed(resp)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(TicketArtifact.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(TicketArtifact.objects.get(
                         pk=self.artifact.pk).name, self.artifact.name)

    def test_delete_artifact_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class APICreateSubTicketArtifactResourceTest(TicketTastypieTestCase):

    def setUp(self):
        super(APICreateSubTicketArtifactResourceTest, self).setUp('/api/v1')
        self.ticket = mommy.make('Ticket')
        self.sourcepackage = sourcepackage_recipe.make()
        self.subticket = mommy.make(
            'SubTicket', ticket=self.ticket, sourcepackage=self.sourcepackage)
        self.artifact = mommy.make(
            'SubTicketArtifact', name='existing_artifact',
            subticket=self.subticket)
        self.resource = 'subticketartifact/'
        self.detail_url = self.resource + '{0}/'.format(self.artifact.pk)
        self.post_artifact_data_subticket = {
            'type': 'SPU',
            'name': 'new_artifact',
            'subticket': '/api/v1/subticket/{0}/'.format(self.subticket.pk),
            'reference': 'http://www.example.com/m1CTLg5FHY',
        }

    def test_post_artifact_subticket(self):
        # `SubTicketArtifact` resource can be created via the API.
        self.post(resource=self.resource,
                  params=self.post_artifact_data_subticket)
        # Retrieve `SubTicketArtifact`s via their context `SubTicket`,
        # this way we can be sure they are properly related.
        [existing, new] = list(self.subticket.subticketartifact_set.all())
        self.assertEqual(existing, self.artifact)
        self.assertEqual('SPU', new.type)
        self.assertEqual('new_artifact', new.name)
        self.assertEqual('http://www.example.com/m1CTLg5FHY', new.reference)

    def test_patch_artifact_not_allowed(self):
        # `SubTicketArtifact` resource cannot be patched via the API.
        resp = self.client.patch(
            '/api/v1/' + self.detail_url, data={'name': 'modified_artifact'})
        self.assertHttpMethodNotAllowed(resp)


class APIUpdateTicketResourceTest(TicketTastypieTestCase):

    def setUp(self):
        super(APIUpdateTicketResourceTest, self).setUp('/api/v1')
        self.ticket = create_ticket()
        self.ticket_detail_url = 'ticket/{0}/'.format(self.ticket.pk)

    def test_patch_ticket_status(self):
        new_data = {
            'current_workflow_step': str(
                TicketWorkflowStep.IMAGE_BUILDING.value),
            'status': str(
                TicketWorkflowStepStatus.COMPLETED.value),
        }
        self.assertEqual(Ticket.objects.count(), 1)
        self.patch('/api/v1/' + self.ticket_detail_url, new_data)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(Ticket.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            int(TicketWorkflowStep.IMAGE_BUILDING))
        self.assertEqual(
            Ticket.objects.get(pk=self.ticket.pk).status,
            int(TicketWorkflowStepStatus.COMPLETED))

    def test_patch_lander_unit(self):
        new_data = {'lander_unit': 123}
        self.patch('/api/v1/' + self.ticket_detail_url, new_data)
        unit = Ticket.objects.get(pk=self.ticket.pk).lander_unit
        self.assertEqual(new_data['lander_unit'], unit)

    def test_update_ticket_status_complete(self):
        new_data = {
            "current_workflow_step": TicketWorkflowStep.COMPLETED.value,
            "status": TicketWorkflowStepStatus.COMPLETED.value,
        }
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            self.ticket.current_workflow_step)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).status,
            self.ticket.status)
        self.patch('/api/v1/' + self.ticket_detail_url, new_data)
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            int(TicketWorkflowStep.COMPLETED))
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).status,
            int(TicketWorkflowStepStatus.COMPLETED))

    def test_update_ticket_status_failed(self):
        '''Ensure we fail a ticket properly

        The workflow step should indicate where the ticket failed. The status
        is how you know if it failed
        '''
        step = TicketWorkflowStep.PKG_BUILDING
        new_data = {
            "current_workflow_step": str(step.value),
            "status": TicketWorkflowStepStatus.FAILED.value,
        }
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            self.ticket.current_workflow_step)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).status,
            self.ticket.status)
        self.patch('/api/v1/' + self.ticket_detail_url, new_data)
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            int(step))
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).status,
            int(TicketWorkflowStepStatus.FAILED))

    def test_update_ticket_status_complete_404(self):
        """
        Test trying to complete the status of a ticket that doesn't exist
        """
        new_data = {
            "current_workflow_step": str(TicketWorkflowStep.COMPLETED.value),
            "status": str(TicketWorkflowStepStatus.COMPLETED.value),
        }
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            self.ticket.current_workflow_step)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).status,
            self.ticket.status)
        resp = self.client.patch('/api/v1/ticket/17394/', data=new_data)
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(resp.status_code, 404)

    def test_update_ticket_just_status(self):
        '''Ensure we can update just the status'''
        new_data = {"status": TicketWorkflowStepStatus.INPROGRESS.value}
        self.patch('/api/v1/' + self.ticket_detail_url, new_data)
        status = Ticket.objects.get(pk=self.ticket.pk).status
        self.assertEqual(new_data['status'], status)

    def test_update_ticket_string(self):
        '''Ensure string status updates work'''
        new_data = {"status": TicketWorkflowStepStatus.INPROGRESS.title}
        self.patch('/api/v1/' + self.ticket_detail_url, new_data)
        status = Ticket.objects.get(pk=self.ticket.pk).status
        self.assertEqual(TicketWorkflowStepStatus.INPROGRESS.value, status)

    def test_update_build_ppa(self):
        # `Ticket.build_ppa` can be updated via PATCH.
        data = {
            'build_ppa': 'ppa:foobar/A_UUID',
        }
        self.patch('/api/v1/' + self.ticket_detail_url, data)
        self.assertEqual(
            'ppa:foobar/A_UUID',
            Ticket.objects.get(pk=self.ticket.pk).build_ppa)


class APIUpdateSubTicketStatuses(TicketTastypieTestCase):

    def setUp(self):
        super(APIUpdateSubTicketStatuses, self).setUp('/api/v1')
        self.sourcepackage = sourcepackage_recipe.make()
        self.subticket = mommy.make(
            'SubTicket', sourcepackage=self.sourcepackage)
        self.subticket_detail_url = (
            'updatesubticketstatus/{0}/'.format(self.subticket.pk))

    def test_patch_subticket_status(self):
        new_data = {
            'current_workflow_step': int(SubTicketWorkflowStep.PKG_BUILDING),
            'status': int(SubTicketWorkflowStepStatus.PKG_BUILDING_WAITING),
        }

        self.assertEqual(SubTicket.objects.count(), 1)
        self.patch('/api/v1/' + self.subticket_detail_url,
                   new_data)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(SubTicket.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(SubTicket.objects.get(
                         pk=self.subticket.pk).current_workflow_step,
                         int(SubTicketWorkflowStep.PKG_BUILDING))
        self.assertEqual(SubTicket.objects.get(pk=self.subticket.pk).status,
                         int(SubTicketWorkflowStepStatus.PKG_BUILDING_WAITING))

    def test_delete_subticket_not_allowed(self):
        resp = self.delete(resource=self.subticket_detail_url)
        self.assertHttpMethodNotAllowed(resp)


class ReviewTestCase(TicketTastypieTestCase):
    def setUp(self):
        super(ReviewTestCase, self).setUp('/api/v1')
        self.ticket = create_ticket()
        self.ticket_uri = '/api/v1/ticket/' + str(self.ticket.pk) + '/'

    def test_create(self):
        params = {
            'ticket': self.ticket_uri,
            'review_type': 'packaging',
            'workflow_step': TicketWorkflowStep.PKG_PUBLISHING.value,
        }
        self.createResource('review/', params=params)

    def test_create_with_workflow_step_title(self):
        params = {
            'ticket': self.ticket_uri,
            'review_type': 'packaging',
            'workflow_step': 'Package publishing',
        }
        self.createResource('review/', params=params)

    def test_update(self):
        t = Review.objects.create(ticket=self.ticket, review_type='a',
                                  workflow_step=0)
        uri = 'review/' + str(t.pk) + '/'
        params = self.get(uri)
        self.assertEqual(ReviewStatus.PENDING.title, params['status'])
        params['status'] = 'Approved'
        self.patch(uri, params)
        self.assertEqual(
            ReviewStatus.APPROVED.value,
            Review.objects.get(pk=t.id).status)

    def test_delete(self):
        t = Review.objects.create(ticket=self.ticket, review_type='a',
                                  workflow_step=0)
        uri = 'review/' + str(t.pk) + '/'
        self.delete(uri)
        self.assertEqual(0, Review.objects.count())

    def test_query(self):
        Review.objects.create(ticket=self.ticket, review_type='a',
                              workflow_step=0)
        resp = self.get('review/', {'ticket': self.ticket.pk})
        self.assertEqual(1, resp['meta']['total_count'])

        resp = self.get('review/', {'status': '10'})
        self.assertEqual(0, resp['meta']['total_count'])
