# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import mock
import unittest

from validator import (
    process_request,
    validate_subticket,
    ValidatorProcessError,
)


class TestValidateSubticket(unittest.TestCase):

    def test_validate_subticket_always_returns_success(self):
        args = ('a_subticket', 'a_series', 'src_archive', 'dst_archive')
        self.assertTrue(validate_subticket(*args))


class TestProcessRequest(unittest.TestCase):

    def setUp(self):
        """Setup testing parameters."""
        super(TestProcessRequest, self).setUp()
        self.params = {
            'ticket_id': 1,
            'progress_trigger': 'A_TRIGGER',
            'subtickets': [{'id': 1,
                            'sourcepackage': {'name': 'foobar'},
                            'source_package_upload': {
                                'version': '1.0'}},
                           {'id': 2,
                            'sourcepackage': {'name': 'barfoo'},
                            'source_package_upload': {
                                'version': '1.0'}}],
            'source_archive': 'SRC_PPA',
            'destination_archive': 'DST_PPA',
            'series': 'A_SERIES',
            'current_workflow_step': 'Queued',
        }

    def test_process_request_missing_variables(self):
        # 'process_request' will raise `ValidatorProcessError` on
        # missing parameters.
        missing_params = self.params.copy()
        del missing_params['series']
        with self.assertRaises(ValidatorProcessError) as cm:
            process_request(missing_params, None, None)
        self.assertEquals(
            "Missing required variables: 'series'.",
            cm.exception.message)

    @mock.patch('validator.validate_subticket')
    def test_process_request_check_only(self, mocked_validate_subticket):
        # 'process_request' validates all given `Subtickets`.
        mocked_ds = mock.MagicMock()
        mocked_cb = mock.MagicMock()
        self.assertEquals({
            'checks': [
                {'subticket_id': 1,
                 'package': 'foobar',
                 'result': 'PASSED',
                 'info': ''},
                {'subticket_id': 2,
                 'package': 'barfoo',
                 'result': 'PASSED',
                 'info': ''},
            ]}, process_request(self.params, mocked_ds, mocked_cb))
        # Validation and status_cb are called for each subticket.
        self.assertEquals(2, mocked_validate_subticket.call_count)
        self.assertEquals(2, mocked_cb.call_count)
        # XXX cprov 2014-06-02: data_store is currently unused.
        self.assertEquals(0, mocked_ds.call_count)

    @mock.patch('validator.validate_subticket')
    def test_validation_error(self, mocked_validate_subticket):
        # `process_request` copes with validation errors.
        mocked_ds = mock.MagicMock()
        mocked_cb = mock.MagicMock()
        mocked_validate_subticket.side_effect = [
            ValidatorProcessError('Cannot Pass!')
        ]
        failed_params = self.params.copy()
        failed_params['subtickets'] = [
            {'id': 1,
             'sourcepackage': {'name': 'foobar'},
             'source_package_upload': {
                 'version': '1.0'}},
        ]
        self.assertEquals({
            'checks': [
                {'subticket_id': 1,
                 'package': 'foobar',
                 'result': 'FAILED',
                 'info': 'Cannot Pass!'},
            ]}, process_request(failed_params, mocked_ds, mocked_cb))
        mocked_validate_subticket.assert_called_once_with(
            failed_params['subtickets'][0],
            failed_params['series'],
            failed_params['source_archive'],
            failed_params['destination_archive'],
        )
        mocked_cb.assert_called_once_with(
            "Validating source 'foobar': FAILED\nERROR: Cannot Pass!")
