# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

import mock

from django.test import TestCase
from ucitests import styles

import status
from status.views import _find_json_health_errors


class TestStatus(TestCase):
    def test_index(self):
        '''Ensure the index is always retrievable.
        We can't really validate the content, but this is a good sanity check
        '''
        content = self.client.get('/status/').content
        self.assertIn('href="/status/" class="active">Engine health', content)

    @mock.patch('status.views.amqp_utils')
    @mock.patch('status.views._get')
    def test_queue_health_not_exist(self, get, amqp_utils):
        '''Ensure we return values when no consumers are found'''
        expected = [
            {'name': 'foo', 'consumers': None},
            {'name': 'bar', 'consumers': None},
        ]
        amqp_utils.MONITORED_QUEUES = [x['name'] for x in expected]
        get.return_value = '[]'
        vals = json.loads(self.client.get('/status/queue_health/').content)
        self.assertEqual(expected, vals)

    @mock.patch('status.views.amqp_utils')
    @mock.patch('status.views._get')
    def test_queue_health_partial(self, get, amqp_utils):
        '''Ensure we detect a single failed worker'''
        expected = [
            {'name': 'foo', 'consumers': 1},
            {'name': 'bar', 'consumers': None},
        ]
        amqp_utils.MONITORED_QUEUES = [x['name'] for x in expected]
        get.return_value = json.dumps([
            {'name': 'foo', 'consumers': 1},
        ])
        vals = json.loads(self.client.get('/status/queue_health/').content)
        self.assertEqual(expected, vals)

    @mock.patch('status.views._find_worker_errors')
    @mock.patch('status.views._find_json_health_errors')
    def test_health_check_ok(self, find_json_errors, find_worker_errors):
        content = self.client.get('/status/health_check/').content
        self.assertEqual('OK', content)

    @mock.patch('status.views._find_worker_errors')
    @mock.patch('status.views._find_json_health_errors')
    def test_health_check_fail(self, find_json_errors, find_worker_errors):
        find_worker_errors.return_value = ['error 1']
        find_json_errors.return_value = ['error 2']
        resp = self.client.get('/status/health_check/')
        self.assertEqual(500, resp.status_code)
        self.assertEqual('error 1\nerror 2', resp.content)

    @mock.patch('status.views._get')
    def test_find_json_health_errors(self, get):
        status_urls = json.dumps({
            'foo': {'url': '/foo', 'unit': 'foo'},
            'bar': {'url': '/bar', 'unit': 'bar'},
            'bam': {'url': '/bam', 'unit': 'bam'},
        })
        get.side_effect = [
            status_urls,
            json.dumps([{'status': 'okay', 'value': True, 'label': 'foo'}]),
            json.dumps([{'status': 'fail', 'value': False, 'label': 'bar'}]),
            RuntimeError('bam'),
        ]
        errors = _find_json_health_errors()
        self.assertEqual(2, len(errors))


class TestPep8(styles.TestPep8):
    packages = [status]


class TestPyflakes(styles.TestPyflakes):
    packages = [status]
